/*|------------------------------------------------------------------------------|
  |                   ITGCommon NetLinx Libraries v33 06/20/19                   |
  |                                                                              |
  | DPSTOA: Format DPS to D:P:S                                                  |
  | fnChannelOff: Turn channel OFF Helper Function                               |
  | fnChannelOffList: Turn channels in list OFF Helper Function                  |
  | fnChannelOn: Turn channel ON Helper Function                                 |
  | fnChannelOnList: Turn channels in list ON Helper Function                    |
  | fnCommandEncode: Returns a command-encoded string                            |
  | fnIntOnOff: Convert integer value to ON/OFF value                            |
  | fnIntTrueFalse: Convert integer value to boolean value                       |
  | fnIsChannelOff: Is channel off?                                              |
  | fnIsChannelOn: Is channel on?                                                |
  | fnIsInRangeFloat: Is value between low and high inclusively                  |
  | fnIsInRangeInteger: Is value between low and high inclusively                |
  | fnIsNotInRangeDouble: Is value not between low and high inclusively          |
  | fnIsNotInRangeFloat: Is value not between low and high inclusively           |
  | fnIsNotRangeInteger: Is value not between low and high inclusively           |
  | fnIsValidDPS: Validates a device DPS device != 0 and Port != 0               |
  | fnMathIsInRangeDouble: Is value between low and high inclusively             |
  | fnOpenIPPort: Opens a server port to listen for client requests.             |
  | fnOpenIPPortResults: Prints message to telnet for the given IP_SERVER_OPEN   |
  |       status                                                                 |
  | fnPulse: Pulse Channel Helper Function                                       |
  | fnRGBToHex: Creates a Hex RGB value for the give values                      |
  | fnSendCommand: Send Command Helper Function                                  |
  | fnSendCommand: Send String Helper Function                                   |
  | fnSendCommandEx: Send Command Helper Function, combines sHeader and sCommand |
  |       to sHeader-sCommand                                                    |
  | fnStringDEVListToDelimString: Create comma seperated DPS values              |
  | fnSystemMACAddress: Get the Systems MAC Address                              |
  | fnToggleState: Toggle state                                                  |
  | fnTPToggleChannel: Toggle channel ON or OFF                                  |
  |                                                                              |
  |                               Array Functions                                |
  |                                                                              |
  | fnArrayDeviceIndex: Finds the index for any matching device entry in an array|
  | fnArrayDoubleIndex: Finds the index for any matching item in the double array|
  | fnArrayIntegerIndex: Finds the index for any matching integer entry          |
  | fnArrayIsInIntegerIndex: Returns true if item is in array                    |
  | fnArrayIntegerToCSString: Create comma seperated sting of array values       |
  | fnArrayInsertionSort: In-place sorting of an integer array using the         |
  |         insertion sort method. Efficient for small data sets. Crude testing  |
  |         on a NI-900 suggests that insertion sort is faster than the          |
  |         non-recursive quicksort for random integer arrays up to a size of    |
  |         about 25 elements.                                                   |
  | fnArrayMaxValue: Finds the maximum value in an integer array                 |
  | fnArrayMaxValueDouble: Finds the maximum value in an double array            |
  | fnArrayMinValue: Finds the maximum value in an integer array                 |
  | fnArrayMinValueDouble: Finds the maximum value in an double array            |
  | fnArrayPartition: In-place partitioning of an integer array.  It partitions  |
  |       the portion of the array between indices left and right inclusive by   |
  |       moving all elements less than list[pivot_index] before the pivot and   |
  |       those equal or greater after it.                                       |
  | fnArrayQuickSort: In-place sorting of an integer array using a non-recursive |
  |       quicksort method                                                       |
  | fnArrayStringIndex: Finds the index for any matching string entry in an      |
  |       array.                                                                 |
  | fnArraySwap: In-place swap of two elements in an integer array               |
  |                                                                              |
  |                               Debug Functions                                |
  |                                                                              |
  | fnDebugDebugMessage: Prints a debug message if debug level DEBUG or greater  |
  | fnDebugDecimal: Prints a debug message forced to decimal                     |
  | fnDebugErrorMessage: Prints a debug message if debug level ERROR or greater  |
  | fnDebugGetLevelFromString: Gets a numerical debug level based on it's        |
  |       equivalent string representation.                                      |
  | fnDebugGetLevelString: Returns the current debug level as String.            |
  | fnDebugHex: Prints a debug message forced to hex - this avoids situations    |
  |       where a hex value is a valid ASCII character                           |
  | fnDebugInfoMessage: Prints a debug message if debug level INFO or greater    |
  | fnDebugIsLevelDebug: Query if debug level it set to DEBUG                    |
  | fnDebugIsLevelError: Query if debug level it set to ERROR or greater         |
  | fnDebugIsLevelInfo: Query if debug level it set to INFO or greater           |
  | fnDebugIsLevelWarning: Query if debug level it set to WARNING or greater     |
  | fnDebugLevelString: Returns the requested debug level as String.             |
  | fnDebugQueryDebugState: Query the given device for it's debug level          |
  | fnDebugSetLevel: Sets the current system debugging level for controlling     |
  |       debug message verbosity.                                               |
  | fnDebugLevelString: Sets the current system debugging level for controlling  |
  |       debug message verbosity.                                               |
  | fnDebugWarningMessage: Prints a debug message if debug level WARNING or      |
  |       greater                                                                |
  |                                                                              |
  |                                 IO Functions                                 |
  |                                                                              |
  | fnIOFileExists: Determines if given file exist on disc                       |
  | fnIOFileNotExists: Determines if given file does not exist on disc           |
  | fnIOPrint: Print a character array to the output stream.                     |
  | fnIOPrintBanner: Prints banner to the output stream.                         |
  | fnIOPrintBannerLine: Prints banner to the output stream.                     |
  |                             fnIOPrintBetaBanner                              |
  | fnIOPrintln: Print a line to the output stream.                              |
  | fnIOWrite: Write a character array to the output device.                     |
  |                                                                              |
  |                                Math Functions                                |
  |                                                                              |
  | fnMathBase10Logarithm: Calculate the base 10 logarithm of the passed number. |
  | fnMathBinaryLogarithm: Calculate the binary logarithm of the passed number.  |
  | fnMathBuildDouble: Build a double using the binary info stored across two    |
  |       longs. It is assumed that the data is stored as per the IEEE754        |
  |       standard.                                                              |
  | fnMathBuildFloat: Build a float using a IEEE754 bit pattern stored in a      |
  |       long.                                                                  |
  | fnMathCeiling: Returns the smallest (closest to negative infinity) long      |
  |       value that is not less than the argument and is equal to a             |
  |       mathematical integer.                                                  |
  | fnMathDecimalToBinaryCodedDecimal: Encode a decimal value as binary-coded    |
  |       decimal (BCD).                                                         |
  | fnMathDoubleHighToBits: Load the raw data stored in bits 63 - 32 of a double |
  |       into a long.                                                           |
  | fnMathDoubleLowToBits: Load the raw data stored in bits 31 - 0 of a double   |
  |       into a long.                                                           |
  | fnMathFastSqrt: Approximate the square root of the passed number based on    |
  |       the inverse square root algorithm in fnMathInverseSqrt(x). This is     |
  |       MUCH faster than fnMathSquareRoot(x) and recommended over              |
  |       fnMathSquareRoot() FOR use anywhere a precise square root is not       |
  |       required. Error is approx +/-0.17%.                                    |
  | fnMathFloatToBits: Load a float value's IEEE 754 bit pattern into a long.    |
  | fnMathFloor: Returns the largest (closest to positive infinity) long value   |
  |       that is not greater than the argument and is equal to a mathematical   |
  |       integer.                                                               |
  | fnMathIEEERemainder: Computes the remainder operation on two arguments as    |
  |       prescribed by the IEEE 754 standard.                                   |
  | fnMathInverseSqrt: Approximate the inverse square root of the passed number. |
  |       This method uses a integer shift and single Newton refinement aka      |
  |       Quake 3 method. Original algorithm by Greg Walsh.                      |
  | fnMathIsInfinite: Checks if a value is either positive infinity or negative  |
  |       infinity.                                                              |
  | fnMathIsInt: Returns TRUE if the argument has no decimal component,          |
  |       otherwise Returns FALSE. +/-Inf and 0 will return TRUE, subnormal and  |
  |       NaN's will return FALSE.                                               |
  | fnMathIsNaN: Checks if a value is NaN.                                       |
  | fnMathIsNear: Compares two numbers and return TRUE if they are within        |
  |       MATH_PRECISION of each other.                                          |
  | fnMathIsNumeric: Checks if a string value contains only a numeric value.     |
  | fnMathLogarithm: Calculate the logarithm of the passed number in the         |
  |       specified base.                                                        |
  | fnMathLShiftDouble: Left shift (<<) a double 1 bit.                          |
  | fnMathMapRange: Map a value within a set of original bounds to its           |
  |       equivalent given a secondary set of bound.                             |
  | fnMathNaturalLogarithm: Calculate the natural logarithm of the passed        |
  |       number.                                                                |
  | fnMathPower: Calculate x raised to the n.                                    |
  | fnMathRandom: Returns a double value with a positive sign, greater than or   |
  |       equal to 0.0 and less than 1.0.                                        |
  | fnMathRawBEtoLong: Load 4 bytes of Big Endian data contained in a character  |
  |       array into a long. Note: Array position 1 should contain MSB.          |
  | fnMathRawLEtoLong: Load 4 bytes of Little Endian data contained in a         |
  |       character array into a long.                                           |
  | fnMathRound: Rounds a flouting point number to it's closest whole number.    |
  | fnMathRShiftDouble: Right shift (>>) a double 1 bit.                         |
  | fnMathSquareRoot: Calculate the square root of the passed number. This       |
  |       function takes a log base 2 approximation then iterates a Babylonian   |
  |       refinement until the answer is within the math libraries defined       |
  |       precision or exceeds 1000 steps of refinement.                         |
  | fnScaleToLevelRange: Scales given value to Level Range (0..255)              |
  |                                                                              |
  |                               String Functions                               |
  |                                                                              |
  | fnStringBoolToString: Converts a boolean value to its string equivalent of   |
  |       either a 'ON' or 'OFF'.                                                |
  | fnStringCiGetBetween: Gets the first instance of a string contained within   |
  |       the bounds of two sub-strings case insensitive                         |
  | fnStringConcatenate: Concatenates two string together                        |
  | fnStringContains: Search through a string for a match strings, case          |
  |       sensistive and return True if found else False                         |
  | fnStringNotContains: Search through a string for a match strings, case       |
  |       sensistive and return Ture if not found else False                     |
  | fnStringContainsIgnoreCase: Search through a string for a match strings,     |
  |       case in-sensistive return True if found else False                     |
  | fnStringNotContainsIgnoreCase: Search through a string for a match strings,  |
  |       case in-sensistive return True if not found else False                 |
  | fnStringDateInvert: Switches the day and month fields of a date string (for  |
  |       converting between US and international standards). for example        |
  |       05/28/2009 becomes 28/05/2009.                                         |
  | fnStringEndsWith: Check is a string end with another string.                 |
  | fnStringEndsWithIgnoreCase: Check is a string ends with another string,      |
  |    ignore case.                                                              |
  | fnStringNotEndsWith: Check is a string end with another string.              |
  | fnStringNotEndsWithIgnoreCase: Check is a string ends with another string,   |
  |    ignore case.                                                              |
  | fnStringNotStartsWithIgnoreCase: Check is a string starts with another       |
  |    string ignore case.                                                       |
	| fnStringNotStartsWith: Check is a string starts with another string.         |
  | fnStringExplode: Explodes a string with a CHAR delimiter into an array of    |
  |       strings. The fnStringExploded data will be placed into ret[]. Due to   |
  |       NetLinx bugs, you must specify the length of ret in ret_len if you     |
  |       want the returned array sanitized.                                     |
  | fnStringExplodeQuoted: Explodes a string with a CHAR delimiter into an array |
  |       of strings. The fnStringExploded data will be placed into ret[].       |
  |       Honours quotes; a character passed as quote (such asdouble quotes) are |
  |       treated as one segment.                                                |
  | fnStringFindMulti: Search through a string for a match against a list of     |
  |       possible sub-strings and return the element index of the matched       |
  |       string                                                                 |
  | fnStringFormatNumberThousands: Foramt the given number ###,###,###           |
  | fnStringGetBetween: Gets the first instance of a string contained within the |
  |       bounds of two sub-strings case sensitive                               |
  | fnStringGetBetweenEx:                                                        |
  | fnStringGetKey: Gets the key from a single key/value pair string with the    |
  |       specified delimiter.                                                   |
  | fnStringGetListItem: Gets an item from a string list. This can be used to    |
  |       grab word n within a string by passing 'space' as the delimiter.       |
  | fnStringGetValue: Gets the value from a key/value pair string with the       |
  |       specified delimiter.                                                   |
  | fnStringImplode: Concatenates elements of an array of strings into a single  |
  |       string with a delimiter string inserted between each item.             |
  | fnStringIntArrayToString: Converts an integer array into a comma separated   |
  |       string list of its values.                                             |
  | fnStringIsEquals: Compare 2 strings                                          |
  | fnStringIsEqualsIgnoreCase: Compare 2 strings using case in-sensistive       |
  | fnStringNotEquals: Compare 2 strings                                         |
  | fnStringNotEqualsIgnoreCase: Compare 2 strings using case in-sensistive      |
  | fnStringIsLengthSizeError: Callback triggered when a function within this    |
  |       string processing library attempts to process anything that will       |
  |       result in a return size greater than that defined by                   |
  |       STRING_RETURN_SIZE_LIMIT.                                              |
  | fnStringIsPrintable: Hecks to see if the passed string is all printable      |
  |       ASCII characters.                                                      |
  | fnStringIsPrintableChar: Checks to see if the passed character is a          |
  |       printable ASCII character.                                             |
  | fnStringIsSizeError: Allback triggered when a function within this string    |
  |       processing library attempts to process anything that will result in a  |
  |       return size greater than that defined by STRING_RETURN_SIZE_LIMIT.     |
  | fnStringIsWhitespace: Checks to see if the passed character is a whitespace  |
  |       character.                                                             |
  | fnStringLastIndexOf: Finds the last index for string entry in a string       |
  | fnStringIndexOf: Finds the index for string entry in an string               |
  | fnStringIndexOfIgnoreCase: Finds the index for string entry in an string     |
  | fnStringLeftTrim: Returns a copy of the string with the left whitespace      |
  |       removed. If no printable characters are found, an empty string will be |
  |       returned.                                                              |
  | fnStringPostfixToLength: Returns a string post fixed with a specified CHAR,  |
  |       to a specified length. If the string is the same size or is larger     |
  |       than the specified length, Returns the original string.                |
  | fnStringPrefixToLength: Returns a string prefixed with a specified value, up |
  |       to a specified length. If the string is the same size or is larger     |
  |       than the specified length, Returns the original string.                |
  | fnStringRemoveStringByLength: Returns the left sub-strings of a string up to |
  |       the specified number of characters.                                    |
  | fnStringRemoveNonPrintable: remove all non-printable ASCII characters.       |
  | fnStringReplace: Replace all occurrences of a sub-strings with another       |
  |       string.                                                                |
  | fnStringReplaceTabWithSpaces: Replace all occurrences of TAB with spaces     |
  | fnStringReverse: Reverse a string.                                           |
  | fnStringRightTrim: returns a copy of the string with the right whitespace    |
  |       removed. If no printable characters are found, an empty string will be |
  |       returned.                                                              |
  | fnStringStartsWith: Check is a string starts with another string.            |
  | fnStringStartsWithIgnoreCase: Check is a string starts with another string,  |
  |    ignore case.                                                              |
  | fnStringSubstr: Wrapper method for MID_STRING to bring inline with other     |
  |       programming languages.                                                 |
  | fnStringSubstring: Alternative to fnStringSubstr which allows an end         |
  |       location to be specified instead of a count                            |
  | fnStringSubstringEx: Returns a substring beginning at specified location to  |
  |       end of string                                                          |
  | fnStringSuffixToLength: Returns a string suffixed with a specified value, up |
  |       to a specified length. If the string is the same size or is larger     |
  |       than the specified length, Returns the original string.                |
  | fnStringToBool: Converts common string representations of boolean values     |
  |       into their boolean value equivalents.                                  |
  | fnStringTrim: Returns a copy of the string with the whitespace removed. If   |
  |       no printable characters are found, an empty string will be returned.   |
  | fnStringTruncate: Returns a string truncated to a specific length. If the    |
  |       string is less than the length specified the original string is        |
  |       returned. If it is truncated an ellipsis will be appended.             |
  | fnStringTruncateEx: Returns a string truncated to a specific length. If the  |
  |       string is less than the length specified the original string is        |
  |       retured. If it is truncated to contents of value is appended to the    |
  |       truncated string.                                                      |
  | fnStringUcFirst: Eturns a copy of a string with the first alpha character    |
  |       capitalized. Non alpha characters are not modified. Pass a             |
  |       lower_string()'d string to lowercase all other characters.             |
  | fnStringUcWords: Returns a copy of a string with the first alpha character   |
  |       in each word capitalized. Non alpha characters are not modified. Pass  |
  |       a lower_string()'d string to lowercae all other characters.            |
  | fnStringUrlEncode: Returns a url-encoded string according to RFC 1736 / RFC  |
  |       2732.                                                                  |
  | fnStripCharsRight: Remove characters from the end of the string.             |
  |                                                                              |
  |                            Touch Panel Functions                             |
  |                                                                              |
  | fnTPConvertValueToPercentage: Converts 0..255 value to %                     |
  | fnTPDisableButton: Disable buttons with a set variable text range            |
  | fnTPEnableButton: Enable buttons with a set variable text range              |
  | fnTPHideAllPopups: Close all popups on all pages.                            |
  | fnTPHideAllPopupsOnPage: Close all popups on a specified page. If the page   |
  |       name is empty, the current page is used.                               |
  | fnTPHideButton: Hide specified button on touch panel                         |
  | fnTPHidePopup: Deactivate a specific popup page on the current page. If the  |
  |       popup page is part of a group, the whole group is deactivated.         |
  | fnTPIsG4: Is this a G4 panel?                                                |
  | fnTPIsG5: Is this a G5 panel?                                                |
  | fnTPSetButtonText:                                                           |
  | fnTPSetButtonTextAllStates: Set button text for all button states            |
  | fnTPShowButton: Show request button(s) in address range                      |
  | fnTPShowPopup: Activate a specific popup page to launch on the current page. |
  |       If the popup page is already on, do not re-draw it.                    |
  | fnTPTextCommand: Text command based upon panel type (G4 vs G5)               |
  |                                                                              |
  |                          Miscellaneous Functions                             |
  |                                                                              |
  | _fnDebugMessage (INTERNAL FUNCTION not intended for direct usage): Voices a  |
  |       debug message IF required by the current debug level. All system       |
  |       messages should pass through here. If the ability to dump to a file or |
  |       networked logging service is required it can be added here.            |
  | _fnIOFileErrorToText (INTERNAL FUNCTION not intended for direct usage):      |
  |       Prints text of file error  to the output stream.                       |
  | _fnLog (INTERNAL FUNCTION not intended for direct usage): Write log          |
  |       statement to disc                                                      |
  | _fnReportParseError (INTERNAL FUNCTION not intended for direct usage): Write |
  |       error code's text description to Telnet session                        |
  |------------------------------------------------------------------------------|*/

PROGRAM_NAME='ITG Libraries Helpers'

#IF_NOT_DEFINED __ITG_LIBRARIES_HELPERS__
#DEFINE __ITG_LIBRARIES_HELPERS__

#INCLUDE 'SNAPI'

DEFINE_DEVICE

  devDebugCommand                       = 33099:1:0

DEFINE_CONSTANT

  PRT_PREFIX                            = 'NetLinx >>>>>> '
  PRT_PREFIX_CONTINUE                   = '>>> '
  LOGFILE_NAME                          =  'SendCommandsLog.xml'
  PRINTLN_MAX                           = 130
  PRINTLN_MAX_BREAK                     = 30
  COMMA_DELM                            =  ','
  EQUAL_DELM                            =  '='

  TXT_ON                                = 'ON'
  TXT_OFF                               = 'OFF'
  TIMEWAIT_SECONDS_1                    =  10
  TIMEWAIT_SECONDS_2                    =  20
  TIMEWAIT_SECONDS_3                    =  30
  TIMEWAIT_SECONDS_4                    =  40
  TIMEWAIT_SECONDS_5                    =  50
  TIMEWAIT_SECONDS_10                   =  100
  TIMEWAIT_SECONDS_20                   =  200
  TIMEWAIT_SECONDS_30                   =  300
  TIMEWAIT_SECONDS_60                   =  600

  BTN_STATE_ALL                         =  '0'
  BTN_STATE_OFF                         =  '1'
  BTN_STATE_ON                          =  '2'

  CMD_DEBUG                              = 'DEBUG'
  DEBUG_OFF                             =  0
  DEBUG_ERROR                           =  1
  DEBUG_WARN                            =  2
  DEBUG_INFO                            =  3
  DEBUG_DEBUG                           =  4

  DEBUG_MAX_LEVEL                       =  4

  CHAR DEBUG_LEVEL_STRINGS[5][16]       =
  {
    'Off',
    'Error',
    'Warn',
    'Info',
    'Debug'
  }

  DEBUG_LEVEL_VALUES[5]                 =
  {
    DEBUG_OFF,
    DEBUG_ERROR,
    DEBUG_WARN,
    DEBUG_INFO,
    DEBUG_DEBUG
  }

  DOUBLE MATH_E                         =  2.718281828459045
  DOUBLE MATH_PI                        =  3.141592653589793

  DAY_OF_WEEK_SUNDAY 									  = 1
	DAY_OF_WEEK_MONDAY 										= 2
	DAY_OF_WEEK_TUESDAY 									= 3
	DAY_OF_WEEK_WEDNESDAY 								= 4
	DAY_OF_WEEK_THURSDAY 									= 5
	DAY_OF_WEEK_FRIDAY 										= 6
	DAY_OF_WEEK_SATURDAY 									= 7

/*|------------------------------------------------------------------------------|
  | Precision required FOR processor intensive math functions. If accuracy is    |
  |       not integral to their use this may be increased to improve             |
  |       performance.                                                           |
  |------------------------------------------------------------------------------|*/
  DOUBLE MATH_PRECISION = 1.0e-13

  STRING_RETURN_SIZE_LIMIT              =  5000              // Maximum string return size for string manipulation functions

  DEV IO_OUT                            =  0:0:0             // output stream device
  CHAR IO_LINE_SEPERATOR                =  $0A               // seperator to insert between lines
  CHAR IO_TAB                           =  $09
  CHAR IO_TAB_AS_SPACES[3]              = '   '
  CHAR ASCII_DBL_QUOTE                  =  $22
  CHAR ASCII_SINGLE_QUOTE               =  $27

DEFINE_VARIABLE

  //  #WARN 'NetLinx debug level is ERROR, to change use Helper Libraries fnDebugSetLevel(0-4)'
  //  #WARN   'Debug Levels: 0-Off, 1-Error, 2-Warning, 3-Info, 4-Debug'
  //  #WARN   'Set Debug Level via COMMAND: SEND_C 33099,"DEBUG-[0-4]"'
  //  #WARN 'Query Debug Level via COMMAND: SEND_C 33099,"?DEBUG"'

  VOLATILE INTEGER nDebugLevel           = DEBUG_ERROR       // Current system debug level, default to Error

  VOLATILE DOUBLE MATH_NaN
  VOLATILE DOUBLE MATH_POSITIVE_INFINITY
  VOLATILE DOUBLE MATH_NEGATIVE_INFINITY
  VOLATILE DOUBLE MATH_TWO_52
  VOLATILE INTEGER nFirstTime           =  0


DEFINE_FUNCTION fnChannelOn(DEV vdvDevice,
  INTEGER      nChannel)
{
/*|------------------------------------------------------------------------------|
  | fnChannelOn: Turn channel ON Helper Function                                 |
  |    Parameters:                                                               |
  |       vdvDevice Virtual device to pulse channel on                           |
  |       nChannel  Channel to turn ON                                           |
  |------------------------------------------------------------------------------|*/

  _fnLog("'ON ',DPSTOA(vdvDevice),':',ITOA(nChannel)")

  ON[vdvDevice,nChannel]
}

DEFINE_FUNCTION fnChannelOnList(DEV vdvDevice,
  INTEGER      nChannels[])
{
/*|------------------------------------------------------------------------------|
  | fnChannelOnList: Turn channel ON Helper Function                             |
  |    Parameters:                                                               |
  |       vdvDevice Virtual device list to pulse channel on                      |
  |       nChannel  Channel to turn ON                                           |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER idx

  FOR(idx = 1; idx <= LENGTH_ARRAY(nChannels); idx++)
  {
    fnChannelOn(vdvDevice, nChannels[idx])
  }
}

DEFINE_FUNCTION fnChannelOff(DEV vdvDevice,
  INTEGER      nChannel)
{
/*|------------------------------------------------------------------------------|
  | fnChannelOff: Turn channel OFF Helper Function                               |
  |    Parameters:                                                               |
  |       vdvDevice Virtual device to pulse channel on                           |
  |       nChannel  Channel to turn OFF                                          |
  |------------------------------------------------------------------------------|*/

  _fnLog("'OFF ',DPSTOA(vdvDevice),':',ITOA(nChannel)")

  OFF[vdvDevice,nChannel]
}

DEFINE_FUNCTION fnChannelOffList(DEV vdvDevice,
  INTEGER      nChannels[])
{
/*|------------------------------------------------------------------------------|
  | fnChannelOffList: Turn channel OFF Helper Function                           |
  |    Parameters:                                                               |
  |       vdvDevice Virtual device list to pulse channel off                     |
  |       nChannel  Channel to turn OFF                                          |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER idx

  FOR(idx = 1; idx <= LENGTH_ARRAY(nChannels); idx++)
  {
    fnChannelOff(vdvDevice, nChannels[idx])
  }
}

DEFINE_FUNCTION INTEGER fnIsChannelOn(DEV dv,
  INTEGER channel)
{
/*|------------------------------------------------------------------------------|
  | fnIsChannelOn: Is channel on?                                                |
  |    Parameters:                                                               |
  |       dv      Device                                                         |
  |       channel Channel to check                                               |
  |    Returns:                                                                  |
  |       True if on, else false                                                 |
  |------------------------------------------------------------------------------|*/

  RETURN [dv,channel]
}

DEFINE_FUNCTION INTEGER fnIsChannelOff(DEV dv,
  INTEGER channel)
{
/*|------------------------------------------------------------------------------|
  | fnIsChannelOff: Is channel off?                                              |
  |    Parameters:                                                               |
  |       dv      Device                                                         |
  |       channel Channel to check                                               |
  |    Returns:                                                                  |
  |       True if off, else false                                                |
  |------------------------------------------------------------------------------|*/
  RETURN !fnIsChannelOn(dv,channel)
}

DEFINE_FUNCTION fnPulse(DEV vdvDevice,
  INTEGER      nChannel)
{
/*|------------------------------------------------------------------------------|
  | fnPulse: Pulse channel Helper Function                                       |
  |    Parameters:                                                               |
  |       vdvDevice Virtual device to pulse channel on                           |
  |       nChannel  Channel to pulse                                             |
  |------------------------------------------------------------------------------|*/

  _fnLog("'PULSE ',DPSTOA(vdvDevice),':',ITOA(nChannel)")

  PULSE[vdvDevice,nChannel]
}

DEFINE_FUNCTION fnToggleChannel(DEV dv,
	INTEGER channel)
{
/*|------------------------------------------------------------------------------|
	| fnToggleChannel: Toggle channel ON or OFF                                    |
	|    Parameters:                                                               |
	|       dv      Device                                                         |
	|       channel Channel to toggle                                              |
	|------------------------------------------------------------------------------|*/

	IF(fnIsChannelOn(dv,channel))
	{
		fnChannelOff(dv,channel)
	}
	ELSE
	{
		fnChannelOn(dv,channel)
	}
}


DEFINE_FUNCTION fnDebugSetDuetDebugLevel(dev vdv,integer level)
{
/*|------------------------------------------------------------------------------|
  | fnDebugSetDuetDebugLevel: SEND_COMMAND vdv,DEBUG-level                       |
  |    Parameters:                                                               |
  |       vdv   Device                                                           |
  |       level level to set to [ DEBUG_OFF,DEBUG_ERROR,DEBUG_WARN,              |
  |             DEBUG_INFO,DEBUG_DEBUG ]                                         |
  |------------------------------------------------------------------------------|*/

  fnSendCommandEx( vdv, CMD_DEBUG,ITOA(level))
}

DEFINE_FUNCTION CHAR[20] DPSTOA (dev dvIn)
{
/*|------------------------------------------------------------------------------|
  | DPSTOA: Format DPS to D:P:S                                                  |
  |    Parameters:                                                               |
  |       dvIn Device                                                            |
  |    Returns:                                                                  |
  |       D:P:S                                                                  |
  |------------------------------------------------------------------------------|*/

  RETURN "ITOA(dvIn.number),':',ITOA(dvIn.port),':',ITOA(dvIn.system)"
}

DEFINE_FUNCTION INTEGER fnArrayDeviceIndex(DEV dvItem,
  DEV dvList[])
{
/*|------------------------------------------------------------------------------|
  | fnArrayDeviceIndex: Finds the index for any matching device entry in an array|
  |      Parameters:                                                             |
  |          dvItem find in the array                                            |
  |          dvList array of items                                               |
  |      Returns:                                                                |
  |           The index of the matching value or 0 if no match                    |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER idx

  FOR(idx = 1; idx <= LENGTH_ARRAY(dvList); idx++)
  {
    IF(dvItem == dvList[idx])
    {
      RETURN idx
    }
  }

  RETURN 0
}

DEFINE_FUNCTION INTEGER fnArrayDoubleIndex(DOUBLE dItem,
  DOUBLE dList[])
{
/*|------------------------------------------------------------------------------|
  | fnArrayDoubleIndex: Finds the index for any matching item in the double array|
  |      Parameters:                                                             |
  |          dItem find in the array                                             |
  |          dList array of items                                                |
  |      Returns:                                                                |
  |          The index of the matching value or 0 if no match                    |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER idx

  FOR(idx = 1; idx <= LENGTH_ARRAY(dList); idx++)
  {
    IF(dItem == dList[idx])
    {
      RETURN idx
    }
  }

  RETURN 0
}

DEFINE_FUNCTION fnArrayInsertionSort(INTEGER nList[])
{
/*|------------------------------------------------------------------------------|
  | fnArrayInsertionSort: In-place sorting of an integer array using the         |
  |         insertion sort method. Efficient for small data sets. Crude testing  |
  |         on a NI-900 suggests that insertion sort is faster than the          |
  |         non-recursive quicksort for random integer arrays up to a size of    |
  |         about 25 elements.                                                   |
  |    Parameters:                                                               |
  |       nList array of integers                                                |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER length
  STACK_VAR INTEGER n
  STACK_VAR INTEGER val
  STACK_VAR INTEGER hole_pos

  length                                = LENGTH_STRING(nList)

  FOR(n = 2; n <= length; n++)
  {
    val                                 = nList[n]
    hole_pos                            = n

    WHILE(val < nList[hole_pos - 1])
    {
      nList[hole_pos]                    = nList[hole_pos - 1]
      hole_pos--

      IF(hole_pos == 1)
      {
        BREAK
      }
    }

    nList[hole_pos]                      = val
  }
}

DEFINE_FUNCTION INTEGER fnArrayIntegerIndex(INTEGER nItem,
  INTEGER nList[])
{
/*|------------------------------------------------------------------------------|
  | fnArrayDoubleIndex: Finds the index for any matching device entry in an      |
  |         array.                                                               |
  |      Parameters:                                                             |
  |          nItem find in the array                                             |
  |          nList array of items                                                |
  |      Returns:                                                                |
  |          The index of the matching value or 0 if no match               |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER idx

  FOR(idx = 1; idx <= LENGTH_ARRAY(nList); idx++)
  {
    IF(nItem == nList[idx])
    {
      RETURN idx
    }
  }

  RETURN 0
}


DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnArrayIntegerToCSString(INTEGER nList[])
{
/*|------------------------------------------------------------------------------|
	| fnArrayIntegerToCSString: Create comma seperated sting of array values       |
	|      Parameters:                                                             |
	|          nList array of items                                                |
	|      Returns:                                                                |
	|          Comma seperated string of array values                              |
	|------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER idx
	STACK_VAR CHAR c[STRING_RETURN_SIZE_LIMIT]

  FOR(idx = 1; idx <= LENGTH_ARRAY(nList); idx++)
  {
		IF (idx > 1)
		{
			c = fnStringConcatenate(c,',')
		}

		c = fnStringConcatenate(c,ITOA(nList[idx]))
  }

  RETURN c
}

DEFINE_FUNCTION INTEGER fnArrayIsInIntegerIndex(INTEGER nItem,
  INTEGER nList[])
{
/*|------------------------------------------------------------------------------|
  | fnArrayIsInIntegerIndex: Returns true if item is in array                    |
  |      Parameters:                                                             |
  |          nItem find in the array                                             |
  |          nList array of items                                                |
  |      Returns:                                                                |
  |          The true it matching value or False if no match                     |
  |------------------------------------------------------------------------------|*/

  RETURN fnArrayIntegerIndex(nItem, nList) > 0
}

DEFINE_FUNCTION INTEGER fnArrayMaxValue(INTEGER nList[])
{
/*|------------------------------------------------------------------------------|
  | fnArrayMaxValue: Finds the maximum value in an integer array                 |
  |    Parameters:                                                               |
  |        nList array of integer                                                |
  |    Returns:                                                                  |
  |        The maximum value in the array                                        |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER idx
  STACK_VAR INTEGER val

  val                                   = 0

  FOR(idx = 1; idx <= LENGTH_ARRAY(nList); idx++)
  {
    val                                 = MAX_VALUE(val, nList[idx])
  }

  RETURN val
}

DEFINE_FUNCTION DOUBLE fnArrayMaxValueDouble(DOUBLE nList[])
{
/*|------------------------------------------------------------------------------|
  | fnArrayMaxValueDouble: Finds the maximum value in an double array            |
  |    Parameters:                                                               |
  |        nList array of double                                                 |
  |    Returns:                                                                  |
  |        The maximum value in the array                                        |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER idx
  STACK_VAR DOUBLE  val

  val                                   = MATH_NEGATIVE_INFINITY

  FOR(idx = 1; idx <= LENGTH_ARRAY(nList); idx++)
  {
    IF(nList[idx] > val)
    {
      val                               = nList[idx]
    }
  }

  RETURN val
}

DEFINE_FUNCTION INTEGER fnArrayMinValue(INTEGER nList[])
{
/*|------------------------------------------------------------------------------|
  | fnArrayMinValue: Finds the maximum value in an integer array                 |
  |    Parameters:                                                               |
  |        nList array of integer                                                |
  |    Returns:                                                                  |
  |        The minimum value in the array                                        |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER i
  STACK_VAR INTEGER val

  val       = 65535;

  FOR(i = 1; i <= LENGTH_ARRAY(nList); i++)
  {
    val       = MIN_VALUE(val, nList[i]);
  }

  RETURN val;
}

DEFINE_FUNCTION DOUBLE fnArrayMinValueDouble(DOUBLE nList[])
{
/*|------------------------------------------------------------------------------|
  | fnArrayMinValueDouble: Finds the maximum value in an double array            |
  |     Parameters:                                                              |
  |          nList array of double                                               |
  |     Returns:                                                                 |
  |          The minimum value in the array                                      |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER i
  STACK_VAR DOUBLE val

  val                                   = MATH_POSITIVE_INFINITY

  FOR(i = 1; i <= LENGTH_ARRAY(nList); i++)
  {
    IF(nList[i] < val)
    {
      val       = nList[i];
    }
  }

  RETURN val;
}

DEFINE_FUNCTION INTEGER fnArrayPartition(INTEGER nList[],
  INTEGER left,
  INTEGER right,
  INTEGER pivot_index)
{
/*|------------------------------------------------------------------------------|
  | fnArrayPartition: In-place partitioning of an integer array.  It partitions  |
  |       the portion of the array between indices left and right inclusive by   |
  |       moving all elements less than list[pivot_index] before the pivot and   |
  |       those equal or greater after it.                                       |
  |    Parameters:                                                               |
  |        array of integers left                                                |
  |        index of first element of the sub-array to partition                  |
  |        right index of second element of the sub-array to partition           |
  |        pivot_index of the pivot element                                      |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER pivot_value
  STACK_VAR INTEGER store_index
  STACK_VAR INTEGER n

  store_index                            = 0

  IF(left <= pivot_index && pivot_index <= right && left && right <= LENGTH_STRING(nList))
  {
    pivot_value                          = nList[pivot_index]
    fnArraySwap(nList, pivot_index, right)
    store_index                          = left

    FOR(n = left; n < right; n++)
    {
      IF(nList[n] <= pivot_value)
      {
        fnArraySwap(nList, n, store_index)
        store_index++
      }
    }

    fnArraySwap(nList, store_index, right)
  }

  RETURN store_index
}

DEFINE_FUNCTION fnArrayQuickSort(INTEGER nList[])
{
/*|------------------------------------------------------------------------------|
  | fnArrayQuickSort: In-place sorting of an integer array using a non-recursive |
  |       quicksort method                                                       |
  |    Parameters:                                                               |
  |       nList array of integers                                                |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER stack_count  // current stack size
  STACK_VAR INTEGER stack[16][2] // stack size with log2(N) slots required: N = max allowable number
  STACK_VAR INTEGER pivot        // current pivot value
  STACK_VAR INTEGER lo, hi       // current lower and upper indices

  stack_count                            = 1
  stack[stack_count][1]                  = 1
  stack[stack_count][2]                  = LENGTH_STRING(nList)

  WHILE(stack_count)
  {
    lo                                  = stack[stack_count][1]
    hi                                    = stack[stack_count][2]

    IF(lo < hi)
    {
      pivot                             = nList[lo]

      WHILE(lo < hi)
      {
        WHILE(nList[hi] >= pivot && lo < hi)
        {
          hi--
        }

        IF(lo < hi)
        {
          nList[lo] = nList[hi]
          lo++
        }

        WHILE(nList[lo] <= pivot && lo < hi)
        {
          lo++
        }

        IF(lo < hi)
        {
          nList[hi]                           = nList[lo]
          hi--
        }
      }
      nList[lo]                               = pivot
      stack_count++
      stack[stack_count][1]                    = lo + 1
      stack[stack_count][2]                    = stack[stack_count - 1][2]
      stack[stack_count - 1][2]               = lo
    }
    ELSE
    {
      stack_count--
    }
  }
}

DEFINE_FUNCTION INTEGER fnArrayStringIndex(CHAR cItem[],
  CHAR cList[][])
{
/*|------------------------------------------------------------------------------|
  | fnArrayStringIndex: Finds the index for any matching string entry in an      |
  |       array.                                                                 |
  |    Parameters:                                                               |
  |       cItem item to find in the array                                        |
  |       cList array of items                                                   |
  |    Returns:                                                                  |
  |       The index of the matching value, or 0 if no match                      |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER idx

  IF(LENGTH_STRING(cItem) > 0)
  {
    FOR(idx = 1; idx <= LENGTH_ARRAY(cList); idx++)
    {
      IF(fnStringIsEqualsIgnoreCase(cItem,cList[idx]))
      {
        RETURN idx
      }
    }
  }

  RETURN 0
}

DEFINE_FUNCTION INTEGER fnStringLastIndexOf(CHAR cSearchFor[],
  CHAR cString[])
{
/*|------------------------------------------------------------------------------|
  | fnStringLastIndexOf: Finds the last index for string entry in an string      |
  |    Parameters:                                                               |
  |       cItem item to find in the string                                       |
  |       cString search string                                                  |
  |    Returns:                                                                  |
  |       The index of the matching value, or 0 if no match                      |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER idx

  idx = FIND_STRING(fnStringReverse(cString),fnStringReverse(cSearchFor),1)

  IF (idx == 0)
  {
    RETURN 0
  }

  RETURN LENGTH_STRING(cString) - idx + 1
}

DEFINE_FUNCTION INTEGER fnStringIndexOf(CHAR cSearchFor[],
  CHAR cString[])
{
/*|------------------------------------------------------------------------------|
  | fnStringIndexOf: Finds the index for string entry in an string               |
  |    Parameters:                                                               |
  |       cItem item to find in the string                                       |
  |       cString search string                                                  |
  |    Returns:                                                                  |
  |       The index of the matching value, or 0 if no match                      |
  |------------------------------------------------------------------------------|*/

  RETURN FIND_STRING(cString,cSearchFor,1)
}

DEFINE_FUNCTION INTEGER fnStringIndexOfIgnoreCase(CHAR cSearchFor[],
  CHAR cString[])
{
/*|------------------------------------------------------------------------------|
  | fnStringIndexOfIgnoreCase: Finds the index for string entry in an string     |
  |    Parameters:                                                               |
  |       cItem item to find in the string                                       |
  |       cString search string                                                  |
  |    Returns:                                                                  |
  |       The index of the matching value, or 0 if no match                      |
  |------------------------------------------------------------------------------|*/

  RETURN FIND_STRING(UPPER_STRING(cString),UPPER_STRING(cSearchFor),1)
}

DEFINE_FUNCTION fnArraySwap(INTEGER nList[],
  INTEGER index1,
  INTEGER index2)
{
/*|------------------------------------------------------------------------------|
  | fnArraySwap: In-place swap of two elements in an integer array               |
  |    Parameters:                                                               |
  |       nList  array of integers                                               |
  |       index1 index of the first element to swap                              |
  |       index2 index of the second element to swap                             |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER length
  STACK_VAR INTEGER temp_store

  length                                = LENGTH_STRING(nList)

  IF(index1 && index1 <= length && index2 && index2 <= length)
  {
    temp_store                           = nList[index1];
    nList[index1]                        = nList[index2]
    nList[index2]                        = temp_store
  }
}

DEFINE_FUNCTION fnDebugDebugMessage(CHAR msg[])
{
/*|------------------------------------------------------------------------------|
  | fnDebugDebugMessage: Prints a debug message if debug level DEBUG or greater  |
  |    Parameters:                                                               |
  |       msg string containing the debug message to be printed                  |
  |------------------------------------------------------------------------------|*/

  IF(fnDebugIsLevelDebug())
  {
    _fnDebugMessage(DEBUG_DEBUG,msg)
  }
}

DEFINE_FUNCTION fnDebugDecimal(CHAR msg_level,
  CHAR msg[])
{
/*|------------------------------------------------------------------------------|
  | fnDebugDecimal: Prints a debug message forced to decimal                     |
  |    Parameters:                                                               |
  |       msg_level char specifying the debug level of the message               |
  |       msg       string containing the debug message to be printed as decimal |
  |------------------------------------------------------------------------------|*/

  STACK_VAR LONG i,l
  STACK_VAR CHAR c
  STACK_VAR CHAR in[255]

  IF(msg_level < DEBUG_ERROR || msg_level > DEBUG_MAX_LEVEL)
  {
    fnDebugErrorMessage("'invalid debug level specified - ', msg")

    RETURN
  }

  IF(msg_level <= nDebugLevel)
  {
    in                                  =  msg
    l                                   = LENGTH_STRING(in)

    fnIOPrintln("'message length: ',ITOA(l)")

    FOR(i = 1; i <= l; i++)
    {
      c                                 = GET_BUFFER_CHAR(in)
      fnIOPrintln("'[',ITOA(i),'] ',ITOA(c)")
    }
  }
}

DEFINE_FUNCTION fnDebugErrorMessage(CHAR msg[])
{
/*|------------------------------------------------------------------------------|
  | fnDebugErrorMessage: Prints a debug message if debug level ERROR or greater  |
  |    Parameters:                                                               |
  |       msg string containing the debug message to be printed                  |
  |------------------------------------------------------------------------------|*/

  IF(fnDebugIsLevelError())
  {
    _fnDebugMessage(DEBUG_ERROR,msg)
  }
}

DEFINE_FUNCTION CHAR fnDebugGetLevelFromString(CHAR x[])
{
/*|------------------------------------------------------------------------------|
  | fnDebugGetLevelFromString: Gets a numerical debug level based on it's        |
  |       equivalent string representation.                                      |
  |    Parameters:                                                               |
  |       x character array containing the string to parse                       |
  |    Returns:                                                                  |
  |       character containing the numerical debug level represented by the      |
  |       content of x                                                           |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR lvl

  IF(LENGTH_STRING(x) == 1)
  {
    lvl                                 = ATOI(x)

    IF(lvl > 4)
    {
      lvl                               = DEBUG_OFF
    }
  }

  FOR(lvl = LENGTH_STRING(DEBUG_LEVEL_STRINGS); lvl; lvl--)
  {
    IF(LOWER_STRING(x) == LOWER_STRING(DEBUG_LEVEL_STRINGS[lvl]))
    {
      lvl--

      BREAK
    }
  }

  RETURN lvl
}

DEFINE_FUNCTION CHAR[16] fnDebugGetLevelString()
{
/*|------------------------------------------------------------------------------|
  | fnDebugGetLevelString: Returns the current debug level as String.            |
  |    Returns:                                                                  |
  |       String representing the level                                          |
  |------------------------------------------------------------------------------|*/

  RETURN DEBUG_LEVEL_STRINGS[nDebugLevel + 1]
}

DEFINE_FUNCTION fnDebugHex(CHAR msg_level,
  CHAR msg[])
{
/*|------------------------------------------------------------------------------|
  | fnDebugHex: Prints a debug message forced to hex - this avoids situations    |
  |       where a hex value is a valid ASCII character                           |
  |    Parameters:                                                               |
  |       msg_level char specifying the debug level of the message               |
  |       msg       string containing the debug message to be printed as hex     |
  |------------------------------------------------------------------------------|*/

  IF(msg_level < DEBUG_ERROR || msg_level > DEBUG_MAX_LEVEL)
  {
    _fnDebugMessage(DEBUG_ERROR, "'invalid debug level specified - ', msg")

    RETURN
  }

  IF(msg_level <= nDebugLevel)
  {
    fnIOPrintln("PRT_PREFIX,UPPER_STRING(fnDebugGetLevelString()),': ',fnStringToASCIIHex(msg)")
  }
}

DEFINE_FUNCTION fnPrintNetLinxMessage(CHAR msg[])
{
	fnIOPrintln("IO_TAB,TIME,' ',DATE,' ',msg")
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT]  fnStringToASCIIHex(CHAR msg[])
{
  STACK_VAR LONG i,l
  STACK_VAR CHAR c
  STACK_VAR CHAR out[1000]
  STACK_VAR CHAR in[1000]

  in                                    = msg
  out                                   = NULL_STR
  l                                     = LENGTH_STRING(in)

  FOR(i = 0; i < l; i++)
  {
    c                                   = GET_BUFFER_CHAR(in)
    out                                 = "out, '$', ITOHEX(c),','"
  }

  RETURN out
}

DEFINE_FUNCTION fnDebugInfoMessage(CHAR msg[])
{
/*|------------------------------------------------------------------------------|
  | fnDebugInfoMessage: Prints a debug message if debug level INFO or greater    |
  |    Parameters:                                                               |
  |       msg string containing the debug message to be printed                  |
  |------------------------------------------------------------------------------|*/

  IF(fnDebugIsLevelInfo())
  {
    _fnDebugMessage(DEBUG_INFO,msg)
  }
}


DEFINE_FUNCTION fnDebugQueryDebugState(DEV vdvDev)
{
/*|------------------------------------------------------------------------------|
  | fnDebugQueryDebugState: Query the given device for it's debug level          |
  |    Parameters:                                                               |
  |       vdvDev Virutal device to send query to                                 |
  |------------------------------------------------------------------------------|*/

  fnSendCommand(vdvDev,'?DEBUGSTATE')
}

DEFINE_FUNCTION INTEGER fnDebugIsLevelDebug()
{
/*|------------------------------------------------------------------------------|
  | fnDebugIsLevelDebug: Query if debug level it set to DEBUG                    |
  |    Returns:                                                                  |
  |       TRUE or FALSE                                                          |
  |------------------------------------------------------------------------------|*/

  RETURN nDebugLevel >= DEBUG_DEBUG
}

DEFINE_FUNCTION INTEGER fnDebugIsLevelError()
{
/*|------------------------------------------------------------------------------|
  | fnDebugIsLevelError: Query if debug level it set to ERROR or greater         |
  |    Returns:                                                                  |
  |       TRUE or FALSE                                                          |
  |------------------------------------------------------------------------------|*/

  RETURN nDebugLevel >= DEBUG_ERROR
}

DEFINE_FUNCTION INTEGER fnDebugIsLevelInfo()
{
/*|------------------------------------------------------------------------------|
  | fnDebugIsLevelInfo: Query if debug level it set to INFO or greater           |
  |    Returns:                                                                  |
  |       TRUE or FALSE                                                          |
  |------------------------------------------------------------------------------|*/

  RETURN nDebugLevel >= DEBUG_INFO
}

DEFINE_FUNCTION INTEGER fnDebugIsLevelWarning()
{
/*|------------------------------------------------------------------------------|
  | fnDebugIsLevelWarning: Query if debug level it set to WARNING or greater     |
  |    Returns:                                                                  |
  |       TRUE or FALSE                                                          |
  |------------------------------------------------------------------------------|*/

  RETURN nDebugLevel >= DEBUG_WARN
}

DEFINE_FUNCTION CHAR[16] fnDebugLevelString(INTEGER nDebugLevel)
{
/*|------------------------------------------------------------------------------|
  | fnDebugLevelString: Returns the requested debug level as String.             |
  |    Parameters:                                                               |
  |       nDebugLevel integer specifying the debug level of the message          |
  |    Returns:                                                                  |
  |       String representing the debug level                                    |
  |------------------------------------------------------------------------------|*/

  RETURN DEBUG_LEVEL_STRINGS[nDebugLevel+1]
}

DEFINE_FUNCTION fnDebugSetLevel(INTEGER x)
{
/*|------------------------------------------------------------------------------|
  | fnDebugSetLevel: Sets the current system debugging level for controlling     |
  |       debug message verbosity.                                               |
  |    Parameters:                                                               |
  |       x integer specifying the debug level to set                            |
  |------------------------------------------------------------------------------|*/

  IF(x >= DEBUG_OFF && x <= DEBUG_MAX_LEVEL)
  {
    nDebugLevel                          = x
  }
  ELSE
  {
    fnDebugWarningMessage( "'Invalid debug level, defaulting to ', fnDebugGetLevelString()")
    fnDebugSetLevel(DEBUG_ERROR)
  }
}

DEFINE_FUNCTION fnDebugSetLevelByString(CHAR x[])
{
/*|------------------------------------------------------------------------------|
  | fnDebugLevelString: Sets the current system debugging level for controlling  |
  |       debug message verbosity.                                               |
  |    Parameters:                                                               |
  |       x char specifying the debug level to set                               |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER idx

  FOR(idx = 1; idx <= LENGTH_STRING(DEBUG_LEVEL_STRINGS); idx++)
  {
    IF(LOWER_STRING(x) == LOWER_STRING(DEBUG_LEVEL_STRINGS[idx]))
    {
      fnDebugSetLevel(DEBUG_LEVEL_VALUES[idx])

      RETURN
    }
  }

  fnIOPrintln("'Invalid debug level, defaulting to ', fnDebugGetLevelString()")
  fnDebugSetLevel(DEBUG_ERROR)
}

DEFINE_FUNCTION fnDebugWarningMessage(CHAR msg[])
{
/*|------------------------------------------------------------------------------|
  | fnDebugWarningMessage: Prints a debug message if debug level WARNING or      |
  |       greater                                                                |
  |    Parameters:                                                               |
  |       msg string containing the debug message to be printed                  |
  |------------------------------------------------------------------------------|*/

  IF(fnDebugIsLevelWarning())
  {
    _fnDebugMessage(DEBUG_WARN,msg)
  }
}

DEFINE_FUNCTION CHAR[3] fnIntOnOff(INTEGER n)
{
/*|------------------------------------------------------------------------------|
  | fnIntOnOff: Convert integer value to ON/OFF value                            |
  |    Parameters:                                                               |
  |       n Integer to convert                                                   |
  |    Returns:                                                                  |
  |       ON or OFF                                                              |
  |------------------------------------------------------------------------------|*/

  IF(n)
  {
    RETURN 'ON'
  }

  RETURN 'OFF'
}

DEFINE_FUNCTION CHAR[5] fnIntTRUEFALSE(INTEGER n)
{
/*|------------------------------------------------------------------------------|
  | fnIntTrueFalse: Convert integer value to boolean value                       |
  |    Parameters:                                                               |
  |       n Integer to convert                                                   |
  |    Returns:                                                                  |
  |       True or False                                                          |
  |------------------------------------------------------------------------------|*/

  IF(n)
  {
    RETURN 'TRUE'
  }

  RETURN 'FALSE'
}

DEFINE_FUNCTION INTEGER fnIOFileExists(CHAR cFileName[])
{
/*|------------------------------------------------------------------------------|
  | fnIOFileExists: Determines if given file exist on disc                       |
  |    Parameters:                                                               |
  |       cFileName file name to validate                                        |
  |    Returns:                                                                  |
  |       TRUE or FALSE                                                          |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR Buffer[1024]
  STACK_VAR SINTEGER NumFiles
  STACK_VAR SINTEGER Entry

  NumFiles = 1
  Entry = 1

  WHILE (NumFiles > 0)
  {
    NumFiles = FILE_DIR ('\', Buffer, Entry)
    Entry = Entry + 1

    IF (fnStringIsEqualsIgnoreCase(Buffer,cFileName))
    {
      RETURN TRUE
    }
  }

  RETURN FALSE
}

DEFINE_FUNCTION INTEGER fnIOFileNotExists(CHAR cFileName[])
{
/*|------------------------------------------------------------------------------|
  | fnIOFileNotExists: Determines if given file does exist on disc               |
  |    Parameters:                                                               |
  |       cFileName file name to validate                                        |
  |    Returns:                                                                  |
  |       TRUE or FALSE                                                          |
  |------------------------------------------------------------------------------|*/

  RETURN !fnIOFileExists(cFileName)
}

DEFINE_FUNCTION fnIOPrint(CHAR x[])
{
/*|------------------------------------------------------------------------------|
  | fnIOPrint: Print a character array to the output stream.                     |
  |    Parameters:                                                               |
  |       x string containing the data to output                                 |
  |------------------------------------------------------------------------------|*/

  fnIOWrite(x, 1, LENGTH_STRING(x))
}

DEFINE_FUNCTION fnIOPrintBanner(CHAR x[], INTEGER w, CHAR y)
{
/*|------------------------------------------------------------------------------|
  | fnIOPrintBanner: Prints banner to the output stream.                         |
  |    Parameters:                                                               |
  |       x banner message                                                       |
  |       w width of banner                                                      |
  |       y banner character ('-','*'...)                                        |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER idx
  STACK_VAR INTEGER width
  STACK_VAR CHAR z[w]

  width                                 =  (w-LENGTH_STRING(x)) / 2

  fnIOPrintln( "x,' width: ',ITOA(width),' w: ',ITOA(w)")

  fnIOPrintln(fnStringPrefixToLength('',y,w))
  fnIOPrintln(fnStringPrefixToLength('',y,w))
  fnIOPrintln(fnStringPrefixToLength('',y,w))
  fnIOPrintln(' ')

  FOR(idx = 1; idx <= width; idx++)
  {
    z                                   =  "z, ' '"
  }
  fnIOPrintln("z,x")

  fnIOPrintln( "x,' width: ',ITOA(width),' w: ',ITOA(w),' z: [',z,'] width z: ',ITOA(LENGTH_STRING(z))")

  fnIOPrintln(' ')
  fnIOPrintln(fnStringPrefixToLength('',y,w))
  fnIOPrintln(fnStringPrefixToLength('',y,w))
  fnIOPrintln(fnStringPrefixToLength('',y,w))
}

DEFINE_FUNCTION fnIOPrintBannerLine(CHAR message[],
  INTEGER bannerWidth,
  CHAR leftFiller,
  CHAR rightFiller,
  CHAR blankBefore,
  CHAR blankAfter)
{
/*|------------------------------------------------------------------------------|
  | fnIOPrintBannerLine: Prints banner to the output stream.                     |
  |    Parameters:                                                               |
  |       message     banner message                                             |
  |       bannerWidth width of banner                                            |
  |       leftFiller  banner character ('-','*'...)                              |
  |       rightFiller banner character ('-','*'...)                              |
  |       blankBefore blanks before?                                             |
  |       blankAfter  blanks after?                                              |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER width

  width                                 =  (bannerWidth-(LENGTH_STRING(message)+2)) / 2

  IF(blankBefore)
  {
    fnIOPrintln(' ')
  }

  fnIOPrintln("fnStringPrefixToLength('',
  leftFiller,width),
  ' ',
  message,
  ' ',
  fnStringPrefixToLength('',rightFiller,width)")

  IF(blankAfter)
  {
    fnIOPrintln(' ')
  }
}

DEFINE_FUNCTION fnIOPrintBetaBanner()
{
/*|------------------------------------------------------------------------------|
  |                             fnIOPrintBetaBanner                              |
  |------------------------------------------------------------------------------|*/

  fnIOPrintln("'x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x'")
  fnIOPrintln("' '")
  fnIOPrintln("'                        ######                     '")
  fnIOPrintln("'                        #     # ###### #####   ##  '")
  fnIOPrintln("'                        #     # #        #    #  # '")
  fnIOPrintln("'                        ######  #####    #   #    #'")
  fnIOPrintln("'                        #     # #        #   ######'")
  fnIOPrintln("'                        #     # #        #   #    #'")
  fnIOPrintln("'                        ######  ######   #   #    #'")
  fnIOPrintln("' '")
  fnIOPrintln("'    ######'")
  fnIOPrintln("'    #     #  ####     #    #  ####  #####    #####  #    # #    #    # #    #'")
  fnIOPrintln("'    #     # #    #    ##   # #    #   #      #    # #    # ##   #    # ##   #'")
  fnIOPrintln("'    #     # #    #    # #  # #    #   #      #    # #    # # #  #    # # #  #'")
  fnIOPrintln("'    #     # #    #    #  # # #    #   #      #####  #    # #  # #    # #  # #'")
  fnIOPrintln("'    #     # #    #    #   ## #    #   #      #   #  #    # #   ##    # #   ##'")
  fnIOPrintln("'    ######   ####     #    #  ####    #      #    #  ####  #    #    # #    #'")
  fnIOPrintln("'     '")
  fnIOPrintln("'        #####  #####   ####  #####  #    #  ####  ##### #  ####  #    #'")
  fnIOPrintln("'        #    # #    # #    # #    # #    # #    #   #   # #    # ##   #'")
  fnIOPrintln("'        #    # #    # #    # #    # #    # #        #   # #    # # #  #'")
  fnIOPrintln("'        #####  #####  #    # #    # #    # #        #   # #    # #  # #'")
  fnIOPrintln("'        #      #   #  #    # #    # #    # #    #   #   # #    # #   ##'")
  fnIOPrintln("'        #      #    #  ####  #####   ####   ####    #   #  ####  #    #'")
  fnIOPrintln("' '")
  fnIOPrintln("'x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x-x'")
}

DEFINE_FUNCTION fnIOPrintln(CHAR msg[])
{
/*|------------------------------------------------------------------------------|
  | fnIOPrintln: Print a line to the output stream.                              |
  |    Parameters:                                                               |
  |       sMessage string containing the data to output                          |
  |------------------------------------------------------------------------------|*/

  STACK_VAR LONG i,l
  STACK_VAR INTEGER idx
  STACK_VAR CHAR sMessage[STRING_RETURN_SIZE_LIMIT]
  STACK_VAR CHAR sMessages[100][PRINTLN_MAX]
  STACK_VAR CHAR sTmp[PRINTLN_MAX]

  // if contains \n then it's user formatted, just write as is
  IF (fnStringContains(msg,"IO_LINE_SEPERATOR"))
  {
    SEND_STRING IO_OUT,msg
    RETURN
  }

  sMessage = fnStringReplaceTabWithSpaces(msg)

  IF (LENGTH_STRING(sMessage) <= PRINTLN_MAX)
  {
    sMessages[1]                        = sMessage
  }
  ELSE
  {
    i = 0

    WHILE(LENGTH_STRING(sMessage) > PRINTLN_MAX)
    {
      i++
      sTmp                               = LEFT_STRING(sMessage,PRINTLN_MAX)
      idx                               = LENGTH_STRING(sTmp) - fnStringLastIndexOf(' ',sTmp)

      IF (idx > 0 && idx <= PRINTLN_MAX_BREAK)
      {
        idx = fnStringLastIndexOf(' ',sTmp)
        sMessages[i]                     = LEFT_STRING(sMessage,idx)
        sMessage                         = MID_STRING(sMessage,idx+1,LENGTH_STRING(sMessage)-idx)
      }
      ELSE
      {
        idx                               = LENGTH_STRING(sTmp) - fnStringLastIndexOf(',',sTmp)

        IF (idx > 0 && idx <= PRINTLN_MAX_BREAK)
        {
          idx = fnStringLastIndexOf(',',sTmp)
          sMessages[i]                     = LEFT_STRING(sMessage,idx)
          sMessage                         = MID_STRING(sMessage,idx+1,LENGTH_STRING(sMessage)-idx)
        }
        ELSE
        {
          sMessages[i]                     = LEFT_STRING(sMessage,PRINTLN_MAX)
          sMessage                         = MID_STRING(sMessage,PRINTLN_MAX+1,LENGTH_STRING(sMessage)-PRINTLN_MAX)
        }
      }
    }

    IF (LENGTH_STRING(sMessage) > 0)
    {
      i++
      sMessages[i]                      = sMessage
    }
  }

  SEND_STRING IO_OUT,sMessages[1]

  FOR(l = 2; l <= i; l++)
  {
    SEND_STRING IO_OUT,"IO_TAB, sMessages[l]"
  }
}

DEFINE_FUNCTION fnIOWrite(CHAR buf[],
  INTEGER offset,
  INTEGER len)
{
/*|------------------------------------------------------------------------------|
  | fnIOWrite: Write a character array to the output device.                     |
  |     Parameters:                                                              |
  |          bufthe character array to write                                     |
  |          offset offset to begin writing from (offset 0 == character 1)       |
  |          len number of characters to write                                   |
  |------------------------------------------------------------------------------|*/

  IF(offset <= 0)
  {
    offset                              =  1
  }

  SEND_STRING IO_OUT, MID_STRING(buf,offset, len)

}

DEFINE_FUNCTION FLOAT fnMathBase10Logarithm(FLOAT x)
{
/*|------------------------------------------------------------------------------|
  | fnMathBase10Logarithm: Calculate the base 10 logarithm of the passed number. |
  |    Parameters:                                                               |
  |       x float to find the natural log of                                     |
  |    Returns:                                                                  |
  |       Float containing the passed numbers log base 10                        |
  |------------------------------------------------------------------------------|*/

  RETURN fnMathLogarithm(x, 10)
}

DEFINE_FUNCTION FLOAT fnMathBinaryLogarithm(FLOAT x)
{
/*|------------------------------------------------------------------------------|
  | fnMathBinaryLogarithm: Calculate the binary logarithm of the passed number.  |
  |    Parameters:                                                               |
  |       x float to find the natural log of                                     |
  |    Returns:                                                                  |
  |       Float containing the passed numbers log base 2                         |
  |------------------------------------------------------------------------------|*/

  RETURN fnMathLogarithm(x, 2)
}

DEFINE_FUNCTION DOUBLE fnMathBuildDouble(LONG hi,
  LONG low)
{
/*|------------------------------------------------------------------------------|
  | fnMathBuildDouble: Build a double using the binary info stored across two    |
  |       longs. It is assumed that the data is stored as per the IEEE754        |
  |       standard.                                                              |
  |    Parameters:                                                               |
  |       hi  Long containing bits 63 - 32                                       |
  |       low Long containing bits 31 - 0                                        |
  |    Returns:                                                                  |
  |       Double built from the passed data                                      |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR serialized[10]// For some reason the buffer
  STACK_VAR DOUBLE ret// passed to string_to_variable()

  serialized                             = "$E4, raw_be(hi), raw_be(low)"
  STRING_TO_VARIABLE(ret, serialized, 1)

  RETURN ret
}

DEFINE_FUNCTION FLOAT fnMathBuildFloat(LONG x)
{
/*|------------------------------------------------------------------------------|
  | fnMathBuildFloat: Build a float using a IEEE754 bit pattern stored in a      |
  |       long.                                                                  |
  |    Parameters:                                                               |
  |       x Long containing the raw data                                         |
  |    Returns:                                                                  |
  |       Float built from the passed data                                       |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR serialized[6]
  STACK_VAR FLOAT ret

  serialized                             = "$E3, raw_be(x)"
  STRING_TO_VARIABLE(ret, serialized, 1)

  RETURN ret
}

DEFINE_FUNCTION DOUBLE fnMathCeiling(DOUBLE x)
{
/*|------------------------------------------------------------------------------|
  | fnMathCeiling: Returns the smallest (closest to negative infinity) long      |
  |       value that is not less than the argument and is equal to a             |
  |       mathematical integer.                                                  |
  |    Parameters:                                                                |
  |       x Double to fnMathRound                                                |
  |    Returns:                                                                  |
  |       Double containing the fnMathRounded number                             |
  |------------------------------------------------------------------------------|*/

  RETURN -fnMathFloor(-x)
}

DEFINE_FUNCTION LONG fnMathDecimalToBinaryCodedDecimal(INTEGER x)
{
/*|------------------------------------------------------------------------------|
  | fnMathDecimalToBinaryCodedDecimal: Encode a decimal value as binary-coded    |
  |       decimal (BCD).                                                         |
  |    Parameters:                                                               |
  |       x Integer to encode as BCD                                             |
  |       |                                                                      |
  |    Returns:                                                                  |
  |       Long containing the BCD representation of x                            |
  |------------------------------------------------------------------------------|*/

  STACK_VAR LONG tmp
  STACK_VAR CHAR i
  STACK_VAR CHAR j

  FOR(i = 16; i; i--)
  {
    FOR(j = 0; j < 5; j++)
    {
      IF((tmp >> (4 * j) & $F) > 4)
      {
        tmp                             = tmp + (3 << (4 * j))
      }
    }

    tmp                                 = tmp << 1 | (x >> (i - 1) & 1)
  }

  RETURN tmp
}

DEFINE_FUNCTION LONG fnMathDoubleHighToBits(DOUBLE x)
{
/*|------------------------------------------------------------------------------|
  | fnMathDoubleHighToBits: Load the raw data stored in bits 63 - 32 of a double |
  |       into a long.                                                           |
  |    Parameters:                                                               |
  |       x Double to load                                                       |
  |    Returns:                                                                  |
  |       Long filled binary data stored in the high DWord of the double         |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR raw[8]

  raw                                   = raw_be(x)

  RETURN fnMathRawBEtoLong("raw[1], raw[2], raw[3], raw[4]")
}

DEFINE_FUNCTION LONG fnMathDoubleLowToBits(DOUBLE x)
{
/*|------------------------------------------------------------------------------|
  | fnMathDoubleLowToBits: Load the raw data stored in bits 31 - 0 of a double   |
  |       into a long.                                                           |
  |    Parameters:                                                               |
  |       x Double to load                                                       |
  |    Returns:                                                                  |
  |       Long filled binary data stored in the low DWord of the double          |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR raw[8]

  raw                                   = raw_be(x)

  RETURN fnMathRawBEtoLong("raw[5], raw[6], raw[7], raw[8]")
}

DEFINE_FUNCTION DOUBLE fnMathFastSqrt(DOUBLE x)
{
/*|------------------------------------------------------------------------------|
  | fnMathFastSqrt: Approximate the square root of the passed number based on    |
  |       the inverse square root algorithm in fnMathInverseSqrt(x). This is     |
  |       MUCH faster than fnMathSquareRoot(x) and recommended over              |
  |       fnMathSquareRoot() FOR use anywhere a precise square root is not       |
  |       required. Error is approx +/-0.17%.                                    |
  |    Parameters:                                                               |
  |       x Double to find the square root of                                    |
  |    Returns:                                                                  |
  |       Double containing an approximation of the square root                  |
  |------------------------------------------------------------------------------|*/

  RETURN x * fnMathInverseSqrt(x)
}

DEFINE_FUNCTION LONG fnMathFloatToBits(FLOAT x)
{
/*|------------------------------------------------------------------------------|
  | fnMathFloatToBits: Load a float value's IEEE 754 bit pattern into a long.    |
  |    Parameters:                                                               |
  |       x Float to load                                                        |
  |    Returns:                                                                  |
  |       Long filled with the IEEE 754 bit pattern of the float                 |
  |------------------------------------------------------------------------------|*/

  RETURN fnMathRawBEtoLong(raw_be(x))
}

DEFINE_FUNCTION DOUBLE fnMathFloor(DOUBLE x)
{
/*|------------------------------------------------------------------------------|
  | fnMathFloor: Returns the largest (closest to positive infinity) long value   |
  |       that is not greater than the argument and is equal to a mathematical   |
  |       integer.                                                               |
  |    Parameters:                                                               |
  |       x Double to fnMathRound                                                |
  |    Returns:                                                                  |
  |       Double containing the fnMathRounded number                             |
  |------------------------------------------------------------------------------|*/

  STACK_VAR DOUBLE tmp
  STACK_VAR SLONG ret

  tmp                                   = ABS_VALUE(x)

  IF(fnMathIsInt(tmp))
  {
    RETURN x
  }

  IF(tmp < 1)
  {
    IF(x >= 0)
    {
      RETURN 0.0 * x
    }
    ELSE
    {
      RETURN -1.0
    }
  }

  IF(x < 0)
  {
    ret                                 = TYPE_CAST(x - 1.0)
  }
  ELSE
  {
    ret                                 = TYPE_CAST(x)
  }

  RETURN ret
}

DEFINE_FUNCTION DOUBLE fnMathIEEERemainder(DOUBLE x,
  DOUBLE y)
{
/*|------------------------------------------------------------------------------|
  | fnMathIEEERemainder: Computes the remainder operation on two arguments as    |
  |       prescribed by the IEEE 754 standard.                                   |
  |    Parameters:                                                               |
  |       x dividend                                                             |
  |       y divisor                                                              |
  |    Returns:                                                                  |
  |       Double equal to x - (y Q), where Q is the quotient of x / y            |
  |       fnMathRounded to the nearest integer (if y = 0, NaN is returned        |
  |------------------------------------------------------------------------------|*/

  IF(y == 0)
  {
    RETURN MATH_NaN
  }

  RETURN x - y * fnMathRound(x / y)
}

DEFINE_FUNCTION DOUBLE fnMathInverseSqrt(DOUBLE x)
{
/*|------------------------------------------------------------------------------|
  | fnMathInverseSqrt: Approximate the inverse square root of the passed number. |
  |       This method uses a integer shift and single Newton refinement aka      |
  |       Quake 3 method. Original algorithm by Greg Walsh.                      |
  |    Parameters:                                                               |
  |       x Double to find the inverse square root of                            |
  |    Returns:                                                                  |
  |       Double containing an approximation of the inverse square root          |
  |------------------------------------------------------------------------------|*/

  STACK_VAR LONG hi
  STACK_VAR LONG low
  STACK_VAR LONG t_hi
  STACK_VAR LONG t_low
  STACK_VAR DOUBLE res
  STACK_VAR DOUBLE tmp

  tmp                                   = fnMathRShiftDouble(x)
  t_hi                                  = fnMathDoubleHighToBits(tmp)
  t_low                                 = fnMathDoubleLowToBits(tmp)
  hi                                    = $5FE6EC85 - t_hi

  IF(t_low > $E7DE30DA)
  {
    hi                                  = hi - (t_low - $E7DE30DA)
    low                                 = 0
  }
  ELSE
  {
    low                                 = $E7DE30DA - t_low
  }

  res                                   = fnMathBuildDouble(hi, low)

  RETURN res * (1.5 - 0.5 * x * res * res)
}

DEFINE_FUNCTION CHAR fnMathIsInfinite(DOUBLE x)
{
/*|------------------------------------------------------------------------------|
  | fnMathIsInfinite: Checks if a value is either positive infinity or negative  |
  |       infinity.                                                              |
  |    Parameters:                                                               |
  |       x Double to check                                                      |
  |    Returns:                                                                  |
  |       Boolean, TRUE is x is infinite                                         |
  |------------------------------------------------------------------------------|*/

  STACK_VAR LONG hi

  hi                                    = fnMathDoubleHighToBits(x)

  RETURN (hi & $7FF00000) >> 20 == $7FF && !(hi & $FFFFF || fnMathDoubleLowToBits(x))
}

DEFINE_FUNCTION CHAR fnMathIsInt(DOUBLE x)
{
/*|------------------------------------------------------------------------------|
  | fnMathIsInt: Returns TRUE if the argument has no decimal component,          |
  |       otherwise Returns FALSE. +/-Inf and 0 will return TRUE, subnormal and  |
  |       NaN's will return FALSE.                                               |
  |    Parameters:                                                               |
  |       x Double to check                                                      |
  |    Returns:                                                                  |
  |       Boolean, TRUE if x is a mathematical integer                           |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR i
  STACK_VAR SINTEGER exp
  STACK_VAR LONG hi
  STACK_VAR LONG m
  STACK_VAR LONG mask

  IF(fnMathIsNaN(x))
  {
    RETURN FALSE
  }

  IF(x >= MATH_TWO_52)
  {
    RETURN TRUE
  }

  IF(ABS_VALUE(x) < 1.0)
  {
    RETURN (ABS_VALUE(x) == 0)
  }

  hi                                    = fnMathDoubleHighToBits(x)

  exp = type_cast((hi & $7FF00000) >> 20 - 1023)

  IF(exp > 20)
  {
    m                                   = fnMathDoubleLowToBits(x)
  }
  ELSE
  {
    m                                   = hi & $FFFFF
  }

  FOR(i = type_cast(32 + (exp > 20) * 20 - exp); i; i--)
  {
    mask                                 = mask + 1 << (i - 1)
  }

  RETURN (m & mask == 0)
}

DEFINE_FUNCTION CHAR fnMathIsNaN(DOUBLE x)
{
/*|------------------------------------------------------------------------------|
  | fnMathIsNaN: Checks if a value is NaN.                                       |
  |    Parameters:                                                               |
  |       x Double to check                                                      |
  |    Returns:                                                                  |
  |       Boolean, TRUE is x is NaN                                              |
  |------------------------------------------------------------------------------|*/

  STACK_VAR LONG hi

  hi                                    = fnMathDoubleHighToBits(x)

  RETURN (hi & $7FF00000) >> 20 == $7FF && (hi & $FFFFF || fnMathDoubleLowToBits(x))
}

DEFINE_FUNCTION CHAR fnMathIsNear(DOUBLE x,
  DOUBLE y)
{
/*|------------------------------------------------------------------------------|
  | fnMathIsNear: Compares two numbers and RETURN TRUE if they are within        |
  |       MATH_PRECISION of each other.                                          |
  |    Parameters:                                                               |
  |       x Number to compare                                                    |
  |       y Another number to compare to x                                       |
  |    Returns:                                                                  |
  |       Boolean specifying if x and y are within MATH_PRECISION of each other  |
  |------------------------------------------------------------------------------|*/

  RETURN ABS_VALUE(x - y) <= MATH_PRECISION
}

DEFINE_FUNCTION CHAR fnMathIsNumeric(CHAR a[])
{
/*|------------------------------------------------------------------------------|
  | fnMathIsNumeric: Checks if a string value contains only a numeric value.     |
  |    Parameters:                                                               |
  |       a String to check                                                      |
  |    Returns:                                                                  |
  |       Boolean, TRUE is a is numeric                                          |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER i

  FOR(i = 1; i <= LENGTH_STRING(a); i++)
  {
    IF(!(a[i] >= $30 && a[i] <= $39))
    {
      RETURN 0
    }
  }

  RETURN 1
}

DEFINE_FUNCTION FLOAT fnMathLogarithm(FLOAT x,
  FLOAT base)
{
/*|------------------------------------------------------------------------------|
  | fnMathLogarithm: Calculate the logarithm of the passed number in the         |
  |       specified base.                                                        |
  |    Parameters:                                                               |
  |       x    Float to find the log of                                          |
  |       base Base to use                                                       |
  |    Returns:                                                                  |
  |       Float containing the passed numbers logarithm                          |
  |------------------------------------------------------------------------------|*/

  STACK_VAR FLOAT tmp
  STACK_VAR INTEGER int
  STACK_VAR FLOAT partial
  STACK_VAR FLOAT decimal

  IF(x < 1 && base < 1)
  {
    RETURN -1.0// cannot compute
  }
  tmp                                   = x + 0.0

  WHILE(tmp < 1)
  {
    int                                 = int - 1
    tmp                                 = tmp * base
  }

  WHILE(tmp >= base)
  {
    int                                 = int + 1
    tmp                                 = tmp / base
  }

  partial                               = 0.5
  tmp                                   = tmp * tmp

  WHILE(!fnMathIsNear(partial, 0))
  {
    IF(tmp >= base)
    {
      decimal                           = decimal + partial
      tmp                               = tmp / base
    }

    partial                             = partial * 0.5
    tmp                                 = tmp * tmp
  }

  RETURN int + decimal
}

DEFINE_FUNCTION DOUBLE fnMathLShiftDouble(DOUBLE x)
{
/*|------------------------------------------------------------------------------|
  | fnMathLShiftDouble: Left shift (<<) a double 1 bit.                          |
  |    Parameters:                                                               |
  |       x Double to shift                                                      |
  |    Returns:                                                                  |
  |       Passed value << 1                                                      |
  |------------------------------------------------------------------------------|*/

  STACK_VAR LONG hi
  STACK_VAR LONG low

  hi                                    = fnMathDoubleHighToBits(x)
  low                                   = fnMathDoubleLowToBits(x)

  hi                                     = (hi & $7FFFFFFF) << 1 + (low & $80000000) >> 15

  low                                   = (low & $7FFFFFFF) << 1

  RETURN fnMathBuildDouble(hi, low)
}

DEFINE_FUNCTION FLOAT fnMathMapRange(FLOAT x,
  FLOAT fromMin,
  FLOAT fromMax,
  FLOAT toMin,
  FLOAT toMax)
{
/*|------------------------------------------------------------------------------|
  | fnMathMapRange: Map a value within a set of original bounds to its           |
  |       equivalent given a secondary set of bound.                             |
  |    Parameters:                                                               |
  |       x       Float to map                                                   |
  |       fromMin Original lower bound                                           |
  |       fromMax Original upper bound                                           |
  |       toMin   Lower bound to map to                                          |
  |       toMax   Upper bound to map to                                          |
  |    Returns:                                                                  |
  |       Float containing x mapped to its equivalent value within he new bounds |
  |------------------------------------------------------------------------------|*/

  RETURN (x - fromMin) * (toMax - toMin) / (fromMax - fromMin) + toMin
}

DEFINE_FUNCTION FLOAT fnMathNaturalLogarithm(FLOAT x)
{
/*|------------------------------------------------------------------------------|
  | fnMathNaturalLogarithm: Calculate the natural logarithm of the passed        |
  |       number.                                                                |
  |    Parameters:                                                               |
  |       x Float to find the natural log of                                     |
  |    Returns:                                                                  |
  |       Float containing the passed numbers log base e                         |
  |------------------------------------------------------------------------------|*/

  RETURN fnMathLogarithm(x, MATH_E)
}

DEFINE_FUNCTION FLOAT fnMathPower(FLOAT x,
  INTEGER n)
{
/*|------------------------------------------------------------------------------|
  | fnMathPower: Calculate x raised to the n.                                    |
  |    Parameters:                                                               |
  |       x Float to find the natural log of                                     |
  |       n Power to raise x to                                                  |
  |    Returns:                                                                  |
  |       Float containing the x^n                                               |
  |------------------------------------------------------------------------------|*/

  STACK_VAR FLOAT result
  STACK_VAR FLOAT base
  STACK_VAR INTEGER exp

  result                                = 1.0
  base                                  = x + 0.0
  exp                                   = n + 0

  WHILE(exp > 0)
  {
    IF(exp & 1)
    {
      result                            = result * base
      exp                               = exp - 1
    }

    base                                = base * base
    exp                                 = TYPE_CAST(fnMathRound(exp * 0.5))
  }

  RETURN result
}

DEFINE_FUNCTION DOUBLE fnMathRandom()
{
/*|------------------------------------------------------------------------------|
  | fnMathRandom: Returns a double value with a positive sign, greater than or   |
  |       equal to 0.0 and less than 1.0.                                        |
  |    Returns:                                                                  |
  |       Double greater than or equal to 0.0 and less than 1.0                  |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR i
  STACK_VAR LONG hi
  STACK_VAR LONG low

  FOR(i = 32; i; i--)
  {
    low                                   = low + RANDOM_NUMBER(2) << (i - 1)
  }

  FOR(i = 20; i; i--)
  {
    hi                                     = hi + RANDOM_NUMBER(2) << (i - 1)
  }

  hi                                       = hi + 1023 << 20

  RETURN fnMathBuildDouble(hi, low) - 1
}

DEFINE_FUNCTION LONG fnMathRawBEtoLong(CHAR x[4])
{
/*|------------------------------------------------------------------------------|
  | fnMathRawBEtoLong: Load 4 bytes of Big Endian data contained in a character  |
  |       array into a long. Note: Array position 1 should contain MSB.          |
  |    Parameters:                                                               |
  |       x 4 byte character array containing the data to load                   |
  |    Returns:                                                                  |
  |       Long filled with the passed data                                       |
  |------------------------------------------------------------------------------|*/

  RETURN x[1] << 24 + x[2] << 16 + x[3] << 8 + x[4]
}

DEFINE_FUNCTION LONG fnMathRawLEtoLong(CHAR x[4])
{
/*|------------------------------------------------------------------------------|
  | fnMathRawLEtoLong: Load 4 bytes of Little Endian data contained in a         |
  |       character array into a long.                                           |
  |                                                                              |
  | Note: Array position 4 should contain MSB / position 1 should contain LSB    |
  |                                                                              |
  |    Parameters:                                                               |
  |       x 4 byte character array containing the data to load                   |
  |    Returns:                                                                  |
  |       Long filled with the passed data                                       |
  |------------------------------------------------------------------------------|*/

  RETURN x[4] << 24 + x[3] << 16 + x[2] << 8 + x[1]
}

DEFINE_FUNCTION DOUBLE fnMathRound(DOUBLE x)
{
/*|------------------------------------------------------------------------------|
  | fnMathRound: Rounds a flouting point number to it's closest whole number.    |
  |    Parameters:                                                               |
  |       x Double to round                                                      |
  |    Returns:                                                                  |
  |       Double containing the rounded number                                   |
  |------------------------------------------------------------------------------|*/

  RETURN fnMathFloor(x + 0.5)
}

DEFINE_FUNCTION DOUBLE fnMathRShiftDouble(DOUBLE x)
{
/*|------------------------------------------------------------------------------|
  | fnMathRShiftDouble: Right shift (>>) a double 1 bit.                         |
  |    Parameters:                                                               |
  |       x Double to shift                                                      |
  |    Returns:                                                                  |
  |       Passed value >> 1                                                      |
  |------------------------------------------------------------------------------|*/

  STACK_VAR LONG hi
  STACK_VAR LONG low

  hi                                    = fnMathDoubleHighToBits(x)
  low                                   = fnMathDoubleLowToBits(x)

  low                                   = low >> 1 + (hi & 1) << 15

  hi                                     = hi >> 1

  RETURN fnMathBuildDouble(hi, low)
}

DEFINE_FUNCTION DOUBLE fnMathSquareRoot(DOUBLE x)
{
/*|------------------------------------------------------------------------------|
  | fnMathSquareRoot: Calculate the square root of the passed number. This       |
  |       function takes a log base 2 approximation then iterates a Babylonian   |
  |       refinement until the answer is within the math libraries defined       |
  |       precision or exceeds 1000 steps of refinement.                         |
  |    Parameters:                                                               |
  |       x Double to find the square root of                                    |
  |    Returns:                                                                  |
  |       Double containing the square root                                      |
  |------------------------------------------------------------------------------|*/

  STACK_VAR LONG hi
  STACK_VAR LONG low
  STACK_VAR DOUBLE i
  STACK_VAR DOUBLE tmp

  IF(x < 0)
  {
    RETURN MATH_NaN
  }

  IF(x == 0 || x == 1 || fnMathIsNaN(x) ||  fnMathIsInfinite(x))
  {
    RETURN x
  }

  tmp                                   = fnMathRShiftDouble(x)

  hi                                     = 1 << 29 + fnMathDoubleHighToBits(tmp) - 1 << 19
  low                                   = fnMathDoubleLowToBits(tmp)
  tmp                                   = fnMathBuildDouble(hi, low)

  WHILE(!fnMathIsNear(tmp * tmp, x) && i < 1000)
  {
    tmp                                 = 0.5 * (tmp + x / tmp)
    i++
  }

  RETURN tmp
}

DEFINE_FUNCTION CHAR[20] fnScaleToLevelRange (CHAR value[20])
{
/*|------------------------------------------------------------------------------|
  | fnScaleToLevelRange: Scales given value to Level Range (0..255)              |
  |    Parameters:                                                               |
  |       value Value to scale                                                   |
  |    Returns:                                                                  |
  |       Scaled value                                                           |
  |------------------------------------------------------------------------------|*/

  RETURN FTOA((ATOI(value) / 100.0) * 255.0)
}

DEFINE_FUNCTION fnSendCommand(DEV vdvDevice,
  CHAR      sCommand[DUET_MAX_CMD_LEN])
{
/*|------------------------------------------------------------------------------|
  | fnSendCommand: Send Command Helper Function                                  |
  |    Parameters:                                                               |
  |       vdvDevice Virtual device to Send Command to                            |
  |       sCommand  Text command to send                                         |
  |------------------------------------------------------------------------------|*/

  _fnLog("'SEND_COMMAND ',DPSTOA(vdvDevice),',',sCommand")

  SEND_COMMAND vdvDevice,sCommand
}

DEFINE_FUNCTION fnSendCommandEx(DEV vdvDevice,
  CHAR      sHeader[DUET_MAX_HDR_LEN],
  CHAR      sCommand[DUET_MAX_CMD_LEN])
{
/*|------------------------------------------------------------------------------|
  | fnSendCommandEx: Send Command Helper Function, combines sHeader and sCommand |
  |       to sHeader-sCommand                                                    |
  |    Parameters:                                                               |
  |       vdvDevice Virtual device to Send Command to                            |
  |       sHeader   Header of command                                            |
  |       sCommand  Text command to send                                         |
  |------------------------------------------------------------------------------|*/

  fnSendCommand( vdvDevice,"sHeader,'-',sCommand")
}

DEFINE_FUNCTION fnSendLevel(DEV vdvDevice,
  INTEGER   level,
  FLOAT     value)
{
/*|------------------------------------------------------------------------------|
  | fnSendLevel: Send Level Helper Function                                      |
  |    Parameters:                                                               |
  |       vdvDevice Virtual device to Send Level to                              |
  |       level     number of the level to receive the new value                 |
  |       value     new level value                                              |
  |------------------------------------------------------------------------------|*/

  _fnLog("'SEND_LEVEL ',DPSTOA(vdvDevice),',',ITOA(level),',',FTOA(value)")

  SEND_LEVEL vdvDevice,level,value
}

DEFINE_FUNCTION fnSendString(DEV vdvDevice,
  CHAR      sCommand[DUET_MAX_CMD_LEN])
{
/*|------------------------------------------------------------------------------|
  | fnSendCommand: Send String Helper Function                                   |
  |    Parameters:                                                               |
  |       vdvDevice Virtual device to Send String to                             |
  |       sCommand  Text String to send                                          |
  |------------------------------------------------------------------------------|*/

  _fnLog("'SEND_STRING ',DPSTOA(vdvDevice),',',sCommand")

  SEND_STRING vdvDevice,sCommand
}


DEFINE_FUNCTION CHAR[3] fnStringBoolToString(CHAR a)
{
/*|------------------------------------------------------------------------------|
  | fnStringBoolToString: Converts a boolean value to its string equivalent of   |
  |       either a 'ON' or 'OFF'.                                                |
  |    Parameters:                                                               |
  |       a Boolean value to convert                                             |
  |    Returns:                                                                  |
  |       String equivalent (as ON/OFF)                                          |
  |------------------------------------------------------------------------------|*/

  IF(a)
  {
    RETURN 'ON'
  }
  ELSE
  {
    RETURN 'OFF'
  }
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringCiGetBetween(CHAR a[],
  CHAR left[], CHAR right[])
{
/*|------------------------------------------------------------------------------|
  | fnStringCiGetBetween: Gets the first instance of a string contained within   |
  |       the bounds of two sub-strings case insensitive                         |
  |    Parameters:                                                               |
  |       a     String to split, max size is 100 kilobytes                       |
  |       left  Character sequence marking the left bound                        |
  |       right Character sequence marking the right bound                       |
  |    Returns:                                                                  |
  |       String contained within the boundary sequences                         |
  |------------------------------------------------------------------------------|*/

  RETURN fnStringGetBetweenEx(a, left, right, FALSE)
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringConcatenate(CHAR a[],
  CHAR b[])
{
/*|------------------------------------------------------------------------------|
  | fnStringConcatenate: Concatenates two string together                        |
  |    Parameters:                                                               |
  |       a First part of to concatenate                                         |
  |       b Second part of to concatenate                                        |
  |    Returns:                                                                  |
  |       Concatenated string                                                    |
  |------------------------------------------------------------------------------|*/

  RETURN "a,b"
}

DEFINE_FUNCTION INTEGER fnStringContains(CHAR haystack[],
  CHAR needle[])
{
/*|------------------------------------------------------------------------------|
  | fnStringContains: Search through a string for a match strings, case          |
  |       sensistive and return True if found else False                         |
  |    Parameters:                                                               |
  |       haystack String to search                                              |
  |       needle   Sub-strings to match                                          |
  |    Returns:                                                                  |
  |       True if found, else False                                              |
  |------------------------------------------------------------------------------|*/

  RETURN FIND_STRING(haystack,needle, 1)
}

DEFINE_FUNCTION INTEGER fnStringNotContains(CHAR haystack[],
  CHAR needle[])
{
/*|------------------------------------------------------------------------------|
  | fnStringNotContains: Search through a string for a match strings, case       |
  |       sensistive and return True if not found else False                     |
  |    Parameters:                                                               |
  |       haystack String to search                                              |
  |       needle   Sub-strings to match                                          |
  |    Returns:                                                                  |
  |       True if found, else False                                              |
  |------------------------------------------------------------------------------|*/

  RETURN !fnStringContains(haystack,needle)
}

DEFINE_FUNCTION INTEGER fnStringContainsIgnoreCase(CHAR haystack[],
  CHAR needle[])
{
/*|------------------------------------------------------------------------------|
  | fnStringContainsIgnoreCase: Search through a string for a match strings,     |
  |       case insensistive and return True if found else False                  |
  |                                                                              |
  |    Parameters:                                                               |
  |       haystack String to search                                              |
  |       needle   Sub-strings to match                                          |
  |    Returns:                                                                  |
  |       True if found, else False                                              |
  |------------------------------------------------------------------------------|*/

  RETURN fnStringContains(UPPER_STRING(haystack),UPPER_STRING(needle))
}

DEFINE_FUNCTION INTEGER fnStringNotContainsIgnoreCase(CHAR haystack[],
  CHAR needle[])
{
/*|------------------------------------------------------------------------------|
  | fnStringContainsIgnoreCase: Search through a string for a match strings,     |
  |       case insensistive and return True if not found else False              |
  |                                                                              |
  |    Parameters:                                                               |
  |       haystack String to search                                              |
  |       needle   Sub-strings to match                                          |
  |    Returns:                                                                  |
  |       True if found, else False                                              |
  |------------------------------------------------------------------------------|*/

  RETURN !fnStringContainsIgnoreCase(haystack, needle)
}

DEFINE_FUNCTION CHAR[10] fnStringDateInvert(CHAR a[])
{
/*|------------------------------------------------------------------------------|
  | fnStringDateInvert: Switches the day and month fields of a date string (for  |
  |       converting between US and international standards). for example        |
  |       05/28/2009 becomes 28/05/2009.                                         |
  |    Parameters:                                                               |
  |       a String representation of a date in the form mm/dd/yyyy               |
  |    Returns:                                                                  |
  |       String representing the same date with the first two components        |
  |       reversed                                                               |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER idx
  STACK_VAR CHAR comp[3][4]

  FOR(idx = 3; idx; idx--)
  {
    comp[idx]                           = fnStringGetListItem(a, "'/'", idx)
  }

  RETURN "comp[2], '/', comp[1], '/', comp[3]"
}

DEFINE_FUNCTION CHAR fnStringEndsWith(CHAR a[],
  CHAR search[])
{
/*|------------------------------------------------------------------------------|
  | fnStringEndsWith: Check is a string end with another string.                 |
  |    Parameters:                                                               |
  |       a      String to check                                                 |
  |       search Sub-strings to search for                                       |
  |    Returns:                                                                  |
  |       True if 'a' ends with 'search', else False                             |
  |------------------------------------------------------------------------------|*/

  RETURN RIGHT_STRING(a, LENGTH_STRING(search)) == search;
}

DEFINE_FUNCTION INTEGER fnStringEndsWithIgnoreCase(CHAR a[],
  CHAR search[])
{
/*|------------------------------------------------------------------------------|
  | fnStringEndsWithIgnoreCase: Check is a string ends with another string,      |
  |    ignore case.                                                              |
  |    Parameters:                                                               |
  |       a      String to check                                                 |
  |       search Sub-strings to search for                                       |
  |    Returns:                                                                  |
  |       True if 'a' ends with 'search', else False                  3           |
  |------------------------------------------------------------------------------|*/

  RETURN fnStringIsEqualsIgnoreCase(RIGHT_STRING(a, LENGTH_STRING(search)),search);
}

DEFINE_FUNCTION INTEGER fnStringNotEndsWithIgnoreCase(CHAR a[],
  CHAR search[])
{
/*|------------------------------------------------------------------------------|
  | fnStringNotEndsWithIgnoreCase: Check is a string ends with another string,   |
  |    ignore case.                                                              |
  |    Parameters:                                                               |
  |       a      String to check                                                 |
  |       search Sub-strings to search for                                       |
  |    Returns:                                                                  |
  |       False if 'a' ends with 'search', else True                             |
  |------------------------------------------------------------------------------|*/

  RETURN !fnStringEndsWithIgnoreCase(a,search);
}

DEFINE_FUNCTION INTEGER fnStringNotEndsWith(CHAR a[],
  CHAR search[])
{
/*|------------------------------------------------------------------------------|
  | fnStringNotEndsWithIgnoreCase: Check is a string ends with another string    |
  |    Parameters:                                                               |
  |       a      String to check                                                 |
  |       search Sub-strings to search for                                       |
  |    Returns:                                                                  |
  |       False if 'a' ends with 'search', else True                             |
  |------------------------------------------------------------------------------|*/

  RETURN !fnStringEndsWith(a,search);
}

DEFINE_FUNCTION INTEGER fnStringExplode(CHAR delim[],
  CHAR a[],
  CHAR ret[][],
  INTEGER ret_len)
{
/*|------------------------------------------------------------------------------|
  | fnStringExplode: Explodes a string with a CHAR delimiter into an array of    |
  |       strings. The fnStringExploded data will be placed into ret[]. Due to   |
  |       NetLinx bugs, you must specify the length of ret in ret_len if you     |
  |       want the returned array sanitized.                                     |
  |    Parameters:                                                               |
  |       delim   Delimiter to use for the exploding                             |
  |       a       String array to explode                                        |
  |       ret     string array of string arrays that have been exploded          |
  |       ret_len Amount of entries in ret[][]; pass 0 if you don't care about   |
  |       sanitizing ret[][]                                                     |
  |    Returns:                                                                  |
  |       The amount of entries stuffed into ret[][]                             |
  |------------------------------------------------------------------------------|*/

  RETURN fnStringExplodeQuoted(delim, a, ret, ret_len, 0)
}

DEFINE_FUNCTION INTEGER fnStringExplodeQuoted(CHAR delim[],
  CHAR a[],
  CHAR ret[][],
  INTEGER ret_len,
  CHAR quote)
{
/*|------------------------------------------------------------------------------|
  | fnStringExplodeQuoted: Explodes a string with a CHAR delimiter into an array |
  |       of strings. The fnStringExploded data will be placed into ret[].       |
  |       Honours quotes; a character passed as quote (such asdouble quotes) are |
  |       treated as one segment.                                                |
  |                                                                              |
  | WARNING: Due to NetLinx bugs, you must specify the length of ret in ret_len  |
  |       if you want the returned array sanitized.                              |
  |                                                                              |
  |    Parameters:                                                               |
  |       delim   Delimiter to use for the exploding                             |
  |       a       String array to fnStringExplode                                |
  |       ret     String array of string arrays that have been exploded          |
  |       ret_len Amount of entries in ret[][]; pass 0 if you don't care about   |
  |       quote   Character to use as a quote                                    |
  |    Returns:                                                                  |
  |       The amount of entries stuffed into ret[][]                             |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER i
  STACK_VAR INTEGER start
  STACK_VAR INTEGER end

  start                                  = 1
  i                                      = 1

  WHILE(start <= LENGTH_STRING(a))
  {
    IF(quote)
    {
      IF(a[start] == quote)
      {
        // handle quotes
        end                             = FIND_STRING(a, "quote", start + 1)

        IF(end)
        {
          ret[i]                        = MID_STRING(a, start + 1, (end - start) - 1)
          i++

          start                         = end + 1
          CONTINUE
        }
      }
    }

    end                                 = FIND_STRING(a, "delim", start)// nothing ELSE stopping us?

    IF(end)
    {
      // then seperate by delimiter
      ret[i]                            = MID_STRING(a, start, (end - start))
      i++

      start                             = end + 1
    }
    ELSE
    {
      ret[i]                            = MID_STRING(a, start, LENGTH_STRING(a))

      start                             = LENGTH_STRING(a) + 1
    }
  }

  FOR(start = i + 1; start <= ret_len; start++)
  {
    ret[start]                           = NULL_STR
  }

  RETURN i
}

DEFINE_FUNCTION INTEGER fnStringFindMulti(CHAR haystack[],
  CHAR needles[][],
  INTEGER start)
{
/*|------------------------------------------------------------------------------|
  | fnStringFindMulti: Search through a string for a match against a list of     |
  |       possible sub-strings and return the element index of the matched       |
  |       string                                                                 |
  |    Parameters:                                                               |
  |       haystack String to search                                              |
  |       needles  List of sub-strings to match                                  |
  |       start    Array element to begin searching from                         |
  |    Returns:                                                                  |
  |       An integer containing the the element index of needles that was        |
  |       matched (0 if not found)                                               |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER i
  STACK_VAR INTEGER len

  len                                   = LENGTH_STRING(needles)

  FOR(i = start; i <= len; i++)
  {
    IF(FIND_STRING(haystack, needles[i], 1))
    {
      RETURN i
    }
  }

  RETURN 0
}

DEFINE_FUNCTION CHAR[20] fnStringFormatNumberThousands(INTEGER value)
{
/*|------------------------------------------------------------------------------|
  | fnStringFormatNumberThousands: Foramt the given number ###,###,###           |
  |    Parameters:                                                               |
  |       value Number to be formatted                                           |
  |    Returns:                                                                  |
  |       Formated value                                                         |
  |------------------------------------------------------------------------------|*/

  IF(value >= 1000)
  {
    STACK_VAR CHAR val[20]

    val                                 =  ITOA(value)

    RETURN "LEFT_STRING(val,LENGTH_STRING(val)-3),',',RIGHT_STRING(val, 3)"
  }
  ELSE
  {
    RETURN ITOA(value)
  }
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringGetBetween(CHAR a[],
  CHAR left[],
  CHAR right[])
{
/*|------------------------------------------------------------------------------|
  | fnStringGetBetween: Gets the first instance of a string contained within the |
  |       bounds of two sub-strings case sensitive                               |
  |    Parameters:                                                               |
  |       a     String to split                                                  |
  |       left  Character sequence marking the left bound                        |
  |       right Character sequence marking the right bound                       |
  |    Returns:                                                                  |
  |       String contained within the boundary sequences                         |
  |------------------------------------------------------------------------------|*/

  RETURN fnStringGetBetweenEx(a, left, right, TRUE)
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringGetBetweenEx(CHAR a[],
  CHAR left[],
  CHAR right[],
  CHAR cs)
{
/*|------------------------------------------------------------------------------|
  | fnStringGetBetweenEx:                                                        |
  |    Parameters:                                                               |
  |       a     String to split                                                  |
  |       left  Character sequence marking the left bound                        |
  |       right Character sequence marking the right bound                       |
  |       cs    True for case sensitive search                                   |
  |    Returns:                                                                  |
  |       String contained within the boundary sequences                         |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER start
  STACK_VAR INTEGER end
  STACK_VAR INTEGER retlen

  IF(TRUE == cs)
  {
    start                               = FIND_STRING(a, left, 1) + LENGTH_STRING(left)
    end                                 = FIND_STRING(a, right, start)
  }
  ELSE
  {
    start                               = FIND_STRING(LOWER_STRING(a), LOWER_STRING(left), 1) + LENGTH_STRING(left)
    end                                 = FIND_STRING(LOWER_STRING(a), LOWER_STRING(right), start)
  }

  retlen                                = end - start

  IF(fnStringIsLengthSizeError(retlen,MID_STRING(a, start, retlen),'fnStringGetBetweenEx'))
  {
    RETURN NULL_STR
  }

  RETURN MID_STRING(a, start, retlen)
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringGetKey(CHAR a[],
  CHAR delim[])
{
/*|------------------------------------------------------------------------------|
  | fnStringGetKey: Gets the key from a single key/value pair string with the    |
  |       specified delimiter.                                                   |
  |    Parameters:                                                               |
  |       a     String containing a key/value pair                               |
  |       delim String which divides the key and value                           |
  |    Returns:                                                                  |
  |       String containing the key component                                    |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER pos
  STACK_VAR INTEGER retlen

  pos                                   = FIND_STRING(a, delim, 1)

  IF(pos)
  {
    retlen                              = pos - 1
  }
  ELSE
  {
    retlen                              = LENGTH_STRING(a)
  }

  IF(fnStringIsLengthSizeError(retlen,LEFT_STRING(a, retlen),'fnStringGetKey'))
  {
    RETURN NULL_STR
  }

  RETURN LEFT_STRING(a, retlen)
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringGetListItem(CHAR a[],
  CHAR delim[],
  INTEGER item)
{
/*|------------------------------------------------------------------------------|
  | fnStringGetListItem: Gets an item from a string list. This can be used to    |
  |       grab word n within a string by passing 'space' as the delimiter.       |
  |    Parameters:                                                               |
  |       a     String to split                                                  |
  |       delim Character string which divides the list entries                  |
  |       item  Item number to return                                            |
  |    Returns:                                                                  |
  |       String array the requested list item                                   |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER retlen
  STACK_VAR INTEGER i
  STACK_VAR INTEGER start
  STACK_VAR INTEGER end

  FOR(i = 1; i <= item; i++)
  {
    start                               = end + 1
    end                                 = FIND_STRING(a, delim, start)
  }

  IF(!end)
  {
    end                               = LENGTH_STRING(a) + 1
  }

  retlen                               = end - start

  IF(fnStringIsLengthSizeError(retlen,MID_STRING(a, start, retlen),'fnStringGetListItem'))
  {
    RETURN NULL_STR
  }

  RETURN MID_STRING(a, start, retlen)
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringGetValue(CHAR a[],
  CHAR delim[])
{
/*|------------------------------------------------------------------------------|
  | fnStringGetValue: Gets the value from a key/value pair string with the       |
  |       specified delimiter.                                                   |
  |    Parameters:                                                               |
  |       a     Containing a key/value pair                                      |
  |       delim String which divides the key and value                           |
  |    Returns:                                                                  |
  |       String containing the value component                                  |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER pos
  STACK_VAR INTEGER retlen

  pos                                   = FIND_STRING(a, delim, 1)
  retlen                                = LENGTH_STRING(a) - (pos + LENGTH_STRING(delim) - 1)

  IF(fnStringIsLengthSizeError(retlen, a,'fnStringGetValue'))
  {
    RETURN NULL_STR
  }

  RETURN RIGHT_STRING(a, retlen)
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringImplode(CHAR strings[][],
  CHAR delim[])
{
/*|------------------------------------------------------------------------------|
  | fnStringImplode: Concatenates elements of an array of strings into a single  |
  |       string with a delimiter string inserted between each item.             |
  |    Parameters:                                                               |
  |       strings String array to implode                                        |
  |       delim   Character string to insert between the imploded elements       |
  |    Returns:                                                                  |
  |       The imploded string                                                    |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR ret[STRING_RETURN_SIZE_LIMIT + 1]
  STACK_VAR INTEGER i
  STACK_VAR INTEGER len

  len                                   = LENGTH_STRING(strings)
  ret                                   = strings[1]

  IF(len > 1)
  {
    FOR(i = len - 1; i; i--)
    {
      ret                               = "ret, delim, strings[(len - i) + 1]"
    }
  }

  IF(fnStringIsSizeError(ret, 'fnStringImplode'))
  {
    RETURN NULL_STR
  }

  RETURN ret
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringIntArrayToString(
  INTEGER ints[],
  CHAR delim[])
{
/*|------------------------------------------------------------------------------|
  | fnStringIntArrayToString: Converts an integer array into a comma separated   |
  |       string list of its values.                                             |
  |    Parameters:                                                               |
  |       ints  Integer array of values to 'listify'                             |
  |       delim String array to insert between entries                           |
  |    Returns:                                                                  |
  |       String list of the values                                              |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR list[STRING_RETURN_SIZE_LIMIT + 1]
  STACK_VAR INTEGER i
  STACK_VAR INTEGER len
  STACK_VAR INTEGER item

  len                                   = LENGTH_STRING(ints)

  list                                  = ITOA(ints[1])

  IF(len > 1)
  {
    FOR(i = len - 1; i; i--)
    {
      item                              = (len - i) + 1
      list                              = "list, delim, ITOA(ints[item])"
    }
  }

  IF(fnStringIsSizeError(list,'fnStringIntArrayToString'))
  {
    RETURN NULL_STR
  }

  RETURN list
}

DEFINE_FUNCTION INTEGER fnStringIsEquals(CHAR c1[],
  CHAR c2[])
{
/*|------------------------------------------------------------------------------|
  | fnStringIsEquals: Compare 2 strings                                          |
  |    Parameters:                                                               |
  |       c1 String to compare                                                   |
  |       c2 String to compare                                                   |
  |    Returns:                                                                  |
  |       True if equal, else False                                              |
  |------------------------------------------------------------------------------|*/

  RETURN (c1 == c2)
}

DEFINE_FUNCTION INTEGER fnStringNotEquals(CHAR c1[],
  CHAR c2[])
{
/*|------------------------------------------------------------------------------|
  | fnStringNotEquals: Compare 2 strings                                         |
  |    Parameters:                                                               |
  |       c1 String to compare                                                   |
  |       c2 String to compare                                                   |
  |    Returns:                                                                  |
  |       False if equal, else True                                              |
  |------------------------------------------------------------------------------|*/

  RETURN !fnStringIsEquals(c1,c2)
}

DEFINE_FUNCTION INTEGER fnStringIsEqualsIgnoreCase(CHAR c1[],
  CHAR c2[])
{
/*|------------------------------------------------------------------------------|
  | fnStringIsEqualsIgnoreCase: Compare 2 strings using case in-sensistive       |
  |    Parameters:                                                               |
  |       c1 String to compare                                                   |
  |       c2 String to compare                                                   |
  |    Returns:                                                                  |
  |       True if equal, else False                                              |
  |------------------------------------------------------------------------------|*/

  RETURN (UPPER_STRING(c1) == UPPER_STRING(c2))
}

DEFINE_FUNCTION INTEGER fnStringNotEqualsIgnoreCase(CHAR c1[],
  CHAR c2[])
{
/*|------------------------------------------------------------------------------|
  | fnStringNotEqualsIgnoreCase: Compare 2 strings using case in-sensistive      |
  |    Parameters:                                                               |
  |       c1 String to compare                                                   |
  |       c2 String to compare                                                   |
  |    Returns:                                                                  |
  |       True if not equal, else False                                          |
  |------------------------------------------------------------------------------|*/

  RETURN (UPPER_STRING(c1) != UPPER_STRING(c2))
}


DEFINE_FUNCTION INTEGER fnStringIsLengthSizeError(INTEGER len, CHAR a[],CHAR caller[])
{
/*|------------------------------------------------------------------------------|
  | fnStringIsLengthSizeError: Callback triggered when a function within this    |
  |       string processing library attempts to process anything that will       |
  |       result in a return size greater than that defined by                   |
  |       STRING_RETURN_SIZE_LIMIT.                                              |
  |    Parameters:                                                               |
  |       len    Length of to be checked                                         |
  |       a      String of which length is to be checked                         |
  |       caller Calling function name                                           |
  |    Returns:                                                                  |
  |       True is valid, else False                                              |
  |------------------------------------------------------------------------------|*/

  IF(len > STRING_RETURN_SIZE_LIMIT || LENGTH_STRING(a) > STRING_RETURN_SIZE_LIMIT)
  {
    fnIOPrintln("'Requested RETURN size,',ITOA(len),' too small from ', caller")

    RETURN TRUE
  }

  RETURN FALSE
}

DEFINE_FUNCTION INTEGER fnStringIsPrintable(CHAR a[])
{
/*|------------------------------------------------------------------------------|
  | fnStringIsPrintable: Hecks to see if the passed string is all printable      |
  |       ASCII characters.                                                      |
  |    Parameters:                                                               |
  |       a String to check                                                      |
  |    Returns:                                                                  |
  |       True if  it is printable, else False                                   |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER i
  STACK_VAR INTEGER len

  len       = LENGTH_STRING(a)

  FOR(i = 1; i <= len; i++)
  {
    IF(!fnStringIsPrintableChar(a[i]))
    {
      RETURN FALSE
    }
  }

  RETURN TRUE
}

DEFINE_FUNCTION INTEGER fnStringIsPrintableChar(CHAR a)
{
/*|------------------------------------------------------------------------------|
  | fnStringIsPrintableChar: Checks to see if the passed character is a          |
  |       printable ASCII character.                                             |
  |    Parameters:                                                               |
  |       a Character to check                                                   |
  |    Returns:                                                                  |
  |       True if it is printable, else False                                    |
  |------------------------------------------------------------------------------|*/

  RETURN a > $20 && a <= $7E
}

DEFINE_FUNCTION INTEGER fnStringIsSizeError(CHAR a[],CHAR caller[])
{
/*|------------------------------------------------------------------------------|
  | fnStringIsSizeError: Allback triggered when a function within this string    |
  |       processing library attempts to process anything that will result in a  |
  |       return size greater than that defined by STRING_RETURN_SIZE_LIMIT.     |
  |    Parameters:                                                               |
  |       a      String of which length is to be checked                         |
  |       caller Calling function name                                           |
  |    Returns:                                                                  |
  |       True is valid, else False                                              |
  |------------------------------------------------------------------------------|*/

  IF(LENGTH_STRING(a) > STRING_RETURN_SIZE_LIMIT)
  {
    fnIOPrintln("'Maximum RETURN size,',ITOA(STRING_RETURN_SIZE_LIMIT),
      ' exceeded, checked string size is ', ITOA(LENGTH_STRING(a)), ' from ', caller")

    RETURN TRUE
  }

  RETURN FALSE
}

DEFINE_FUNCTION CHAR fnStringIsWhitespace(CHAR a)
{
/*|------------------------------------------------------------------------------|
  | fnStringIsWhitespace: Checks to see if the passed character is a whitespace  |
  |       character.                                                             |
  |    Parameters:                                                               |
  |       a Character to check                                                   |
  |    Returns:                                                                  |
  |       True if it is whitespace, else False                                   |
  |------------------------------------------------------------------------------|*/

  RETURN (a >= $09 && a <= $0D) || (a >= $1C && a <= $20)
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringLeftTrim(CHAR a[])
{
/*|------------------------------------------------------------------------------|
  | fnStringLeftTrim: Returns a copy of the string with the left whitespace      |
  |       removed. If no printable characters are found, an empty string will be |
  |       returned.                                                              |
  |    Parameters:                                                               |
  |       a String to Left Trim                                                  |
  |    Returns:                                                                  |
  |       String with left whitespace removed                                    |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR ret[STRING_RETURN_SIZE_LIMIT + 1]
  STACK_VAR INTEGER i
  STACK_VAR INTEGER len

  len                                   = LENGTH_STRING(a)

  FOR(i = 1; i <= len; i++)
  {
    IF(!fnStringIsWhitespace(a[i]))
    {
      ret                               = RIGHT_STRING(a, len - i + 1)

      IF(fnStringIsSizeError(ret,'fnStringLeftTrim'))
      {
        RETURN NULL_STR
      }

      RETURN ret

    }
  }
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringPostfixToLength(
  CHAR a[], CHAR value, INTEGER len)
{
/*|------------------------------------------------------------------------------|
  | fnStringPostfixToLength: Returns a string post fixed with a specified CHAR,  |
  |       to a specified length. If the string is the same size or is larger     |
  |       than the specified length, Returns the original string.                |
  |    Parameters:                                                               |
  |       a     String to postfix                                                |
  |       value Value to postfix on the string                                   |
  |       len   Requested length of the string                                   |
  |    Returns:                                                                  |
  |       string postfix to length len with value                                |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR ret[STRING_RETURN_SIZE_LIMIT]
  STACK_VAR INTEGER i

  IF(fnStringIsLengthSizeError(len,a,'fnStringPostfixToLength'))
  {
    RETURN NULL_STR
  }

  IF(LENGTH_STRING(a) < len)
  {
    FOR(i = LENGTH_STRING(a); i < len; i++)
    {
      ret                               = "value, ret"
    }
  }

  RETURN "a, ret"
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringPrefixToLength(
  CHAR a[], CHAR value, INTEGER len)
{
/*|------------------------------------------------------------------------------|
  | fnStringPrefixToLength: Returns a string prefixed with a specified value, up |
  |       to a specified length. If the string is the same size or is larger     |
  |       than the specified length, Returns the original string.                |
  |    Parameters:                                                               |
  |       a     String to prefix                                                 |
  |       value Value to prefix on the string                                    |
  |       len   Requested length of the string                                   |
  |    Returns:                                                                  |
  |       string prefixed to length len with value                               |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR ret[STRING_RETURN_SIZE_LIMIT]
  STACK_VAR INTEGER i

  IF(fnStringIsLengthSizeError(len,a,'fnStringPrefixToLength'))
  {
    RETURN NULL_STR
  }

  IF(LENGTH_STRING(a) < len)
  {
    FOR(i = LENGTH_STRING(a); i < len; i++)
    {
      ret                               = "value, ret"
    }
  }

  RETURN "ret, a"
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringRemoveStringByLength(
  CHAR a[],
  INTEGER len)
{
/*|------------------------------------------------------------------------------|
  | fnStringRemoveStringByLength: Returns the left sub-strings of a string up to |
  |       the specified number of characters.                                    |
  |                                                                              |
  | WARNING: this is a destructive removal - the returned sub-strings will be    |
  |       removed from string 'a'.                                               |
  |                                                                              |
  |    Parameters:                                                               |
  |       a   String to remove the sub-string from                               |
  |       len Number of characters to remove                                     |
  |    Returns:                                                                  |
  |       String containing the first 'len' characters of 'a'                    |
  |------------------------------------------------------------------------------|*/

  IF(fnStringIsLengthSizeError(len,a,'fnStringRemoveStringByLength'))
  {
    RETURN NULL_STR
  }

  RETURN REMOVE_STRING(a, LEFT_STRING(a, len), 1)
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringRemoveNonPrintable(CHAR a[])
{
/*|------------------------------------------------------------------------------|
  | fnStringRemoveNonPrintable: remove all non-printable ASCII characters.       |
  |    Parameters:                                                               |
  |       a String to check                                                      |
  |    Returns:                                                                  |
  |       String containing only printable characters                                   |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER i
  STACK_VAR INTEGER len
  STACK_VAR CHAR ret[STRING_RETURN_SIZE_LIMIT + 1]

  len                                      = LENGTH_STRING(a)
  ret = NULL_STR

  FOR(i = 1; i <= len; i++)
  {
    IF(fnStringIsPrintableChar(a[i]))
    {
      ret = fnStringConcatenate(ret, a[i])
    }
  }

  RETURN ret
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringReplace(CHAR a[],
  CHAR search[], CHAR replace[])
{
/*|------------------------------------------------------------------------------|
  | fnStringReplace: Replace all occurrences of a sub-strings with another       |
  |       string.                                                                |
  |    Parameters:                                                               |
  |       a       String to search                                               |
  |       search  Sub-strings to replace                                         |
  |       replace Replacement sub-string                                         |
  |    Returns:                                                                  |
  |       'a' with all occurrences of 'search' replaced by the contents of       |
  |       'replace'                                                              |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER start
  STACK_VAR INTEGER end
  STACK_VAR CHAR ret[STRING_RETURN_SIZE_LIMIT]

  IF(fnStringIsSizeError(a,'fnStringReplace'))
  {
    RETURN NULL_STR
  }

  start                                  = 1
  end                                    = FIND_STRING(a, search, start)

  WHILE(end)
  {
    ret                                  = "ret, MID_STRING(a, start, end - start), replace"
    start                               = end + LENGTH_STRING(search)
    end                                  = FIND_STRING(a, search, start)
  }

  ret                                    = "ret, RIGHT_STRING(a, LENGTH_STRING(a) - start + 1)"

  RETURN ret
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringReplaceTabWithSpaces(CHAR a[])
{
/*|------------------------------------------------------------------------------|
  | fnStringReplaceTabWithSpaces: Replace all occurrences of TAB with spaces     |
  |    Parameters:                                                               |
  |       a       String to search                                               |
  |    Returns:                                                                  |
  |       'a' with all occurrences of TAB replaced with spaces                   |
  |------------------------------------------------------------------------------|*/

  RETURN fnStringReplace(a,"IO_TAB",IO_TAB_AS_SPACES)
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringReverse(CHAR a[])
{
/*|------------------------------------------------------------------------------|
  | fnStringReverse: Reverse a string.                                           |
  |    Parameters:                                                               |
  |       a String to reverse                                                    |
  |    Returns:                                                                  |
  |       The contents of 'a' with the character order reversed                  |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER i
  STACK_VAR INTEGER len
  STACK_VAR CHAR ret[STRING_RETURN_SIZE_LIMIT]

  len                                    = LENGTH_STRING(a)

  FOR(i =  len; i; i--)
  {
    ret[(len - i) + 1]  = a[i];
  }

  SET_LENGTH_STRING(ret, len);

  RETURN ret
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringRightTrim(CHAR a[])
{
/*|------------------------------------------------------------------------------|
  | fnStringRightTrim: Eturns a copy of the string with the right whitespace     |
  |       removed. If no printable characters are found, an empty string will be |
  |       returned.                                                              |
  |    Parameters:                                                               |
  |       a String to Right Trim                                                 |
  |    Returns:                                                                  |
  |       He string with right whitespace removed                                |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER i
  STACK_VAR INTEGER len

  len                                   = LENGTH_STRING(a)

  FOR(i =  len; i; i--)
  {
    IF(!fnStringIsWhitespace(a[i]))
    {
      IF(fnStringIsLengthSizeError(i,a, 'fnStringRightTrim'))
      {
        RETURN NULL_STR
      }
      ELSE
      {
        RETURN LEFT_STRING(a, i)
      }
    }
  }
}

DEFINE_FUNCTION CHAR fnStringStartsWith(CHAR a[],
  CHAR search[])
{
/*|------------------------------------------------------------------------------|
  | fnStringStartsWith: Check is a string starts with another string.            |
  |    Parameters:                                                               |
  |       a      String to check                                                 |
  |       search Sub-strings to search for                                       |
  |    Returns:                                                                  |
  |       True if 'a' begins with 'search', else False                           |
  |------------------------------------------------------------------------------|*/

  RETURN LEFT_STRING(a, LENGTH_STRING(search)) == search;
}

DEFINE_FUNCTION INTEGER fnStringStartsWithIgnoreCase(CHAR a[],
  CHAR search[])
{
/*|------------------------------------------------------------------------------|
  | fnStringStartsWithIgnoreCase: Check is a string starts with another string,  |
  |    ignore case.                                                              |
  |    Parameters:                                                               |
  |       a      String to check                                                 |
  |       search Sub-strings to search for                                       |
  |    Returns:                                                                  |
  |       True if 'a' begins with 'search', else False                           |
  |------------------------------------------------------------------------------|*/

  RETURN fnStringIsEqualsIgnoreCase(LEFT_STRING(a, LENGTH_STRING(search)),search);
}

DEFINE_FUNCTION INTEGER fnStringNotStartsWithIgnoreCase(CHAR a[],
  CHAR search[])
{
/*|------------------------------------------------------------------------------|
  | fnStringNotStartsWithIgnoreCase: Check is a string starts with another       |
  |    string ignore case.                                                       |
  |    Parameters:                                                               |
  |       a      String to check                                                 |
  |       search Sub-strings to search for                                       |
  |    Returns:                                                                  |
  |       False if 'a' begins with 'search', else True                           |
  |------------------------------------------------------------------------------|*/

  RETURN !fnStringStartsWithIgnoreCase(a,search);
}

DEFINE_FUNCTION INTEGER fnStringNotStartsWith(CHAR a[],
  CHAR search[])
{
/*|------------------------------------------------------------------------------|
  | fnStringNotStartsWith: Check is a string starts with another string.         |
  |    Parameters:                                                               |
  |       a      String to check                                                 |
  |       search Sub-strings to search for                                       |
  |    Returns:                                                                  |
  |       False if 'a' begins with 'search', else True                           |
  |------------------------------------------------------------------------------|*/

  RETURN !fnStringStartsWith(a,search);
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringSubstr(CHAR a[],
  INTEGER start,
  INTEGER count)
{
/*|------------------------------------------------------------------------------|
  | fnStringSubstr: Wrapper method for MID_STRING to bring inline with other     |
  |       programming languages.                                                 |
  |    Parameters:                                                               |
  |       a     Input string                                                     |
  |       start Start location of the substrings                                 |
  |       count Number of characters to extract                                  |
  |    Returns:                                                                  |
  |       MID_STRING(a, start, count)                                            |
  |------------------------------------------------------------------------------|*/

  RETURN MID_STRING(a, start, count);
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringSubstring(CHAR a[],
  INTEGER start,
  INTEGER end)
{
/*|------------------------------------------------------------------------------|
  | fnStringSubstring: Alternative to fnStringSubstr which allows an end         |
  |       location to be specified instead of a count                            |
  |    Parameters:                                                               |
  |       a     Input string                                                     |
  |       start Start location of the sub-strings                                |
  |       end   End location of the sub-strings                                  |
  |    Returns:                                                                  |
  |       MID_STRING(a, start, end-start+1)                                      |
  |------------------------------------------------------------------------------|*/

  RETURN fnStringSubstr(a, start, end-start+1);
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringSubstringEx(CHAR a[],
  INTEGER start)
{
/*|------------------------------------------------------------------------------|
  | fnStringSubstringEx: Returns a substring beginning at specified location to  |
  |       end of string                                                          |
  |    Parameters:                                                               |
  |       a     Input string                                                     |
  |       start Start location of the sub-strings                                |
  |    Returns:                                                                  |
  |       MID_STRING(a, start, end-start+1)                                      |
  |------------------------------------------------------------------------------|*/

  RETURN fnStringSubstring(a, start, LENGTH_STRING(a));
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringSuffixToLength(
  CHAR a[],
  CHAR value,
  INTEGER len)
{
/*|------------------------------------------------------------------------------|
  | fnStringSuffixToLength: Returns a string suffixed with a specified value, up |
  |       to a specified length. If the string is the same size or is larger     |
  |       than the specified length, Returns the original string.                |
  |    Parameters:                                                               |
  |       a     String to suffix                                                 |
  |       value Value to suffix on the string                                    |
  |       len   Requested length of the string                                   |
  |    Returns:                                                                  |
  |       String suffixed to length len with value                               |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR ret[STRING_RETURN_SIZE_LIMIT]
  STACK_VAR INTEGER i

  IF(fnStringIsLengthSizeError(len,a,'fnStringSuffixToLength'))
  {
    RETURN NULL_STR
  }

  IF(LENGTH_STRING(a) < len)
  {
    FOR(i = LENGTH_STRING(a); i < len; i++)
    {
      ret                               = "value, ret"
    }
  }

  RETURN "a, ret"
}

DEFINE_FUNCTION CHAR fnStringToBool(CHAR a[])
{
/*|------------------------------------------------------------------------------|
  | fnStringToBool: Converts common string representations of boolean values     |
  |       into their boolean value equivalents.                                  |
  |    Parameters:                                                               |
  |       a String representing a boolean value                                  |
  |    Returns:                                                                  |
  |       Boolean value equivalent                                               |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR tmp[8]

  tmp                                   = LOWER_STRING(fnStringLeftTrim(a))

  IF(tmp == 'on' || tmp == 'TRUE' || tmp == 'yes'  || tmp == 'y' || tmp == '1')
  {
    RETURN TRUE
  }
  ELSE
  {
    RETURN FALSE
  }
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringTrim(CHAR a[])
{
/*|------------------------------------------------------------------------------|
  | fnStringTrim: Returns a copy of the string with the whitespace removed. If   |
  |       no printable characters are found, an empty string will be returned.   |
  |    Parameters:                                                               |
  |       a String to trim                                                       |
  |    Returns:                                                                  |
  |       The original string with whitespace removed                            |
  |------------------------------------------------------------------------------|*/

  RETURN fnStringLeftTrim(fnStringRightTrim(a))
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringTruncate(CHAR a[],
  INTEGER len)
{
/*|------------------------------------------------------------------------------|
  | fnStringTruncate: Returns a string truncated to a specific length. If the    |
  |       string is less than the length specified the original string is        |
  |       returned. If it is truncated an ellipsis will be appended.             |
  |    Parameters:                                                               |
  |       a   String to truncate                                                 |
  |       len Requested length of the string                                     |
  |    Returns:                                                                  |
  |       String truncated to a maximum len characters                           |
  |------------------------------------------------------------------------------|*/

  RETURN fnStringTruncateEx(a, NULL_STR, len)
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringTruncateEx(
  CHAR a[],
  CHAR value[],
  INTEGER len)
{
/*|------------------------------------------------------------------------------|
  | fnStringTruncateEx: Returns a string truncated to a specific length. If the  |
  |       string is less than the length specified the original string is        |
  |       retured. If it is truncated to contents of value is appended to the    |
  |       truncated string.                                                      |
  |    Parameters:                                                               |
  |       a     String to truncate                                               |
  |       value Value to suffix on the string if truncated                       |
  |       len   Requested length of the string                                   |
  |    Returns:                                                                  |
  |       String truncated to a maximum len characters                           |
  |------------------------------------------------------------------------------|*/

  IF(fnStringIsLengthSizeError(len,a,'fnStringTruncateEx'))
  {
    RETURN NULL_STR
  }

  IF(LENGTH_STRING(a) > len)
  {
    RETURN "LEFT_STRING(a, len - LENGTH_STRING(value)), value"
  }
  ELSE
  {
    RETURN a
  }
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringUcFirst(CHAR a[])
{
/*|------------------------------------------------------------------------------|
  | fnStringUcFirst: Eturns a copy of a string with the first alpha character    |
  |       capitalized. Non alpha characters are not modified. Pass a             |
  |       lower_string()'d string to lowercase all other characters.             |
  |    Parameters:                                                               |
  |       a String to capitalize first characters of                             |
  |    Returns:                                                                  |
  |       Capitalized string                                                     |
  |------------------------------------------------------------------------------|*/

  IF(fnStringIsSizeError(a,'fnStringUcFirst'))
  {
    RETURN NULL_STR
  }

  IF(a[1] >= $61 && a[1] <= $7A)
  {
    RETURN "a[1] - $20, MID_STRING(a, 2, STRING_RETURN_SIZE_LIMIT)"
  }

  RETURN a
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringUcWords(CHAR a[])
{
/*|------------------------------------------------------------------------------|
  | fnStringUcWords: Returns a copy of a string with the first alpha character   |
  |       in each word capitalized. Non alpha characters are not modified. Pass  |
  |       a lower_string()'d string to lowercae all other characters.            |
  |    Parameters:                                                               |
  |       a String to capitalize first characters of                             |
  |    Returns:                                                                  |
  |       Capitalized string                                                     |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR ret[STRING_RETURN_SIZE_LIMIT]
  STACK_VAR INTEGER i

  IF(fnStringIsSizeError(a,'fnStringUcWords'))
  {
    RETURN NULL_STR
  }

  ret                                   = a

  FOR(i = 1; i < LENGTH_STRING(ret); i++)
  {
    IF(fnStringIsWhitespace(ret[i]))
    {
      IF(ret[i + 1] >= $61 && ret[i + 1] <= $7a)
      {
        ret[i + 1]                       = ret[i + 1] - $20
      }
    }
  }

  RETURN ret
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringUrlEncode(CHAR a[])
{
/*|------------------------------------------------------------------------------|
  | fnStringUrlEncode: Returns a url-encoded string according to RFC 1736 / RFC  |
  |       2732.                                                                  |
  |    Parameters:                                                               |
  |       a String to encode                                                     |
  |    Returns:                                                                  |
  |       String encoded                                                         |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR ret[STRING_RETURN_SIZE_LIMIT + 1]
  STACK_VAR INTEGER i

  FOR(i = 1; i <= LENGTH_STRING(a); i++)
  {
    IF((a[i] >= $30 && a[i] <= $39) ||      // numerics
    (a[i] >= $41 && a[i] <= $5a) ||          // uppercase
    (a[i] >= $61 && a[i] <= $7a) ||          // lowercase
    a[i] == '$' || a[i] == '-' || a[i] == '_' ||
    a[i] == '.' || a[i] == '+' || a[i] == '!' ||
    a[i] == '*' || a[i] == $27 || a[i] == '(' ||
    a[i] == ')' || a[i] == ',' || a[i] == '[' ||
    a[i] == ']')
    {
      ret                               = "ret, a[i]"
    }
    ELSE
    {
      ret                                 = "ret, '%', fnStringPrefixToLength(itohex(a[i]), '0', 2)"
    }
  }

  IF(fnStringIsSizeError(ret,'fnStringUrlEncode'))
  {
    RETURN NULL_STR
  }

  RETURN ret
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnCommandEncode(CHAR a[])
{
/*|------------------------------------------------------------------------------|
  | fnCommandEncode: Returns a command-encoded string                            |
  |    Parameters:                                                               |
  |       a String to encode                                                     |
  |    Returns:                                                                  |
  |       String encoded                                                         |
  |------------------------------------------------------------------------------|*/
  STACK_VAR CHAR ret[STRING_RETURN_SIZE_LIMIT + 1]
  STACK_VAR INTEGER i
  IF (fnStringContains(a,','))
  {
    RETURN "'"',a,'"'"
  }
  RETURN a
}
DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStripCharsRight(CHAR a[],
  INTEGER count)
{
/*|------------------------------------------------------------------------------|
  | fnStripCharsRight: Remove characters from the end of the string.             |
  |    Parameters:                                                               |
  |       a     Input string                                                     |
  |       count Number of characters to remove                                   |
  |    Returns:                                                                  |
  |       The contents of 'a' with the characters removed                        |
  |------------------------------------------------------------------------------|*/

  IF(LENGTH_STRING(a) > count)
  {
    RETURN LEFT_STRING(a, LENGTH_STRING(a) - count)
  }
  ELSE
  {
    RETURN NULL_STR
  }
}

DEFINE_FUNCTION INTEGER fnToggleState(INTEGER state)
{
/*|------------------------------------------------------------------------------|
  | fnToggleState: Toggle state                                                  |
  |    Parameters:                                                               |
  |       state 1 or 0, TRUE or FALSE                                            |
  |    Returns:                                                                  |
  |       !state                                                                 |
  |------------------------------------------------------------------------------|*/

  RETURN !state
}

DEFINE_FUNCTION CHAR[3] fnTPConvertValueToPercentage(SINTEGER value)
{
/*|------------------------------------------------------------------------------|
  | fnTPConvertValueToPercentage: Converts a TP level value to %                 |
  |    Parameters:                                                               |
  |       value Value in range of 0..255 to convert to percentage                |
  |    Returns:                                                                  |
  |       percentage                                                             |
  |------------------------------------------------------------------------------|*/

  RETURN FORMAT('%3.0f',(ABS_VALUE(value) / 2.55))
}

DEFINE_FUNCTION fnTPDisableButton (DEV dvTP,
  CHAR addressRange[])
{
/*|------------------------------------------------------------------------------|
  | fnTPDisableButton: Disable buttons with a set variable text range            |
  |    Parameters:                                                               |
  |       dvTP         Touch panel device                                        |
  |       addressRange Button address or range Address codes of buttons to       |
  |           affect. A '.' between addresses includes the range, and an '&'     |
  |           between addresses includes each address                            |
  |------------------------------------------------------------------------------|*/

  SEND_COMMAND dvTP,"'^ENA-',addressRange,',0'"
}

DEFINE_FUNCTION fnTPEnableButton (DEV dvTP,
  CHAR addressRange[])
{
/*|------------------------------------------------------------------------------|
  | fnTPEnableButton: Enable buttons with a set variable text range              |
  |    Parameters:                                                               |
  |       dvTP         Touch panel device                                        |
  |       addressRange Button address or range Address codes of buttons to       |
  |           affect. A '.' between addresses includes the range, and an '&'     |
  |           between addresses includes each address                            |
  |------------------------------------------------------------------------------|*/

  SEND_COMMAND dvTP,"'^ENA-',addressRange,',1'"
}

DEFINE_FUNCTION fnTPHideAllPopups (DEV dvTP)
{
/*|------------------------------------------------------------------------------|
  | fnTPHideAllPopups: Close all popups on all pages.                            |
  |    Parameters:                                                               |
  |       dvTP Touch panel device                                                |
  |------------------------------------------------------------------------------|*/

  SEND_COMMAND dvTP,"'@PPX'"
}

DEFINE_FUNCTION fnTPHideAllPopupsOnPage (DEV dvTP,
  CHAR page[])
{
/*|------------------------------------------------------------------------------|
  | fnTPHideAllPopupsOnPage: Close all popups on a specified page. If the page   |
  |       name is empty, the current page is used.                               |
  |    Parameters:                                                               |
  |       dvTP Touch panel device                                                |
  |       page Page to close all popups                                          |
  |------------------------------------------------------------------------------|*/

  SEND_COMMAND dvTP,"'@PPA-',page"
}

DEFINE_FUNCTION fnTPSetOpacityAllStates(DEV dvTP,
  CHAR addr[],
  CHAR value[])
{
/*|------------------------------------------------------------------------------|
  | fnTPSetOpacityAllStates: Set button opacity for all button states            |
  |    Parameters:                                                               |
  |       dvTP  Touch panel device                                               |
  |       addr  Button address or range Address codes of buttons to affect. A    |
  |       '.' between addresses includes the range, and an '&' between addresses |
  |       includes each address                                                  |
  |       value Opacity 0-255 range                                              |
  |------------------------------------------------------------------------------|*/

  fnTPSetOpacityStates(dvTP,addr,'0',value)
}

DEFINE_FUNCTION fnTPSetOpacityStates(DEV dvTP,
  CHAR addr[],
  CHAR states[],
  CHAR value[])
{
/*|------------------------------------------------------------------------------|
  | fnTPSetOpacityStates: Set button opacity for requested button states         |
  |    Parameters:                                                               |
  |       dvTP   Touch panel device                                              |
  |       addr   Button address or range Address codes of buttons to affect. A   |
  |       '.' between addresses includes the range, and an '&' between addresses |
  |       includes each address                                                  |
  |       states 1 - 256 for multi-state buttons 0 = All states General buttons, |
  |       1 = Off state and 2 = On state Can be combined using &, ie... 1&2&3    |
  |       value  Opacity 0-255 range                                             |
  |------------------------------------------------------------------------------|*/

  fnSendCommand(dvTP,"'^BOP-',addr,',',states,',',value")
}

DEFINE_FUNCTION fnTPSetFillColorAllStates(DEV dvTP,
  CHAR addr[],
  CHAR value[])
{
/*|------------------------------------------------------------------------------|
  | fnTPSetFillColorAllStates: Set button fill color for all button states       |
  |    Parameters:                                                               |
  |       dvTP   Touch panel device                                              |
  |       addr   Button address or range Address codes of buttons to affect. A   |
  |       '.' between addresses includes the range, and an '&' between addresses |
  |       includes each address                                                  |
  |       value  Color as RGB or Named Color                                     |
  |------------------------------------------------------------------------------|*/

  fnTPSetFillColorStates(dvTP,addr,'0',value)
}

DEFINE_FUNCTION fnTPSetFillColorStates(DEV dvTP,
  CHAR addr[],
  CHAR states[],
  CHAR value[])
{
/*|------------------------------------------------------------------------------|
  | fnTPSetFillColorStates: Set button fill color for requested button states    |
  |    Parameters:                                                               |
  |       dvTP   Touch panel device                                              |
  |       addr   Button address or range Address codes of buttons to affect. A   |
  |       '.' between addresses includes the range, and an '&' between addresses |
  |       includes each address                                                  |
  |       states 1 - 256 for multi-state buttons 0 = All states General buttons, |
  |       1 = Off state and 2 = On state Can be combined using &, ie... 1&2&3    |
  |       value  Color as RGB or Named Color                                     |
  |------------------------------------------------------------------------------|*/

  fnSendCommand(dvTP,"'^BCF-',addr,',',states,',',value")
}

DEFINE_FUNCTION fnTPSetBorderColorAllStates(DEV dvTP,
  CHAR addr[],
  CHAR value[])
{
/*|------------------------------------------------------------------------------|
  | fnTPSetBorderColorStates: Set button boarder color for all button states     |
  |    Parameters:                                                               |
  |       dvTP   Touch panel device                                              |
  |       addr   Button address or range Address codes of buttons to affect. A   |
  |       '.' between addresses includes the range, and an '&' between addresses |
  |       includes each address                                                  |
  |       value  Color as RGB or Named Color                                     |
  |------------------------------------------------------------------------------|*/

  fnTPSetBorderColorStates(dvTP,addr,'0',value)
}

DEFINE_FUNCTION fnTPSetBorderColorStates(DEV dvTP,
  CHAR addr[],
  CHAR states[],
  CHAR value[])
{
/*|------------------------------------------------------------------------------|
  | fnTPSetBorderColorStates: Set button border color for requested button       |
  |    states                                                                    |
  |    Parameters:                                                               |
  |       dvTP   Touch panel device                                              |
  |       addr   Button address or range Address codes of buttons to affect. A   |
  |       '.' between addresses includes the range, and an '&' between addresses |
  |       includes each address                                                  |
  |       states 1 - 256 for multi-state buttons 0 = All states General buttons, |
  |       1 = Off state and 2 = On state Can be combined using &, ie... 1&2&3    |
  |       value  Color as RGB or Named Color                                     |
  |------------------------------------------------------------------------------|*/

  fnSendCommand(dvTP,"'^BCB-',addr,',',states,',',value")
}

DEFINE_FUNCTION fnTPHideButton (DEV dvTP,
  CHAR addressRange[])
{
/*|------------------------------------------------------------------------------|
  | fnTPHideButton: Hide specified button on touch panel                         |
  |    Parameters:                                                               |
  |       dvTP         Touch panel device                                        |
  |       addressRange Button address or range Address codes of buttons to       |
  |          affect. A '.' between addresses includes the range, and an '&'      |
  |          between addresses includes each address                             |
  |------------------------------------------------------------------------------|*/

  SEND_COMMAND dvTP,"'^SHO-',addressRange,',0'"
}

DEFINE_FUNCTION fnTPHidePopup (DEV dvTP,
  CHAR popupPage[])
{
/*|------------------------------------------------------------------------------|
  | fnTPHidePopup: Deactivate a specific popup page on the current page. If the  |
  |       popup page is part of a group, the whole group is deactivated.         |
  |    Parameters:                                                               |
  |       dvTP      Popup page to deactivate                                     |
  |       popupPage                                                              |
  |------------------------------------------------------------------------------|*/

  SEND_COMMAND dvTP,"'@PPF-',popupPage"
}

DEFINE_FUNCTION INTEGER fnTPIsG4(DEV dvTP)
{
/*|------------------------------------------------------------------------------|
  | fnTPIsG4: Is this a G4 panel?                                                |
  |    Parameters:                                                               |
  |       dvTP Touch panel device                                                |
  |    Returns:                                                                  |
  |       True if G4 panel, else False                                           |
  |------------------------------------------------------------------------------|*/

  RETURN !fnTPisG5(dvTP)
}

DEFINE_FUNCTION INTEGER fnTPIsG5(DEV dvTP)
{
/*|------------------------------------------------------------------------------|
  | fnTPIsG5: Is this a G5 panel?                                                |
  |    Parameters:                                                               |
  |       dvTP Touch panel device                                                |
  |    Returns:                                                                  |
  |       True if G5 panel, else False                                           |
  |------------------------------------------------------------------------------|*/

  RETURN ((DEVICE_ID(dvTP) > 404 ) AND (DEVICE_ID(dvTP) < 418 )) // G5
}

DEFINE_FUNCTION fnTPSetButtonText (DEV dvTP,
  CHAR addressRange[],
  CHAR btnStates[],
  CHAR text[])
{
/*|------------------------------------------------------------------------------|
  | fnTPSetButtonText:                                                           |
  |    Parameters:                                                               |
  |       dvTP         Touch panel device                                        |
  |       addressRange Button address or range Address codes of buttons to       |
  |          affect. A '.' between addresses includes the range, and an '&'      |
  |          between addresses includes each address                             |
  |       btnStates    1 - 256 for multi-state buttons 0 = All states General    |
  |          buttons, 1 = Off state and 2 = On state Can be combined using &,    |
  |          ie... 1&2&3                                                         |
  |       text         Text as ASCII characters                                  |
  |------------------------------------------------------------------------------|*/

  SEND_COMMAND dvTP,"fnTPTextCommand(dvTP),addressRange,',',btnStates,',',text"
}

DEFINE_FUNCTION fnTPSetButtonTextAllStates (DEV dvTP,
  CHAR addressRange[],
  CHAR text[])
{
/*|------------------------------------------------------------------------------|
  | fnTPSetButtonTextAllStates: Set button text for all button states            |
  |    Parameters:                                                               |
  |       dvTP         Text as ASCII characters                                  |
  |       addressRange Button address or range Address codes of buttons to       |
  |          affect. A '.' between addresses includes the range, and an '&'      |
  |          between addresses includes each address                             |
  |       text         Text as ASCII characters                                  |
  |------------------------------------------------------------------------------|*/

  SEND_COMMAND dvTP,"fnTPTextCommand(dvTP),addressRange,',0,',text"
}

DEFINE_FUNCTION fnTPShowButton (DEV dvTP,
  CHAR addressRange[])
{
/*|------------------------------------------------------------------------------|
  | fnTPShowButton: Show request button(s) in address range                      |
  |    Parameters:                                                               |
  |       dvTP         Touch panel device                                        |
  |       addressRange Button address or range Address codes of buttons to       |
  |          affect. A '.' between addresses includes the range, and an '&'      |
  |          between addresses includes each address                             |
  |------------------------------------------------------------------------------|*/

  SEND_COMMAND dvTP,"'^SHO-',addressRange,',1'"
}

DEFINE_FUNCTION fnTPShowPopup (DEV dvTP,
  CHAR popupPage[])
{
/*|------------------------------------------------------------------------------|
  | fnTPShowPopup: Activate a specific popup page to launch on the current page. |
  |       If the popup page is already on, do not re-draw it.                    |
  |    Parameters:                                                               |
  |       dvTP      Touch panel device                                           |
  |       popupPage Popup page to activate                                       |
  |------------------------------------------------------------------------------|*/

  SEND_COMMAND dvTP,"'@PPN-',popupPage"
}

DEFINE_FUNCTION CHAR[5] fnTPTextCommand(DEV dvTP)
{
/*|------------------------------------------------------------------------------|
  | fnTPTextCommand: Text command based upon panel type (G4 vs G5)               |
  |    Parameters:                                                               |
  |       dvTP Touch panel device                                                |
  |    Returns:                                                                  |
  |       G5 == ^UTF-, G4 == ^TXT-                                               |
  |------------------------------------------------------------------------------|*/

  IF( fnTPisG5(dvTP))
  {
    RETURN "'^UTF-'"
  }

  RETURN "'^TXT-'"

}

DEFINE_FUNCTION fnTPToggleChannel(DEV dvTP,
  INTEGER channel)
{
/*|------------------------------------------------------------------------------|
  | fnTPToggleChannel: Toggle channel ON or OFF                                  |
  |    Parameters:                                                               |
  |       dvTP    Touch panel device                                             |
  |       channel Channel to toggle                                              |
  |------------------------------------------------------------------------------|*/

  IF([dvTP,channel])
  {
    OFF[dvTP,channel]
  }
  ELSE
  {
    ON[dvTP,channel]
  }
}

DEFINE_FUNCTION _fnDebugMessage(INTEGER msg_level,
  CHAR msg[])
{
/*|------------------------------------------------------------------------------|
  | _fnDebugMessage (INTERNAL FUNCTION not intended for direct usage): Voices a  |
  |       debug message IF required by the current debug level. All system       |
  |       messages should pass through here. If the ability to dump to a file or |
  |       networked logging service is required it can be added here.            |
  |    Parameters:                                                               |
  |       msg_level char specifying the debug level of the message               |
  |       msg       string containing the debug message                          |
  |------------------------------------------------------------------------------|*/

  STACK_VAR LONG i,l
  STACK_VAR CHAR c
  STACK_VAR CHAR out[5000]
  STACK_VAR CHAR in[5000]
  STACK_VAR CHAR sErrorMessage[5000]

  IF(msg_level < DEBUG_ERROR || msg_level > DEBUG_MAX_LEVEL)
  {
    fnDebugErrorMessage("'invalid debug level specified - ', msg")

    RETURN
  }

  IF(msg_level <= nDebugLevel)
  {
    IF(FIND_STRING(msg, "$00", 1) > 0)
    {
      in                                = msg
      out                               = ""
      l                                 = LENGTH_STRING(in)

      FOR(i = 0; i < l; i++)
      {
        c                               = GET_BUFFER_CHAR(in)

        IF(c == "$00")
        {
          out                           = "out,'$00'"
        }
        ELSE
        {
          out                           = "out,c"
        }
      }

      sErrorMessage                     = out
    }
    ELSE
    {
      sErrorMessage                     = msg
    }

    fnIOPrintln("IO_TAB,TIME,' ',DATE,' ',PRT_PREFIX,UPPER_STRING(fnDebugGetLevelString()),': ', sErrorMessage")
  }
}

DEFINE_FUNCTION _fnIOFileErrorToText(CHAR cFileName[],SLONG slFileHandle)
{
/*|------------------------------------------------------------------------------|
  | _fnIOFileErrorToText (INTERNAL FUNCTION not intended for direct usage):      |
  |       Prints text of file error  to the output stream.                       |
  |    Parameters:                                                               |
  |       cFileName    file name that caused error                               |
  |       slFileHandle handle of the file                                        |
  |------------------------------------------------------------------------------|*/

  SWITCH(ITOA(slFileHandle))
  {
    CASE '-2':
    {
      fnIOPrintln("'Invalid file path or name - ',cFileName")
    }

    CASE '-3':
    {
      fnIOPrintln("'Invalid value supplied for IOFlag - ',cFileName")
    }

    CASE '-5':
    {
      fnIOPrintln("'Disk I/O error - ',cFileName")
    }

    CASE '-6':
    {
      fnIOPrintln("'invalid parameter (buffer length must be greater than zero) - ',cFileName")
    }

    CASE '-9':
    {
      fnIOPrintln("'EOF (end-of-file) reached - ',cFileName")
    }

    CASE '-14':
    {
      fnIOPrintln("'Maximum number of files are already open (max is 10) - ',cFileName")
    }

    CASE '-15':
    {
      fnIOPrintln("'Invalid file format - ',cFileName")
    }
  }
}

DEFINE_FUNCTION _fnLog(CHAR cLogString[])
{
 /*|------------------------------------------------------------------------------|
  | _fnLog (INTERNAL FUNCTION not intended for direct usage): Write log          |
  |       statement to disc                                                      |
  |    Parameters:                                                               |
  |       cLogString Log message                                                 |
  |------------------------------------------------------------------------------|*/

  STACK_VAR SLONG slFileHandle

  IF(!fnDebugIsLevelDebug())   RETURN

  IF(nFirstTime)
  {
    FILE_DELETE(LOGFILE_NAME)
    nFirstTime                          =  FALSE
  }

  slFileHandle                          =  FILE_OPEN(LOGFILE_NAME,FILE_RW_APPEND)

  IF(slFileHandle>0)
  {
    FILE_WRITE_LINE(slFileHandle,cLogString,LENGTH_STRING(cLogString))
    FILE_CLOSE(slFileHandle)
  }
}

DEFINE_FUNCTION _fnReportParseError(SLONG slResult,
  CHAR sfnName[])
{
/*|------------------------------------------------------------------------------|
  | _fnReportParseError (INTERNAL FUNCTION not intended for direct usage): Write |
  |       error code's text description to Telnet session                        |
  |    Parameters:                                                               |
  |       slResult Error code                                                    |
  |       sfnName  Function name calling                                         |
  |------------------------------------------------------------------------------|*/

  IF(slResult == 3)
  {
    fnIOPrintln("sfnName,': XML decode data type mismatch'")
  }

  ELSE IF(slResult == 2)
  {
    fnIOPrintln("sfnName,': XML decode data too small, more members in structure'")
  }

  ELSE IF(slResult == 1)
  {
    fnIOPrintln("sfnName,': XML decode data too small, more members in structure'")
  }

  ELSE IF(slResult == -1)
  {
    fnIOPrintln("sfnName,': decode variable type mismatch'")
  }

  ELSE IF(slResult == -2 )
  {
    fnIOPrintln("sfnName,': decode data too small, decoder ran out of data. Most likely poorly formed XML.'")
  }

  ELSE IF(slResult == -3)
  {
    fnIOPrintln("sfnName,': output character buffer was too small'")
  }
}

  DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnRGBToHex(INTEGER nRed, INTEGER nGreen, INTEGER nBlue)
  {
  /*|------------------------------------------------------------------------------|
    | fnRGBToHex: Creates a Hex RGB value for the give values                      |
    |    Parameters:                                                               |
    |       nRed   Red value                                                       |
    |       nGreen Green value                                                     |
    |       nBlue  Blue value                                                      |
    |    Returns:                                                                  |
    |       HEX RBG value                                                          |
    |------------------------------------------------------------------------------|*/

    RETURN "'#',fnStringPrefixToLength(ITOHEX(nRed),'0',2),
      fnStringPrefixToLength(ITOHEX(nGreen),'0',2),
      fnStringPrefixToLength(ITOHEX(nBlue),'0',2),
      'FF'"
  }

  DEFINE_FUNCTION SINTEGER fnOpenIPPort(INTEGER localPort, INTEGER serverPort, INTEGER protocol)
  {
  /*|------------------------------------------------------------------------------|
    | fnOpenIPPort: Opens a server port to listen for client requests.             |
    |    Parameters:                                                               |
    |       localPort  The local port number to open.                              |
    |       serverPort The number of the server port to listen on                  |
    |       protocol   The transport protocol to use: 1 = TCP 2 = UDP              |
    |    Returns:                                                                  |
    |       0  = operation was successfu                                           |
    |       -1 = invalid server port                                               |
    |       -2 = invalid value for Protoco                                         |
    |       -3 = unable to open communication port                                 |
    |------------------------------------------------------------------------------|*/

    STACK_VAR SINTEGER result

    result = IP_SERVER_OPEN(localPort,serverPort,protocol)

    IF (result == -1)
    {
      IP_SERVER_CLOSE(localPort)

      WAIT 2
      {
        fnOpenIPPortResults(IP_SERVER_OPEN(localPort,serverPort,protocol),localPort,serverPort,protocol)
      }
    }
    ELSE
    {
      fnOpenIPPortResults(result,localPort,serverPort,protocol)
    }

    RETURN result
  }

  DEFINE_FUNCTION fnOpenIPPortResults(SINTEGER status,INTEGER localPort, INTEGER serverPort, INTEGER protocol)
  {
  /*|------------------------------------------------------------------------------|
    | fnOpenIPPortResults: Prints message to telnet for the given IP_SERVER_OPEN   |
    |       status                                                                 |
    |    Parameters:                                                               |
    |       status                                                                 |
    |       localPort  The local port number to open                               |
    |       serverPort The number of the server port to listen on                  |
    |       protocol   The transport protocol to use: 1 = TCP 2 = UDP              |
    |------------------------------------------------------------------------------|*/

    fnIOPrintln('-------------------------------------------------------------------------------')

    IF (status == 0)
    {
      STACK_VAR IP_ADDRESS_STRUCT MyIPAddress
      GET_IP_ADDRESS(0:0:0,MyIPAddress)

      fnIOPrintln("'IP Port ',ITOA(serverPort),' opened on ',MyIPAddress.IPADDRESS,' (',MyIPAddress.HOSTNAME,')'")
    }
    ELSE IF (status == -1)
    {
      fnIOPrintln("'fnOpenIPPort FAILED: invalid server port, port used ',ITOA(serverPort)")
    }
    ELSE IF (status == -2)
    {
      fnIOPrintln("'fnOpenIPPort FAILED: invalid value for Protocol, valid values [IP_TCP==1,IP_UDP==2], got ',ITOA(protocol)")
    }
    ELSE IF (status == -3)
    {
      fnIOPrintln("'fnOpenIPPort FAILED: unable to open communication port, port used ',ITOA(localPort)")
    }
    ELSE
    {
      fnIOPrintln("'fnOpenIPPort UNEXPECTED RESULT OF ',ITOA(status)")
    }

    fnIOPrintln('-------------------------------------------------------------------------------')
  }

  DEFINE_FUNCTION INTEGER fnIsValidDPS(DEV dv)
  {
  /*|------------------------------------------------------------------------------|
    | fnIsValidDPS: Validates a device DPS device != 0 and Port != 0               |
    |    Parameters:                                                               |
    |       dv Device to validate                                                  |
    |    Returns:                                                                  |
    |       True if valid, else False                                              |
    |------------------------------------------------------------------------------|*/
    RETURN dv.NUMBER != 0 && dv.PORT != 0
  }

  DEFINE_FUNCTION INTEGER fnIsInRangeDouble(DOUBLE val,
  DOUBLE low,
  DOUBLE high)
{
/*|------------------------------------------------------------------------------|
  | fnMathIsInRangeDouble: Is value between low and high inclusively             |
  |    Parameters:                                                               |
  |       val  Value to test                                                     |
  |       low  Range low                                                         |
  |       high Range high                                                        |
  |    Returns:                                                                  |
  |       True if in range, else False                                           |
  |------------------------------------------------------------------------------|*/

  RETURN (val >= low && val <= high)
}

DEFINE_FUNCTION INTEGER fnIsNotInRangeDouble(DOUBLE val,
  DOUBLE low,
  DOUBLE high)
{
/*|------------------------------------------------------------------------------|
  | fnIsNotInRangeDouble: Is value not between low and high inclusively          |
  |    Parameters:                                                               |
  |       val  Value to test                                                     |
  |       low  Range low                                                         |
  |       high Range high                                                        |
  |    Returns:                                                                  |
  |       True if not in range, else False                                       |
  |------------------------------------------------------------------------------|*/

  RETURN !fnIsInRangeDouble(val,low,high)
}

DEFINE_FUNCTION INTEGER fnIsInRangeFloat(FLOAT val,
  FLOAT low,
  FLOAT high)
{
/*|------------------------------------------------------------------------------|
  | fnIsInRangeFloat: Is value between low and high inclusively                  |
  |    Parameters:                                                               |
  |       val  Value to test                                                     |
  |       low  Range low                                                         |
  |       high Range high                                                        |
  |    Returns:                                                                  |
  |       True if in range, else False                                           |
  |------------------------------------------------------------------------------|*/

  RETURN (val >= low && val <= high)
}

DEFINE_FUNCTION INTEGER fnIsNotInRangeFloat(FLOAT val,
  FLOAT low,
  FLOAT high)
{
/*|------------------------------------------------------------------------------|
  | fnIsNotInRangeFloat: Is value not between low and high inclusively           |
  |    Parameters:                                                               |
  |       val  Value to test                                                     |
  |       low  Range low                                                         |
  |       high Range high                                                        |
  |    Returns:                                                                  |
  |       True if not in range, else False                                       |
  |------------------------------------------------------------------------------|*/

  RETURN !fnIsInRangeFloat(val,low,high)
}

DEFINE_FUNCTION INTEGER fnIsInRangeInteger(INTEGER val,
  INTEGER low,
  INTEGER high)
{
/*|------------------------------------------------------------------------------|
  | fnIsInRangeInteger: Is value between low and high inclusively                |
  |    Parameters:                                                               |
  |       val  Value to test                                                     |
  |       low  Range low                                                         |
  |       high Range high                                                        |
  |    Returns:                                                                  |
  |       True if in range, else False                                           |
  |------------------------------------------------------------------------------|*/

  RETURN (val >= low && val <= high)
}

DEFINE_FUNCTION INTEGER fnIsNotRangeInteger(INTEGER val,
  INTEGER low,
  INTEGER high)
{
/*|------------------------------------------------------------------------------|
  | fnIsNotRangeInteger: Is value not between low and high inclusively           |
  |    Parameters:                                                               |
  |       val  Value to test                                                     |
  |       low  Range low                                                         |
  |       high Range high                                                        |
  |    Returns:                                                                  |
  |       True if not in range, else False                                       |
  |------------------------------------------------------------------------------|*/

  RETURN !fnIsInRangeInteger(val,low,high)
}

DEFINE_FUNCTION INTEGER fnIsInRangeSInteger(SINTEGER val,
  SINTEGER low,
  SINTEGER high)
{
/*|------------------------------------------------------------------------------|
  | fnIsInRangeSInteger: Is value between low and high inclusively               |
  |    Parameters:                                                               |
  |       val  Value to test                                                     |
  |       low  Range low                                                         |
  |       high Range high                                                        |
  |    Returns:                                                                  |
  |       True if in range, else False                                           |
  |------------------------------------------------------------------------------|*/

  RETURN (val >= low && val <= high)
}

DEFINE_FUNCTION INTEGER fnIsNotRangeSInteger(SINTEGER val,
  SINTEGER low,
  SINTEGER high)
{
/*|------------------------------------------------------------------------------|
  | fnIsNotRangeSInteger: Is value not between low and high inclusively          |
  |    Parameters:                                                               |
  |       val  Value to test                                                     |
  |       low  Range low                                                         |
  |       high Range high                                                        |
  |    Returns:                                                                  |
  |       True if not in range, else False                                       |
  |------------------------------------------------------------------------------|*/

  RETURN !fnIsInRangeSInteger(val,low,high)
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnStringDEVListToDelimString(DEV  dvList[])
{
/*|------------------------------------------------------------------------------|
  | fnStringDEVListToDelimString: Create comma seperated DPS values              |
  |    Parameters:                                                               |
  |       dvList Array of devices                                                |
  |    Returns:                                                                  |
  |       Comma seperated string                                                 |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER idx
  STACK_VAR CHAR    cList[1000][20]
  FOR(idx = 1; idx <=  LENGTH_ARRAY(dvList); idx++)
  {
    cList[idx] = dpstoa(dvList[idx])
  }
  SET_LENGTH_ARRAY(cList,LENGTH_ARRAY(dvList))
  RETURN fnStringImplode(cList,',')
}

DEFINE_FUNCTION CHAR[STRING_RETURN_SIZE_LIMIT] fnSystemMACAddress()
{
/*|------------------------------------------------------------------------------|
  | fnSystemMACAddress: Get the Systems MAC Address                              |
  |    Returns:                                                                  |
  |       MAC Address string                                                     |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR    cMAC[6][3]
  fnStringExplode(',',fnStringToASCIIHex(GET_UNIQUE_ID()),cMAC,0)
  RETURN "fnStringPrefixToLength(fnStringReplace(cMAC[1],'$',NULL_STR),'0',2),':',
  fnStringPrefixToLength(fnStringReplace(cMAC[2],'$',NULL_STR),'0',2),':',
  fnStringPrefixToLength(fnStringReplace(cMAC[3],'$',NULL_STR),'0',2),':',
  fnStringPrefixToLength(fnStringReplace(cMAC[4],'$',NULL_STR),'0',2),':',
  fnStringPrefixToLength(fnStringReplace(cMAC[5],'$',NULL_STR),'0',2),':',
  fnStringPrefixToLength(fnStringReplace(cMAC[6],'$',NULL_STR),'0',2)"
}

DEFINE_START

  {
    MATH_NaN                            = fnMathBuildDouble($7FFFFFFF, $FFFFFFFF)
    MATH_POSITIVE_INFINITY               = fnMathBuildDouble($7FF00000, $00000000)
    MATH_NEGATIVE_INFINITY               = fnMathBuildDouble($FFF00000, $00000000)

    MATH_TWO_52                         = 1 << 52
  }

DEFINE_EVENT

  DATA_EVENT[devDebugCommand]
  {
    COMMAND:
    {
      STACK_VAR CHAR    cCmd[DUET_MAX_CMD_LEN]
      STACK_VAR CHAR    cHeader[DUET_MAX_HDR_LEN]
      STACK_VAR CHAR    cValue[DUET_MAX_HDR_LEN]

      cCmd                              = DATA.TEXT
      cHeader                           = DuetParseCmdHeader(cCmd)
      cValue                            = DuetParseCmdParam(cCmd)

      SWITCH(UPPER_STRING(cHeader))
      {
        CASE 'DEBUG':
        {
          IF(fnMathIsNumeric(cValue))
          {
            fnDebugSetLevel(ATOI(cValue))
          }

          ELSE IF(LENGTH_STRING(cValue)>1)
          {
            fnDebugSetLevelByString(cValue)
          }
        }

        CASE '?DEBUG':
        {
          fnIOPrintln("'current debug level ',fnDebugGetLevelString()")
        }
      }
    }
  }


#END_IF  // __ITG_LIBRARIES_HELPERS__

/*|------------------------------------------------------------------------------|
  |                                                                              |
  |      !!! END OF PROGRAM --- DO NOT PUT ANY CODE BELOW THIS COMMENT  !!!      |
  |                                                                              |
  |------------------------------------------------------------------------------|*/
