/*|------------------------------------------------------------------------------|
  |                   ITGCommon NetLinx Libraries v33 06/20/19                   |
  |------------------------------------------------------------------------------|*/

/*|------------------------------------------------------------------------------|
  |                            ITG Navigation Include                            |
  |                                                                              |
  | Copy following into NetLinx module before DEFINE_START or DEFINE_EVENTS      |
  |                                                                              |
  |   DEFINE_FUNCTION fnOnDeviceChange()                                         |
  |    {                                                                         |
  |       // NOP                                                                 |
  |    }                                                                         |
  |                                                                              |
  |    DEFINE_FUNCTION fnOnPageChange()                                          |
  |    {                                                                         |
  |       // NOP                                                                 |
  |    }                                                                         |
  |                                                                              |
  |    DEFINE_FUNCTION fnOnZoneChange()                                          |
  |    {                                                                         |
  |       // NOP                                                                 |
  |    }                                                                         |
  |                                                                              |
/*|------------------------------------------------------------------------------|
  |                            ITG Navigation Include                            |
  |                                                                              |
  | Must be included after defines of OnDeviceChanged(), OnPageChanged(),        |
  |       OnPageChanged(), OnZoneChange().                                       |
  |------------------------------------------------------------------------------|*/
  |                                                                              |
  |    #INCLUDE 'ITG Navigation Include'                                         |

  |                                                                              |
  |------------------------------------------------------------------------------|*/

/*|------------------------------------------------------------------------------|
  |                       ITG Navigation Include Functions                       |
  |                                                                              |
  | fnOnDeviceChanged: Notify the module that a device selection has occurred    |
  | fnOnPageChanged: Notify the module that a page flip has occurred             |
  | fnOnZoneChange: Notify the module that the zone has changed                  |
  |------------------------------------------------------------------------------|*/

PROGRAM_NAME = 'ITG Navigation Include'

  #IF_NOT_DEFINED __ITGNAVIGATIONINCLUDE__
  #DEFINE __ITGNAVIGATIONINCLUDE__

DEFINE_EVENT

  BUTTON_EVENT[dvTP, nNavigationBtns]
  {
    /*|------------------------------------------------------------------------------|
      |                       Get the currently ACTIVE device                        |
      |------------------------------------------------------------------------------|*/

    PUSH:
    {
      bActiveComponent                  = FALSE
      nActiveDevice                     = GET_LAST(nNavigationBtns);
      nActiveDeviceID                   = nNavigationBtns[nActiveDevice]

      nActivePage                       = 0
      nActivePageID                     = 0

    /*|------------------------------------------------------------------------------|
      | Disable the zone selection button                                            |
      |------------------------------------------------------------------------------|*/
      fnTPDisableButton( dvTP,ITOA(ZONE_PAGE))

    /*|------------------------------------------------------------------------------|
      | Notify the module that a device selection has occurred                       |
      |------------------------------------------------------------------------------|*/
      fnOnDeviceChange()

    /*|------------------------------------------------------------------------------|
      | This is a hack to clear all component popup pages                            |
      |------------------------------------------------------------------------------|*/
      fnTPHideAllPopupsOnPage( dvTP, 'Popup Pages')
    }
  }

  BUTTON_EVENT[dvTP, nSubNavBtns]
  {
    /*|------------------------------------------------------------------------------|
      |                        Get the currently ACTIVE page                         |
      |------------------------------------------------------------------------------|*/

    PUSH:
    {
      bActiveComponent                  = FALSE
      nActivePage                       = BUTTON.INPUT.CHANNEL;
      nActivePageID                     = fnArrayIntegerIndex(BUTTON.INPUT.CHANNEL,nSubNavBtns)

    /*|------------------------------------------------------------------------------|
      | If this is from the device we were expecting                                 |
      |------------------------------------------------------------------------------|*/

      IF(nActiveDevice == nDevice && fnArrayIsInIntegerIndex(BUTTON.INPUT.CHANNEL, nPages) && nActivePageID > 0)
      {
      /*|------------------------------------------------------------------------------|
        | Enable the Zone selection button                                             |
        |------------------------------------------------------------------------------|*/
        fnTPEnableButton( dvTP,ITOA(ZONE_PAGE))
        bActiveComponent                = TRUE

        fnChannelOn(dvTP, nActivePage)

      /*|------------------------------------------------------------------------------|
        | Notify the module that a page flip has occurred                              |
        |------------------------------------------------------------------------------|*/
        fnOnPageChange()
      }
    }
  }

  BUTTON_EVENT[dvTP, nZoneBtns]
  BUTTON_EVENT[dvTP,ZONE_PAGE]
  {
    /*|------------------------------------------------------------------------------|
      | Get the current zone                                                         |
      |------------------------------------------------------------------------------|*/

    PUSH:
    {

    /*|------------------------------------------------------------------------------|
      | If this zone was changed from the expected device                            |
      |------------------------------------------------------------------------------|*/

      IF(GET_LAST(nZoneBtns) && GET_LAST(nZoneBtns) <= LENGTH_ARRAY(vdvDev))
      {

      /*|------------------------------------------------------------------------------|
        | If zone changed from ZONE sub nav, update nCurrentZone for all components    |
        |------------------------------------------------------------------------------|*/

        IF(bActiveComponent || [dvTP,ZONE_PAGE])
        {
          nCurrentZone                  = GET_LAST(nZoneBtns)
        }
      }

    /*|------------------------------------------------------------------------------|
      |                           Zone page sub nav button                           |
      |------------------------------------------------------------------------------|*/

      IF(BUTTON.INPUT.CHANNEL == ZONE_PAGE)
      {
        fnChannelOn(dvTP, ZONE_PAGE)
      }

      nActiveZoneID                     = nZoneBtns[nCurrentZone]
      fnChannelOn(dvTP, nActiveZoneID)

      IF(bActiveComponent)
      {
      /*|------------------------------------------------------------------------------|
        | Disable the all zone buttons                                                 |
        |------------------------------------------------------------------------------|*/
        fnTPDisableButton( dvTP,"ITOA(nZoneBtns[1]),'..',ITOA(nZoneBtns[LENGTH_ARRAY(nZoneBtns)])")

      /*|------------------------------------------------------------------------------|
        | Enable only those available for the current component                        |
        |------------------------------------------------------------------------------|*/
        fnTPEnableButton( dvTP,"ITOA(nZoneBtns[1]),'..',ITOA(nZoneBtns[LENGTH_ARRAY(nZoneBtns)])")
      }

    /*|------------------------------------------------------------------------------|
      | Notify the module that the zone has changed                                  |
      |------------------------------------------------------------------------------|*/
      fnOnZoneChange()
    }
  }

/*|------------------------------------------------------------------------------|
  |                                                                              |
  |      !!! END OF PROGRAM --- DO NOT PUT ANY CODE BELOW THIS COMMENT  !!!      |
  |                                                                              |
  |------------------------------------------------------------------------------|*/
  #END_IF // __ITGNAVIGATIONINCLUDE__
