/*|------------------------------------------------------------------------------|
  |                   ITGCommon NetLinx Libraries v33 06/20/19                   |
  |------------------------------------------------------------------------------|*/

/*|------------------------------------------------------------------------------|
  |                             ITGComponentInclude                              |
  |                                                                              |
  | Handles page navagation and zone control                                     |
  |------------------------------------------------------------------------------|*/

/*|------------------------------------------------------------------------------|
  |                        ITGComponentInclude Functions                         |
  |                                                                              |
  | getFeedbackZone: Get the zone which matches the responding device            |
  |------------------------------------------------------------------------------|*/

PROGRAM_NAME='ITGComponentInclude'

#INCLUDE 'ITG Libraries Mainline TimeLine'
#INCLUDE 'ITG Libraries Helpers'
#INCLUDE 'SNAPI'
#INCLUDE 'G4API'

#IF_NOT_DEFINED __ITGCOMPONENTINCLUDE__
#DEFINE __ITGCOMPONENTINCLUDE__

DEFINE_VARIABLE

  VOLATILE INTEGER nDbg           = 1 // The current debug level

  VOLATILE CHAR bActiveComponent  = FALSE // Used to determine if the current component is ACTIVE or not

  VOLATILE INTEGER nActiveDevice  = 0 // The currently ACTIVE device number
  VOLATILE INTEGER nActiveDeviceID= 0 // The currently ACTIVE device button ID

  VOLATILE INTEGER nActivePage    = 0 // The currently ACTIVE page number
  VOLATILE INTEGER nActivePageID  = 0 // The currently ACTIVE page button ID

  VOLATILE INTEGER nCurrentZone   = 1 // The currently ACTIVE zone
  VOLATILE INTEGER nActiveZoneID  = 0 // The currently ACTIVE zone button ID
  VOLATILE INTEGER nZoneLoop      = 0 // Zone feedback loop

  VOLATILE integer bNoLevelReset = 0

DEFINE_CONSTANT

  INTEGER MAX_ZONE                = 20
  INTEGER ZONE_PAGE               = 550

  // List of device navigation buttons
  INTEGER nNavigationBtns[]                =
  {
    500,501,502,503,504,505,506,507,508,509,
    510,511,512,513,514,515,516,517,518,519
  }

  // List of page sub navigation buttons
  INTEGER nSubNavBtns[]                    =
  {
    520,521,522,523,524,525,526,527,528,529,
    530,531,532,533,534,535,536,537,538,539
  }

  INTEGER nZoneBtns[]                      =
  {
    901,902,903,904,905,906,907,908,909,910,
    911,912,913,914,915,916,917,918,919,920
  }

DEFINE_MUTUALLY_EXCLUSIVE

  ([dvTP,nSubNavBtns[1]]..[dvTP,nSubNavBtns[LENGTH_ARRAY(nSubNavBtns)]], [dvTP,ZONE_PAGE])
  ([dvTP,nZoneBtns[1]]..[dvTP,nZoneBtns[MAX_ZONE]])

DEFINE_FUNCTION INTEGER getFeedbackZone(dev dvIn)
{
/*|------------------------------------------------------------------------------|
  | getFeedbackZone: Get the zone which matches the responding device            |
  |    Parameters:                                                               |
  |       dvIn The responding device                                             |
  |    Returns:                                                                  |
  |       The zone which the dvIn's DPS is at, else 0 if not found               |
  |------------------------------------------------------------------------------|*/

  STACK_VAR INTEGER zone

  FOR(zone = 1; zone <= LENGTH_ARRAY(vdvDev); zone++)
  {
    IF(vdvDev[zone] == dvIn)
    {
      RETURN zone;
    }
  }

  RETURN 0;
}

DEFINE_START

  nActiveDevice                   = 1
  nActiveDeviceID                 = nNavigationBtns[1]
  nCurrentZone                    = 1
  nActiveZoneID                   = nZoneBtns[1]

DEFINE_EVENT

  DATA_EVENT[dvTP]
  {
    ONLINE:
    {
    /*|------------------------------------------------------------------------------|
      |                               Reset the panel                                |
      |------------------------------------------------------------------------------|*/

    /*|------------------------------------------------------------------------------|
      | This is a hack to clear all component popup pages                            |
      |------------------------------------------------------------------------------|*/
      fnTPHideAllPopupsOnPage( dvTP, 'Popup Pages')

    /*|------------------------------------------------------------------------------|
      | Disable the zone selection button                                            |
      |------------------------------------------------------------------------------|*/
      fnTPDisableButton( dvTP,ITOA(ZONE_PAGE))
    }
  }

  TIMELINE_EVENT[DEFINE_PROGRAM_TIMELINE_ID]
  {
    IF(bActiveComponent)
    {
      FOR(nZoneLoop = 1; nZoneLoop <= MAX_ZONE; nZoneLoop++)
      {
        [dvTP, nZoneBtns[nZoneLoop]]    =(nZoneLoop == nCurrentZone)
      }
    }
  }

#END_IF // __ITGCOMPONENTINCLUDE__

/*|------------------------------------------------------------------------------|
  |                                                                              |
  |      !!! END OF PROGRAM --- DO NOT PUT ANY CODE BELOW THIS COMMENT  !!!      |
  |                                                                              |
  |------------------------------------------------------------------------------|*/
