/*|------------------------------------------------------------------------------|
  |                   ITGCommon NetLinx Libraries v33 06/20/19                   |
  |------------------------------------------------------------------------------|*/

MODULE_NAME='ITGModuleComponent' (DEV vdvDev[], DEV dvTP, INTEGER nDevice, INTEGER nPages[],CHAR cDiscoverCommands[][])

#INCLUDE 'ITGComponentInclude'

DEFINE_CONSTANT

  BTN_QUERY_DEBUG                       = 1605
  BTN_REINITIALIZE                      = 1600
  BTN_REFRESH                           = 1609
  CH_DEBUG_LEVEL                        = 1610
  BTN_DISCOVER_PROPERTIES               = 1611
	TXT_CONNECTED_TO                      = '1612'

  CMD_REFRESH                           = 'REFRESH'
  INTEGER BTN_DEBUG[]                   =
  {
    1601, 1602, 1603, 1604
  }
  INTEGER BTN_DEBUG_FB[]                =
  {
    1601, 1602, 1603, 1604
  }
  TXT_DEBUG_LEVEL                       = 'Debug Level '

DEFINE_VARIABLE

  INTEGER bEchoCommands                 = FALSE
  CHAR cLastHeader[DUET_MAX_HDR_LEN]
  CHAR cLastDebugCommand[DUET_MAX_CMD_LEN]
	

/*|------------------------------------------------------------------------------|
  |                       ITG Navigation Include Functions                       |
  |                                                                              |
  | fnOnDeviceChanged: Notify the module that a device selection has occurred    |
  | fnOnPageChanged: Notify the module that a page flip has occurred             |
  | fnOnZoneChange: Notify the module that the zone has changed                  |
  |------------------------------------------------------------------------------|*/
DEFINE_FUNCTION fnOnDeviceChange()
{
  // NOP
}

DEFINE_FUNCTION fnOnPageChange()
{
  // NOP
}

DEFINE_FUNCTION fnOnZoneChange()
{
  // NOP
}

DEFINE_FUNCTION fnPopulateTPConnectedTo()
{		
	STACK_VAR SINTEGER slResult
	STACK_VAR IP_ADDRESS_STRUCT _IPAddress
	STACK_VAR CHAR txt[50]
	
	slResult = GET_IP_ADDRESS(0:0:0,_IPAddress)
	
	IF (slResult == 0)
	{
		txt = 'Connected to '
		
		IF (LENGTH_STRING(fnStringTrim(_IPAddress.HOSTNAME)) > 0 && fnStringNotEqualsIgnoreCase(_IPAddress.IPADDRESS,_IPAddress.HOSTNAME))
		{
			txt = "txt,_IPAddress.HOSTNAME,' (',_IPAddress.IPADDRESS,')'"
		}
		ELSE
		{
			txt = "txt,_IPAddress.IPADDRESS"
		}
		
		/*
		fnIOPrintln('================================================================================')
		fnIOPrintln("PRT_PREFIX,'IP_ADDRESS_STRUCT.FLAGS      [',_IPAddress.FLAGS,']'")
		fnIOPrintln("PRT_PREFIX,'IP_ADDRESS_STRUCT.GATEWAY    [',_IPAddress.GATEWAY,']'")
		fnIOPrintln("PRT_PREFIX,'IP_ADDRESS_STRUCT.HOSTNAME   [',_IPAddress.HOSTNAME,']'")
		fnIOPrintln("PRT_PREFIX,'IP_ADDRESS_STRUCT.IPADDRESS  [',_IPAddress.IPADDRESS,']'")
		fnIOPrintln("PRT_PREFIX,'IP_ADDRESS_STRUCT.SUBNETMASK [',_IPAddress.SUBNETMASK,']'")
		fnIOPrintln("PRT_PREFIX,'SETTING TP TEXT TO           [',txt,']'")
		fnIOPrintln('================================================================================')
		*/
	}
	
	ELSE
	{
		IF(slResult == -1)
		{
			txt = 'specified device is invalid or is not online'
		}

		ELSE IF(slResult == -2 )
		{
			txt = 'request timed out'
		}

		ELSE 
		{
			txt = "'return code ',ITOA(slResult)"
		}
		
		fnIOPrintln("PRT_PREFIX, 'GET_IP_ADDRESS failed with result ',txt")
		
		txt = NULL_STR
	}
	
	fnTPSetButtonTextAllStates(dvTP,TXT_CONNECTED_TO,txt)
	
	/*
	fnIOPrintln('================================================================================')
	fnIOPrintln("PRT_PREFIX,'Set Connected to label       [',txt,']'")
	fnIOPrintln('================================================================================')
	*/
}

#INCLUDE 'ITG Navigation Include'

DEFINE_MUTUALLY_EXCLUSIVE

  ([dvTp,BTN_DEBUG[1]]..[dvTp,BTN_DEBUG[LENGTH_ARRAY(BTN_DEBUG)]])

DEFINE_EVENT

  DATA_EVENT [dvTP]
  {
    ONLINE:
    {

      bACTIVEComponent                  = FALSE
      nACTIVEDevice                     = 1
      nACTIVEPage                       = 0
      nACTIVEDeviceID                   = nNavigationBtns[1]
      nACTIVEPageID                     = 0
      nCurrentZone                      = 1
      bNoLevelReset                     = 0
      fnDebugQueryDebugState(vdvDev[1])
			

			fnPopulateTPConnectedTo()
    }

    OFFLINE:
    {
      bNoLevelReset                     = 1
    }
  }

  DATA_EVENT[vdvDev]
  {
    ONLINE:
    {
      fnDebugQueryDebugState(vdvDev[1])
    }

    COMMAND:
    {
      IF(fnStringNotEquals(DATA.TEXT, cLastDebugCommand))
      {
        STACK_VAR CHAR    cCmd[DUET_MAX_CMD_LEN]

        cLastDebugCommand               = DATA.TEXT
        cCmd                            = DATA.TEXT

        IF(fnStringStartsWithIgnoreCase(DuetParseCmdHeader(cCmd),DEBUGSTATE_CMD))
        {
          STACK_VAR INTEGER nParameter

          nParameter                      = ATOI(DuetParseCmdParam(cCmd))

          IF(nParameter >= DEBUG_ERROR && nParameter <= DEBUG_DEBUG)
          {
            fnTPSetButTOnTextAllStates(dvTP,ITOA(CH_DEBUG_LEVEL),"TXT_DEBUG_LEVEL, fnDebugLevelString(nParameter)")
            fnChannelOn(dvTP,BTN_DEBUG_FB[nParameter])
            fnChannelOn(dvTP,CH_DEBUG_LEVEL)
          }
          ELSE
          {
            fnTPSetButTOnTextAllStates(dvTP,ITOA(CH_DEBUG_LEVEL),"TXT_DEBUG_LEVEL, '?'")
            fnChannelOff(dvTP,CH_DEBUG_LEVEL)
          }
        }
      }
    }
  }

  BUTTON_EVENT[dvTP, BTN_REINITIALIZE]
  {
    PUSH:
    {
      IF(bACTIVEComponent)
      {
        fnSendCommand(vdvDev[1], REINITIALIZE_CMD)
      }
    }
  }

  BUTTON_EVENT[dvTP, BTN_REFRESH]
  {
    PUSH:
    {
      fnSendCommand(vdvDev[1], CMD_REFRESH)
    }
  }

  BUTTON_EVENT[dvTP, BTN_QUERY_DEBUG]
  {
    PUSH:
    {
      IF(bACTIVEComponent)
      {
        fnDebugQueryDebugState(vdvDev[1])
      }
    }
  }

  BUTTON_EVENT[dvTp, BTN_DEBUG]
  {
    PUSH:
    {
      fnSendCommandEx( vdvDev[1], DEBUG_CMD,ITOA(GET_LAST(BTN_DEBUG)))
    }
  }

  DATA_EVENT[vdvDev[1]]
  {
    COMMAND:
    {
      STACK_VAR CHAR    cCmd[DUET_MAX_CMD_LEN]
      STACK_VAR CHAR    cHeader[DUET_MAX_HDR_LEN]

      cCmd                              = DATA.TEXT
      cHeader                           = DuetParseCmdHeader(cCmd)

      IF(bEchoCommands)
      {
        IF(fnStringNotEndsWithIgnoreCase(cHeader,DEBUGSTATE_CMD) && cHeader != cLastHeader)
        {
          IF(LENGTH_ARRAY(cLastHeader) > 0)
          {
            fnIOPrint(' ')
          }
          ELSE
          {
            fnIOPrint('>>>>>>>>>>>>>>>>>>>>>>>>>>> Module Query Results <<<<<<<<<<<<<<<<<<<<<<<<<<<')
          }
          cLastHeader                   = cHeader
        }
        fnIOPrintln("'    ', DATA.TEXT")
      }
    }
  }

  BUTTON_EVENT[dvTP, BTN_DISCOVER_PROPERTIES]
  {
    PUSH:
    {
      INTEGER idx
      bEchoCommands                     = TRUE
      cLastHeader                       = NULL_STR

      FOR(idx = 1; idx <= LENGTH_ARRAY(cDiscoverCommands); idx++)
      {
        fnSendCommand(vdvDev[1],cDiscoverCommands[idx])
      }
      WAIT 50
      {
        bEchoCommands                   = FALSE
        fnIOPrint('>>>>>>>>>>>>>>>>>>>>>>>>>>> Module Query Results <<<<<<<<<<<<<<<<<<<<<<<<<<<')
      }
    }
  }

  CHANNEL_EVENT[vdvDev[1],DEVICE_COMMUNICATING]
  {
    ON:
    {
      fnTPEnableButTOn(dvTP,ITOA(BTN_DISCOVER_PROPERTIES))
    }

    OFF:
    {
      fnTPDisableButTOn(dvTP,ITOA(BTN_DISCOVER_PROPERTIES))
    }
  }

  TIMELINE_EVENT[DEFINE_PROGRAM_TIMELINE_ID]
  {
    [dvTP,DEVICE_COMMUNICATING]         = [vdvDev[1],DEVICE_COMMUNICATING]
    [dvTP,DATA_INITIALIZED]             = [vdvDev[1],DATA_INITIALIZED]
  }
