/*|------------------------------------------------------------------------------|
  |                               SNAPI Constants                                |
  |                                                                              |
  |------------------------------------------------------------------------------|*/

/*|------------------------------------------------------------------------------|
  |                          SNAPI Constants Functions                           |
  |                                                                              |
  | DuetPackCmdHeader: Package header for module SEND_COMMAND or SEND_STRING     |
  |                                                                              |
  | Note: Adds the command header to the string and adds the command if missing. |
  |       This function assumes the standard Duet command separator '-'          |
  |                                                                              |
  | DuetPackCmdParam: Package parameter for module SEND_COMMAND or SEND_STRING   |
  |                                                                              |
  | Note: Wraps the parameter in double-quotes if it contains the separator.     |
  |       This function assumes the standard Duet parameter separator ','        |
  |                                                                              |
  | DuetPackCmdParamArray: To package parameters for module SEND_COMMAND or      |
  |       SEND_STRING                                                            |
  |                                                                              |
  |    Notes : Wraps the parameter in double-quotes if it contains the           |
  |       separator and separates them using the separator sequence This         |
  |       function assumes the standard Duet parameter separator ','             |
  |                                                                              |
  | DuetPackCmdSimple: To package header and 1 parameter for module SEND_COMMAND |
  |       or SEND_STRING                                                         |
  |                                                                              |
  |    Notes : Adds the command header to the string and adds the command if     |
  |       missing. This function assumes the standard Duet command separator     |
  |       '-'. This function also adds a parameter to the command                |
  |                                                                              |
  | DuetParseCmdHeader: Parse out parameters from module SEND_COMMAND or         |
  |       SEND_STRING                                                            |
  |                                                                              |
  |    Notes : Parses the strings sent to or from modules extracting the command |
  |       header. Command separating CHARacter assumed to be '-', Duet standard  |
  |                                                                              |
  | DuetParseCmdParam: Parse out parameters from module SEND_COMMAND or          |
  |       SEND_STRING                                                            |
  |                                                                              |
  |    Notes : Parses the strings sent to or from modules extracting the         |
  |       parameters. A single param is picked of the cmd string and removed,    |
  |       through the separator. The separator is not returned from the          |
  |       function. If the first character of the param is a double quote, the   |
  |       function will remove up to (and including) the next double-quote and   |
  |       the separator without spaces. The double quotes will then be stripped  |
  |       from the parameter before it is returned. If the                       |
  |       double-quote/separator sequence is not found, the function will remove |
  |       up to (and including) the separator CHARacter and the leading double   |
  |       quote will not be removed. If the separator is not found, the entire   |
  |       remained of the command is removed. Command separating character       |
  |       assumed to be ',', Duet standard                                       |
  |                                                                              |
  |------------------------------------------------------------------------------|*/

PROGRAM_NAME='SNAPIConstants'

#IF_NOT_DEFINED __SNAPI_CONST__
#DEFINE __SNAPI_CONST__

DEFINE_CONSTANT
/*|------------------------------------------------------------------------------|
  |                                SNAPI Version                                 |
  |------------------------------------------------------------------------------|*/

  CHAR SNAPI_AXI_VERSION[]              =	'1.14 ITG'

/*|------------------------------------------------------------------------------|
  |                                Invalid Values                                |
  |------------------------------------------------------------------------------|*/

  LONG INVALID_NUMBER                   =	$80000000
  CHAR INVALID_STRING[]                 =	''
  CHAR INVALID_ENUM[]                   =	'INVALID'

/*|------------------------------------------------------------------------------|
  |                         Limits/Command/Param Lengths                         |
  |------------------------------------------------------------------------------|*/

  DUET_MIN_LEVEL_VALUE                  =	0
  DUET_MAX_LEVEL_VALUE                  =	255
  DUET_MAX_CMD_LEN                      = 1000                          // Maximum command length for parsing/packing functions
  DUET_MAX_PARAM_LEN                    = 100                           // Maximum parameter length for parsing/packing functions
  DUET_MAX_HDR_LEN                      = 100                           // Maximum command header length for parsing/packing functions

/*|------------------------------------------------------------------------------|
  |                          Status Channels : Channels                          |
  |------------------------------------------------------------------------------|*/

  DEVICE_COMMUNICATING                  = 251                           // Feedback:  Device online event
  DATA_INITIALIZED                      = 252                           // Feedback:  Data initialized event
  LAMP_WARMING_FB                       = 253                           // Feedback:  Lamp is warming up
  LAMP_COOLING_FB                       = 254                           // Feedback:  Lamp is cooling down
  POWER_ON                              = 255                           // Discrete:  Set lamp power state
  POWER_FB                              = 255                           // Feedback:  Power feedback
  LAMP_POWER_FB                         = 255                           // Feedback:  Lamp power state feedback

DEFINE_VARIABLE
/*|------------------------------------------------------------------------------|
  |                             Ramping Repeat Count                             |
  |------------------------------------------------------------------------------|*/

  INTEGER DUET_RAMPING_REPEAT           =	3

DEFINE_CONSTANT	//	----- Amplifier : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle power
  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback

  VOL_LVL                               = 1                             // Level: Volume level (0-255)

  #IF_DEFINED __AMPLIFIER_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
  #END_IF // __AMPLIFIER_COMMANDS__

DEFINE_CONSTANT	//	----- Audio Conferencer : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  ACONF_PRIVACY                         = 145                           // Momentary: Cycle privacy
  ACONF_PRIVACY_ON                      = 146                           // Discrete:  Set privacy on or off
  ACONF_PRIVACY_FB                      = 146                           // Feedback:  Privacy feedback
  ACONF_TRAIN                           = 147                           // Momentary: Execute train
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback
  DIAL_REDIAL                           = 201                           // Momentary: Redial
  DIAL_OFF_HOOK                         = 202                           // Momentary: Cycle off hook state
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  DIAL_AUTO_ANSWER                      = 204                           // Momentary: Cycle auto answer state
  DIAL_AUDIBLE_RING                     = 205                           // Momentary: Cycle audible ring state
  DIAL_FLASH_HOOK                       = 208                           // Momentary: Flash hook
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay
  DIAL_OFF_HOOK_ON                      = 238                           // Discrete:  Set the dialer off or on hook
  DIAL_OFF_HOOK_FB                      = 238                           // Feedback:  Dialer's hook state feedback
  DIAL_AUTO_ANSWER_ON                   = 239                           // Discrete:  Set auto answer on or off
  DIAL_AUTO_ANSWER_FB                   = 239                           // Feedback:  Auto answer state feedback
  DIAL_AUDIBLE_RING_ON                  = 240                           // Discrete:  Set the audible ring state on or off
  DIAL_AUDIBLE_RING_FB                  = 240                           // Feedback:  Audible ring state feedback

  VOL_LVL                               = 1                             // Level: Volume level (0-255)

  #IF_DEFINED __AUDIOCONFERENCER_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    CLOSESEARCHDB_CMD                   = 'PHONEBOOKCLOSESEARCH'        // closeSearchDB, 'SPEEDDIALCLOSESEARCH-<key>'
    DBNEXT_QRY                          = 'PHONEBOOKNEXT'               // queryDBNext, 'SPEEDDIALNEXT-<key>[,<count>]'
    DBPREVIOUS_QRY                      = 'PHONEBOOKPREV'               // queryDBPrevious, 'SPEEDDIALPREV-<key>[,<count>]'
    DELETERECORD_CMD                    = 'PHONEBOOKDELETE'             // deleteRecord, 'SPEEDDIALDELETE-<id>'
    DIAL1_CMD                           = 'DIALID'                      // dial, 'DIALID-<id>'
    DIALDTMF_CMD                        = 'DTMF'                        // dialDTMF, 'DTMF-<digit>'
    DIALERSTATUS_CMD                    = 'DIALERSTATUS'                // Dialer Status, 'DIALERSTATUS'
    DIALERSTATUS_QRY                    = '?DIALERSTATUS'               // getDialerStatus, '?DIALERSTATUS'
    DIALNUMBER_CMD                      = 'DIALNUMBER'                  // dialNumber, 'DIALNUMBER-<number>'
    DIAL_CMD                            = 'DIALINDEX'                   // dial, 'DIALINDEX-<index>'
    INCOMINGCALL_CMD                    = 'INCOMINGCALL'                // incoming call, 'INCOMINGCALL'
    PHONEBOOKCAPACITY_CMD               = 'PHONEBOOKCAPACITY'           // PHONEBOOKCAPACITTY-<capacity>
    PHONEBOOKCAPACITY_QRY               = '?PHONEBOOKCAPACITY'          // getPhonebookCapacity, '?SPEEDDIALCAPACITY'
    PHONEBOOKRECORD_CMD                 = 'PHONEBOOKRECORD'             // PHONEBOOKRECORD
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    REFRESH_CMD                         = 'PHONEBOOKREFRESH'            // queryDB, 'PHONEBOOKREFESH-<key>[,<count>]'
    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    SEARCHDB_CMD                        = 'PHONEBOOKSEARCH'             // searchDB, 'SPEEDDIALSEARCH-<key>,ID=<id>'
    UPDATERECORD_CMD                    = 'PHONEBOOKUPDATE'             // updateRecord, 'SPEEDDIALUPDATE-<id>,<name>,<number>'
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
  #END_IF // __AUDIOCONFERENCER_COMMANDS__

DEFINE_CONSTANT	//	----- Audio Mixer : Channels and Levels

  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  GAIN_UP                               = 140                           // Ramping:   Ramp gain up
  GAIN_UP_FB                            = 140                           // Feedback:  Gain ramping up feedback
  GAIN_DN                               = 141                           // Ramping:   Ramp gain down
  GAIN_DN_FB                            = 141                           // Feedback:  Gain ramping down feedback
  GAIN_MUTE_ON                          = 143                           // Discrete:  Set gain mute on
  GAIN_MUTE_FB                          = 143                           // Feedback:  Gain mute feedback
  GAIN_MUTE                             = 144                           // Momentary: Cycle gain mute
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback

  VOL_LVL                               = 1                             // Level: Volume level (0-255)
  GAIN_LVL                              = 5                             // Level: Gain level (0-255)

  #IF_DEFINED __AUDIOMIXER_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    AUDIOMIXERCROSSPOINTMUTEON_CMD      = 'XPOINTMUTE'                  // setAudioMixerCrosspointMuteOn, 'XPOINTMUTE-<state>,<input>,<output>'
    AUDIOMIXERCROSSPOINTMUTEON_QRY      = '?XPOINTMUTE'                 // isAudioMixerCrosspointMuteOn, '?XPOINTMUTE-<input>,<output>'
    AUDIOMIXERCROSSPOINT_CMD            = 'XPOINT'                      // setAudioMixerCrosspoint, 'XPOINT-<value>,<input>,<output>[,<output>,<output>]'
    AUDIOMIXERCROSSPOINT_QRY            = '?XPOINT'                     // getAudioMixerCrosspoint, '?XPOINT-<input>,<output>'
    AUDIOMIXERPRESET_CMD                = 'MIXERPRESET'                 // setAudioMixerPreset, 'MIXERPRESET-<preset>'
    AUDIOMIXERPRESET_QRY                = '?MIXERPRESET'                // getAudioMixerPreset, '?MIXERPRESET'
    INPUT_QRY                           = '?INPUT'                      // getInput, '?INPUT-<sl>,<output>'
    OUTPUT_QRY                          = '?OUTPUT'                     // getOutput, '?OUTPUT-<sl>,<input>'
    SAVEAUDIOMIXERPRESET_CMD            = 'MIXERPRESETSAVE'             // saveAudioMixerPreset, 'MIXERPRESETSAVE-<preset>'
    SAVESWITCHERPRESET_CMD              = 'SWITCHPRESETSAVE'            // saveSwitcherPreset, 'SWICTHPRESETSAVE-<preset>'
    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    SWITCHERPRESET_CMD                  = 'SWITCHPRESET'                // setSwitcherPreset, 'SWITCHPRESET-<preset>'
    SWITCHERPRESET_QRY                  = '?SWITCHPRESET'               // getSwitcherPreset, '?SWITCHPRESET'
    SWITCHINPUTTOOUTPUT1_CMD            = 'AI<input>O<output,output,>'
    SWITCHINPUTTOOUTPUT2_CMD            = 'CL<sl>I<input>O<output,output>'
    SWITCHINPUTTOOUTPUT3_CMD            = 'VI<input>O<output,output,>'
    SWITCHINPUTTOOUTPUT_CMD             = 'CI<input>O<output,output>'
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
  #END_IF // __AUDIOMIXER_COMMANDS__

DEFINE_CONSTANT	//	----- Audio Processor : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle power
  AUDIOPROC_LEVEL_UP                    = 24                            // Ramping:   Increment the audio processor level
  AUDIOPROC_LEVEL_DN                    = 25                            // Ramping:   Decrement the audio processor level
  AUDIOPROC_STATE                       = 26                            // Momentary: Cycle the audio processor state
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  AUDIOPROC_STATE_ON                    = 199                           // Discrete:  Set the audio processor state on or off
  AUDIOPROC_STATE_FB                    = 199                           // Feedback:  Audio processor state feedback
  AUDIOPROC_PRESET                      = 209                           // Momentary: Cycle the audio processor preset

  AUDIOPROC_LVL                         = 1                             // Level: Audio processor level (0-255)

  #IF_DEFINED __AUDIOPROCESSOR_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADDAUDIOPROCESSORCOMPONENT_CMD      = 'AUDIOPROCADD'                // addAudioProcessorComponent, 'AUDIOPROCADD-<index>,<address>'
    AUDIOPROCESSORCOMPONENTADDRESS_QRY  = '?AUDIOPROCADDR'              // getAudioProcessorComponentAddress, '?AUDIOPROCADDR-<index>'
    AUDIOPROCESSORCOMPONENTINDEX_QRY    = '?AUDIOPROCIDX'               // getAudioProcessorComponentIndex, '?AUDIOPROCIDX-<address>'
    AUDIOPROCESSORCROSSPOINTMUTEON_CMD  = 'XPOINTMUTE'                  // setAudioProcessorCrosspointMuteOn, 'XPOINTMUTE-<state>,<input>,<output>'
    AUDIOPROCESSORCROSSPOINTMUTEON_QRY  = '?XPOINTMUTE'                 // isAudioProcessorCrosspointMuteOn, '?XPOINTMUTE-<input>,<output>'
    AUDIOPROCESSORCROSSPOINT_CMD        = 'XPOINT'                      // setAudioProcessorCrosspoint, 'XPOINT-<value>,<input>,<output>[,<output>,<output>]'
    AUDIOPROCESSORCROSSPOINT_QRY        = '?XPOINT'                     // getAudioProcessorCrosspoint, '?XPOINT-<input>,<output>'
    AUDIOPROCESSORINPUT_QRY             = '?INPUT'                      // getAudioProcessorInput, '?INPUT-<output>'
    AUDIOPROCESSOROUTPUT_QRY            = '?OUTPUT'                     // getAudioProcessorOutput, '?OUTPUT-<input>'
    AUDIOPROCESSORPRESET_CMD            = 'AUDIOPROCPRESET'             // setAudioProcessorPreset, 'AUDIOPROCPRESET-<preset>'
    AUDIOPROCESSORPRESET_QRY            = '?AUDIOPROCPRESET'            // getAudioProcessorPreset, '?AUDIOPROCPRESET'
    REMOVEAUDIOPROCESSORCOMPONENT1_CMD  = 'AUDIOPROCREMOVEIDX'          // removeAudioProcessorComponent, 'AUDIOPROCREMOVEIDX-<index>'
    REMOVEAUDIOPROCESSORCOMPONENT_CMD   = 'AUDIOPROCREMOVEADDR'         // removeAudioProcessorComponent, 'AUDIOPROCREMOVEADDR-<address>'
    SAVEAUDIOPROCESSORPRESET_CMD        = 'AUDIOPROCPRESETSAVE'         // saveAudioProcessorPreset, 'AUDIOPROCPRESETSAVE-<preset>'
    SWITCHAUDIOPROCESSORINPUTTOOUTPUT_CMD  = 'AI<input>O<output,output,>'
    SWITCH_LAUDIOI_CMD                  = 'SWITCH-LAUDIOI'
  #END_IF // __AUDIOPROCESSOR_COMMANDS__

DEFINE_CONSTANT	//	----- Audio Tape : Channels

  PLAY                                  = 1                             // Momentary: Play
  STOP                                  = 2                             // Momentary: Stop
  PAUSE                                 = 3                             // Momentary: Pause
  FFWD                                  = 4                             // Momentary: Fast forward
  REW                                   = 5                             // Momentary: Rewind
  SFWD                                  = 6                             // Momentary: Search forward
  SREV                                  = 7                             // Momentary: Search reverse
  RECORD                                = 8                             // Momentary: Record
  POWER                                 = 9                             // Momentary: Cycle power
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  CASS_REVERSE_PLAY                     = 41                            // Momentary: Reverse play
  CASS_TAPE_SIDE                        = 42                            // Momentary: Cycle the audio tape side
  SEARCH_SPEED                          = 119                           // Momentary: Cycle search speed
  EJECT                                 = 120                           // Momentary: Eject tape
  RESET_COUNTER                         = 121                           // Momentary: Reset counter
  TAPE_LOADED_FB                        = 122                           // Feedback:  Tape is loaded
  RECORD_LOCK_FB                        = 123                           // Feedback:  Tape record is locked
  CASS_TAPE_SIDE_A                      = 126                           // Momentary: Set the tape side to A
  CASS_TAPE_SIDE_A_FB                   = 126                           // Feedback:  Tape side set to A feedback
  CASS_TAPE_SIDE_B                      = 127                           // Momentary: Set the tape side to B
  CASS_TAPE_SIDE_B_FB                   = 127                           // Feedback:  Tape side set to B feedback
  CASS_RECORD_MUTE                      = 128                           // Momentary: Cycle record mute
  SLOW_FWD                              = 188                           // Momentary: Slow forward
  SLOW_REV                              = 189                           // Momentary: Slow reverse
  CASS_RECORD_MUTE_ON                   = 200                           // Discrete:  Set the record mute on or off
  CASS_RECORD_MUTE_FB                   = 200                           // Feedback:  Record mute feedback
  PLAY_FB                               = 241                           // Feedback:  Play feedback
  STOP_FB                               = 242                           // Feedback:  Stop feedback
  PAUSE_FB                              = 243                           // Feedback:  Pause feedback
  FFWD_FB                               = 244                           // Feedback:  Fast forward feedback
  REW_FB                                = 245                           // Feedback:  Rewind feedback
  SFWD_FB                               = 246                           // Feedback:  Search forward feedback
  SREV_FB                               = 247                           // Feedback:  Search reverse feedback
  RECORD_FB                             = 248                           // Feedback:  Record feedback
  SLOW_FWD_FB                           = 249                           // Feedback:  Slow forward feedback
  SLOW_REV_FB                           = 250                           // Feedback:  Slow reverse feedback

  #IF_DEFINED __AUDIOTAPE_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    TAPECOUNTERNOTIFICATION_CMD         = 'TAPECOUNTERNOTIFY'           // setTapeCounterNotification, 'TAPECOUNTERNOTIFY-<state>'
  #END_IF // __AUDIOTAPE_COMMANDS__

DEFINE_CONSTANT	//	----- Audio Tuner Device : Channels

  POWER                                 = 9                             // Momentary: Cycle power
  CHAN_UP                               = 22                            // Momentary: Next station preset
  CHAN_DN                               = 23                            // Momentary: Previous station preset
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  TUNER_BAND                            = 40                            // Momentary: Cycle tuner band
  TUNER_PRESET_GROUP                    = 224                           // Momentary: Cycle station preset group
  TUNER_STATION_UP                      = 225                           // Momentary: Increment tuner station
  TUNER_STATION_DN                      = 226                           // Momentary: Decrement tuner station
  TUNER_SCAN_FWD                        = 227                           // Momentary: Station scan forward
  TUNER_SCAN_REV                        = 228                           // Momentary: Station scan backward/reverse
  TUNER_SEEK_FWD                        = 229                           // Momentary: Station seek forward
  TUNER_SEEK_REV                        = 230                           // Momentary: Station seek backward/reverse
  TUNER_OSD                             = 234                           // Momentary: Cycle on-screen or front panel display info
  TUNER_PREV                            = 235                           // Momentary: Goto previous tuner station

  #IF_DEFINED __AUDIOTUNERDEVICE_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    BAND_CMD                            = 'BAND'                        // setBand, 'BAND-<tb>'
    BAND_QRY                            = '?BAND'                       // getBand, '?BAND'
    STATIONPRESETCOUNT_CMD              = 'STATIONPRESETCOUNT'          // 'STATIONPRESETCOUNT-<count>'
    STATIONPRESETCOUNT_QRY              = '?STATIONPRESETCOUNT'         // getStationPresetCount, '?STATIONPRESETCOUNT'
    STATIONPRESETPROPERTIES_CMD         = 'STATIONPRESETPROPERTIES'     // '[STATIONPRESETPROPERTIES-<index>,<displayName>,<station>]*'
    STATIONPRESETPROPERTIES_QRY         = '?STATIONPRESETPROPERTIES'    // getStationPresetProperties, '?STATIONPRESETPROPERTIES'
    STATIONPRESETPROPERTY_CMD           = 'STATIONPRESETPROPERTY'       // 'STATIONPRESETPROPERTY-<index>,<displayName>,<station>'
    STATIONPRESETPROPERTY_QRY           = '?STATIONPRESETPROPERTY'      // getStationPresetProperty(index), '?STATIONPRESETPROPERTY-<index>'
    STATIONPRESETSELECT_CMD             = 'STATIONPRESETSELECT'         // 'STATIONPRESETSELECT-<index>'
    STATIONPRESETSELECT_QRY             = '?STATIONPRESETSELECT'        // getStationPresetSelect, '?STATIONPRESETSELECT'
    STATIONPRESET_CMD                   = 'TUNERPRESET'                 // setStationPreset, 'TUNERPRESET-<preset>'
    STATIONPRESET_QRY                   = '?TUNERPRESET'                // getStationPreset, '?TUNERPRESET'
    STATION_CMD                         = 'XCH'                         // setStation, 'XCH-<station>'
    STATION_QRY                         = '?XCH'                        // getStation, '?XCH'
    TUNERBANDPROPERTIES_CMD             = 'TUNERBANDPROPERTIES'         // '[TUNERBANDPROPERTIES-<index>,<displayName>,<tunerBand>]*'
    TUNERBANDPROPERTIES_QRY             = '?TUNERBANDPROPERTIES'        // getTunerBandProperties, '?TUNERBANDPROPERTIES'
    TUNERBANDPROPERTYCOUNT_CMD          = 'TUNERBANDPROPERTYCOUNT'      // 'TUNERBANDPROPERTYCOUNT-<count>'
    TUNERBANDPROPERTYCOUNT_QRY          = '?TUNERBANDPROPERTYCOUNT'     // getTunerBandPropertyCount, '?TUNERBANDPROPERTYCOUNT'
    TUNERBANDPROPERTY_CMD               = 'TUNERBANDPROPERTY'           // 'TUNERBANDPROPERTY-<index>,<displayName>,<tunerBand>'
    TUNERBANDPROPERTY_QRY               = '?TUNERBANDPROPERTY'          // getTunerBandProperty(index), '?TUNERBANDPROPERTY-<index>'
    TUNERBANDSELECT_CMD                 = 'TUNERBANDSELECT'             // 'TUNERBANDSELECT-<index>'
    TUNERBANDSELECT_QRY                 = '?TUNERBANDSELECT'            // getTunerBandSelect, '?TUNERBANDSELECT'
    TUNERCOMPONENTPROPERTIES_CMD        = 'TUNERCOMPONENTPROPERTIES'    // '[TUNERCOMPONENTPROPERTIES-<index>,<displayName>,<value>]*'
    TUNERCOMPONENTPROPERTIES_QRY        = '?TUNERCOMPONENTPROPERTIES'   // getTunerComponentProperties, '?TUNERCOMPONENTPROPERTIES'
    TUNERCOMPONENTPROPERTYCOUNT_CMD     = 'TUNERCOMPONENTPROPERTYCOUNT' // 'TUNERCOMPONENTPROPERTYCOUNT-<count>'
    TUNERCOMPONENTPROPERTYCOUNT_QRY     = '?TUNERCOMPONENTPROPERTYCOUNT'// getTunerComponentPropertyCount, '?TUNERCOMPONENTPROPERTYCOUNT'
    TUNERCOMPONENTPROPERTY_CMD          = 'TUNERCOMPONENTPROPERTY'      // 'TUNERCOMPONENTPROPERTY-<index>,<displayName>,<value>'
    TUNERCOMPONENTPROPERTY_QRY          = '?TUNERCOMPONENTPROPERTY'     // getTunerComponentProperty(index), '?TUNERCOMPONENTPROPERTY-<index>'
    TUNERCOMPONENTSELECT_CMD            = 'TUNERCOMPONENTSELECT'        // 'TUNERCOMPONENTSELECT-<index>'
    TUNERCOMPONENTSELECT_QRY            = '?TUNERCOMPONENTSELECT'       // getTunerComponentSelect, '?TUNERCOMPONENTSELECT'
  #END_IF // __AUDIOTUNERDEVICE_COMMANDS__

DEFINE_CONSTANT	//	----- Camera : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle power
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  TILT_UP                               = 132                           // Ramping:   Ramp tilt up
  TILT_UP_FB                            = 132                           // Feedback:  Tilt ramp up feedback
  TILT_DN                               = 133                           // Ramping:   Ramp tilt down
  TILT_DN_FB                            = 133                           // Feedback:  Tilt ramp down feedback
  PAN_LT                                = 134                           // Ramping:   Ramp pan left
  PAN_LT_FB                             = 134                           // Feedback:  Ramp pan left feedback
  PAN_RT                                = 135                           // Ramping:   Ramp pan right
  PAN_RT_FB                             = 135                           // Feedback:  Ramp pan right feedback
  ZOOM_OUT                              = 158                           // Ramping:   Ramp zoom out
  ZOOM_OUT_FB                           = 158                           // Feedback:  Ramp zoom out feedback
  ZOOM_IN                               = 159                           // Ramping:   Ramp zoom in
  ZOOM_IN_FB                            = 159                           // Feedback:  Ramp zoom in feedback
  FOCUS_NEAR                            = 160                           // Ramping:   Ramp focus near
  FOCUS_NEAR_FB                         = 160                           // Feedback:  Ramp focus near feedback
  FOCUS_FAR                             = 161                           // Ramping:   Ramp focus far
  FOCUS_FAR_FB                          = 161                           // Feedback:  Focus ramp far feedback
  AUTO_FOCUS_ON                         = 162                           // Discrete:  Set auto focus on or off
  AUTO_FOCUS_FB                         = 162                           // Feedback:  Auto focus feedback
  AUTO_IRIS_ON                          = 163                           // Discrete:  Set auto iris on or off
  AUTO_IRIS_FB                          = 163                           // Feedback:  Auto iris feedback
  AUTO_FOCUS                            = 172                           // Momentary: Cycle auto focus
  AUTO_IRIS                             = 173                           // Momentary: Cycle auto iris
  IRIS_OPEN                             = 174                           // Ramping:   Ramp iris open
  IRIS_OPEN_FB                          = 174                           // Feedback:  Ramp iris open feedback
  IRIS_CLOSE                            = 175                           // Ramping:   Ramp iris closed
  IRIS_CLOSE_FB                         = 175                           // Feedback:  Ramp iris closed feedback
  CAM_PRESET                            = 177                           // Momentary: Cycle camera preset

  ZOOM_LVL                              = 15                            // Level: Zoom position ( 0=out/Wide, 255= in/Tele)
  FOCUS_LVL                             = 16                            // Level: Focus position (0=near, 255=far)
  IRIS_LVL                              = 17                            // Level: Iris position (0=closed, 255=open)
  ZOOM_SPEED_LVL                        = 18                            // Level: Zoom ramp speed (0-255)
  FOCUS_SPEED_LVL                       = 19                            // Level: Focus ramp speed (0-255)
  IRIS_SPEED_LVL                        = 20                            // Level: Iris ramp speed (0-255)
  PAN_LVL                               = 27                            // Level: Pan position (0=left,255=right)
  TILT_LVL                              = 28                            // Level: Tilt position (0=down, 255=up)
  PAN_SPEED_LVL                         = 29                            // Level: Pan ramp speed (0-255)
  TILT_SPEED_LVL                        = 30                            // Level: Tilt ramp speed (0-255)

  #IF_DEFINED __CAMERA_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    CAMERACOMPONENTCOUNT_CMD            = 'CAMERACOMPONENTCOUNT'        // CAMERACOMPONENTCOUNT-<count>'
    CAMERACOMPONENTCOUNT_QRY            = '?CAMERACOMPONENTCOUNT'       // getCameraComponentCount, '?CAMERACOMPONENTCOUNT'
    CAMERAPRESETCOUNT_CMD               = 'CAMERAPRESETCOUNT'           // CAMERAPRESETCOUNT-<count>'
    CAMERAPRESETCOUNT_QRY               = '?CAMERAPRESETCOUNT'          // getCameraPresetCount, '?CAMERAPRESETCOUNT'
    CAMERAPRESETPROPERTIES_CMD          = 'CAMERAPRESETPROPERTIES'      // [CAMERAPRESETPROPERTIES-<index>,<displayName>]*'
    CAMERAPRESETPROPERTIES_QRY          = '?CAMERAPRESETPROPERTIES'     // getCameraPresetProperties, '?CAMERAPRESETPROPERTIES'
    CAMERAPRESETPROPERTY_CMD            = 'CAMERAPRESETPROPERTY'        // CAMERAPRESETPROPERTY-<index>,<displayName>'
    CAMERAPRESETPROPERTY_QRY            = '?CAMERAPRESETPROPERTY'       // getCameraPresetProperty, '?CAMERAPRESETPROPERTY-<index>'
    CAMERAPRESET_CMD                    = 'CAMERAPRESET'                // CAMERAPRESET-<index>
    CAMERAPRESET_CMD                    = 'CAMERAPRESET'                // setCameraPreset, 'CAMERAPRESET-<preset>'
    CAMERAPRESET_QRY                    = '?CAMERAPRESET'               // getCameraPreset, '?CAMERAPRESET'
    SAVECAMERAPRESET_CMD                = 'CAMERAPRESETSAVE'            // saveCameraPreset, 'CAMERAPRESETSAVE-<preset>'
  #END_IF // __CAMERA_COMMANDS__

DEFINE_CONSTANT	//	----- Digital Media Decoder : Channels and Levels

  PLAY                                  = 1                             // Momentary: Play
  STOP                                  = 2                             // Momentary: Stop
  PAUSE                                 = 3                             // Momentary: Pause
  FFWD                                  = 4                             // Momentary: Goto the next track
  REW                                   = 5                             // Momentary: Goto the previous track
  SFWD                                  = 6                             // Momentary: Scan forward
  SREV                                  = 7                             // Momentary: Scan reverse
  RECORD                                = 8                             // Momentary: Record
  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MEDIA_RANDOM                          = 124                           // Momentary: Cycle the random state
  MEDIA_REPEAT                          = 125                           // Momentary: Cycle the repeat state
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  MEDIA_RANDOM_DISC_ON                  = 178                           // Momentary: Set the random mode to disc
  MEDIA_RANDOM_DISC_FB                  = 178                           // Feedback:  Random mode disc feedback
  MEDIA_RANDOM_ALL_ON                   = 179                           // Momentary: Set the random mode to all
  MEDIA_RANDOM_ALL_FB                   = 179                           // Feedback:  Random mode all feedback
  MEDIA_RANDOM_OFF_ON                   = 180                           // Momentary: Set the random mode to off
  MEDIA_RANDOM_OFF_FB                   = 180                           // Feedback:  Random mode off feedback
  MEDIA_REPEAT_DISC_ON                  = 181                           // Momentary: Set the repeat mode to disc
  MEDIA_REPEAT_DISC_FB                  = 181                           // Feedback:  Repeat mode disc feedback
  MEDIA_REPEAT_TRACK_ON                 = 182                           // Momentary: Set the repeat mode to track
  MEDIA_REPEAT_TRACK_FB                 = 182                           // Feedback:  Repeat mode track feedback
  MEDIA_REPEAT_ALL_ON                   = 183                           // Momentary: Set the repeat mode to all
  MEDIA_REPEAT_ALL_FB                   = 183                           // Feedback:  Repeat mode all feedback
  MEDIA_REPEAT_OFF_ON                   = 184                           // Momentary: Set the repeat mode to off
  MEDIA_REPEAT_OFF_FB                   = 184                           // Feedback:  Repeat mode off feedback
  FRAME_FWD                             = 185                           // Momentary: Frame forward
  FRAME_REV                             = 186                           // Momentary: Frame reverse
  SLOW_FWD                              = 188                           // Momentary: Slow forward
  SLOW_REV                              = 189                           // Momentary: Slow reverse
  SCAN_SPEED                            = 192                           // Momentary: Cycle the scanning speed
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay
  PLAY_FB                               = 241                           // Feedback:  Play feedback
  STOP_FB                               = 242                           // Feedback:  Stop feedback
  PAUSE_FB                              = 243                           // Feedback:  Pause feedback
  SFWD_FB                               = 246                           // Feedback:  Scan forward feedback
  SREV_FB                               = 247                           // Feedback:  Scan reverse feedback
  RECORD_FB                             = 248                           // Feedback:  Record feedback
  SLOW_FWD_FB                           = 249                           // Feedback:  Slow forward feedback
  SLOW_REV_FB                           = 250                           // Feedback:  Slow reverse feedback

  VOL_LVL                               = 1                             // Level: Volume level (0-255)

  #IF_DEFINED __DIGITALMEDIADECODER_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    MEDIADEVICECOUNTERNOTIFICATIONON_CMD= 'MEDIACOUNTERNOTIFY'          // setMediaDeviceCounterNotificationOn, 'MEDIACOUNTERNOTIFY-<state>'
    MEDIADEVICEPROPERTIES_QRY           = '?DECODEPROPS'                // queryMediaDeviceProperties, '?DECODEPROPS'
    MEDIADEVICEPROPERTY_QRY             = '?DECODEPROP'                 // queryMediaDeviceProperty, '?DECODEPROP-<key>'
    MEDIADEVICESOURCE_CMD               = 'DECODESOURCE'                // setMediaDeviceSource, 'DECODESOURCE-<url>'
    MEDIADEVICESOURCE_QRY               = '?DECODESOURCE'               // getMediaDeviceSource, '?DECODESOURCE'
    PLAYPOSITION_CMD                    = 'PLAYPOSITION'                // setPlayPosition, 'PLAYPOSITION-<counter>'
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    TRACKCOUNTERNOTIFICATIONON_CMD      = 'TRACKCOUNTERNOTIFY'          // setTrackCounterNotificationOn, 'TRACKCOUNTERNOTIFY-<state>'
    TRACKINFO_QRY                       = '?TRACKINFO'                  // getTrackInfo, '?TRACKINFO'
    TRACKPROPERTIES_QRY                 = '?TRACKPROPS'                 // queryTrackProperties, '?TRACKPROPS'
    TRACKPROPERTY_QRY                   = '?TRACKPROP'                  // queryTrackProperty, '?TRACKPROP'
    TRACKPROP_CMD                       = 'TRACKPROP'                   // Track Property
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
  #END_IF // __DIGITALMEDIADECODER_COMMANDS__

DEFINE_CONSTANT	//	----- Digital Media Encoder : Channels and Levels

  PLAY                                  = 1                             // Momentary: Play
  STOP                                  = 2                             // Momentary: Stop
  PAUSE                                 = 3                             // Momentary: Pause
  FFWD                                  = 4                             // Momentary: Goto the next track
  REW                                   = 5                             // Momentary: Goto the previous track
  SFWD                                  = 6                             // Momentary: Scan forward
  SREV                                  = 7                             // Momentary: Scan reverse
  RECORD                                = 8                             // Momentary: Record
  POWER                                 = 9                             // Momentary: Cycle power
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  GAIN_UP                               = 140                           // Ramping:   Ramp gain up
  GAIN_UP_FB                            = 140                           // Feedback:  Gain ramping up feedback
  GAIN_DN                               = 141                           // Ramping:   Ramp gain down
  GAIN_DN_FB                            = 141                           // Feedback:  Gain ramping down feedback
  GAIN_MUTE_ON                          = 143                           // Discrete:  Set gain mute on
  GAIN_MUTE_FB                          = 143                           // Feedback:  Gain mute feedback
  GAIN_MUTE                             = 144                           // Momentary: Cycle gain mute
  FRAME_FWD                             = 185                           // Momentary: Frame forward
  FRAME_REV                             = 186                           // Momentary: Frame reverse
  SLOW_FWD                              = 188                           // Momentary: Slow forward
  SLOW_REV                              = 189                           // Momentary: Slow reverse
  SCAN_SPEED                            = 192                           // Momentary: Cycle the scanning speed
  PLAY_FB                               = 241                           // Feedback:  Play feedback
  STOP_FB                               = 242                           // Feedback:  Stop feedback
  PAUSE_FB                              = 243                           // Feedback:  Pause feedback
  SFWD_FB                               = 246                           // Feedback:  Scan forward feedback
  SREV_FB                               = 247                           // Feedback:  Scan reverse feedback
  RECORD_FB                             = 248                           // Feedback:  Record feedback
  SLOW_FWD_FB                           = 249                           // Feedback:  Slow forward feedback
  SLOW_REV_FB                           = 250                           // Feedback:  Slow reverse feedback

  GAIN_LVL                              = 5                             // Level: Gain level (0-255)

  #IF_DEFINED __DIGITALMEDIAENCODER_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADDOUTPUTSTREAMSINK_CMD             = 'ENCODESINKADD'               // addOutputStreamSink, 'ENCODESINKADD-<url>'
    OUTPUTSTREAMPROPERTIES_QRY          = '?ENCODEPROPS'                // queryOutputStreamProperties, '?ENCODEPROPS'
    OUTPUTSTREAMPROPERTY_CMD            = 'ENCODEPROP'                  // setOutputStreamProperty, 'ENCODEPROP-<key>,<value>'
    OUTPUTSTREAMPROPERTY_QRY            = '?ENCODEPROP'                 // queryOutputStreamProperty, '?ENCODEPROP-<key>'
    OUTPUTSTREAMSINKS_QRY               = '?ENCODESINK'                 // getOutputStreamSinks, '?ENCODESINK'
    OUTPUTSTREAMSOURCE_CMD              = 'ENCODESOURCE'                // setOutputStreamSource, 'ENCODESOURCE-<url>'
    OUTPUTSTREAMSOURCE_QRY              = '?ENCODESOURCE'               // getOutputStreamSource, '?ENCODESOURCE'
    PLAYPOSITION_CMD                    = 'PLAYPOSITION'                // setPlayPosition, 'PLAYPOSITION-<counter>'
    REMOVEOUTPUTSTREAMSINK_CMD          = 'ENCODESINKREMOVE'            // removeOutputStreamSink, 'ENCODESINKREMOVE-<url>'
    TRACKCOUNTERNOTIFICATIONON_CMD      = 'TRACKCOUNTERNOTIFY'          // setTrackCounterNotificationOn, 'TRACKCOUNTERNOTIFY-<state>'
    TRACKINFO_QRY                       = '?TRACKINFO'                  // getTrackInfo, '?TRACKINFO'
    TRACKPROPERTIES_QRY                 = '?TRACKPROPS'                 // queryTrackProperties, '?TRACKPROPS'
    TRACKPROPERTY_QRY                   = '?TRACKPROP'                  // queryTrackProperty, '?TRACKPROP'
    TRACKPROP_CMD                       = 'TRACKPROP'                   // Track Property
  #END_IF // __DIGITALMEDIAENCODER_COMMANDS__

DEFINE_CONSTANT	//	----- Digital Media Server : Channels and Levels

  PLAY                                  = 1                             // Momentary: Play
  STOP                                  = 2                             // Momentary: Stop
  PAUSE                                 = 3                             // Momentary: Pause
  FFWD                                  = 4                             // Momentary: Goto the next track
  REW                                   = 5                             // Momentary: Goto the previous track
  SFWD                                  = 6                             // Momentary: Scan forward
  SREV                                  = 7                             // Momentary: Scan reverse
  RECORD                                = 8                             // Momentary: Record
  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MEDIA_RANDOM                          = 124                           // Momentary: Cycle the random state
  MEDIA_REPEAT                          = 125                           // Momentary: Cycle the repeat state
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  GAIN_UP                               = 140                           // Ramping:   Ramp gain up
  GAIN_UP_FB                            = 140                           // Feedback:  Gain ramping up feedback
  GAIN_DN                               = 141                           // Ramping:   Ramp gain down
  GAIN_DN_FB                            = 141                           // Feedback:  Gain ramping down feedback
  GAIN_MUTE_ON                          = 143                           // Discrete:  Set gain mute on
  GAIN_MUTE_FB                          = 143                           // Feedback:  Gain mute feedback
  GAIN_MUTE                             = 144                           // Momentary: Cycle gain mute
  MEDIA_RANDOM_DISC_ON                  = 178                           // Momentary: Set the random mode to disc
  MEDIA_RANDOM_DISC_FB                  = 178                           // Feedback:  Random mode disc feedback
  MEDIA_RANDOM_ALL_ON                   = 179                           // Momentary: Set the random mode to all
  MEDIA_RANDOM_ALL_FB                   = 179                           // Feedback:  Random mode all feedback
  MEDIA_RANDOM_OFF_ON                   = 180                           // Momentary: Set the random mode to off
  MEDIA_RANDOM_OFF_FB                   = 180                           // Feedback:  Random mode off feedback
  MEDIA_REPEAT_DISC_ON                  = 181                           // Momentary: Set the repeat mode to disc
  MEDIA_REPEAT_DISC_FB                  = 181                           // Feedback:  Repeat mode disc feedback
  MEDIA_REPEAT_TRACK_ON                 = 182                           // Momentary: Set the repeat mode to track
  MEDIA_REPEAT_TRACK_FB                 = 182                           // Feedback:  Repeat mode track feedback
  MEDIA_REPEAT_ALL_ON                   = 183                           // Momentary: Set the repeat mode to all
  MEDIA_REPEAT_ALL_FB                   = 183                           // Feedback:  Repeat mode all feedback
  MEDIA_REPEAT_OFF_ON                   = 184                           // Momentary: Set the repeat mode to off
  MEDIA_REPEAT_OFF_FB                   = 184                           // Feedback:  Repeat mode off feedback
  FRAME_FWD                             = 185                           // Momentary: Frame forward
  FRAME_REV                             = 186                           // Momentary: Frame reverse
  SLOW_FWD                              = 188                           // Momentary: Slow forward
  SLOW_REV                              = 189                           // Momentary: Slow reverse
  SCAN_SPEED                            = 192                           // Momentary: Cycle the scanning speed
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay
  PLAY_FB                               = 241                           // Feedback:  Play feedback
  STOP_FB                               = 242                           // Feedback:  Stop feedback
  PAUSE_FB                              = 243                           // Feedback:  Pause feedback
  SFWD_FB                               = 246                           // Feedback:  Scan forward feedback
  SREV_FB                               = 247                           // Feedback:  Scan reverse feedback
  RECORD_FB                             = 248                           // Feedback:  Record feedback
  SLOW_FWD_FB                           = 249                           // Feedback:  Slow forward feedback
  SLOW_REV_FB                           = 250                           // Feedback:  Slow reverse feedback

  VOL_LVL                               = 1                             // Level: Volume level (0-255)
  GAIN_LVL                              = 5                             // Level: Gain level (0-255)

  #IF_DEFINED __DIGITALMEDIASERVER_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADDOUTPUTSTREAMSINK_CMD             = 'ENCODESINKADD'               // addOutputStreamSink, 'ENCODESINKADD-<url>'
    CLOSESEARCHDB_CMD                   = 'MEDIADBCLOSESEARCH'          // closeSearchDB, 'MEDIADBCLOSESEARCH-<key>'
    DBNEXT_QRY                          = 'MEDIADBNEXT'                 // queryDB, 'MEDIADBNEXT-<key>[,<count>]'
    DBPREVIOUS_QRY                      = 'MEDIADBPREV'                 // queryDB, 'MEDIADBPREV-<key>[,<count>]'
    DELETERECORD_CMD                    = 'MEDIADBDELETE'               // deleteRecord, 'MEDIADBDELETE-<id>'
    MEDIADBPROPERTIES_QRY               = '?MEDIADBPROPS'               // queryMediaDBProperties, '?MEDIADBPROPS-<id>'
    MEDIADBPROPERTY_CMD                 = 'MEDIADBPROP'                 // setMediaDBProperty, 'MEDIADBPROP-<id>,<key>,<value>'
    MEDIADBPROPERTY_QRY                 = '?MEDIADBPROP'                // queryMediaDBProperty, '?MEDIADBPROP-<id>,<key>'
    MEDIADEVICECOUNTERNOTIFICATIONON_CMD= 'MEDIACOUNTERNOTIFY'          // setMediaDeviceCounterNotificationOn, 'MEDIACOUNTERNOTIFY-<state>'
    MEDIADEVICEPROPERTIES_QRY           = '?DECODEPROPS'                // queryMediaDeviceProperties, '?DECODEPROPS'
    MEDIADEVICEPROPERTY_QRY             = '?DECODEPROP'                 // queryMediaDeviceProperty, '?DECODEPROP-<key>'
    MEDIADEVICESOURCE_CMD               = 'DECODESOURCE'                // setMediaDeviceSource, 'DECODESOURCE-<url>'
    MEDIADEVICESOURCE_QRY               = '?DECODESOURCE'               // getMediaDeviceSource, '?DECODESOURCE'
    OUTPUTSTREAMPROPERTIES_QRY          = '?ENCODEPROPS'                // queryOutputStreamProperties, '?ENCODEPROPS'
    OUTPUTSTREAMPROPERTY_CMD            = 'ENCODEPROP'                  // setOutputStreamProperty, 'ENCODEPROP-<key>,<value>'
    OUTPUTSTREAMPROPERTY_QRY            = '?ENCODEPROP'                 // queryOutputStreamProperty, '?ENCODEPROP-<key>'
    OUTPUTSTREAMSINKS_QRY               = '?ENCODESINK'                 // getOutputStreamSinks, '?ENCODESINK'
    OUTPUTSTREAMSOURCE_CMD              = 'ENCODESOURCE'                // setOutputStreamSource, 'ENCODESOURCE-<url>'
    OUTPUTSTREAMSOURCE_QRY              = '?ENCODESOURCE'               // getOutputStreamSource, '?ENCODESOURCE'
    PLAYPOSITION_CMD                    = 'PLAYPOSITION'                // setPlayPosition, 'PLAYPOSITION-<counter>'
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    REFRESH_CMD                         = 'MEDIADBREFRESH'              // queryDB, 'MEDIADBREFESH-<key>[,<count>]'
    REMOVEOUTPUTSTREAMSINK_CMD          = 'ENCODESINKREMOVE'            // removeOutputStreamSink, 'ENCODESINKREMOVE-<url>'
    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    SEARCHDB_CMD                        = 'MEDIADBSEARCH'               // searchDB, 'MEDIADBSEARCH-<key>,<search type>=<search string>,RETURN=<return type>'
    TRACKCOUNTERNOTIFICATIONON_CMD      = 'TRACKCOUNTERNOTIFY'          // setTrackCounterNotificationOn, 'TRACKCOUNTERNOTIFY-<state>'
    TRACKINFO_QRY                       = '?TRACKINFO'                  // getTrackInfo, '?TRACKINFO'
    TRACKPROPERTIES_QRY                 = '?TRACKPROPS'                 // queryTrackProperties, '?TRACKPROPS'
    TRACKPROPERTY_QRY                   = '?TRACKPROP'                  // queryTrackProperty, '?TRACKPROP'
    TRACKPROP_CMD                       = 'TRACKPROP'                   // Track Property
    UPDATERECORD_CMD                    = 'MEDIADBUPDATE'               // updateRecord, 'MEDIADBUPDATE-<id>,<name>,<record type>[,<url>]'
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
  #END_IF // __DIGITALMEDIASERVER_COMMANDS__

DEFINE_CONSTANT	//	----- Digital Satellite System : Channels

  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  CHAN_UP                               = 22                            // Momentary: Next station preset
  CHAN_DN                               = 23                            // Momentary: Previous station preset
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  TUNER_BAND                            = 40                            // Momentary: Cycle tuner band
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay
  TUNER_PRESET_GROUP                    = 224                           // Momentary: Cycle station preset group
  TUNER_STATION_UP                      = 225                           // Momentary: Increment tuner station
  TUNER_STATION_DN                      = 226                           // Momentary: Decrement tuner station
  TUNER_SCAN_FWD                        = 227                           // Momentary: Station scan forward
  TUNER_SCAN_REV                        = 228                           // Momentary: Station scan backward/reverse
  TUNER_SEEK_FWD                        = 229                           // Momentary: Station seek forward
  TUNER_SEEK_REV                        = 230                           // Momentary: Station seek backward/reverse
  TUNER_OSD                             = 234                           // Momentary: Cycle on-screen or front panel display info
  TUNER_PREV                            = 235                           // Momentary: Goto previous tuner station

  #IF_DEFINED __DIGITALSATELLITESYSTEM_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    BAND_CMD                            = 'BAND'                        // setBand, 'BAND-<tb>'
    BAND_QRY                            = '?BAND'                       // getBand, '?BAND'
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    STATIONPRESETCOUNT_CMD              = 'STATIONPRESETCOUNT'          // 'STATIONPRESETCOUNT-<count>'
    STATIONPRESETCOUNT_QRY              = '?STATIONPRESETCOUNT'         // getStationPresetCount, '?STATIONPRESETCOUNT'
    STATIONPRESETPROPERTIES_CMD         = 'STATIONPRESETPROPERTIES'     // '[STATIONPRESETPROPERTIES-<index>,<displayName>,<station>]*'
    STATIONPRESETPROPERTIES_QRY         = '?STATIONPRESETPROPERTIES'    // getStationPresetProperties, '?STATIONPRESETPROPERTIES'
    STATIONPRESETPROPERTY_CMD           = 'STATIONPRESETPROPERTY'       // 'STATIONPRESETPROPERTY-<index>,<displayName>,<station>'
    STATIONPRESETPROPERTY_QRY           = '?STATIONPRESETPROPERTY'      // getStationPresetProperty(index), '?STATIONPRESETPROPERTY-<index>'
    STATIONPRESETSELECT_CMD             = 'STATIONPRESETSELECT'         // 'STATIONPRESETSELECT-<index>'
    STATIONPRESETSELECT_QRY             = '?STATIONPRESETSELECT'        // getStationPresetSelect, '?STATIONPRESETSELECT'
    STATIONPRESET_CMD                   = 'TUNERPRESET'                 // setStationPreset, 'TUNERPRESET-<preset>'
    STATIONPRESET_QRY                   = '?TUNERPRESET'                // getStationPreset, '?TUNERPRESET'
    STATION_CMD                         = 'XCH'                         // setStation, 'XCH-<station>'
    STATION_QRY                         = '?XCH'                        // getStation, '?XCH'
    TUNERBANDPROPERTIES_CMD             = 'TUNERBANDPROPERTIES'         // '[TUNERBANDPROPERTIES-<index>,<displayName>,<tunerBand>]*'
    TUNERBANDPROPERTIES_QRY             = '?TUNERBANDPROPERTIES'        // getTunerBandProperties, '?TUNERBANDPROPERTIES'
    TUNERBANDPROPERTYCOUNT_CMD          = 'TUNERBANDPROPERTYCOUNT'      // 'TUNERBANDPROPERTYCOUNT-<count>'
    TUNERBANDPROPERTYCOUNT_QRY          = '?TUNERBANDPROPERTYCOUNT'     // getTunerBandPropertyCount, '?TUNERBANDPROPERTYCOUNT'
    TUNERBANDPROPERTY_CMD               = 'TUNERBANDPROPERTY'           // 'TUNERBANDPROPERTY-<index>,<displayName>,<tunerBand>'
    TUNERBANDPROPERTY_QRY               = '?TUNERBANDPROPERTY'          // getTunerBandProperty(index), '?TUNERBANDPROPERTY-<index>'
    TUNERBANDSELECT_CMD                 = 'TUNERBANDSELECT'             // 'TUNERBANDSELECT-<index>'
    TUNERBANDSELECT_QRY                 = '?TUNERBANDSELECT'            // getTunerBandSelect, '?TUNERBANDSELECT'
    TUNERCOMPONENTPROPERTIES_CMD        = 'TUNERCOMPONENTPROPERTIES'    // '[TUNERCOMPONENTPROPERTIES-<index>,<displayName>,<value>]*'
    TUNERCOMPONENTPROPERTIES_QRY        = '?TUNERCOMPONENTPROPERTIES'   // getTunerComponentProperties, '?TUNERCOMPONENTPROPERTIES'
    TUNERCOMPONENTPROPERTYCOUNT_CMD     = 'TUNERCOMPONENTPROPERTYCOUNT' // 'TUNERCOMPONENTPROPERTYCOUNT-<count>'
    TUNERCOMPONENTPROPERTYCOUNT_QRY     = '?TUNERCOMPONENTPROPERTYCOUNT'// getTunerComponentPropertyCount, '?TUNERCOMPONENTPROPERTYCOUNT'
    TUNERCOMPONENTPROPERTY_CMD          = 'TUNERCOMPONENTPROPERTY'      // 'TUNERCOMPONENTPROPERTY-<index>,<displayName>,<value>'
    TUNERCOMPONENTPROPERTY_QRY          = '?TUNERCOMPONENTPROPERTY'     // getTunerComponentProperty(index), '?TUNERCOMPONENTPROPERTY-<index>'
    TUNERCOMPONENTSELECT_CMD            = 'TUNERCOMPONENTSELECT'        // 'TUNERCOMPONENTSELECT-<index>'
    TUNERCOMPONENTSELECT_QRY            = '?TUNERCOMPONENTSELECT'       // getTunerComponentSelect, '?TUNERCOMPONENTSELECT'
  #END_IF // __DIGITALSATELLITESYSTEM_COMMANDS__

DEFINE_CONSTANT	//	----- Digital Video Recorder : Channels

  PLAY                                  = 1                             // Momentary: Play
  STOP                                  = 2                             // Momentary: Stop
  PAUSE                                 = 3                             // Momentary: Pause
  FFWD                                  = 4                             // Momentary: Goto the next track
  REW                                   = 5                             // Momentary: Goto the previous track
  SFWD                                  = 6                             // Momentary: Scan forward
  SREV                                  = 7                             // Momentary: Scan reverse
  RECORD                                = 8                             // Momentary: Record
  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  CHAN_UP                               = 22                            // Momentary: Next station preset
  CHAN_DN                               = 23                            // Momentary: Previous station preset
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  TUNER_BAND                            = 40                            // Momentary: Cycle tuner band
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  FRAME_FWD                             = 185                           // Momentary: Frame forward
  FRAME_REV                             = 186                           // Momentary: Frame reverse
  SLOW_FWD                              = 188                           // Momentary: Slow forward
  SLOW_REV                              = 189                           // Momentary: Slow reverse
  SCAN_SPEED                            = 192                           // Momentary: Cycle the scanning speed
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay
  TUNER_PRESET_GROUP                    = 224                           // Momentary: Cycle station preset group
  TUNER_STATION_UP                      = 225                           // Momentary: Increment tuner station
  TUNER_STATION_DN                      = 226                           // Momentary: Decrement tuner station
  TUNER_SCAN_FWD                        = 227                           // Momentary: Station scan forward
  TUNER_SCAN_REV                        = 228                           // Momentary: Station scan backward/reverse
  TUNER_SEEK_FWD                        = 229                           // Momentary: Station seek forward
  TUNER_SEEK_REV                        = 230                           // Momentary: Station seek backward/reverse
  TUNER_OSD                             = 234                           // Momentary: Cycle on-screen or front panel display info
  TUNER_PREV                            = 235                           // Momentary: Goto previous tuner station
  PLAY_FB                               = 241                           // Feedback:  Play feedback
  STOP_FB                               = 242                           // Feedback:  Stop feedback
  PAUSE_FB                              = 243                           // Feedback:  Pause feedback
  SFWD_FB                               = 246                           // Feedback:  Scan forward feedback
  SREV_FB                               = 247                           // Feedback:  Scan reverse feedback
  RECORD_FB                             = 248                           // Feedback:  Record feedback
  SLOW_FWD_FB                           = 249                           // Feedback:  Slow forward feedback
  SLOW_REV_FB                           = 250                           // Feedback:  Slow reverse feedback

  #IF_DEFINED __DIGITALVIDEORECORDER_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    BAND_CMD                            = 'BAND'                        // setBand, 'BAND-<tb>'
    BAND_QRY                            = '?BAND'                       // getBand, '?BAND'
    CYCLEINPUTSELECT_CMD                = 'CYCLESELECT'                 // cycleInputSelect(), CYCLESELECT
    INPUTCOUNT_CMD                      = 'INPUTCOUNT'                  // INPUTCOUNT-<count>
    INPUTCOUNT_QRY                      = '?INPUTCOUNT'                 // getInputCount(), INPUTCOUNT-<count>
    INPUTGROUPSELECT_CMD                = 'INPUTGROUPSELECT'            // setInputGroupSelect(inputGroup), INPUTGROUPSELECT-<inputGroup>
    INPUTGROUPSELECT_QRY                = '?INPUTGROUPSELECT'           // getInputGroupSelect(), ?INPUTGROUPSELECT
    INPUTPROPERTIES_CMD                 = 'INPUTPROPERTIES'             // [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTIES_QRY                 = '?INPUTPROPERTIES'            // getInputProperties(), [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTY_CMD                   = 'INPUTPROPERTY'               // INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTPROPERTY_QRY                   = '?INPUTPROPERTY'              // getInputProperty(index), INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTSELECT_CMD                     = 'INPUTSELECT'                 // setInputSelect(index), INPUTSELECT-<index>
    INPUTSELECT_QRY                     = '?INPUTSELECT'                // getInputSelect(), INPUTSELECT-<index>
    INPUTSOURCE_CMD                     = 'INPUT'                       // setInputSource, INPUT-<sourceSelect>,<inputNumber>
    INPUTSOURCE_QRY                     = '?INPUT'                      // getInputSource, ?INPUT
    PLAYPOSITION_CMD                    = 'PLAYPOSITION'                // setPlayPosition, 'PLAYPOSITION-<counter>'
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    STATIONPRESETCOUNT_CMD              = 'STATIONPRESETCOUNT'          // 'STATIONPRESETCOUNT-<count>'
    STATIONPRESETCOUNT_QRY              = '?STATIONPRESETCOUNT'         // getStationPresetCount, '?STATIONPRESETCOUNT'
    STATIONPRESETPROPERTIES_CMD         = 'STATIONPRESETPROPERTIES'     // '[STATIONPRESETPROPERTIES-<index>,<displayName>,<station>]*'
    STATIONPRESETPROPERTIES_QRY         = '?STATIONPRESETPROPERTIES'    // getStationPresetProperties, '?STATIONPRESETPROPERTIES'
    STATIONPRESETPROPERTY_CMD           = 'STATIONPRESETPROPERTY'       // 'STATIONPRESETPROPERTY-<index>,<displayName>,<station>'
    STATIONPRESETPROPERTY_QRY           = '?STATIONPRESETPROPERTY'      // getStationPresetProperty(index), '?STATIONPRESETPROPERTY-<index>'
    STATIONPRESETSELECT_CMD             = 'STATIONPRESETSELECT'         // 'STATIONPRESETSELECT-<index>'
    STATIONPRESETSELECT_QRY             = '?STATIONPRESETSELECT'        // getStationPresetSelect, '?STATIONPRESETSELECT'
    STATIONPRESET_CMD                   = 'TUNERPRESET'                 // setStationPreset, 'TUNERPRESET-<preset>'
    STATIONPRESET_QRY                   = '?TUNERPRESET'                // getStationPreset, '?TUNERPRESET'
    STATION_CMD                         = 'XCH'                         // setStation, 'XCH-<station>'
    STATION_QRY                         = '?XCH'                        // getStation, '?XCH'
    TRACKCOUNTERNOTIFICATIONON_CMD      = 'TRACKCOUNTERNOTIFY'          // setTrackCounterNotificationOn, 'TRACKCOUNTERNOTIFY-<state>'
    TRACKINFO_QRY                       = '?TRACKINFO'                  // getTrackInfo, '?TRACKINFO'
    TRACKPROPERTIES_QRY                 = '?TRACKPROPS'                 // queryTrackProperties, '?TRACKPROPS'
    TRACKPROPERTY_QRY                   = '?TRACKPROP'                  // queryTrackProperty, '?TRACKPROP'
    TRACKPROP_CMD                       = 'TRACKPROP'                   // Track Property
    TUNERBANDPROPERTIES_CMD             = 'TUNERBANDPROPERTIES'         // '[TUNERBANDPROPERTIES-<index>,<displayName>,<tunerBand>]*'
    TUNERBANDPROPERTIES_QRY             = '?TUNERBANDPROPERTIES'        // getTunerBandProperties, '?TUNERBANDPROPERTIES'
    TUNERBANDPROPERTYCOUNT_CMD          = 'TUNERBANDPROPERTYCOUNT'      // 'TUNERBANDPROPERTYCOUNT-<count>'
    TUNERBANDPROPERTYCOUNT_QRY          = '?TUNERBANDPROPERTYCOUNT'     // getTunerBandPropertyCount, '?TUNERBANDPROPERTYCOUNT'
    TUNERBANDPROPERTY_CMD               = 'TUNERBANDPROPERTY'           // 'TUNERBANDPROPERTY-<index>,<displayName>,<tunerBand>'
    TUNERBANDPROPERTY_QRY               = '?TUNERBANDPROPERTY'          // getTunerBandProperty(index), '?TUNERBANDPROPERTY-<index>'
    TUNERBANDSELECT_CMD                 = 'TUNERBANDSELECT'             // 'TUNERBANDSELECT-<index>'
    TUNERBANDSELECT_QRY                 = '?TUNERBANDSELECT'            // getTunerBandSelect, '?TUNERBANDSELECT'
    TUNERCOMPONENTPROPERTIES_CMD        = 'TUNERCOMPONENTPROPERTIES'    // '[TUNERCOMPONENTPROPERTIES-<index>,<displayName>,<value>]*'
    TUNERCOMPONENTPROPERTIES_QRY        = '?TUNERCOMPONENTPROPERTIES'   // getTunerComponentProperties, '?TUNERCOMPONENTPROPERTIES'
    TUNERCOMPONENTPROPERTYCOUNT_CMD     = 'TUNERCOMPONENTPROPERTYCOUNT' // 'TUNERCOMPONENTPROPERTYCOUNT-<count>'
    TUNERCOMPONENTPROPERTYCOUNT_QRY     = '?TUNERCOMPONENTPROPERTYCOUNT'// getTunerComponentPropertyCount, '?TUNERCOMPONENTPROPERTYCOUNT'
    TUNERCOMPONENTPROPERTY_CMD          = 'TUNERCOMPONENTPROPERTY'      // 'TUNERCOMPONENTPROPERTY-<index>,<displayName>,<value>'
    TUNERCOMPONENTPROPERTY_QRY          = '?TUNERCOMPONENTPROPERTY'     // getTunerComponentProperty(index), '?TUNERCOMPONENTPROPERTY-<index>'
    TUNERCOMPONENTSELECT_CMD            = 'TUNERCOMPONENTSELECT'        // 'TUNERCOMPONENTSELECT-<index>'
    TUNERCOMPONENTSELECT_QRY            = '?TUNERCOMPONENTSELECT'       // getTunerComponentSelect, '?TUNERCOMPONENTSELECT'
  #END_IF // __DIGITALVIDEORECORDER_COMMANDS__

DEFINE_CONSTANT	//	----- Disc Device : Channels

  PLAY                                  = 1                             // Momentary: Play
  STOP                                  = 2                             // Momentary: Stop
  PAUSE                                 = 3                             // Momentary: Pause
  FFWD                                  = 4                             // Momentary: Goto the next track
  REW                                   = 5                             // Momentary: Goto the previous track
  SFWD                                  = 6                             // Momentary: Scan forward
  SREV                                  = 7                             // Momentary: Scan reverse
  RECORD                                = 8                             // Momentary: Record
  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  DISC_NEXT                             = 55                            // Momentary: Goto the next disc
  DISC_PREV                             = 56                            // Momentary: Goto the previous disc
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  DISC_TRAY                             = 120                           // Momentary: Open/Close disc tray
  DISC_RANDOM                           = 124                           // Momentary: Cycle the random state
  DISC_REPEAT                           = 125                           // Momentary: Cycle the repeat state
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  DISC_RANDOM_DISC_ON                   = 178                           // Momentary: Set the random mode to disc
  DISC_RANDOM_DISC_FB                   = 178                           // Feedback:  Random mode disc feedback
  DISC_RANDOM_ALL_ON                    = 179                           // Momentary: Set the random mode to all
  DISC_RANDOM_ALL_FB                    = 179                           // Feedback:  Random mode all feedback
  DISC_RANDOM_OFF_ON                    = 180                           // Momentary: Set the random mode to off
  DISC_RANDOM_OFF_FB                    = 180                           // Feedback:  Random mode off feedback
  DISC_REPEAT_DISC_ON                   = 181                           // Momentary: Set the repeat mode to disc
  DISC_REPEAT_DISC_FB                   = 181                           // Feedback:  Repeat mode disc feedback
  DISC_REPEAT_TRACK_ON                  = 182                           // Momentary: Set the repeat mode to track
  DISC_REPEAT_TRACK_FB                  = 182                           // Feedback:  Repeat mode track feedback
  DISC_REPEAT_ALL_ON                    = 183                           // Momentary: Set the repeat mode to all
  DISC_REPEAT_ALL_FB                    = 183                           // Feedback:  Repeat mode all feedback
  DISC_REPEAT_OFF_ON                    = 184                           // Momentary: Set the repeat mode to off
  DISC_REPEAT_OFF_FB                    = 184                           // Feedback:  Repeat mode off feedback
  FRAME_FWD                             = 185                           // Momentary: Frame forward
  FRAME_REV                             = 186                           // Momentary: Frame reverse
  SLOW_FWD                              = 188                           // Momentary: Slow forward
  SLOW_REV                              = 189                           // Momentary: Slow reverse
  SCAN_SPEED                            = 192                           // Momentary: Cycle the scanning speed
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay
  PLAY_FB                               = 241                           // Feedback:  Play feedback
  STOP_FB                               = 242                           // Feedback:  Stop feedback
  PAUSE_FB                              = 243                           // Feedback:  Pause feedback
  SFWD_FB                               = 246                           // Feedback:  Scan forward feedback
  SREV_FB                               = 247                           // Feedback:  Scan reverse feedback
  RECORD_FB                             = 248                           // Feedback:  Record feedback
  SLOW_FWD_FB                           = 249                           // Feedback:  Slow forward feedback
  SLOW_REV_FB                           = 250                           // Feedback:  Slow reverse feedback

  #IF_DEFINED __DISCDEVICE_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    DISCCAPACITY_QRY                    = '?DISCCAPACITY'               // getDiscCapacity, '?DISCCAPACITY'
    DISCINFO_CMD                        = 'DISCINFO'                    // Disc Info
    DISCINFO_QRY                        = '?DISCINFO'                   // getDiscInfo, '?DISCINFO'
    DISCPROPERTIES_QRY                  = '?DISCPROPS'                  // queryDiscProperties, '?DISCPROPS'
    DISCPROPERTY_QRY                    = '?DISCPROP'                   // queryDiscProperty, '?DISCPROP'
    DISCPROP_CMD                        = 'DISCPROP'                    // Disc Property
    DISC_CMD                            = 'SETDISC'                     // setDisc, 'SETDISC-<disc>'
    PLAYPOSITION1_CMD                   = 'PLAYPOSITION'                // setPlayPosition, 'PLAYPOSITION-<title>,<track>,<counter>'
    PLAYPOSITION2_CMD                   = 'PLAYPOSITION'                // setPlayPosition, 'PLAYPOSITION-<track>'
    PLAYPOSITION3_CMD                   = 'PLAYPOSITION'                // setPlayPosition, 'PLAYPOSITION-<track>,<counter>'
    PLAYPOSITION_CMD                    = 'PLAYPOSITION'                // setPlayPosition, 'PLAYPOSITION-<counter>'
    PLAYPOSITION_CMD                    = 'PLAYPOSITION'                // setPlayPosition, 'PLAYPOSITION-<title>,<track>'
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    TITLECOUNTERNOTIFICATIONON_CMD      = 'TITLECOUNTERNOTIFY'          // setTitleCounterNotificationOn, 'TITLECOUNTERNOTIFY-<state>'
    TITLECOUNTER_CMD                    = 'TITLECOUNTER'                // Title Counter
    TITLEINFO_CMD                       = 'TITLEINFO'                   // Title Info
    TITLEINFO_QRY                       = '?TITLEINFO'                  // getTitleInfo, '?TITLEINFO'
    TITLEPROPERTIES_QRY                 = '?TITLEPROPS'                 // queryTitleProperties, '?TITLEPROPS'
    TITLEPROPERTY_QRY                   = '?TITLEPROP'                  // queryTitleProperty, '?TITLEPROP'
    TITLEPROP_CMD                       = 'TITLEPROP'                   // Disc Property
    TRACKCOUNTERNOTIFICATIONON_CMD      = 'TRACKCOUNTERNOTIFY'          // setTrackCounterNotificationOn, 'TRACKCOUNTERNOTIFY-<state>'
    TRACKINFO_QRY                       = '?TRACKINFO'                  // getTrackInfo, '?TRACKINFO'
    TRACKPROPERTIES_QRY                 = '?TRACKPROPS'                 // queryTrackProperties, '?TRACKPROPS'
    TRACKPROPERTY_QRY                   = '?TRACKPROP'                  // queryTrackProperty, '?TRACKPROP'
    TRACKPROP_CMD                       = 'TRACKPROP'                   // Track Property
  #END_IF // __DISCDEVICE_COMMANDS__

DEFINE_CONSTANT	//	----- Display : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle lamp power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  CHAN_UP                               = 22                            // Momentary: Next station preset
  CHAN_DN                               = 23                            // Momentary: Previous station preset
  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  PWR_ON                                = 27                            // Momentary: Set lamp power on
  PWR_OFF                               = 28                            // Momentary: Set lamp power off
  TUNER_BAND                            = 40                            // Momentary: Cycle tuner band
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  ASPECT_RATIO                          = 142                           // Momentary: Cycle aspect ratio
  BRIGHT_UP                             = 148                           // Momentary: Increment brightness
  BRIGHT_DN                             = 149                           // Momentary: Decrement brightness
  COLOR_UP                              = 150                           // Momentary: Increment color
  COLOR_DN                              = 151                           // Momentary: Decrement color
  CONTRAST_UP                           = 152                           // Momentary: Increment contrast
  CONTRAST_DN                           = 153                           // Momentary: Decrement contrast
  SHARP_UP                              = 154                           // Momentary: Increment sharpness
  SHARP_DN                              = 155                           // Momentary: Decrement sharpness
  TINT_UP                               = 156                           // Momentary: Increment tint
  TINT_DN                               = 157                           // Momentary: Decrement tint
  PIP_POS                               = 191                           // Momentary: Cycle pip position
  PIP_SWAP                              = 193                           // Momentary: Swap pip
  PIP                                   = 194                           // Momentary: Cycle pip
  PIP_ON                                = 195                           // Discrete:  Set pip on
  PIP_FB                                = 195                           // Feedback:  Pip feedback
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  PIC_MUTE                              = 210                           // Momentary: Cycle picture/video mute
  PIC_MUTE_ON                           = 211                           // Discrete:  Set picture/video mute on
  PIC_MUTE_FB                           = 211                           // Feedback:  Picture/video mute feedback
  PIC_FREEZE                            = 213                           // Momentary: Cycle freeze
  PIC_FREEZE_ON                         = 214                           // Discrete:  Set freeze on
  PIC_FREEZE_FB                         = 214                           // Feedback:  Freeze feedback
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay
  TUNER_PRESET_GROUP                    = 224                           // Momentary: Cycle station preset group
  TUNER_STATION_UP                      = 225                           // Momentary: Increment tuner station
  TUNER_STATION_DN                      = 226                           // Momentary: Decrement tuner station
  TUNER_SCAN_FWD                        = 227                           // Momentary: Station scan forward
  TUNER_SCAN_REV                        = 228                           // Momentary: Station scan backward/reverse
  TUNER_SEEK_FWD                        = 229                           // Momentary: Station seek forward
  TUNER_SEEK_REV                        = 230                           // Momentary: Station seek backward/reverse
  TUNER_OSD                             = 234                           // Momentary: Cycle on-screen or front panel display info
  TUNER_PREV                            = 235                           // Momentary: Goto previous tuner station

  VOL_LVL                               = 1                             // Level: Volume level (0-255)
  BRIGHT_LVL                            = 10                            // Level: Brightness level (0-255)
  COLOR_LVL                             = 11                            // Level: Color level (0-255)
  CONTRAST_LVL                          = 12                            // Level: Contrast level (0-255)
  SHARP_LVL                             = 13                            // Level: Sharpness level (0-255)
  TINT_LVL                              = 14                            // Level: Tint level (0-255)

  #IF_DEFINED __DISPLAY_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ACTIVEWINDOW_CMD                    = 'ACTIVEWINDOW'                // setActiveWindow, 'ACTIVEWINDOW-<mss>'
    ACTIVEWINDOW_QRY                    = '?ACTIVEWINDOW'               // getActiveWindow, '?ACTIVEWINDOW'
    ASPECTRATIOCOUNT_CMD                = 'ASPECTRATIOCOUNT'            // ASPECTRATIOCOUNT-<count>'
    ASPECTRATIOCOUNT_QRY                = '?ASPECTRATIOCOUNT'           // getAspectRatioCount(), '?ASPECTRATIOCOUNT'
    ASPECTRATIOPROPERTIES_CMD           = 'ASPECTRATIOPROPERTIES'       // '[ASPECTRATIOPROPERTIES-<index>,<displayName>,<value>]*'
    ASPECTRATIOPROPERTIES_QRY           = '?ASPECTRATIOPROPERTIES'      // getAspectRatioProperties(), '?ASPECTRATIOPROPERTIES'
    ASPECTRATIOPROPERTY_CMD             = 'ASPECTRATIOPROPERTY'         // 'ASPECTRATIOPROPERTY-<index>,<displayName>,<value>'
    ASPECTRATIOPROPERTY_QRY             = '?ASPECTRATIOPROPERTY'        // getAspectRatioProperty(index), '?ASPECTRATIOPROPERTY-<index>'
    ASPECTRATIOSELECT_CMD               = 'ASPECTRATIOSELECT'           // setAspectRatioSelect(index), 'ASPECTRATIOSELECT-<index>'
    ASPECTRATIOSELECT_QRY               = '?ASPECTRATIOSELECT'          // getAspectRatioSelect(), '?ASPECTRATIOSELECT'
    ASPECTRATIO_CMD                     = 'ASPECT'                      // setAspectRatio, 'ASPECT-<aspectRatio>'
    ASPECTRATIO_QRY                     = '?ASPECT'                     // getAspectRatio, '?ASPECT'
    BAND_CMD                            = 'BAND'                        // setBand, 'BAND-<tb>'
    BAND_QRY                            = '?BAND'                       // getBand, '?BAND'
    COOLDOWNCOUNTER_CMD                 = 'COOLING'                     // 'COOLING-<int>'
    COOLDOWNTIME_CMD                    = 'COOLDOWN'                    // setCoolDownTime, 'COOLDOWN-<time>'
    COOLDOWNTIME_QRY                    = '?COOLDOWN'                   // getCoolDownTime, '?COOLDOWN'
    COUNTERNOTIFICATIONON_CMD           = 'COUNTERNOTIFY'               // setCounterNotificationOn, 'COUNTERNOTIFY-<state>'
    CYCLEINPUTSELECT_CMD                = 'CYCLESELECT'                 // cycleInputSelect(), CYCLESELECT
    INPUTCOUNT_CMD                      = 'INPUTCOUNT'                  // INPUTCOUNT-<count>
    INPUTCOUNT_QRY                      = '?INPUTCOUNT'                 // getInputCount(), INPUTCOUNT-<count>
    INPUTGROUPSELECT_CMD                = 'INPUTGROUPSELECT'            // setInputGroupSelect(inputGroup), INPUTGROUPSELECT-<inputGroup>
    INPUTGROUPSELECT_QRY                = '?INPUTGROUPSELECT'           // getInputGroupSelect(), ?INPUTGROUPSELECT
    INPUTPROPERTIES_CMD                 = 'INPUTPROPERTIES'             // [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTIES_QRY                 = '?INPUTPROPERTIES'            // getInputProperties(), [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTY_CMD                   = 'INPUTPROPERTY'               // INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTPROPERTY_QRY                   = '?INPUTPROPERTY'              // getInputProperty(index), INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTSELECT_CMD                     = 'INPUTSELECT'                 // setInputSelect(index), INPUTSELECT-<index>
    INPUTSELECT_QRY                     = '?INPUTSELECT'                // getInputSelect(), INPUTSELECT-<index>
    INPUTSOURCE_CMD                     = 'INPUT'                       // setInputSource, INPUT-<sourceSelect>,<inputNumber>
    INPUTSOURCE_QRY                     = '?INPUT'                      // getInputSource, ?INPUT
    LAMPTIME_CMD                        = 'LAMPTIME'                    // setLampTime, 'LAMPTIME-<time>'
    LAMPTIME_QRY                        = '?LAMPTIME'                   // getLampTime, '?LAMPTIME'
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    PROJECTORPRESET_CMD                 = 'VPROJPRESET'                 // setProjectorPreset, 'VPROJPRESET-<preset>'
    PROJECTORPRESET_QRY                 = '?VPROJPRESET'                // getProjectorPreset, '?VPROJPRESET'
    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    STATIONPRESETCOUNT_CMD              = 'STATIONPRESETCOUNT'          // 'STATIONPRESETCOUNT-<count>'
    STATIONPRESETCOUNT_QRY              = '?STATIONPRESETCOUNT'         // getStationPresetCount, '?STATIONPRESETCOUNT'
    STATIONPRESETPROPERTIES_CMD         = 'STATIONPRESETPROPERTIES'     // '[STATIONPRESETPROPERTIES-<index>,<displayName>,<station>]*'
    STATIONPRESETPROPERTIES_QRY         = '?STATIONPRESETPROPERTIES'    // getStationPresetProperties, '?STATIONPRESETPROPERTIES'
    STATIONPRESETPROPERTY_CMD           = 'STATIONPRESETPROPERTY'       // 'STATIONPRESETPROPERTY-<index>,<displayName>,<station>'
    STATIONPRESETPROPERTY_QRY           = '?STATIONPRESETPROPERTY'      // getStationPresetProperty(index), '?STATIONPRESETPROPERTY-<index>'
    STATIONPRESETSELECT_CMD             = 'STATIONPRESETSELECT'         // 'STATIONPRESETSELECT-<index>'
    STATIONPRESETSELECT_QRY             = '?STATIONPRESETSELECT'        // getStationPresetSelect, '?STATIONPRESETSELECT'
    STATIONPRESET_CMD                   = 'TUNERPRESET'                 // setStationPreset, 'TUNERPRESET-<preset>'
    STATIONPRESET_QRY                   = '?TUNERPRESET'                // getStationPreset, '?TUNERPRESET'
    STATION_CMD                         = 'XCH'                         // setStation, 'XCH-<station>'
    STATION_QRY                         = '?XCH'                        // getStation, '?XCH'
    TUNERBANDPROPERTIES_CMD             = 'TUNERBANDPROPERTIES'         // '[TUNERBANDPROPERTIES-<index>,<displayName>,<tunerBand>]*'
    TUNERBANDPROPERTIES_QRY             = '?TUNERBANDPROPERTIES'        // getTunerBandProperties, '?TUNERBANDPROPERTIES'
    TUNERBANDPROPERTYCOUNT_CMD          = 'TUNERBANDPROPERTYCOUNT'      // 'TUNERBANDPROPERTYCOUNT-<count>'
    TUNERBANDPROPERTYCOUNT_QRY          = '?TUNERBANDPROPERTYCOUNT'     // getTunerBandPropertyCount, '?TUNERBANDPROPERTYCOUNT'
    TUNERBANDPROPERTY_CMD               = 'TUNERBANDPROPERTY'           // 'TUNERBANDPROPERTY-<index>,<displayName>,<tunerBand>'
    TUNERBANDPROPERTY_QRY               = '?TUNERBANDPROPERTY'          // getTunerBandProperty(index), '?TUNERBANDPROPERTY-<index>'
    TUNERBANDSELECT_CMD                 = 'TUNERBANDSELECT'             // 'TUNERBANDSELECT-<index>'
    TUNERBANDSELECT_QRY                 = '?TUNERBANDSELECT'            // getTunerBandSelect, '?TUNERBANDSELECT'
    TUNERCOMPONENTPROPERTIES_CMD        = 'TUNERCOMPONENTPROPERTIES'    // '[TUNERCOMPONENTPROPERTIES-<index>,<displayName>,<value>]*'
    TUNERCOMPONENTPROPERTIES_QRY        = '?TUNERCOMPONENTPROPERTIES'   // getTunerComponentProperties, '?TUNERCOMPONENTPROPERTIES'
    TUNERCOMPONENTPROPERTYCOUNT_CMD     = 'TUNERCOMPONENTPROPERTYCOUNT' // 'TUNERCOMPONENTPROPERTYCOUNT-<count>'
    TUNERCOMPONENTPROPERTYCOUNT_QRY     = '?TUNERCOMPONENTPROPERTYCOUNT'// getTunerComponentPropertyCount, '?TUNERCOMPONENTPROPERTYCOUNT'
    TUNERCOMPONENTPROPERTY_CMD          = 'TUNERCOMPONENTPROPERTY'      // 'TUNERCOMPONENTPROPERTY-<index>,<displayName>,<value>'
    TUNERCOMPONENTPROPERTY_QRY          = '?TUNERCOMPONENTPROPERTY'     // getTunerComponentProperty(index), '?TUNERCOMPONENTPROPERTY-<index>'
    TUNERCOMPONENTSELECT_CMD            = 'TUNERCOMPONENTSELECT'        // 'TUNERCOMPONENTSELECT-<index>'
    TUNERCOMPONENTSELECT_QRY            = '?TUNERCOMPONENTSELECT'       // getTunerComponentSelect, '?TUNERCOMPONENTSELECT'
    VIDEOTYPE_CMD                       = 'VIDEOTYPE'                   // setVideoType, 'VIDEOTYPE-<vt>'
    VIDEOTYPE_QRY                       = '?VIDEOTYPE'                  // getVideoType, '?VIDEOTYPE'
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
    WARMUPCOUNTER_CMD                   = 'WARMING'                     // 'WARMING-<int>'
    WARMUPTIME_CMD                      = 'WARMUP'                      // setWarmUpTime, 'WARMUP-<time>'
    WARMUPTIME_QRY                      = '?WARMUP'                     // getWarmUpTime, '?WARMUP'
  #END_IF // __DISPLAY_COMMANDS__

DEFINE_CONSTANT	//	----- Document Camera : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle power
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  ZOOM_OUT                              = 158                           // Ramping:   Ramp zoom out
  ZOOM_OUT_FB                           = 158                           // Feedback:  Ramp zoom out feedback
  ZOOM_IN                               = 159                           // Ramping:   Ramp zoom in
  ZOOM_IN_FB                            = 159                           // Feedback:  Ramp zoom in feedback
  FOCUS_NEAR                            = 160                           // Ramping:   Ramp focus near
  FOCUS_NEAR_FB                         = 160                           // Feedback:  Ramp focus near feedback
  FOCUS_FAR                             = 161                           // Ramping:   Ramp focus far
  FOCUS_FAR_FB                          = 161                           // Feedback:  Focus ramp far feedback
  AUTO_FOCUS_ON                         = 162                           // Discrete:  Set auto focus on or off
  AUTO_FOCUS_FB                         = 162                           // Feedback:  Auto focus feedback
  AUTO_IRIS_ON                          = 163                           // Discrete:  Set auto iris on or off
  AUTO_IRIS_FB                          = 163                           // Feedback:  Auto iris feedback
  AUTO_FOCUS                            = 172                           // Momentary: Cycle auto focus
  AUTO_IRIS                             = 173                           // Momentary: Cycle auto iris
  IRIS_OPEN                             = 174                           // Ramping:   Ramp iris open
  IRIS_OPEN_FB                          = 174                           // Feedback:  Ramp iris open feedback
  IRIS_CLOSE                            = 175                           // Ramping:   Ramp iris closed
  IRIS_CLOSE_FB                         = 175                           // Feedback:  Ramp iris closed feedback
  DOCCAM_LIGHT                          = 176                           // Momentary: Cycle light
  DOCCAM_LOWER_LIGHT_ON                 = 197                           // Discrete:  Set the lower light on or off
  DOCCAM_LOWER_LIGHT_FB                 = 197                           // Feedback:  Lower light feedback
  DOCCAM_UPPER_LIGHT_ON                 = 198                           // Discrete:  Set the upper light on or off
  DOCCAM_UPPER_LIGHT_FB                 = 198                           // Feedback:  Upper light feedback

  ZOOM_LVL                              = 15                            // Level: Zoom position (0=out/Wide, 255=in/Tele)
  FOCUS_LVL                             = 16                            // Level: Focus position (0=near, 255=far)
  IRIS_LVL                              = 17                            // Level: Iris position (0=closed, 255=open)
  ZOOM_SPEED_LVL                        = 18                            // Level: Zoom ramp speed (0-255)
  FOCUS_SPEED_LVL                       = 19                            // Level: Focus ramp speed (0-255)
  IRIS_SPEED_LVL                        = 20                            // Level: Iris ramp speed (0-255)

  #IF_DEFINED __DOCUMENTCAMERA_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    CYCLEINPUTSELECT_CMD                = 'CYCLESELECT'                 // cycleInputSelect(), CYCLESELECT
    INPUTCOUNT_CMD                      = 'INPUTCOUNT'                  // INPUTCOUNT-<count>
    INPUTCOUNT_QRY                      = '?INPUTCOUNT'                 // getInputCount(), INPUTCOUNT-<count>
    INPUTGROUPSELECT_CMD                = 'INPUTGROUPSELECT'            // setInputGroupSelect(inputGroup), INPUTGROUPSELECT-<inputGroup>
    INPUTGROUPSELECT_QRY                = '?INPUTGROUPSELECT'           // getInputGroupSelect(), ?INPUTGROUPSELECT
    INPUTPROPERTIES_CMD                 = 'INPUTPROPERTIES'             // [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTIES_QRY                 = '?INPUTPROPERTIES'            // getInputProperties(), [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTY_CMD                   = 'INPUTPROPERTY'               // INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTPROPERTY_QRY                   = '?INPUTPROPERTY'              // getInputProperty(index), INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTSELECT_CMD                     = 'INPUTSELECT'                 // setInputSelect(index), INPUTSELECT-<index>
    INPUTSELECT_QRY                     = '?INPUTSELECT'                // getInputSelect(), INPUTSELECT-<index>
    INPUTSOURCE_CMD                     = 'INPUT'                       // setInputSource, INPUT-<sourceSelect>,<inputNumber>
    INPUTSOURCE_QRY                     = '?INPUT'                      // getInputSource, ?INPUT
  #END_IF // __DOCUMENTCAMERA_COMMANDS__

DEFINE_CONSTANT	//	----- HVAC : Channels and Levels

  HVAC_COOL_UP                          = 140                           // Momentary: Increment cooling set point
  HVAC_COOL_DN                          = 141                           // Momentary: Decrement cooling set point
  HVAC_HEAT_UP                          = 143                           // Momentary: Increment heating set point
  HVAC_HEAT_DN                          = 144                           // Momentary: Decrement heating set point
  HVAC_HUMIDIFY_UP                      = 148                           // Momentary: Increment humidifier set point
  HVAC_HUMIDIFY_DN                      = 149                           // Momentary: Decrement humidifier set point
  HVAC_DEHUMIDIFY_UP                    = 150                           // Momentary: Increment dehumidifier set point
  HVAC_DEHUMIDIFY_DN                    = 151                           // Momentary: Decrement dehumidifier set point
  HVAC_HOLD_ON                          = 211                           // Discrete:  Set thermostat hold mode
  HVAC_HOLD_FB                          = 211                           // Feedback:  Thermostat hold feedback
  HVAC_LOCK_ON                          = 212                           // Discrete:  Set thermostat lock state
  HVAC_LOCK_FB                          = 212                           // Feedback:  Thermostat lock state feedback
  HVAC_FAN                              = 213                           // Momentary: Cycle fan state
  HVAC_FAN_ON                           = 214                           // Momentary: Set fan state to on
  HVAC_FAN_ON_FB                        = 214                           // Feedback:  Fan state on feedback
  HVAC_FAN_AUTO                         = 215                           // Momentary: Set fan state to auto
  HVAC_FAN_AUTO_FB                      = 215                           // Feedback:  Auto fan state feedback
  HVAC_FAN_STATUS_FB                    = 216                           // Feedback:  Fan status state feedback
  HVAC_HUMIDIFY_STATE                   = 217                           // Momentary: Cycle the humidify state
  HVAC_STATE                            = 218                           // Momentary: Cycle HVAC state
  HVAC_AUTO                             = 219                           // Momentary: Set HVAC state to automatic
  HVAC_AUTO_FB                          = 219                           // Feedback:  HVAC auto feedback
  HVAC_COOL                             = 220                           // Momentary: Set HVAC state to cooling
  HVAC_COOL_FB                          = 220                           // Feedback:  HVAC cool feedback
  HVAC_HEAT                             = 221                           // Momentary: Set HVAC state to heating
  HVAC_HEAT_FB                          = 221                           // Feedback:  HVAC heat feedback
  HVAC_OFF                              = 222                           // Momentary: Set HVAC state to off
  HVAC_OFF_FB                           = 222                           // Feedback:  HVAC off feedback
  HVAC_EHEAT                            = 223                           // Momentary: Set HVAC state to emergency heat
  HVAC_EHEAT_FB                         = 223                           // Feedback:  HVAC emergency heat feedback
  HVAC_COOLING_FB                       = 224                           // Feedback:  HVAC cooling status feedback
  HVAC_HEATING_FB                       = 225                           // Feedback:  HVAC heating status feedback
  HVAC_COOLING2_FB                      = 226                           // Feedback:  HVAC cooling 2 status feedback
  HVAC_EHEATING_FB                      = 227                           // Feedback:  HVAC emergency heating status feedback
  HVAC_HUMIDIFY_AUTO                    = 228                           // Momentary: Set the humidify state to auto
  HVAC_HUMIDIFY_AUTO_FB                 = 228                           // Feedback:  Humidify state auto feedback
  HVAC_DEHUMIDIFY                       = 229                           // Momentary: Set the humidify state to dehumidify
  HVAC_DEHUMIDIFY_FB                    = 229                           // Feedback:  Dehumidify state feedback
  HVAC_HUMIDIFY                         = 230                           // Momentary: Set the humidify state to humidify
  HVAC_HUMIDIFY_FB                      = 230                           // Feedback:  Humidify state feedback
  HVAC_HUMIDIFY_OFF                     = 231                           // Momentary: Set the humidify state to off
  HVAC_HUMIDIFY_OFF_FB                  = 231                           // Feedback:  Humidify state off feedback
  HVAC_DEHUMIDIFING_FB                  = 232                           // Feedback:  Dehumidify state status feedback
  HVAC_HUMIDIFING_FB                    = 233                           // Feedback:  Humidify state status feedback

  HVAC_COOL_LVL                         = 31                            // Level: Cooling set point, range is n..m degrees F or C
  HVAC_HEAT_LVL                         = 32                            // Level: Heating set point, range is n..m degrees F or C
  INDOOR_TEMP_LVL                       = 33                            // Level: Indoor temperature, range is n..m degrees F or C
  OUTDOOR_TEMP_LVL                      = 34                            // Level: Outdoor temperature, range is n..m degrees F or C
  INDOOR_HUMID_LVL                      = 35                            // Level: Indoor humidity, range is 0..100 percent
  OUTDOOR_HUMID_LVL                     = 36                            // Level: Outdoor humidity, range is 0..100 percent
  HVAC_HUMIDIFY_LVL                     = 37                            // Level: Humidifier set point, range is 0..100 percent
  HVAC_DEHUMIDIFY_LVL                   = 38                            // Level: Dehumidifier set point, range is 0..100 percent

  #IF_DEFINED __HVAC_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADDHVACCOMPONENT_CMD                = 'HVACADD'                     // addHVACComponent, 'HVACADD-<index>,<hvacAddress>'
    HUMIDIFYSTATE_CMD                   = 'HVACHUMID'                   // setHumidifyState, 'HVACHUMID-<hs>'
    HUMIDIFYSTATE_QRY                   = '?HVACHUMID'                  // getHumidifyState, '?HVACHUMID'
    HUMIDIFYSTATUS_CMD                  = 'HVACHUMIDSTATUS'             // 'HVACHUMIDSTATUS-<state>'
    HUMIDIFYSTATUS_QRY                  = '?HVACHUMIDSTATUS'            // getHumidifyStatus, '?HVACHUMIDSTATUS'
    HVACCOMPONENTADDRESS_CMD            = 'HVACADDR'                    // 'HVACADDR-<index>,<address>'
    HVACCOMPONENTADDRESS_QRY            = '?HVACADDR'                   // getHVACComponentAddress, '?HVACADDR-<index>'
    HVACCOMPONENTINDEX_QRY              = '?HVACIDX'                    // getHVACComponentIndex, '?HVACIDX-<hvacAddress>'
    REMOVEHVACCOMPONENT1_CMD            = 'HVACREMOVEADDR'              // removeHVACComponent, 'HVACREMOVEADDR-<hvacAddress>'
    REMOVEHVACCOMPONENT_CMD             = 'HVACREMOVEIDX'               // removeHVACComponent, 'HVACREMOVEIDX-<index>'
    TEMPERATURESCALE_CMD                = 'HVACSCALE'                   // setTemperatureScale, 'HVACSCALE-<ts>'
    TEMPERATURESCALE_QRY                = '?HVACSCALE'                  // getTemperatureScale, '?HVACSCALE'
  #END_IF // __HVAC_COMMANDS__

DEFINE_CONSTANT	//	----- IODevice : Commands Reference

  #IF_DEFINED __IODEVICE_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    IOCHANNELCOUNT_CMD                  = 'IOCHANNELCOUNT'              // setIOChannelCount, 'IOCHANNELCOUNT-<count>'
    IOCHANNELCOUNT_QRY                  = '?IOCHANNELCOUNT'             // getIOChannelCount, '?IOCHANNELCOUNT'
    IOCHANNELDIRECTION_CMD              = 'IOCHANNELDIRECTION'          // setIOChannelDirection, 'IOCHANNELDIRECTION-<channel>,<direction>'
    IOCHANNELDIRECTION_QRY              = '?IOCHANNELDIRECTION'         // getIOChannelDirection, '?IOCHANNELDIRECTION-<channel>'
    IOCHANNELINPUTSENSE_CMD             = 'IOCHANNELINPUTSENSE'         // setIOChannelInputSense, 'IOCHANNELINPUTSENSE-<channel>,<sense>'
    IOCHANNELINPUTSENSE_QRY             = '?IOCHANNELINPUTSENSE'        // getIOChannelState, '?IOCHANNELINPUTSENSE-<channel>'
    IOCHANNELSTATE_CMD                  = 'IOCHANNELSTATE'              // setIOChannelState, 'IOCHANNELSTATE-<channel>,<state>'
    IOCHANNELSTATE_QRY                  = '?IOCHANNELSTATE'             // getIOChannelState, '?IOCHANNELSTATE-<channel>'
  #END_IF // __IODEVICE_COMMANDS__

DEFINE_CONSTANT	//	----- Keypad : Channels

  KEYPAD_BTN1                           = 1                             // Discrete:  Keypad button 1 on
  KEYPAD_BTN1_FB                        = 1                             // Input:     Keypad button 1 feedback
  KEYPAD_BTN2                           = 2                             // Discrete:  Keypad button 2 on
  KEYPAD_BTN2_FB                        = 2                             // Input:     Keypad button 2 feedback
  KEYPAD_BTN3                           = 3                             // Discrete:  Keypad button 3 on
  KEYPAD_BTN3_FB                        = 3                             // Input:     Keypad button 3 feedback
  KEYPAD_BTN4                           = 4                             // Discrete:  Keypad button 4 on
  KEYPAD_BTN4_FB                        = 4                             // Input:     Keypad button 4 feedback
  KEYPAD_BTN5                           = 5                             // Discrete:  Keypad button 5 on
  KEYPAD_BTN5_FB                        = 5                             // Input:     Keypad button 5 feedback
  KEYPAD_BTN6                           = 6                             // Discrete:  Keypad button 6 on
  KEYPAD_BTN6_FB                        = 6                             // Input:     Keypad button 6 feedback
  KEYPAD_BTN7                           = 7                             // Discrete:  Keypad button 7 on
  KEYPAD_BTN7_FB                        = 7                             // Input:     Keypad button 7 feedback
  KEYPAD_BTN8                           = 8                             // Discrete:  Keypad button 8 on
  KEYPAD_BTN8_FB                        = 8                             // Input:     Keypad button 8 feedback
  KEYPAD_BTN9                           = 9                             // Discrete:  Keypad button 9 on
  KEYPAD_BTN9_FB                        = 9                             // Input:     Keypad button 9 feedback
  KEYPAD_BTN10                          = 10                            // Discrete:  Keypad button 10 on
  KEYPAD_BTN10_FB                       = 10                            // Input:     Keypad button 10 feedback
  KEYPAD_BTN1_BLINK                     = 101                           // Discrete:  Keypad button 1 blink
  KEYPAD_BTN1_BLINK_FB                  = 101                           // Feedback:  Keypad button 1 blink feedback
  KEYPAD_BTN2_BLINK                     = 102                           // Discrete:  Keypad button 2 blink
  KEYPAD_BTN2_BLINK_FB                  = 102                           // Feedback:  Keypad button 2 blink feedback
  KEYPAD_BTN3_BLINK                     = 103                           // Discrete:  Keypad button 3 blink
  KEYPAD_BTN3_BLINK_FB                  = 103                           // Feedback:  Keypad button 3 blink feedback
  KEYPAD_BTN4_BLINK                     = 104                           // Discrete:  Keypad button 4 blink
  KEYPAD_BTN4_BLINK_FB                  = 104                           // Feedback:  Keypad button 4 blink feedback
  KEYPAD_BTN5_BLINK                     = 105                           // Discrete:  Keypad button 5 blink
  KEYPAD_BTN5_BLINK_FB                  = 105                           // Feedback:  Keypad button 5 blink feedback
  KEYPAD_BTN6_BLINK                     = 106                           // Discrete:  Keypad button 6 blink
  KEYPAD_BTN6_BLINK_FB                  = 106                           // Feedback:  Keypad button 6 blink feedback
  KEYPAD_BTN7_BLINK                     = 107                           // Discrete:  Keypad button 7 blink
  KEYPAD_BTN7_BLINK_FB                  = 107                           // Feedback:  Keypad button 7 blink feedback
  KEYPAD_BTN8_BLINK                     = 108                           // Discrete:  Keypad button 8 blink
  KEYPAD_BTN8_BLINK_FB                  = 108                           // Feedback:  Keypad button 8 blink feedback
  KEYPAD_BTN9_BLINK                     = 109                           // Discrete:  Keypad button 9 blink
  KEYPAD_BTN9_BLINK_FB                  = 109                           // Feedback:  Keypad button 9 blink feedback
  KEYPAD_BTN10_BLINK                    = 110                           // Discrete:  Keypad button 10 blink
  KEYPAD_BTN10_BLINK_FB                 = 110                           // Feedback:  Keypad button 10 blink feedback

  #IF_DEFINED __KEYPAD_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADDKEYPADCOMPONENT_CMD              = 'KEYPADADD'                   // addKeypadComponent, 'KEYPADADD-<index>,<keypadaddr>'
    ADDKEYPADR_CMD                      = 'KEYPADADDR'                  // setButtonState, 'KEYPADBTN-<index>,<buttonaddr>,<state>'
    BUTTONSTATE_CMD                     = 'KEYPADBTN'                   // setButtonState, 'KEYPADBTN-<index>,<buttonaddr>,<state>'
    KEYPADCOMPONENTADDRESS_QRY          = '?KEYPADADDR'                 // getKeypadComponentAddress, '?KEYPADADDR-<index>'
    KEYPADCOMPONENTINDEX_QRY            = '?KEYPADIDX'                  // getKeypadComponentIndex, '?KEYPADIDX-<keypadaddr>'
    REMOVEKEYPADCOMPONENT1_CMD          = 'KEYPADREMOVEADDR'            // removeKeypadComponent, 'KEYPADREMOVEADDR-<keypadaddr>'
    REMOVEKEYPADCOMPONENT_CMD           = 'KEYPADREMOVEIDX'             // removeKeypadComponent, 'KEYPADREMOVEIDX-<index>'
  #END_IF // __KEYPAD_COMMANDS__

DEFINE_CONSTANT	//	----- KeypadSystem

  #IF_DEFINED __KEYPADSYSTEM_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADDKEYPADSYSTEMADDRESS_CMD          = 'KEYPADSYSTEMADDRESS'         // addKeypadSystemAddress, 'KEYPADSYSTEMADDRESS-<index>,<address>'
    ADDKEYPADSYSTEMCOMPONENT_CMD        = 'KEYPADSYSTEMINDEX'           // addKeypadSystemComponent, 'KEYPADSYSTEMADD-<index>,<address>'
    KEYPADSYSTEMBUTTONSTATE_CMD         = 'KEYPADSYSTEMBUTTONSTATE'     // setKeypadSystemButtonState, 'KEYPADSYSTEMBUTTONSTATE-<address>,<state>'
    KEYPADSYSTEMBUTTONSTATE_QRY         = '?KEYPADSYSTEMBUTTONSTATE'    // getKeypadSystemButtonState, '?KEYPADSYSTEMBUTTONSTATE-<address>'
    KEYPADSYSTEMBUTTONSTATUS_CMD        = 'KEYPADSYSTEMBUTTONSTATUS'    // setKeypadSystemButtonStatus, 'KEYPADSYSTEMBUTTONSTATUS-<address>,<status>'
    KEYPADSYSTEMBUTTONSTATUS_QRY        = '?KEYPADSYSTEMBUTTONSTATUS'   // getKeypadSystemButtonStatus, '?KEYPADSYSTEMBUTTONSTATUS-<address>'
    KEYPADSYSTEMCOMPONENTADDRESS_QRY    = '?KEYPADSYSTEMADDRESS'        // getKeypadSystemComponentAddress, '?KEYPADSYSTEMADDRESS-<index>'
    KEYPADSYSTEMCOMPONENTINDEX_QRY      = '?KEYPADSYSTEMINDEX'          // getKeypadSystemComponentIndex, '?KEYPADSYSTEMINDEX-<address>'
    REMOVEKEYPADSYSTEMCOMPONENT1_CMD    = 'KEYPADSYSTEMREMOVEADDRESS'   // removeKeypadSystemComponent, 'KEYPADSYSTEMREMOVEADDRESS-<address>'
    REMOVEKEYPADSYSTEMCOMPONENT_CMD     = 'KEYPADSYSTEMREMOVEINDEX'     // removeKeypadSystemComponent, 'KEYPADSYSTEMREMOVEINDEX-<index>'
  #END_IF // __KEYPADSYSTEM_COMMANDS__

DEFINE_CONSTANT	//	----- Light : Channels

  KEYPAD_BTN1                           = 1                             // Discrete:  Keypad button 1 on
  KEYPAD_BTN1_FB                        = 1                             // Input:     Keypad button 1 feedback
  KEYPAD_BTN2                           = 2                             // Discrete:  Keypad button 2 on
  KEYPAD_BTN2_FB                        = 2                             // Input:     Keypad button 2 feedback
  KEYPAD_BTN3                           = 3                             // Discrete:  Keypad button 3 on
  KEYPAD_BTN3_FB                        = 3                             // Input:     Keypad button 3 feedback
  KEYPAD_BTN4                           = 4                             // Discrete:  Keypad button 4 on
  KEYPAD_BTN4_FB                        = 4                             // Input:     Keypad button 4 feedback
  KEYPAD_BTN5                           = 5                             // Discrete:  Keypad button 5 on
  KEYPAD_BTN5_FB                        = 5                             // Input:     Keypad button 5 feedback
  KEYPAD_BTN6                           = 6                             // Discrete:  Keypad button 6 on
  KEYPAD_BTN6_FB                        = 6                             // Input:     Keypad button 6 feedback
  KEYPAD_BTN7                           = 7                             // Discrete:  Keypad button 7 on
  KEYPAD_BTN7_FB                        = 7                             // Input:     Keypad button 7 feedback
  KEYPAD_BTN8                           = 8                             // Discrete:  Keypad button 8 on
  KEYPAD_BTN8_FB                        = 8                             // Input:     Keypad button 8 feedback
  KEYPAD_BTN9                           = 9                             // Discrete:  Keypad button 9 on
  KEYPAD_BTN9_FB                        = 9                             // Input:     Keypad button 9 feedback
  KEYPAD_BTN10                          = 10                            // Discrete:  Keypad button 10 on
  KEYPAD_BTN10_FB                       = 10                            // Input:     Keypad button 10 feedback
  KEYPAD_BTN1_BLINK                     = 101                           // Discrete:  Keypad button 1 blink
  KEYPAD_BTN1_BLINK_FB                  = 101                           // Feedback:  Keypad button 1 blink feedback
  KEYPAD_BTN2_BLINK                     = 102                           // Discrete:  Keypad button 2 blink
  KEYPAD_BTN2_BLINK_FB                  = 102                           // Feedback:  Keypad button 2 blink feedback
  KEYPAD_BTN3_BLINK                     = 103                           // Discrete:  Keypad button 3 blink
  KEYPAD_BTN3_BLINK_FB                  = 103                           // Feedback:  Keypad button 3 blink feedback
  KEYPAD_BTN4_BLINK                     = 104                           // Discrete:  Keypad button 4 blink
  KEYPAD_BTN4_BLINK_FB                  = 104                           // Feedback:  Keypad button 4 blink feedback
  KEYPAD_BTN5_BLINK                     = 105                           // Discrete:  Keypad button 5 blink
  KEYPAD_BTN5_BLINK_FB                  = 105                           // Feedback:  Keypad button 5 blink feedback
  KEYPAD_BTN6_BLINK                     = 106                           // Discrete:  Keypad button 6 blink
  KEYPAD_BTN6_BLINK_FB                  = 106                           // Feedback:  Keypad button 6 blink feedback
  KEYPAD_BTN7_BLINK                     = 107                           // Discrete:  Keypad button 7 blink
  KEYPAD_BTN7_BLINK_FB                  = 107                           // Feedback:  Keypad button 7 blink feedback
  KEYPAD_BTN8_BLINK                     = 108                           // Discrete:  Keypad button 8 blink
  KEYPAD_BTN8_BLINK_FB                  = 108                           // Feedback:  Keypad button 8 blink feedback
  KEYPAD_BTN9_BLINK                     = 109                           // Discrete:  Keypad button 9 blink
  KEYPAD_BTN9_BLINK_FB                  = 109                           // Feedback:  Keypad button 9 blink feedback
  KEYPAD_BTN10_BLINK                    = 110                           // Discrete:  Keypad button 10 blink
  KEYPAD_BTN10_BLINK_FB                 = 110                           // Feedback:  Keypad button 10 blink feedback

  #IF_DEFINED __LIGHT_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADDKEYPADCOMPONENT_CMD              = 'KEYPADADD'                   // addKeypadComponent, 'KEYPADADD-<index>,<keypadaddr>'
    ADDKEYPADR_CMD                      = 'KEYPADADDR'                  // setButtonState, 'KEYPADBTN-<index>,<buttonaddr>,<state>'
    ADDLIGHTCOMPONENT_CMD               = 'LIGHTADD'                    // addLightComponent, 'LIGHTADD-<index>,<address>'
    BUTTONSTATE_CMD                     = 'KEYPADBTN'                   // setButtonState, 'KEYPADBTN-<index>,<buttonaddr>,<state>'
    CYCLELIGHT_CMD                      = 'LIGHTSTATE'                  // cycleLight, 'LIGHTSTATE-<index>,TOGGLE'
    KEYPADCOMPONENTADDRESS_QRY          = '?KEYPADADDR'                 // getKeypadComponentAddress, '?KEYPADADDR-<index>'
    KEYPADCOMPONENTINDEX_QRY            = '?KEYPADIDX'                  // getKeypadComponentIndex, '?KEYPADIDX-<keypadaddr>'
    LIGHTADDRESS_CMD                    = 'LIGHTADDR'
    LIGHTCOMPONENTADDRESS_QRY           = '?LIGHTADDR'                  // getLightComponentAddress, '?LIGHTADDR-<index>'
    LIGHTCOMPONENTINDEX_QRY             = '?LIGHTIDX'                   // getLightComponentIndex, '?LIGHTIDX-<address>'
    LIGHTLEVEL1_CMD                     = 'LIGHTLEVEL'                  // setLightLevel, 'LIGHTLEVEL-<index>,<level>,<time>'
    LIGHTLEVELRAMP1_CMD                 = 'LIGHTRAMP'                   // setLightLevelRamp, 'LIGHTRAMP-<index>,DOWN'
    LIGHTLEVELRAMP2_CMD                 = 'LIGHTRAMP'                   // setLightLevelRamp, 'LIGHTRAMP-<index>,STOP'
    LIGHTLEVELRAMP_CMD                  = 'LIGHTRAMP'                   // setLightLevelRamp, 'LIGHTRAMP-<index>,UP'
    LIGHTLEVEL_CMD                      = 'LIGHTLEVEL'                  // setLightLevel, 'LIGHTLEVEL-<index>,<level>'
    LIGHTLEVEL_QRY                      = '?LIGHTLEVEL'                 // getLightLevel, '?LIGHTLEVEL-<index>'
    LIGHTON_CMD                         = 'LIGHTSTATE'                  // setLightOn, 'LIGHTSTATE-<index>,<state>'
    LIGHTON_QRY                         = '?LIGHTSTATE'                 // isLightOn, '?LIGHTSTATE-<index>'
    REMOVEKEYPADCOMPONENT1_CMD          = 'KEYPADREMOVEADDR'            // removeKeypadComponent, 'KEYPADREMOVEADDR-<keypadaddr>'
    REMOVEKEYPADCOMPONENT_CMD           = 'KEYPADREMOVEIDX'             // removeKeypadComponent, 'KEYPADREMOVEIDX-<index>'
    REMOVELIGHTCOMPONENT1_CMD           = 'LIGHTREMOVEADDR'             // removeLightComponent, 'LIGHTREMOVEADDR-<address>'
    REMOVELIGHTCOMPONENT_CMD            = 'LIGHTREMOVEIDX'              // removeLightComponent, 'LIGHTREMOVEIDX-<index>'
  #END_IF // __LIGHT_COMMANDS__

DEFINE_CONSTANT	//	----- LightSystem : Commands Reference

  #IF_DEFINED __LIGHTSYSTEM_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADDKEYPADSYSTEMADDRESS_CMD          = 'KEYPADSYSTEMADDRESS'         // addKeypadSystemAddress, 'KEYPADSYSTEMADDRESS-<index>,<address>'
    ADDKEYPADSYSTEMCOMPONENT_CMD        = 'KEYPADSYSTEMINDEX'           // addKeypadSystemComponent, 'KEYPADSYSTEMADD-<index>,<address>'
    ADDLIGHTSYSTEMADDRESS_CMD           = 'LIGHTSYSTEMADDR'             // addLightSystemComponent, 'LIGHTSYSTEMADDR-<index>,<address>'
    CYCLELIGHTSYSTEMSTATE_CMD           = 'LIGHTSYSTEMSTATE'            // cycleLight, 'LIGHTSYSTEMSTATE-<address>,TOGGLE' toggles state
    KEYPADSYSTEMBUTTONSTATE_CMD         = 'KEYPADSYSTEMBUTTONSTATE'     // setKeypadSystemButtonState, 'KEYPADSYSTEMBUTTONSTATE-<address>,<state>'
    KEYPADSYSTEMBUTTONSTATE_QRY         = '?KEYPADSYSTEMBUTTONSTATE'    // getKeypadSystemButtonState, '?KEYPADSYSTEMBUTTONSTATE-<address>'
    KEYPADSYSTEMBUTTONSTATUS_CMD        = 'KEYPADSYSTEMBUTTONSTATUS'    // setKeypadSystemButtonStatus, 'KEYPADSYSTEMBUTTONSTATUS-<address>,<status>'
    KEYPADSYSTEMBUTTONSTATUS_QRY        = '?KEYPADSYSTEMBUTTONSTATUS'   // getKeypadSystemButtonStatus, '?KEYPADSYSTEMBUTTONSTATUS-<address>'
    KEYPADSYSTEMCOMPONENTADDRESS_QRY    = '?KEYPADSYSTEMADDRESS'        // getKeypadSystemComponentAddress, '?KEYPADSYSTEMADDRESS-<index>'
    KEYPADSYSTEMCOMPONENTINDEX_QRY      = '?KEYPADSYSTEMINDEX'          // getKeypadSystemComponentIndex, '?KEYPADSYSTEMINDEX-<address>'
    LIGHTSYSTEMADDRESS_CMD              = 'LIGHTSYSTEMADDR'
    LIGHTSYSTEMADDRESS_QRY              = '?LIGHTSYSTEMADDR'            // getLightSystemComponentAddress, '?LIGHTSYSTEMADDR-<index>'
    LIGHTSYSTEMINDEX_CMD                = 'LIGHTSYSTEMINDEX'
    LIGHTSYSTEMINDEX_QRY                = '?LIGHTSYSTEMINDEX'           // getLightSystemComponentIndex, '?LIGHTSYSTEMINDEX-<address>'
    LIGHTSYSTEMLEVELRAMP_CMD            = 'LIGHTSYSTEMRAMP'             // setLightLevelRamp, 'LIGHTSYSTEMRAMP-<address>,UP|DOWN|STOP'
    LIGHTSYSTEMLEVEL_CMD                = 'LIGHTSYSTEMLEVEL'            // setLightLevel, 'LIGHTSYSTEMLEVEL-<address>,<level>[,<interval>]'
    LIGHTSYSTEMLEVEL_QRY                = '?LIGHTSYSTEMLEVEL'           // getLightLevel, '?LIGHTSYSTEMLEVEL-<address>'
    LIGHTSYSTEMON_QRY                   = '?LIGHTSYSTEMSTATE'           // isLightOn, '?LIGHTSYSTEMSTATE-<address>'
    LIGHTSYSTEMSTATE_CMD                = 'LIGHTSYSTEMSTATE'            // setLightState, 'LIGHTSYSTEMSTATE-<address>,ON|OFF'
    REMOVEKEYPADSYSTEMCOMPONENT1_CMD    = 'KEYPADSYSTEMREMOVEADDRESS'   // removeKeypadSystemComponent, 'KEYPADSYSTEMREMOVEADDRESS-<address>'
    REMOVEKEYPADSYSTEMCOMPONENT_CMD     = 'KEYPADSYSTEMREMOVEINDEX'     // removeKeypadSystemComponent, 'KEYPADSYSTEMREMOVEINDEX-<index>'
    REMOVELIGHTSYSTEMADDRESS_CMD        = 'LIGHTSYSTEMREMOVEADDRESS'    // removeLightSystemComponent, 'LIGHTSYSTEMREMOVEADDRESS-<address>'
    REMOVELIGHTSYSTEMINDEX_CMD          = 'LIGHTSYSTEMREMOVEINDEX'      // removeLightSystemComponent, 'LIGHTSYSTEMREMOVEINDEX-<index>'
  #END_IF // __LIGHTSYSTEM_COMMANDS__

DEFINE_CONSTANT	//	----- Monitor : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  ASPECT_RATIO                          = 142                           // Momentary: Cycle aspect ratio
  BRIGHT_UP                             = 148                           // Momentary: Increment brightness
  BRIGHT_DN                             = 149                           // Momentary: Decrement brightness
  COLOR_UP                              = 150                           // Momentary: Increment color
  COLOR_DN                              = 151                           // Momentary: Decrement color
  CONTRAST_UP                           = 152                           // Momentary: Increment contrast
  CONTRAST_DN                           = 153                           // Momentary: Decrement contrast
  SHARP_UP                              = 154                           // Momentary: Increment sharpness
  SHARP_DN                              = 155                           // Momentary: Decrement sharpness
  TINT_UP                               = 156                           // Momentary: Increment tint
  TINT_DN                               = 157                           // Momentary: Decrement tint
  PIP_POS                               = 191                           // Momentary: Cycle pip position
  PIP_SWAP                              = 193                           // Momentary: Swap pip
  PIP                                   = 194                           // Momentary: Cycle pip
  PIP_ON                                = 195                           // Discrete:  Set pip on
  PIP_FB                                = 195                           // Feedback:  Pip feedback
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  PIC_MUTE                              = 210                           // Momentary: Cycle picture/video mute
  PIC_MUTE_ON                           = 211                           // Discrete:  Set picture/video mute on
  PIC_MUTE_FB                           = 211                           // Feedback:  Picture/video mute feedback
  PIC_FREEZE                            = 213                           // Momentary: Cycle freeze
  PIC_FREEZE_ON                         = 214                           // Discrete:  Set freeze on
  PIC_FREEZE_FB                         = 214                           // Feedback:  Freeze feedback
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay

  VOL_LVL                               = 1                             // Level: Volume level (0-255)
  BRIGHT_LVL                            = 10                            // Level: Brightness level (0-255)
  COLOR_LVL                             = 11                            // Level: Color level (0-255)
  CONTRAST_LVL                          = 12                            // Level: Contrast level (0-255)
  SHARP_LVL                             = 13                            // Level: Sharpness level (0-255)
  TINT_LVL                              = 14                            // Level: Tint level (0-255)

  #IF_DEFINED __MONITOR_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ACTIVEWINDOW_CMD                    = 'ACTIVEWINDOW'                // setActiveWindow, 'ACTIVEWINDOW-<mss>'
    ACTIVEWINDOW_QRY                    = '?ACTIVEWINDOW'               // getActiveWindow, '?ACTIVEWINDOW'
    ASPECTRATIOCOUNT_CMD                = 'ASPECTRATIOCOUNT'            // ASPECTRATIOCOUNT-<count>'
    ASPECTRATIOCOUNT_QRY                = '?ASPECTRATIOCOUNT'           // getAspectRatioCount(), '?ASPECTRATIOCOUNT'
    ASPECTRATIOPROPERTIES_CMD           = 'ASPECTRATIOPROPERTIES'       // '[ASPECTRATIOPROPERTIES-<index>,<displayName>,<value>]*'
    ASPECTRATIOPROPERTIES_QRY           = '?ASPECTRATIOPROPERTIES'      // getAspectRatioProperties(), '?ASPECTRATIOPROPERTIES'
    ASPECTRATIOPROPERTY_CMD             = 'ASPECTRATIOPROPERTY'         // 'ASPECTRATIOPROPERTY-<index>,<displayName>,<value>'
    ASPECTRATIOPROPERTY_QRY             = '?ASPECTRATIOPROPERTY'        // getAspectRatioProperty(index), '?ASPECTRATIOPROPERTY-<index>'
    ASPECTRATIOSELECT_CMD               = 'ASPECTRATIOSELECT'           // setAspectRatioSelect(index), 'ASPECTRATIOSELECT-<index>'
    ASPECTRATIOSELECT_QRY               = '?ASPECTRATIOSELECT'          // getAspectRatioSelect(), '?ASPECTRATIOSELECT'
    ASPECTRATIO_CMD                     = 'ASPECT'                      // setAspectRatio, 'ASPECT-<aspectRatio>'
    ASPECTRATIO_QRY                     = '?ASPECT'                     // getAspectRatio, '?ASPECT'
    CYCLEINPUTSELECT_CMD                = 'CYCLESELECT'                 // cycleInputSelect(), CYCLESELECT
    INPUTCOUNT_CMD                      = 'INPUTCOUNT'                  // INPUTCOUNT-<count>
    INPUTCOUNT_QRY                      = '?INPUTCOUNT'                 // getInputCount(), INPUTCOUNT-<count>
    INPUTGROUPSELECT_CMD                = 'INPUTGROUPSELECT'            // setInputGroupSelect(inputGroup), INPUTGROUPSELECT-<inputGroup>
    INPUTGROUPSELECT_QRY                = '?INPUTGROUPSELECT'           // getInputGroupSelect(), ?INPUTGROUPSELECT
    INPUTPROPERTIES_CMD                 = 'INPUTPROPERTIES'             // [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTIES_QRY                 = '?INPUTPROPERTIES'            // getInputProperties(), [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTY_CMD                   = 'INPUTPROPERTY'               // INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTPROPERTY_QRY                   = '?INPUTPROPERTY'              // getInputProperty(index), INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTSELECT_CMD                     = 'INPUTSELECT'                 // setInputSelect(index), INPUTSELECT-<index>
    INPUTSELECT_QRY                     = '?INPUTSELECT'                // getInputSelect(), INPUTSELECT-<index>
    INPUTSOURCE_CMD                     = 'INPUT'                       // setInputSource, INPUT-<sourceSelect>,<inputNumber>
    INPUTSOURCE_QRY                     = '?INPUT'                      // getInputSource, ?INPUT
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    VIDEOTYPE_CMD                       = 'VIDEOTYPE'                   // setVideoType, 'VIDEOTYPE-<vt>'
    VIDEOTYPE_QRY                       = '?VIDEOTYPE'                  // getVideoType, '?VIDEOTYPE'
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
  #END_IF // __MONITOR_COMMANDS__

DEFINE_CONSTANT	//	----- Motor : Channels and Levels

  MOTOR_STOP                            = 2                             // Momentary: Motor stop
  MOTOR_STOP_FB                         = 2                             // Feedback:  Motor stop feedback
  MOTOR_OPEN                            = 4                             // Momentary: Motor open
  MOTOR_OPEN_FB                         = 4                             // Feedback:  Motor open feedback
  MOTOR_CLOSE                           = 5                             // Momentary: Motor close
  MOTOR_CLOSE_FB                        = 5                             // Feedback:  Motor close feedback
  MOTOR_PRESET                          = 187                           // Momentary: Cycle motor preset

  MOTOR_POS_LVL                         = 6                             // Level: Motor position (0=close, 255=open)

  #IF_DEFINED __MOTOR_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    MOTORPRESET_CMD                     = 'MOTORPRESET'                 // setMotorPreset, 'MOTORPRESET-<preset>'
    MOTORPRESET_QRY                     = '?MOTORPRESET'                // getMotorPreset, '?MOTORPRESET'
  #END_IF // __MOTOR_COMMANDS__

DEFINE_CONSTANT	//	----- Multi Window : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  PAN_UP                                = 132                           // Momentary: Pan up
  PAN_DN                                = 133                           // Momentary: Pan down
  PAN_LT                                = 134                           // Momentary: Pan left
  PAN_RT                                = 135                           // Momentary: Pan right
  MULTIWIN_PRESET                       = 136                           // Momentary: Cycle multi-window preset
  BRIGHT_UP                             = 148                           // Momentary: Increment brightness
  BRIGHT_DN                             = 149                           // Momentary: Decrement brightness
  COLOR_UP                              = 150                           // Momentary: Increment color
  COLOR_DN                              = 151                           // Momentary: Decrement color
  CONTRAST_UP                           = 152                           // Momentary: Increment contrast
  CONTRAST_DN                           = 153                           // Momentary: Decrement contrast
  SHARP_UP                              = 154                           // Momentary: Increment sharpness
  SHARP_DN                              = 155                           // Momentary: Decrement sharpness
  TINT_UP                               = 156                           // Momentary: Increment tint
  TINT_DN                               = 157                           // Momentary: Decrement tint
  ZOOM_OUT                              = 158                           // Momentary: Zoom out
  ZOOM_IN                               = 159                           // Momentary: Zoom in
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  PIC_MUTE                              = 210                           // Momentary: Cycle picture/video mute
  PIC_MUTE_ON                           = 211                           // Discrete:  Set picture/video mute on
  PIC_MUTE_FB                           = 211                           // Feedback:  Picture/video mute feedback
  PIC_FREEZE                            = 213                           // Momentary: Cycle freeze
  PIC_FREEZE_ON                         = 214                           // Discrete:  Set freeze on
  PIC_FREEZE_FB                         = 214                           // Feedback:  Freeze feedback
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay

  BRIGHT_LVL                            = 10                            // Level: Brightness level (0-255)
  COLOR_LVL                             = 11                            // Level: Color level (0-255)
  CONTRAST_LVL                          = 12                            // Level: Contrast level (0-255)
  SHARP_LVL                             = 13                            // Level: Sharpness level (0-255)
  TINT_LVL                              = 14                            // Level: Tint level (0-255)

  #IF_DEFINED __MULTIWINDOW_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADDWINDOWCOMPONENT_CMD              = 'WINDOWADD'                   // addWindowComponent, 'WINDOWADD-<index>,<address>'
    IMAGEPOSITION_QRY                   = '?IMAGE'                      // getImagePosition, '?IMAGE'
    IMAGESIZE_QRY                       = '?IMAGE'                      // getImageSize, '?IMAGE'
    IMAGESOURCE_CMD                     = 'IMAGESOURCE'                 // setImageSource, 'IMAGESOURCE-<source>'
    IMAGESOURCE_QRY                     = '?IMAGE'                      // getImageSource, '?IMAGE'
    IMAGE_CMD                           = 'IMAGE'                       // setImage, 'IMAGE-<source>,<x>,<y>,<height>,<width>'
    MULTIWINDOWPRESET_CMD               = 'MULTIWINPRESET'              // setMultiWindowPreset, 'MULTIWINPRESET-<preset>'
    MULTIWINDOWPRESET_QRY               = '?MULTIWINPRESET'             // getMultiWindowPreset, '?MULTIWINPRESET'
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    REMOVEWINDOWCOMPONENT1_CMD          = 'WINDOWREMOVEADDR'            // removeWindowComponent, 'WINDOWREMOVEADDR-<address>'
    REMOVEWINDOWCOMPONENT_CMD           = 'WINDOWREMOVEIDX'             // removeWindowComponent, 'WINDOWREMOVEIDX-<index>'
    SAVEMULTIWINDOWPRESET_CMD           = 'MULTIWINPRESETSAVE'          // saveMultiWindowPreset, 'MULTIWINPRESETSAVE-<preset>'
    WINDOWADDR_CMD                      = 'WINDOWADDR'
    WINDOWCOMPONENTADDRESS_QRY          = '?WINDOWADDR'                 // getWindowComponentAddress, '?WINDOWADDR-<index>'
    WINDOWCOMPONENTINDEX_QRY            = '?WINDOWIDX'                  // getWindowComponentIndex, '?WINDOWIDX-<address>'
    WINDOWPOSITION_QRY                  = '?WINDOW'                     // getWindowPosition, '?WINDOW'
    WINDOWSIZE_QRY                      = '?WINDOW'                     // getWindowSize, '?WINDOW'
    WINDOW_CMD                          = 'WINDOW'                      // setWindow, 'WINDOW-<x>,<y>,<height>,<width>'
    ZORDER1_CMD                         = 'WINDOWZORDER'                // setZOrder, 'WINDOWZORDER-<value>'
    ZORDER_CMD                          = 'WINDOWZORDER'                // setZOrder, 'WINDOWZORDER-<position>'
    ZORDER_QRY                          = '?WINDOWZORDER'               // getZOrder, '?WINDOWZORDER'
  #END_IF // __MULTIWINDOW_COMMANDS__

DEFINE_CONSTANT	//	----- Pool Spa : Channels and Levels

  POOL_HEAT                             = 123                           // Momentary: Cycle pool heat state
  SPA_HEAT                              = 124                           // Momentary: Cycle spa heat state
  SPA_JETS                              = 125                           // Momentary: Cycle spa jets
  POOL_HEAT_UP                          = 152                           // Momentary: Increment pool setpoint
  POOL_HEAT_DN                          = 153                           // Momentary: Decrement pool setpoint
  SPA_HEAT_UP                           = 154                           // Momentary: Increment spa setpoint
  SPA_HEAT_DN                           = 155                           // Momentary: Decrement spa setpoint
  POOL_PUMP_ON                          = 170                           // Discrete:  Set pool pump state on or off
  POOL_PUMP_FB                          = 170                           // Feedback:  Pool pump feedback
  SPA_PUMP_ON                           = 171                           // Discrete:  Set spa pump state on or off
  SPA_PUMP_FB                           = 171                           // Feedback:  Spa pump feedback
  POOL_LIGHT_ON                         = 172                           // Discrete:  Set pool lights on or off
  POOL_LIGHT_FB                         = 172                           // Feedback:  Pool light feedback
  SPA_LIGHT_ON                          = 173                           // Discrete:  Set spa lights on or off
  SPA_LIGHT_FB                          = 173                           // Feedback:  Spa light feedback
  POOL_HEAT_OFF                         = 174                           // Momentary: Set pool heat to off
  POOL_HEAT_OFF_FB                      = 174                           // Feedback:  Pool heat set to off feedback
  POOL_HEATER                           = 175                           // Momentary: Set pool heat to heater
  POOL_HEATER_FB                        = 175                           // Feedback:  Pool heat set to heater feedback
  POOL_SOLAR                            = 176                           // Momentary: Set pool heat to solar preferred
  POOL_SOLAR_FB                         = 176                           // Feedback:  Pool heat set to solar feedback
  POOL_SOLAR_PREF                       = 177                           // Momentary: Set pool heat to solar preferred
  POOL_SOLAR_PREF_FB                    = 177                           // Feedback:  Pool heat set to solar preferred feedback
  SPA_HEAT_OFF                          = 178                           // Momentary: Set spa heat to off
  SPA_HEAT_OFF_FB                       = 178                           // Feedback:  Set spa heat to off feedback
  SPA_HEATER                            = 179                           // Momentary: Set spa heat to heater
  SPA_HEATER_FB                         = 179                           // Feedback:  Set spa heat to heater feedback
  SPA_SOLAR                             = 180                           // Momentary: Set spa heat to solar
  SPA_SOLAR_FB                          = 180                           // Feedback:  Set spa heat to solar feedback
  SPA_SOLAR_PREF                        = 181                           // Momentary: Set spa heat to solar preferred
  SPA_SOLAR_PREF_FB                     = 181                           // Feedback:  Set spa heat to solar preferred feedback
  SPA_JETS_OFF                          = 182                           // Momentary: Set spa jets to off
  SPA_JETS_OFF_FB                       = 182                           // Feedback:  Spa jets off feedback
  SPA_JETS_LO                           = 183                           // Momentary: Set spa jets to low
  SPA_JETS_LO_FB                        = 183                           // Feedback:  Spa jets on low feedback
  SPA_JETS_MED                          = 184                           // Momentary: Set spa jets to medium
  SPA_JETS_MED_FB                       = 184                           // Feedback:  Spa jets on medium feedback
  SPA_JETS_HI                           = 185                           // Momentary: Set spa jets to high
  SPA_JETS_HI_FB                        = 185                           // Feedback:  Spa jets on high feedback
  SPA_BLOWER_ON                         = 186                           // Discrete:  Set spa blower on or off
  SPA_BLOWER_FB                         = 186                           // Feedback:  Spa blower feedback
  POOL_HEATING                          = 187                           // Feedback:  Pool is heating using heater
  POOL_HEATING_SOLAR                    = 188                           // Feedback:  Pool is heating using solar
  SPA_HEATING                           = 189                           // Feedback:  Spa is heating using heater
  SPA_HEATING_SOLAR                     = 190                           // Feedback:  Spa is heating using solar

  OUTDOOR_TEMP_LVL                      = 34                            // Level: Pool/Spa outdoor temperature, range is n..m degrees F or C
  POOL_HEAT_LVL                         = 39                            // Level: Pool heating set point, range is n..m degrees F or C
  SPA_HEAT_LVL                          = 40                            // Level: Spa heating set point, range is n..m degrees F or C
  POOL_TEMP_LVL                         = 41                            // Level: Pool water temperature, range is n..m degrees F or C
  SPA_TEMP_LVL                          = 42                            // Level: Spa water temperature, range is n..m degrees F or C

  #IF_DEFINED __POOLSPA_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    POOLSPAAUXON_CMD                    = 'POOLAUX'                     // setPoolSpaAuxOn, 'POOLAUX-<aux name>,<state>'
    POOLSPAAUXON_QRY                    = '?POOLAUX'                    // isPoolSpaAuxOn, '?POOLAUX-<aux name>'
    POOLSPATEMPERATURESCALE_CMD         = 'POOLSCALE'                   // setPoolSpaTemperatureScale, 'POOLSCALE-<ts>'
    POOLSPATEMPERATURESCALE_QRY         = '?POOLSCALE'                  // getPoolSpaTemperatureScale, '?POOLSCALE'
  #END_IF // __POOLSPA_COMMANDS__

DEFINE_CONSTANT	//	----- Pre Amp Surround Sound Processor : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  BALANCE_UP                            = 164                           // Ramping:   Increment balance
  BALANCE_DN                            = 165                           // Ramping:   Decrement balance
  BASS_UP                               = 166                           // Ramping:   Increment bass
  BASS_DN                               = 167                           // Ramping:   Decrement bass
  TREBLE_UP                             = 168                           // Ramping:   Increment treble
  TREBLE_DN                             = 169                           // Ramping:   Decrement treble
  SURROUND_NEXT                         = 170                           // Momentary: Next surround sound mode
  SURROUND_PREV                         = 171                           // Momentary: Previous surround sound mode
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  LOUDNESS                              = 206                           // Momentary: Cycle loudness
  LOUDNESS_ON                           = 207                           // Discrete:  Set loudness state
  LOUDNESS_FB                           = 207                           // Feedback:  Loudness state feedback
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay

  VOL_LVL                               = 1                             // Level: Volume level (0-255)
  BALANCE_LVL                           = 2                             // Level: Balance level (0=left, 255=right)
  BASS_LVL                              = 3                             // Level: Bass level (0-255)
  TREBLE_LVL                            = 4                             // Level: Treble level (0-255)

  #IF_DEFINED __PREAMPSURROUNDSOUNDPROCESSOR_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADJUSTEQUALIZEROFFSET_CMD           = 'EQUALIZER_OFFSET'            // setEqualizerOffset overloads, 'EQUALIZER_OFFSET-(band)=(offset),[,EQUALIZER-(band)=(offset)]+'
    CYCLEINPUTSELECT_CMD                = 'CYCLESELECT'                 // cycleInputSelect(), CYCLESELECT
    EQUALIZERBANDS_CMD                  = 'EQUALIZER_BANDS'             // return 'EQUALIZER_BANDS-(band)[,(band)]+'
    EQUALIZERBANDS_QRY                  = '?EQUALIZER_BANDS'            // getEqualizerBands, '?EQUALIZER_BANDS'
    EQUALIZER_CMD                       = 'EQUALIZER'                   // setEqualizer overloads, 'EQUALIZER-(band)=(gain),[,EQUALIZER-(band)=(gain)]+'
    EQUALIZER_QRY                       = '?EQUALIZER'                  // getEqualizer overloads, '?EQUALIZER', '?EQUALIZER-(band)[,(band)]+'
    INPUTCOUNT_CMD                      = 'INPUTCOUNT'                  // INPUTCOUNT-<count>
    INPUTCOUNT_QRY                      = '?INPUTCOUNT'                 // getInputCount(), INPUTCOUNT-<count>
    INPUTGROUPSELECT_CMD                = 'INPUTGROUPSELECT'            // setInputGroupSelect(inputGroup), INPUTGROUPSELECT-<inputGroup>
    INPUTGROUPSELECT_QRY                = '?INPUTGROUPSELECT'           // getInputGroupSelect(), ?INPUTGROUPSELECT
    INPUTPROPERTIES_CMD                 = 'INPUTPROPERTIES'             // [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTIES_QRY                 = '?INPUTPROPERTIES'            // getInputProperties(), [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTY_CMD                   = 'INPUTPROPERTY'               // INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTPROPERTY_QRY                   = '?INPUTPROPERTY'              // getInputProperty(index), INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTSELECT_CMD                     = 'INPUTSELECT'                 // setInputSelect(index), INPUTSELECT-<index>
    INPUTSELECT_QRY                     = '?INPUTSELECT'                // getInputSelect(), INPUTSELECT-<index>
    INPUTSOURCE_CMD                     = 'INPUT'                       // setInputSource, INPUT-<sourceSelect>,<inputNumber>
    INPUTSOURCE_QRY                     = '?INPUT'                      // getInputSource, ?INPUT
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    SURROUNDMODE_CMD                    = 'SURROUND'                    // setSurroundMode, 'SURROUND-<mode>'
    SURROUNDMODE_QRY                    = '?SURROUND'                   // getSurroundMode, '?SURROUND'
		SURROUNDMODES_CMD                   = 'SURROUNDMODES'               // SurroundModes, 'SURROUNDMODES-<mode>,<mode>...'
	  SURROUNDMODES_QRY                   = '?SURROUNDMODES'              // getSurroundMode, '?SURROUND'
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
  #END_IF // __PREAMPSURROUNDSOUNDPROCESSOR_COMMANDS__

DEFINE_CONSTANT	//	----- Receiver : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  CHAN_UP                               = 22                            // Momentary: Next station preset
  CHAN_DN                               = 23                            // Momentary: Previous station preset
  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  TUNER_BAND                            = 40                            // Momentary: Cycle tuner band
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  BALANCE_UP                            = 164                           // Ramping:   Increment balance
  BALANCE_DN                            = 165                           // Ramping:   Decrement balance
  BASS_UP                               = 166                           // Ramping:   Increment bass
  BASS_DN                               = 167                           // Ramping:   Decrement bass
  TREBLE_UP                             = 168                           // Ramping:   Increment treble
  TREBLE_DN                             = 169                           // Ramping:   Decrement treble
  SURROUND_NEXT                         = 170                           // Momentary: Next surround sound mode
  SURROUND_PREV                         = 171                           // Momentary: Previous surround sound mode
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  LOUDNESS                              = 206                           // Momentary: Cycle loudness
  LOUDNESS_ON                           = 207                           // Discrete:  Set loudness state
  LOUDNESS_FB                           = 207                           // Feedback:  Loudness state feedback
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay
  TUNER_PRESET_GROUP                    = 224                           // Momentary: Cycle station preset group
  TUNER_STATION_UP                      = 225                           // Momentary: Increment tuner station
  TUNER_STATION_DN                      = 226                           // Momentary: Decrement tuner station
  TUNER_SCAN_FWD                        = 227                           // Momentary: Station scan forward
  TUNER_SCAN_REV                        = 228                           // Momentary: Station scan backward/reverse
  TUNER_SEEK_FWD                        = 229                           // Momentary: Station seek forward
  TUNER_SEEK_REV                        = 230                           // Momentary: Station seek backward/reverse
  TUNER_OSD                             = 234                           // Momentary: Cycle on-screen or front panel display info
  TUNER_PREV                            = 235                           // Momentary: Goto previous tuner station

  VOL_LVL                               = 1                             // Level: Volume level (0-255)
  BALANCE_LVL                           = 2                             // Level: Balance level (0=left, 255=right)
  BASS_LVL                              = 3                             // Level: Bass level (0-255)
  TREBLE_LVL                            = 4                             // Level: Treble level (0-255)

  #IF_DEFINED __RECEIVER_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADJUSTEQUALIZEROFFSET_CMD           = 'EQUALIZER_OFFSET'            // setEqualizerOffset overloads, 'EQUALIZER_OFFSET-(band)=(offset),[,EQUALIZER-(band)=(offset)]+'
    BAND_CMD                            = 'BAND'                        // setBand, 'BAND-<tb>'
    BAND_QRY                            = '?BAND'                       // getBand, '?BAND'
    CYCLEINPUTSELECT_CMD                = 'CYCLESELECT'                 // cycleInputSelect(), CYCLESELECT
    EQUALIZERBANDS_CMD                  = 'EQUALIZER_BANDS'             // return 'EQUALIZER_BANDS-(band)[,(band)]+'
    EQUALIZERBANDS_QRY                  = '?EQUALIZER_BANDS'            // getEqualizerBands, '?EQUALIZER_BANDS'
    EQUALIZER_CMD                       = 'EQUALIZER'                   // setEqualizer overloads, 'EQUALIZER-(band)=(gain),[,EQUALIZER-(band)=(gain)]+'
    EQUALIZER_QRY                       = '?EQUALIZER'                  // getEqualizer overloads, '?EQUALIZER', '?EQUALIZER-(band)[,(band)]+'
    INPUTCOUNT_CMD                      = 'INPUTCOUNT'                  // INPUTCOUNT-<count>
    INPUTCOUNT_QRY                      = '?INPUTCOUNT'                 // getInputCount(), INPUTCOUNT-<count>
    INPUTGROUPSELECT_CMD                = 'INPUTGROUPSELECT'            // setInputGroupSelect(inputGroup), INPUTGROUPSELECT-<inputGroup>
    INPUTGROUPSELECT_QRY                = '?INPUTGROUPSELECT'           // getInputGroupSelect(), ?INPUTGROUPSELECT
    INPUTPROPERTIES_CMD                 = 'INPUTPROPERTIES'             // [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTIES_QRY                 = '?INPUTPROPERTIES'            // getInputProperties(), [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTY_CMD                   = 'INPUTPROPERTY'               // INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTPROPERTY_QRY                   = '?INPUTPROPERTY'              // getInputProperty(index), INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTSELECT_CMD                     = 'INPUTSELECT'                 // setInputSelect(index), INPUTSELECT-<index>
    INPUTSELECT_QRY                     = '?INPUTSELECT'                // getInputSelect(), INPUTSELECT-<index>
    INPUTSOURCE_CMD                     = 'INPUT'                       // setInputSource, INPUT-<sourceSelect>,<inputNumber>
    INPUTSOURCE_QRY                     = '?INPUT'                      // getInputSource, ?INPUT
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    STATIONPRESETCOUNT_CMD              = 'STATIONPRESETCOUNT'          // 'STATIONPRESETCOUNT-<count>'
    STATIONPRESETCOUNT_QRY              = '?STATIONPRESETCOUNT'         // getStationPresetCount, '?STATIONPRESETCOUNT'
    STATIONPRESETPROPERTIES_CMD         = 'STATIONPRESETPROPERTIES'     // '[STATIONPRESETPROPERTIES-<index>,<displayName>,<station>]*'
    STATIONPRESETPROPERTIES_QRY         = '?STATIONPRESETPROPERTIES'    // getStationPresetProperties, '?STATIONPRESETPROPERTIES'
    STATIONPRESETPROPERTY_CMD           = 'STATIONPRESETPROPERTY'       // 'STATIONPRESETPROPERTY-<index>,<displayName>,<station>'
    STATIONPRESETPROPERTY_QRY           = '?STATIONPRESETPROPERTY'      // getStationPresetProperty(index), '?STATIONPRESETPROPERTY-<index>'
    STATIONPRESETSELECT_CMD             = 'STATIONPRESETSELECT'         // 'STATIONPRESETSELECT-<index>'
    STATIONPRESETSELECT_QRY             = '?STATIONPRESETSELECT'        // getStationPresetSelect, '?STATIONPRESETSELECT'
    STATIONPRESET_CMD                   = 'TUNERPRESET'                 // setStationPreset, 'TUNERPRESET-<preset>'
    STATIONPRESET_QRY                   = '?TUNERPRESET'                // getStationPreset, '?TUNERPRESET'
    STATION_CMD                         = 'XCH'                         // setStation, 'XCH-<station>'
    STATION_QRY                         = '?XCH'                        // getStation, '?XCH'
    SURROUNDMODE_CMD                    = 'SURROUND'                    // setSurroundMode, 'SURROUND-<mode>'
    SURROUNDMODE_QRY                    = '?SURROUND'                   // getSurroundMode, '?SURROUND'
		SURROUNDMODES_CMD                   = 'SURROUNDMODES'               // SurroundModes, 'SURROUNDMODES-<mode>,<mode>...'
	  SURROUNDMODES_QRY                   = '?SURROUNDMODES'              // getSurroundMode, '?SURROUND'
    TUNERBANDPROPERTIES_CMD             = 'TUNERBANDPROPERTIES'         // '[TUNERBANDPROPERTIES-<index>,<displayName>,<tunerBand>]*'
    TUNERBANDPROPERTIES_QRY             = '?TUNERBANDPROPERTIES'        // getTunerBandProperties, '?TUNERBANDPROPERTIES'
    TUNERBANDPROPERTYCOUNT_CMD          = 'TUNERBANDPROPERTYCOUNT'      // 'TUNERBANDPROPERTYCOUNT-<count>'
    TUNERBANDPROPERTYCOUNT_QRY          = '?TUNERBANDPROPERTYCOUNT'     // getTunerBandPropertyCount, '?TUNERBANDPROPERTYCOUNT'
    TUNERBANDPROPERTY_CMD               = 'TUNERBANDPROPERTY'           // 'TUNERBANDPROPERTY-<index>,<displayName>,<tunerBand>'
    TUNERBANDPROPERTY_QRY               = '?TUNERBANDPROPERTY'          // getTunerBandProperty(index), '?TUNERBANDPROPERTY-<index>'
    TUNERBANDSELECT_CMD                 = 'TUNERBANDSELECT'             // 'TUNERBANDSELECT-<index>'
    TUNERBANDSELECT_QRY                 = '?TUNERBANDSELECT'            // getTunerBandSelect, '?TUNERBANDSELECT'
    TUNERCOMPONENTPROPERTIES_CMD        = 'TUNERCOMPONENTPROPERTIES'    // '[TUNERCOMPONENTPROPERTIES-<index>,<displayName>,<value>]*'
    TUNERCOMPONENTPROPERTIES_QRY        = '?TUNERCOMPONENTPROPERTIES'   // getTunerComponentProperties, '?TUNERCOMPONENTPROPERTIES'
    TUNERCOMPONENTPROPERTYCOUNT_CMD     = 'TUNERCOMPONENTPROPERTYCOUNT' // 'TUNERCOMPONENTPROPERTYCOUNT-<count>'
    TUNERCOMPONENTPROPERTYCOUNT_QRY     = '?TUNERCOMPONENTPROPERTYCOUNT'// getTunerComponentPropertyCount, '?TUNERCOMPONENTPROPERTYCOUNT'
    TUNERCOMPONENTPROPERTY_CMD          = 'TUNERCOMPONENTPROPERTY'      // 'TUNERCOMPONENTPROPERTY-<index>,<displayName>,<value>'
    TUNERCOMPONENTPROPERTY_QRY          = '?TUNERCOMPONENTPROPERTY'     // getTunerComponentProperty(index), '?TUNERCOMPONENTPROPERTY-<index>'
    TUNERCOMPONENTSELECT_CMD            = 'TUNERCOMPONENTSELECT'        // 'TUNERCOMPONENTSELECT-<index>'
    TUNERCOMPONENTSELECT_QRY            = '?TUNERCOMPONENTSELECT'       // getTunerComponentSelect, '?TUNERCOMPONENTSELECT'
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
  #END_IF // __RECEIVER_COMMANDS__

DEFINE_CONSTANT	//	----- RelayDevice : Commands Reference

  #IF_DEFINED __RELAYDEVICE_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    RELAYCHANNELCOUNT_CMD               = 'RELAYCHANNELCOUNT'           // setRELAYChannelCount, 'RELAYCHANNELCOUNT-<count>'
    RELAYCHANNELCOUNT_QRY               = '?RELAYCHANNELCOUNT'          // getRELAYChannelCount, '?RELAYCHANNELCOUNT'
    RELAYCHANNELSTATE_CMD               = 'RELAYCHANNELSTATE'           // setRELAYChannelState, 'RELAYCHANNELSTATE-<channel>,<state>'
    RELAYCHANNELSTATE_QRY               = '?RELAYCHANNELSTATE'          // getRELAYChannelState, '?RELAYCHANNELSTATE-<channel>'
  #END_IF // __RELAYDEVICE_COMMANDS__

DEFINE_CONSTANT	//	----- RFIDSystem : Commands Reference

  #IF_DEFINED __RFIDSYSTEM_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADDRFIDSYSTEMADDRESS_CMD            = 'RFIDSYSTEMADDR'              // addRFIDSystemComponent, 'RFIDSYSTEMADDR-<index>,<address>'
    PROCESSREADERSBYTAG_CMD             = 'READERSBYTAG'                // processReadersByTagEvent, 'READERSBYTAG- READERSBYTAG-<tagId>,<tagName>,<tagInfo>,<tagPercentPower>,<readerCount>,<readerIndex> [,<readerAddress>,<tagSignalStrength>]*'
    PROCESSREADERSTATUS_CMD             = 'READERSTATUS'                // processReaderStatusEvent, 'READERSTATUS-<readerAddress>,<readerStatus>,<lastError>'
    PROCESSREADERS_CMD                  = 'READERS'                     // processReadersEvent, 'READERS-<readerAddress1>[,readerAddressN>]*'
    PROCESSTAGACQUIRED_CMD              = 'TAGACQUIRED'                 // processTagAcquiredEvent, 'TAGACQUIRED-<readerAddress>,<tagId>,<tagName>,<tagInfo>,<timestamp>, <tagSignalStrength>,<tagPercentPower>'
    PROCESSTAGBUTTON_CMD                = 'TAGBUTTON'                   // processTagLostEvent, 'TAGBUTTON-<tagId>,<tagName>,<tagInfo>,<state>'
    PROCESSTAGINFO_CMD                  = 'TAGINFO'                     // processTagInfoEvent, 'TAGINFO-<tagId>,<tagName>,<tagInfo>'
    PROCESSTAGLOST_CMD                  = 'TAGLOST'                     // processTagLostEvent, 'TAGLOST-<readerAddress>,<tagId>,<tagName>,<tagInfo>,<lastTimestamp>'
    PROCESSTAGSBYREADER_CMD             = 'TAGSBYREADER'                // processTagsByReaderEvent, 'TAGSBYREADER-<readerAddress>,<tagCount>,<tagIndex> [,<tagId>,<tagName>,<tagInfo>,<tagSignalStrength>,<tagPercentPower>]*'
    PROCESSTAGSIGNALSTRENGTH_CMD        = 'TAGSIGNALSTRENGTH'           // processTagAcquiredEvent, 'TAGSIGNALSTRENGTH-<readerAddress>,<tagId>,<tagName>,<tagInfo>,<timestamp>, <tagSignalStrength>,<tagPercentPower>'
    READERAUTOPOLL_CMD                  = 'READERAUTOPOLL'              // SetReaderAutoPoll, 'READERAUTOPOLL-<readerAddress>,<autoPoll>[,<interval>]'
    READERSBYTAG_QRY                    = '?READERSBYTAG'               // getReadersByTag, '?READERSBYTAG-<tagId>'
    READERSTATUS_QRY                    = '?READERSTATUS'               // getReaderStatus, '?READERSTATUS-<readerAddress>'
    READERS_QRY                         = '?READERS'                    // getReaders, '?READERS'
    REMOVERFIDSYSTEMADDRESS_CMD         = 'RFIDSYSTEMREMOVEADDRESS'     // removeRFIDSystemComponent, 'RFIDSYSTEMREMOVEADDRESS-<address>'
    REMOVERFIDSYSTEMINDEX_CMD           = 'RFIDSYSTEMREMOVEINDEX'       // removeRFIDSystemComponent, 'RFIDSYSTEMREMOVEINDEX-<index>'
    RFIDSYSTEMADDRESS_CMD               = 'RFIDSYSTEMADDR'
    RFIDSYSTEMADDRESS_QRY               = '?RFIDSYSTEMADDR'             // getRFIDSystemComponentAddress, '?RFIDSYSTEMADDR-<index>'
    RFIDSYSTEMINDEX_CMD                 = 'RFIDSYSTEMINDEX'
    RFIDSYSTEMINDEX_QRY                 = '?RFIDSYSTEMINDEX'            // getRFIDSystemComponentIndex, '?RFIDSYSTEMINDEX-<address>'
    TAGINFO_QRY                         = '?TAGINFO'                    // getTagInfo, '?TAGINFO-<tagId>'
    TAGSBYREADER_QRY                    = '?TAGSBYREADER'               // getTagsByReader, '?TAGSBYREADER-<readerAddress>'
  #END_IF // __RFIDSYSTEM_COMMANDS__

DEFINE_CONSTANT	//	----- SecuritySystem : Commands Reference

  #IF_DEFINED __SECURITYSYSTEM_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    OKTOARM_QRY                         = '?SECARMABLE'                 // isOKToArm, '?SECARMABLE'
    POINTSTATE_CMD                      = 'SECPOINTSTATE'               // setPointState, 'SECPOINTSTATE-<point>,<state>,<password>'
    POINTSTATE_QRY                      = '?SECPOINTSTATE'              // getPointState, '?SECPOINTSTATE-<point>'
    POINTSTATUS_QRY                     = '?SECPOINTSTATUS'             // getPointStatus, '?SECPOINTSTATUS-<point>'
    SECURITYSTATE_CMD                   = 'SECSTATE'                    // setSecurityState, 'SECSTATE-<state>,<password>'
    SECURITYSTATE_QRY                   = '?SECSTATE'                   // getSecurityState, '?SECSTATE'
    SECURITYSTATUS_QRY                  = '?SECSTATUS'                  // getSecurityStatus, '?SECSTATUS'
  #END_IF // __SECURITYSYSTEM_COMMANDS__

DEFINE_CONSTANT	//	----- Sensor Device : Channels and Levels

  SENSOR_FB                             = 255                           // Feedback:  Sensor feedback

  SENSOR_VALUE                          = 7                             // Level: Sensor value

DEFINE_CONSTANT	//	----- Settop Box : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  CHAN_UP                               = 22                            // Momentary: Next station preset
  CHAN_DN                               = 23                            // Momentary: Previous station preset
  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  TUNER_BAND                            = 40                            // Momentary: Cycle tuner band
  CABLE_AB                              = 42                            // Momentary: Cycle AB SWITCH
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  CABLE_B_ON                            = 212                           // Discrete:  Set B SWITCH on or off
  CABLE_B_FB                            = 212                           // Feedback:  AB SWITCH feedback
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay
  TUNER_PRESET_GROUP                    = 224                           // Momentary: Cycle station preset group
  TUNER_STATION_UP                      = 225                           // Momentary: Increment tuner station
  TUNER_STATION_DN                      = 226                           // Momentary: Decrement tuner station
  TUNER_SCAN_FWD                        = 227                           // Momentary: Station scan forward
  TUNER_SCAN_REV                        = 228                           // Momentary: Station scan backward/reverse
  TUNER_SEEK_FWD                        = 229                           // Momentary: Station seek forward
  TUNER_SEEK_REV                        = 230                           // Momentary: Station seek backward/reverse
  TUNER_OSD                             = 234                           // Momentary: Cycle on-screen or front panel display info
  TUNER_PREV                            = 235                           // Momentary: Goto previous tuner station

  VOL_LVL                               = 1                             // Level: Volume level (0-255)

  #IF_DEFINED __SETTOPBOX_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    BAND_CMD                            = 'BAND'                        // setBand, 'BAND-<tb>'
    BAND_QRY                            = '?BAND'                       // getBand, '?BAND'
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    STATIONPRESETCOUNT_CMD              = 'STATIONPRESETCOUNT'          // 'STATIONPRESETCOUNT-<count>'
    STATIONPRESETCOUNT_QRY              = '?STATIONPRESETCOUNT'         // getStationPresetCount, '?STATIONPRESETCOUNT'
    STATIONPRESETPROPERTIES_CMD         = 'STATIONPRESETPROPERTIES'     // '[STATIONPRESETPROPERTIES-<index>,<displayName>,<station>]*'
    STATIONPRESETPROPERTIES_QRY         = '?STATIONPRESETPROPERTIES'    // getStationPresetProperties, '?STATIONPRESETPROPERTIES'
    STATIONPRESETPROPERTY_CMD           = 'STATIONPRESETPROPERTY'       // 'STATIONPRESETPROPERTY-<index>,<displayName>,<station>'
    STATIONPRESETPROPERTY_QRY           = '?STATIONPRESETPROPERTY'      // getStationPresetProperty(index), '?STATIONPRESETPROPERTY-<index>'
    STATIONPRESETSELECT_CMD             = 'STATIONPRESETSELECT'         // 'STATIONPRESETSELECT-<index>'
    STATIONPRESETSELECT_QRY             = '?STATIONPRESETSELECT'        // getStationPresetSelect, '?STATIONPRESETSELECT'
    STATIONPRESET_CMD                   = 'TUNERPRESET'                 // setStationPreset, 'TUNERPRESET-<preset>'
    STATIONPRESET_QRY                   = '?TUNERPRESET'                // getStationPreset, '?TUNERPRESET'
    STATION_CMD                         = 'XCH'                         // setStation, 'XCH-<station>'
    STATION_QRY                         = '?XCH'                        // getStation, '?XCH'
    TUNERBANDPROPERTIES_CMD             = 'TUNERBANDPROPERTIES'         // '[TUNERBANDPROPERTIES-<index>,<displayName>,<tunerBand>]*'
    TUNERBANDPROPERTIES_QRY             = '?TUNERBANDPROPERTIES'        // getTunerBandProperties, '?TUNERBANDPROPERTIES'
    TUNERBANDPROPERTYCOUNT_CMD          = 'TUNERBANDPROPERTYCOUNT'      // 'TUNERBANDPROPERTYCOUNT-<count>'
    TUNERBANDPROPERTYCOUNT_QRY          = '?TUNERBANDPROPERTYCOUNT'     // getTunerBandPropertyCount, '?TUNERBANDPROPERTYCOUNT'
    TUNERBANDPROPERTY_CMD               = 'TUNERBANDPROPERTY'           // 'TUNERBANDPROPERTY-<index>,<displayName>,<tunerBand>'
    TUNERBANDPROPERTY_QRY               = '?TUNERBANDPROPERTY'          // getTunerBandProperty(index), '?TUNERBANDPROPERTY-<index>'
    TUNERBANDSELECT_CMD                 = 'TUNERBANDSELECT'             // 'TUNERBANDSELECT-<index>'
    TUNERBANDSELECT_QRY                 = '?TUNERBANDSELECT'            // getTunerBandSelect, '?TUNERBANDSELECT'
    TUNERCOMPONENTPROPERTIES_CMD        = 'TUNERCOMPONENTPROPERTIES'    // '[TUNERCOMPONENTPROPERTIES-<index>,<displayName>,<value>]*'
    TUNERCOMPONENTPROPERTIES_QRY        = '?TUNERCOMPONENTPROPERTIES'   // getTunerComponentProperties, '?TUNERCOMPONENTPROPERTIES'
    TUNERCOMPONENTPROPERTYCOUNT_CMD     = 'TUNERCOMPONENTPROPERTYCOUNT' // 'TUNERCOMPONENTPROPERTYCOUNT-<count>'
    TUNERCOMPONENTPROPERTYCOUNT_QRY     = '?TUNERCOMPONENTPROPERTYCOUNT'// getTunerComponentPropertyCount, '?TUNERCOMPONENTPROPERTYCOUNT'
    TUNERCOMPONENTPROPERTY_CMD          = 'TUNERCOMPONENTPROPERTY'      // 'TUNERCOMPONENTPROPERTY-<index>,<displayName>,<value>'
    TUNERCOMPONENTPROPERTY_QRY          = '?TUNERCOMPONENTPROPERTY'     // getTunerComponentProperty(index), '?TUNERCOMPONENTPROPERTY-<index>'
    TUNERCOMPONENTSELECT_CMD            = 'TUNERCOMPONENTSELECT'        // 'TUNERCOMPONENTSELECT-<index>'
    TUNERCOMPONENTSELECT_QRY            = '?TUNERCOMPONENTSELECT'       // getTunerComponentSelect, '?TUNERCOMPONENTSELECT'
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
  #END_IF // __SETTOPBOX_COMMANDS__

DEFINE_CONSTANT	//	----- Slide Projector : Channels

  SLIDE_NEXT                            = 4                             // Momentary: Goto next slide
  SLIDE_PREV                            = 5                             // Momentary: Goto previous slide
  POWER                                 = 9                             // Momentary: Cycle lamp power
  PWR_ON                                = 27                            // Momentary: Set lamp power on
  PWR_OFF                               = 28                            // Momentary: Set lamp power off
  FOCUS_NEAR                            = 160                           // Ramping:   Ramp focus near
  FOCUS_NEAR_FB                         = 160                           // Feedback:  Ramp focus near feedback
  FOCUS_FAR                             = 161                           // Ramping:   Ramp focus far
  FOCUS_FAR_FB                          = 161                           // Feedback:  Ramp focus far feedback

  #IF_DEFINED __SLIDEPROJECTOR_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    COOLDOWNCOUNTER_CMD                 = 'COOLING'                     // 'COOLING-<int>'
    COOLDOWNTIME_CMD                    = 'COOLDOWN'                    // setCoolDownTime, 'COOLDOWN-<time>'
    COOLDOWNTIME_QRY                    = '?COOLDOWN'                   // getCoolDownTime, '?COOLDOWN'
    COUNTERNOTIFICATIONON_CMD           = 'COUNTERNOTIFY'               // setCounterNotificationOn, 'COUNTERNOTIFY-<state>'
    LAMPTIME_CMD                        = 'LAMPTIME'                    // setLampTime, 'LAMPTIME-<time>'
    LAMPTIME_QRY                        = '?LAMPTIME'                   // getLampTime, '?LAMPTIME'
    SLIDE_CMD                           = 'SLIDE'                       // setSlide, 'SLIDE-<slide>'
    SLIDE_QRY                           = '?SLIDE'                      // getSlide, '?SLIDE'
    WARMUPCOUNTER_CMD                   = 'WARMING'                     // 'WARMING-<int>'
    WARMUPTIME_CMD                      = 'WARMUP'                      // setWarmUpTime, 'WARMUP-<time>'
    WARMUPTIME_QRY                      = '?WARMUP'                     // getWarmUpTime, '?WARMUP'
  #END_IF // __SLIDEPROJECTOR_COMMANDS__

DEFINE_CONSTANT	//	----- Switcher : Channels and Levels

  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  GAIN_UP                               = 140                           // Ramping:   Ramp gain up
  GAIN_UP_FB                            = 140                           // Feedback:  Gain ramping up feedback
  GAIN_DN                               = 141                           // Ramping:   Ramp gain down
  GAIN_DN_FB                            = 141                           // Feedback:  Gain ramping down feedback
  GAIN_MUTE_ON                          = 143                           // Discrete:  Set gain mute on
  GAIN_MUTE_FB                          = 143                           // Feedback:  Gain mute feedback
  GAIN_MUTE                             = 144                           // Momentary: Cycle gain mute
  BALANCE_UP                            = 164                           // Ramping:   Increment balance
  BALANCE_DN                            = 165                           // Ramping:   Decrement balance
  BASS_UP                               = 166                           // Ramping:   Increment bass
  BASS_DN                               = 167                           // Ramping:   Decrement bass
  TREBLE_UP                             = 168                           // Ramping:   Increment treble
  TREBLE_DN                             = 169                           // Ramping:   Decrement treble
  SURROUND_NEXT                         = 170                           // Momentary: Next surround sound mode
  SURROUND_PREV                         = 171                           // Momentary: Previous surround sound mode
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback
  LOUDNESS                              = 206                           // Momentary: Cycle loudness
  LOUDNESS_ON                           = 207                           // Discrete:  Set loudness state
  LOUDNESS_FB                           = 207                           // Feedback:  Loudness state feedback

  VOL_LVL                               = 1                             // Level: Volume level (0-255)
  BALANCE_LVL                           = 2                             // Level: Balance level (0=left, 255=right)
  BASS_LVL                              = 3                             // Level: Bass level (0-255)
  TREBLE_LVL                            = 4                             // Level: Treble level (0-255)
  GAIN_LVL                              = 5                             // Level: Gain level (0-255)

  #IF_DEFINED __SWITCHER_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADJUSTEQUALIZEROFFSET_CMD           = 'EQUALIZER_OFFSET'            // setEqualizerOffset overloads, 'EQUALIZER_OFFSET-(band)=(offset),[,EQUALIZER-(band)=(offset)]+'
    EQUALIZERBANDS_CMD                  = 'EQUALIZER_BANDS'             // return 'EQUALIZER_BANDS-(band)[,(band)]+'
    EQUALIZERBANDS_QRY                  = '?EQUALIZER_BANDS'            // getEqualizerBands, '?EQUALIZER_BANDS'
    EQUALIZER_CMD                       = 'EQUALIZER'                   // setEqualizer overloads, 'EQUALIZER-(band)=(gain),[,EQUALIZER-(band)=(gain)]+'
    EQUALIZER_QRY                       = '?EQUALIZER'                  // getEqualizer overloads, '?EQUALIZER', '?EQUALIZER-(band)[,(band)]+'
    INPUT_QRY                           = '?INPUT'                      // getInput, '?INPUT-<sl>,<output>'
    OUTPUT_QRY                          = '?OUTPUT'                     // getOutput, '?OUTPUT-<sl>,<input>'
    SAVESWITCHERPRESET_CMD              = 'SWITCHPRESETSAVE'            // saveSwitcherPreset, 'SWICTHPRESETSAVE-<preset>'
    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    SURROUNDMODE_CMD                    = 'SURROUND'                    // setSurroundMode, 'SURROUND-<mode>'
    SURROUNDMODE_QRY                    = '?SURROUND'                   // getSurroundMode, '?SURROUND'
		SURROUNDMODES_CMD                   = 'SURROUNDMODES'               // SurroundModes, 'SURROUNDMODES-<mode>,<mode>...'
	  SURROUNDMODES_QRY                   = '?SURROUNDMODES'              // getSurroundMode, '?SURROUND'
    SWITCHERPRESET_CMD                  = 'SWITCHPRESET'                // setSwitcherPreset, 'SWITCHPRESET-<preset>'
    SWITCHERPRESET_QRY                  = '?SWITCHPRESET'               // getSwitcherPreset, '?SWITCHPRESET'
    SWITCHINPUTTOOUTPUT1_CMD            = 'AI<input>O<output,output,>'
    SWITCHINPUTTOOUTPUT2_CMD            = 'CL<sl>I<input>O<output,output>'
    SWITCHINPUTTOOUTPUT3_CMD            = 'VI<input>O<output,output,>'
    SWITCHINPUTTOOUTPUT_CMD             = 'CI<input>O<output,output>'
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
  #END_IF // __SWITCHER_COMMANDS__

DEFINE_CONSTANT	//	----- Text Keypad : Channels

  KEYPAD_BTN1                           = 1                             // Discrete:  Keypad button 1 on
  KEYPAD_BTN1_FB                        = 1                             // Input:     Keypad button 1 feedback
  KEYPAD_BTN2                           = 2                             // Discrete:  Keypad button 2 on
  KEYPAD_BTN2_FB                        = 2                             // Input:     Keypad button 2 feedback
  KEYPAD_BTN3                           = 3                             // Discrete:  Keypad button 3 on
  KEYPAD_BTN3_FB                        = 3                             // Input:     Keypad button 3 feedback
  KEYPAD_BTN4                           = 4                             // Discrete:  Keypad button 4 on
  KEYPAD_BTN4_FB                        = 4                             // Input:     Keypad button 4 feedback
  KEYPAD_BTN5                           = 5                             // Discrete:  Keypad button 5 on
  KEYPAD_BTN5_FB                        = 5                             // Input:     Keypad button 5 feedback
  KEYPAD_BTN6                           = 6                             // Discrete:  Keypad button 6 on
  KEYPAD_BTN6_FB                        = 6                             // Input:     Keypad button 6 feedback
  KEYPAD_BTN7                           = 7                             // Discrete:  Keypad button 7 on
  KEYPAD_BTN7_FB                        = 7                             // Input:     Keypad button 7 feedback
  KEYPAD_BTN8                           = 8                             // Discrete:  Keypad button 8 on
  KEYPAD_BTN8_FB                        = 8                             // Input:     Keypad button 8 feedback
  KEYPAD_BTN9                           = 9                             // Discrete:  Keypad button 9 on
  KEYPAD_BTN9_FB                        = 9                             // Input:     Keypad button 9 feedback
  KEYPAD_BTN10                          = 10                            // Discrete:  Keypad button 10 on
  KEYPAD_BTN10_FB                       = 10                            // Input:     Keypad button 10 feedback
  KEYPAD_BTN1_BLINK                     = 101                           // Discrete:  Keypad button 1 blink
  KEYPAD_BTN1_BLINK_FB                  = 101                           // Feedback:  Keypad button 1 blink feedback
  KEYPAD_BTN2_BLINK                     = 102                           // Discrete:  Keypad button 2 blink
  KEYPAD_BTN2_BLINK_FB                  = 102                           // Feedback:  Keypad button 2 blink feedback
  KEYPAD_BTN3_BLINK                     = 103                           // Discrete:  Keypad button 3 blink
  KEYPAD_BTN3_BLINK_FB                  = 103                           // Feedback:  Keypad button 3 blink feedback
  KEYPAD_BTN4_BLINK                     = 104                           // Discrete:  Keypad button 4 blink
  KEYPAD_BTN4_BLINK_FB                  = 104                           // Feedback:  Keypad button 4 blink feedback
  KEYPAD_BTN5_BLINK                     = 105                           // Discrete:  Keypad button 5 blink
  KEYPAD_BTN5_BLINK_FB                  = 105                           // Feedback:  Keypad button 5 blink feedback
  KEYPAD_BTN6_BLINK                     = 106                           // Discrete:  Keypad button 6 blink
  KEYPAD_BTN6_BLINK_FB                  = 106                           // Feedback:  Keypad button 6 blink feedback
  KEYPAD_BTN7_BLINK                     = 107                           // Discrete:  Keypad button 7 blink
  KEYPAD_BTN7_BLINK_FB                  = 107                           // Feedback:  Keypad button 7 blink feedback
  KEYPAD_BTN8_BLINK                     = 108                           // Discrete:  Keypad button 8 blink
  KEYPAD_BTN8_BLINK_FB                  = 108                           // Feedback:  Keypad button 8 blink feedback
  KEYPAD_BTN9_BLINK                     = 109                           // Discrete:  Keypad button 9 blink
  KEYPAD_BTN9_BLINK_FB                  = 109                           // Feedback:  Keypad button 9 blink feedback
  KEYPAD_BTN10_BLINK                    = 110                           // Discrete:  Keypad button 10 blink
  KEYPAD_BTN10_BLINK_FB                 = 110                           // Feedback:  Keypad button 10 blink feedback

  #IF_DEFINED __TEXTKEYPAD_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADDKEYPADCOMPONENT_CMD              = 'KEYPADADD'                   // addKeypadComponent, 'KEYPADADD-<index>,<keypadaddr>'
    ADDKEYPADR_CMD                      = 'KEYPADADDR'                  // setButtonState, 'KEYPADBTN-<index>,<buttonaddr>,<state>'
    BACKLIGHTON_CMD                     = 'TEXTBACKLIGHT'               // setBacklightOn, 'TEXTBACKLIGHT-<state>'
    BUTTONSTATE_CMD                     = 'KEYPADBTN'                   // setButtonState, 'KEYPADBTN-<index>,<buttonaddr>,<state>'
    CLEARDISPLAY_CMD                    = 'TEXT'                        // clearDisplay, 'TEXT-'
    KEYPADCOMPONENTADDRESS_QRY          = '?KEYPADADDR'                 // getKeypadComponentAddress, '?KEYPADADDR-<index>'
    KEYPADCOMPONENTINDEX_QRY            = '?KEYPADIDX'                  // getKeypadComponentIndex, '?KEYPADIDX-<keypadaddr>'
    REMOVEKEYPADCOMPONENT1_CMD          = 'KEYPADREMOVEADDR'            // removeKeypadComponent, 'KEYPADREMOVEADDR-<keypadaddr>'
    REMOVEKEYPADCOMPONENT_CMD           = 'KEYPADREMOVEIDX'             // removeKeypadComponent, 'KEYPADREMOVEIDX-<index>'
    TEXTDISPLAY_CMD                     = 'TEXT'                        // setTextDisplay, 'TEXT-<text>'
    TEXTLINE_CMD                        = 'TEXT'                        // setTextLine, 'TEXT-<line>,<text>'
    TEXT_CMD                            = 'TEXT'                        // setText, 'TEXT-<line>,<column>,<text>'
  #END_IF // __TEXTKEYPAD_COMMANDS__

DEFINE_CONSTANT	//	----- TV : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  CHAN_UP                               = 22                            // Momentary: Next station preset
  CHAN_DN                               = 23                            // Momentary: Previous station preset
  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  TUNER_BAND                            = 40                            // Momentary: Cycle tuner band
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  ASPECT_RATIO                          = 142                           // Momentary: Cycle aspect ratio
  BRIGHT_UP                             = 148                           // Momentary: Increment brightness
  BRIGHT_DN                             = 149                           // Momentary: Decrement brightness
  COLOR_UP                              = 150                           // Momentary: Increment color
  COLOR_DN                              = 151                           // Momentary: Decrement color
  CONTRAST_UP                           = 152                           // Momentary: Increment contrast
  CONTRAST_DN                           = 153                           // Momentary: Decrement contrast
  SHARP_UP                              = 154                           // Momentary: Increment sharpness
  SHARP_DN                              = 155                           // Momentary: Decrement sharpness
  TINT_UP                               = 156                           // Momentary: Increment tint
  TINT_DN                               = 157                           // Momentary: Decrement tint
  PIP_POS                               = 191                           // Momentary: Cycle pip position
  PIP_SWAP                              = 193                           // Momentary: Swap pip
  PIP                                   = 194                           // Momentary: Cycle pip
  PIP_ON                                = 195                           // Discrete:  Set pip on
  PIP_FB                                = 195                           // Feedback:  Pip feedback
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  PIC_MUTE                              = 210                           // Momentary: Cycle picture/video mute
  PIC_MUTE_ON                           = 211                           // Discrete:  Set picture/video mute on
  PIC_MUTE_FB                           = 211                           // Feedback:  Picture/video mute feedback
  PIC_FREEZE                            = 213                           // Momentary: Cycle freeze
  PIC_FREEZE_ON                         = 214                           // Discrete:  Set freeze on
  PIC_FREEZE_FB                         = 214                           // Feedback:  Freeze feedback
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay
  TUNER_PRESET_GROUP                    = 224                           // Momentary: Cycle station preset group
  TUNER_STATION_UP                      = 225                           // Momentary: Increment tuner station
  TUNER_STATION_DN                      = 226                           // Momentary: Decrement tuner station
  TUNER_SCAN_FWD                        = 227                           // Momentary: Station scan forward
  TUNER_SCAN_REV                        = 228                           // Momentary: Station scan backward/reverse
  TUNER_SEEK_FWD                        = 229                           // Momentary: Station seek forward
  TUNER_SEEK_REV                        = 230                           // Momentary: Station seek backward/reverse
  TUNER_OSD                             = 234                           // Momentary: Cycle on-screen or front panel display info
  TUNER_PREV                            = 235                           // Momentary: Goto previous tuner station

  VOL_LVL                               = 1                             // Level: Volume level (0-255)
  BRIGHT_LVL                            = 10                            // Level: Brightness level (0-255)
  COLOR_LVL                             = 11                            // Level: Color level (0-255)
  CONTRAST_LVL                          = 12                            // Level: Contrast level (0-255)
  SHARP_LVL                             = 13                            // Level: Sharpness level (0-255)
  TINT_LVL                              = 14                            // Level: Tint level (0-255)

  #IF_DEFINED __TV_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ACTIVEWINDOW_CMD                    = 'ACTIVEWINDOW'                // setActiveWindow, 'ACTIVEWINDOW-<mss>'
    ACTIVEWINDOW_QRY                    = '?ACTIVEWINDOW'               // getActiveWindow, '?ACTIVEWINDOW'
    ASPECTRATIOCOUNT_CMD                = 'ASPECTRATIOCOUNT'            // ASPECTRATIOCOUNT-<count>'
    ASPECTRATIOCOUNT_QRY                = '?ASPECTRATIOCOUNT'           // getAspectRatioCount(), '?ASPECTRATIOCOUNT'
    ASPECTRATIOPROPERTIES_CMD           = 'ASPECTRATIOPROPERTIES'       // '[ASPECTRATIOPROPERTIES-<index>,<displayName>,<value>]*'
    ASPECTRATIOPROPERTIES_QRY           = '?ASPECTRATIOPROPERTIES'      // getAspectRatioProperties(), '?ASPECTRATIOPROPERTIES'
    ASPECTRATIOPROPERTY_CMD             = 'ASPECTRATIOPROPERTY'         // 'ASPECTRATIOPROPERTY-<index>,<displayName>,<value>'
    ASPECTRATIOPROPERTY_QRY             = '?ASPECTRATIOPROPERTY'        // getAspectRatioProperty(index), '?ASPECTRATIOPROPERTY-<index>'
    ASPECTRATIOSELECT_CMD               = 'ASPECTRATIOSELECT'           // setAspectRatioSelect(index), 'ASPECTRATIOSELECT-<index>'
    ASPECTRATIOSELECT_QRY               = '?ASPECTRATIOSELECT'          // getAspectRatioSelect(), '?ASPECTRATIOSELECT'
    ASPECTRATIO_CMD                     = 'ASPECT'                      // setAspectRatio, 'ASPECT-<aspectRatio>'
    ASPECTRATIO_QRY                     = '?ASPECT'                     // getAspectRatio, '?ASPECT'
    BAND_CMD                            = 'BAND'                        // setBand, 'BAND-<tb>'
    BAND_QRY                            = '?BAND'                       // getBand, '?BAND'
    CYCLEINPUTSELECT_CMD                = 'CYCLESELECT'                 // cycleInputSelect(), CYCLESELECT
    INPUTCOUNT_CMD                      = 'INPUTCOUNT'                  // INPUTCOUNT-<count>
    INPUTCOUNT_QRY                      = '?INPUTCOUNT'                 // getInputCount(), INPUTCOUNT-<count>
    INPUTGROUPSELECT_CMD                = 'INPUTGROUPSELECT'            // setInputGroupSelect(inputGroup), INPUTGROUPSELECT-<inputGroup>
    INPUTGROUPSELECT_QRY                = '?INPUTGROUPSELECT'           // getInputGroupSelect(), ?INPUTGROUPSELECT
    INPUTPROPERTIES_CMD                 = 'INPUTPROPERTIES'             // [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTIES_QRY                 = '?INPUTPROPERTIES'            // getInputProperties(), [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTY_CMD                   = 'INPUTPROPERTY'               // INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTPROPERTY_QRY                   = '?INPUTPROPERTY'              // getInputProperty(index), INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTSELECT_CMD                     = 'INPUTSELECT'                 // setInputSelect(index), INPUTSELECT-<index>
    INPUTSELECT_QRY                     = '?INPUTSELECT'                // getInputSelect(), INPUTSELECT-<index>
    INPUTSOURCE_CMD                     = 'INPUT'                       // setInputSource, INPUT-<sourceSelect>,<inputNumber>
    INPUTSOURCE_QRY                     = '?INPUT'                      // getInputSource, ?INPUT
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    STATIONPRESETCOUNT_CMD              = 'STATIONPRESETCOUNT'          // 'STATIONPRESETCOUNT-<count>'
    STATIONPRESETCOUNT_QRY              = '?STATIONPRESETCOUNT'         // getStationPresetCount, '?STATIONPRESETCOUNT'
    STATIONPRESETPROPERTIES_CMD         = 'STATIONPRESETPROPERTIES'     // '[STATIONPRESETPROPERTIES-<index>,<displayName>,<station>]*'
    STATIONPRESETPROPERTIES_QRY         = '?STATIONPRESETPROPERTIES'    // getStationPresetProperties, '?STATIONPRESETPROPERTIES'
    STATIONPRESETPROPERTY_CMD           = 'STATIONPRESETPROPERTY'       // 'STATIONPRESETPROPERTY-<index>,<displayName>,<station>'
    STATIONPRESETPROPERTY_QRY           = '?STATIONPRESETPROPERTY'      // getStationPresetProperty(index), '?STATIONPRESETPROPERTY-<index>'
    STATIONPRESETSELECT_CMD             = 'STATIONPRESETSELECT'         // 'STATIONPRESETSELECT-<index>'
    STATIONPRESETSELECT_QRY             = '?STATIONPRESETSELECT'        // getStationPresetSelect, '?STATIONPRESETSELECT'
    STATIONPRESET_CMD                   = 'TUNERPRESET'                 // setStationPreset, 'TUNERPRESET-<preset>'
    STATIONPRESET_QRY                   = '?TUNERPRESET'                // getStationPreset, '?TUNERPRESET'
    STATION_CMD                         = 'XCH'                         // setStation, 'XCH-<station>'
    STATION_QRY                         = '?XCH'                        // getStation, '?XCH'
    TUNERBANDPROPERTIES_CMD             = 'TUNERBANDPROPERTIES'         // '[TUNERBANDPROPERTIES-<index>,<displayName>,<tunerBand>]*'
    TUNERBANDPROPERTIES_QRY             = '?TUNERBANDPROPERTIES'        // getTunerBandProperties, '?TUNERBANDPROPERTIES'
    TUNERBANDPROPERTYCOUNT_CMD          = 'TUNERBANDPROPERTYCOUNT'      // 'TUNERBANDPROPERTYCOUNT-<count>'
    TUNERBANDPROPERTYCOUNT_QRY          = '?TUNERBANDPROPERTYCOUNT'     // getTunerBandPropertyCount, '?TUNERBANDPROPERTYCOUNT'
    TUNERBANDPROPERTY_CMD               = 'TUNERBANDPROPERTY'           // 'TUNERBANDPROPERTY-<index>,<displayName>,<tunerBand>'
    TUNERBANDPROPERTY_QRY               = '?TUNERBANDPROPERTY'          // getTunerBandProperty(index), '?TUNERBANDPROPERTY-<index>'
    TUNERBANDSELECT_CMD                 = 'TUNERBANDSELECT'             // 'TUNERBANDSELECT-<index>'
    TUNERBANDSELECT_QRY                 = '?TUNERBANDSELECT'            // getTunerBandSelect, '?TUNERBANDSELECT'
    TUNERCOMPONENTPROPERTIES_CMD        = 'TUNERCOMPONENTPROPERTIES'    // '[TUNERCOMPONENTPROPERTIES-<index>,<displayName>,<value>]*'
    TUNERCOMPONENTPROPERTIES_QRY        = '?TUNERCOMPONENTPROPERTIES'   // getTunerComponentProperties, '?TUNERCOMPONENTPROPERTIES'
    TUNERCOMPONENTPROPERTYCOUNT_CMD     = 'TUNERCOMPONENTPROPERTYCOUNT' // 'TUNERCOMPONENTPROPERTYCOUNT-<count>'
    TUNERCOMPONENTPROPERTYCOUNT_QRY     = '?TUNERCOMPONENTPROPERTYCOUNT'// getTunerComponentPropertyCount, '?TUNERCOMPONENTPROPERTYCOUNT'
    TUNERCOMPONENTPROPERTY_CMD          = 'TUNERCOMPONENTPROPERTY'      // 'TUNERCOMPONENTPROPERTY-<index>,<displayName>,<value>'
    TUNERCOMPONENTPROPERTY_QRY          = '?TUNERCOMPONENTPROPERTY'     // getTunerComponentProperty(index), '?TUNERCOMPONENTPROPERTY-<index>'
    TUNERCOMPONENTSELECT_CMD            = 'TUNERCOMPONENTSELECT'        // 'TUNERCOMPONENTSELECT-<index>'
    TUNERCOMPONENTSELECT_QRY            = '?TUNERCOMPONENTSELECT'       // getTunerComponentSelect, '?TUNERCOMPONENTSELECT'
    VIDEOTYPE_CMD                       = 'VIDEOTYPE'                   // setVideoType, 'VIDEOTYPE-<vt>'
    VIDEOTYPE_QRY                       = '?VIDEOTYPE'                  // getVideoType, '?VIDEOTYPE'
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
  #END_IF // __TV_COMMANDS__

DEFINE_CONSTANT	//	----- UPS : Commands Reference

  #IF_DEFINED __UPS_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    UPSALARMS_CMD                       = 'UPSALARMS'                   // UPSALARMS-<alarms>
    UPSALARMS_QRY                       = '?UPSALARMS'                  // getAlarms, '?UPSALARMS'
    UPSBACKUPPROPERTY_CMD               = 'UPSBACKUPPROPERTY'           // UPSBACKUPPROPERTY-<bu-prop>,<bu-value>
    UPSBACKUPPROPERTY_QRY               = '?UPSBACKUPPROPERTY'          // getBackupProperty, '?UPSBACKUPPROPERTY-<bu-prop>'
    UPSBACKUPSTATUS_CMD                 = 'UPSBACKUPSTATUS'             // UPSBACKUPSTATUS-<backupstatus>
    UPSBACKUPSTATUS_QRY                 = '?UPSBACKUPSTATUS'            // getBackupStatus, '?UPSBACKUPSTATUS'
    UPSBACKUPSUPPORTED_CMD              = 'UPSBACKUPSUPPORTED'          // UPSBACKUPSUPPORTED-<TRUE|FALSE>
    UPSBACKUPSUPPORTED_QRY              = '?UPSBACKUPSUPPORTED'         // isBackupSupported, '?UPSBACKUPSUPPORTED'
    UPSBACKUPTIMER_CMD                  = 'UPSBACKUPTIMER'              // UPSBACKUPTIMER-<timer>,<value>
    UPSBACKUPTIMER_QRY                  = '?UPSBACKUPTIMER'             // getBackupTimer, '?UPSBACKUPTIMER-<timer>'
    UPSINPUTPROPERTY_CMD                = 'UPSINPUTPROPERTY'            // UPSINPUTPROPERTY-<input-prop>,<value>
    UPSINPUTPROPERTY_QRY                = '?UPSINPUTPROPERTY'           // getInputProperty, '?UPSINPUTPROPERTY-<input-prop>'
    UPSOUTLETCOUNT_CMD                  = 'UPSOUTLETCOUNT'              // UPSOUTLETCOUNT-<count>
    UPSOUTLETCOUNT_QRY                  = '?UPSOUTLETCOUNT'             // getOutletCount, '?UPSOUTLETCOUNT'
    UPSOUTLETPROPERTIES_CMD             = 'UPSOUTLETPROPERTIES'         // UPSOUTLETPROPERTIES-[<outlet_number>,<outlet-prop>,<value>]
    UPSOUTLETPROPERTIES_QRY             = '?UPSOUTLETPROPERTIES'        // getOutletProperties, '?UPSOUTLETPROPERTIES-<outlet_number>'
    UPSOUTLETPROPERTY_CMD               = 'UPSOUTLETPROPERTY'           // UPSOUTLETPROPERTY-<outlet_number>,<outlet-prop>,<value>    1+
    UPSOUTLETPROPERTY_QRY               = '?UPSOUTLETPROPERTY'          // getOutletProperty, '?UPSOUTLETPROPERTY-<outlet_number>,<outlet-prop>'
    UPSOUTLETSTATE_CMD                  = 'UPSOUTLETSTATE'              // setOutletState, 'UPSOUTLETSTATE-<outlet_number>,<outlet-state>'
    UPSOUTLETSTATE_QRY                  = '?UPSOUTLETSTATE'             // getOutletState, '?UPSOUTLETSTATE-<outlet_number>'
    UPSSTATE_CMD                        = 'UPSSTATE'                    // setState, 'UPSSTATE-<ups-state>[,<delay>]'
    UPSSTATE_QRY                        = '?UPSSTATE'                   // getState, '?UPSSTATE-<ups_state>'
    UPSSTATUS_CMD                       = 'UPSSTATUS'                   // UPSSTATUS-<state>
    UPSSTATUS_QRY                       = '?UPSSTATUS'                  // getStatus, '?UPSSTATUS'
    UPSTEMPERATURESCALE_CMD             = 'UPSTEMPERATURESCALE'         // setTemperatureScale, 'UPSTEMPERATURESCALE-<tempscale>'
    UPSTEMPERATURESCALE_QRY             = '?UPSTEMPERATURESCALE'        // getTemperatureScale, '?UPSTEMPERATURESCALE'
    UPSTESTRESULT_CMD                   = 'UPSTESTRESULT'               // UPSTESTRESULT-<test>,<testresult>
    UPSTESTRESULT_QRY                   = '?UPSTESTRESULT'              // getTestResult, '?UPSTESTRESULT'
    UPSTESTSSUPPORTED_CMD               = 'UPSTESTSSUPPORTED'           // UPSTESTSSUPPORTED-<test>[,<test>]*
    UPSTESTSSUPPORTED_QRY               = '?UPSTESTSSUPPORTED'          // getTestsSupported, '?UPSTESTSSUPPORTED'
    UPSTEST_CMD                         = 'UPSTEST'                     // setTest, 'UPSTEST-<test>,<action>'
  #END_IF // __UPS_COMMANDS__

DEFINE_CONSTANT	//	----- VCR : Channels

  PLAY                                  = 1                             // Momentary: Play
  STOP                                  = 2                             // Momentary: Stop
  PAUSE                                 = 3                             // Momentary: Pause
  FFWD                                  = 4                             // Momentary: Fast forward
  REW                                   = 5                             // Momentary: Rewind
  SFWD                                  = 6                             // Momentary: Search forward
  SREV                                  = 7                             // Momentary: Search reverse
  RECORD                                = 8                             // Momentary: Record
  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  CHAN_UP                               = 22                            // Momentary: Next station preset
  CHAN_DN                               = 23                            // Momentary: Previous station preset
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  TUNER_BAND                            = 40                            // Momentary: Cycle tuner band
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  SEARCH_SPEED                          = 119                           // Momentary: Cycle search speed
  EJECT                                 = 120                           // Momentary: Eject tape
  RESET_COUNTER                         = 121                           // Momentary: Reset counter
  TAPE_LOADED_FB                        = 122                           // Feedback:  Tape is loaded
  RECORD_LOCK_FB                        = 123                           // Feedback:  Tape record is locked
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  SLOW_FWD                              = 188                           // Momentary: Slow forward
  SLOW_REV                              = 189                           // Momentary: Slow reverse
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay
  TUNER_PRESET_GROUP                    = 224                           // Momentary: Cycle station preset group
  TUNER_STATION_UP                      = 225                           // Momentary: Increment tuner station
  TUNER_STATION_DN                      = 226                           // Momentary: Decrement tuner station
  TUNER_SCAN_FWD                        = 227                           // Momentary: Station scan forward
  TUNER_SCAN_REV                        = 228                           // Momentary: Station scan backward/reverse
  TUNER_SEEK_FWD                        = 229                           // Momentary: Station seek forward
  TUNER_SEEK_REV                        = 230                           // Momentary: Station seek backward/reverse
  TUNER_OSD                             = 234                           // Momentary: Cycle on-screen or front panel display info
  TUNER_PREV                            = 235                           // Momentary: Goto previous tuner station
  PLAY_FB                               = 241                           // Feedback:  Play feedback
  STOP_FB                               = 242                           // Feedback:  Stop feedback
  PAUSE_FB                              = 243                           // Feedback:  Pause feedback
  FFWD_FB                               = 244                           // Feedback:  Fast forward feedback
  REW_FB                                = 245                           // Feedback:  Rewind feedback
  SFWD_FB                               = 246                           // Feedback:  Search forward feedback
  SREV_FB                               = 247                           // Feedback:  Search reverse feedback
  RECORD_FB                             = 248                           // Feedback:  Record feedback
  SLOW_FWD_FB                           = 249                           // Feedback:  Slow forward feedback
  SLOW_REV_FB                           = 250                           // Feedback:  Slow reverse feedback

  #IF_DEFINED __VCR_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    BAND_CMD                            = 'BAND'                        // setBand, 'BAND-<tb>'
    BAND_QRY                            = '?BAND'                       // getBand, '?BAND'
    CYCLEINPUTSELECT_CMD                = 'CYCLESELECT'                 // cycleInputSelect(), CYCLESELECT
    INPUTCOUNT_CMD                      = 'INPUTCOUNT'                  // INPUTCOUNT-<count>
    INPUTCOUNT_QRY                      = '?INPUTCOUNT'                 // getInputCount(), INPUTCOUNT-<count>
    INPUTGROUPSELECT_CMD                = 'INPUTGROUPSELECT'            // setInputGroupSelect(inputGroup), INPUTGROUPSELECT-<inputGroup>
    INPUTGROUPSELECT_QRY                = '?INPUTGROUPSELECT'           // getInputGroupSelect(), ?INPUTGROUPSELECT
    INPUTPROPERTIES_CMD                 = 'INPUTPROPERTIES'             // [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTIES_QRY                 = '?INPUTPROPERTIES'            // getInputProperties(), [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTY_CMD                   = 'INPUTPROPERTY'               // INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTPROPERTY_QRY                   = '?INPUTPROPERTY'              // getInputProperty(index), INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTSELECT_CMD                     = 'INPUTSELECT'                 // setInputSelect(index), INPUTSELECT-<index>
    INPUTSELECT_QRY                     = '?INPUTSELECT'                // getInputSelect(), INPUTSELECT-<index>
    INPUTSOURCE_CMD                     = 'INPUT'                       // setInputSource, INPUT-<sourceSelect>,<inputNumber>
    INPUTSOURCE_QRY                     = '?INPUT'                      // getInputSource, ?INPUT
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    STATIONPRESETCOUNT_CMD              = 'STATIONPRESETCOUNT'          // 'STATIONPRESETCOUNT-<count>'
    STATIONPRESETCOUNT_QRY              = '?STATIONPRESETCOUNT'         // getStationPresetCount, '?STATIONPRESETCOUNT'
    STATIONPRESETPROPERTIES_CMD         = 'STATIONPRESETPROPERTIES'     // '[STATIONPRESETPROPERTIES-<index>,<displayName>,<station>]*'
    STATIONPRESETPROPERTIES_QRY         = '?STATIONPRESETPROPERTIES'    // getStationPresetProperties, '?STATIONPRESETPROPERTIES'
    STATIONPRESETPROPERTY_CMD           = 'STATIONPRESETPROPERTY'       // 'STATIONPRESETPROPERTY-<index>,<displayName>,<station>'
    STATIONPRESETPROPERTY_QRY           = '?STATIONPRESETPROPERTY'      // getStationPresetProperty(index), '?STATIONPRESETPROPERTY-<index>'
    STATIONPRESETSELECT_CMD             = 'STATIONPRESETSELECT'         // 'STATIONPRESETSELECT-<index>'
    STATIONPRESETSELECT_QRY             = '?STATIONPRESETSELECT'        // getStationPresetSelect, '?STATIONPRESETSELECT'
    STATIONPRESET_CMD                   = 'TUNERPRESET'                 // setStationPreset, 'TUNERPRESET-<preset>'
    STATIONPRESET_QRY                   = '?TUNERPRESET'                // getStationPreset, '?TUNERPRESET'
    STATION_CMD                         = 'XCH'                         // setStation, 'XCH-<station>'
    STATION_QRY                         = '?XCH'                        // getStation, '?XCH'
    TAPECOUNTERNOTIFICATION_CMD         = 'TAPECOUNTERNOTIFY'           // setTapeCounterNotification, 'TAPECOUNTERNOTIFY-<state>'
    TUNERBANDPROPERTIES_CMD             = 'TUNERBANDPROPERTIES'         // '[TUNERBANDPROPERTIES-<index>,<displayName>,<tunerBand>]*'
    TUNERBANDPROPERTIES_QRY             = '?TUNERBANDPROPERTIES'        // getTunerBandProperties, '?TUNERBANDPROPERTIES'
    TUNERBANDPROPERTYCOUNT_CMD          = 'TUNERBANDPROPERTYCOUNT'      // 'TUNERBANDPROPERTYCOUNT-<count>'
    TUNERBANDPROPERTYCOUNT_QRY          = '?TUNERBANDPROPERTYCOUNT'     // getTunerBandPropertyCount, '?TUNERBANDPROPERTYCOUNT'
    TUNERBANDPROPERTY_CMD               = 'TUNERBANDPROPERTY'           // 'TUNERBANDPROPERTY-<index>,<displayName>,<tunerBand>'
    TUNERBANDPROPERTY_QRY               = '?TUNERBANDPROPERTY'          // getTunerBandProperty(index), '?TUNERBANDPROPERTY-<index>'
    TUNERBANDSELECT_CMD                 = 'TUNERBANDSELECT'             // 'TUNERBANDSELECT-<index>'
    TUNERBANDSELECT_QRY                 = '?TUNERBANDSELECT'            // getTunerBandSelect, '?TUNERBANDSELECT'
    TUNERCOMPONENTPROPERTIES_CMD        = 'TUNERCOMPONENTPROPERTIES'    // '[TUNERCOMPONENTPROPERTIES-<index>,<displayName>,<value>]*'
    TUNERCOMPONENTPROPERTIES_QRY        = '?TUNERCOMPONENTPROPERTIES'   // getTunerComponentProperties, '?TUNERCOMPONENTPROPERTIES'
    TUNERCOMPONENTPROPERTYCOUNT_CMD     = 'TUNERCOMPONENTPROPERTYCOUNT' // 'TUNERCOMPONENTPROPERTYCOUNT-<count>'
    TUNERCOMPONENTPROPERTYCOUNT_QRY     = '?TUNERCOMPONENTPROPERTYCOUNT'// getTunerComponentPropertyCount, '?TUNERCOMPONENTPROPERTYCOUNT'
    TUNERCOMPONENTPROPERTY_CMD          = 'TUNERCOMPONENTPROPERTY'      // 'TUNERCOMPONENTPROPERTY-<index>,<displayName>,<value>'
    TUNERCOMPONENTPROPERTY_QRY          = '?TUNERCOMPONENTPROPERTY'     // getTunerComponentProperty(index), '?TUNERCOMPONENTPROPERTY-<index>'
    TUNERCOMPONENTSELECT_CMD            = 'TUNERCOMPONENTSELECT'        // 'TUNERCOMPONENTSELECT-<index>'
    TUNERCOMPONENTSELECT_QRY            = '?TUNERCOMPONENTSELECT'       // getTunerComponentSelect, '?TUNERCOMPONENTSELECT'
  #END_IF // __VCR_COMMANDS__

DEFINE_CONSTANT	//	----- Video Conferencer : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  TILT_UP                               = 132                           // Ramping:   Ramp tilt up
  TILT_UP_FB                            = 132                           // Feedback:  Tilt ramp up feedback
  TILT_DN                               = 133                           // Ramping:   Ramp tilt down
  TILT_DN_FB                            = 133                           // Feedback:  Tilt ramp down feedback
  PAN_LT                                = 134                           // Ramping:   Ramp pan left
  PAN_LT_FB                             = 134                           // Feedback:  Ramp pan left feedback
  PAN_RT                                = 135                           // Ramping:   Ramp pan right
  PAN_RT_FB                             = 135                           // Feedback:  Ramp pan right feedback
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  VCONF_PRIVACY                         = 145                           // Momentary: Cycle privacy
  VCONF_PRIVACY_ON                      = 146                           // Discrete:  Set privacy on or off
  VCONF_PRIVACY_FB                      = 146                           // Feedback:  Privacy feedback
  VCONF_TRAIN                           = 147                           // Momentary: Execute train
  ZOOM_OUT                              = 158                           // Ramping:   Ramp zoom out
  ZOOM_OUT_FB                           = 158                           // Feedback:  Ramp zoom out feedback
  ZOOM_IN                               = 159                           // Ramping:   Ramp zoom in
  ZOOM_IN_FB                            = 159                           // Feedback:  Ramp zoom in feedback
  FOCUS_NEAR                            = 160                           // Ramping:   Ramp focus near
  FOCUS_NEAR_FB                         = 160                           // Feedback:  Ramp focus near feedback
  FOCUS_FAR                             = 161                           // Ramping:   Ramp focus far
  FOCUS_FAR_FB                          = 161                           // Feedback:  Focus ramp far feedback
  AUTO_FOCUS_ON                         = 162                           // Discrete:  Set auto focus on or off
  AUTO_FOCUS_FB                         = 162                           // Feedback:  Auto focus feedback
  AUTO_IRIS_ON                          = 163                           // Discrete:  Set auto iris on or off
  AUTO_IRIS_FB                          = 163                           // Feedback:  Auto iris feedback
  AUTO_FOCUS                            = 172                           // Momentary: Cycle auto focus
  AUTO_IRIS                             = 173                           // Momentary: Cycle auto iris
  IRIS_OPEN                             = 174                           // Ramping:   Ramp iris open
  IRIS_OPEN_FB                          = 174                           // Feedback:  Ramp iris open feedback
  IRIS_CLOSE                            = 175                           // Ramping:   Ramp iris closed
  IRIS_CLOSE_FB                         = 175                           // Feedback:  Ramp iris closed feedback
  CAM_PRESET                            = 177                           // Momentary: Cycle camera preset
  PIP_POS                               = 191                           // Momentary: Cycle pip position
  PIP_SWAP                              = 193                           // Momentary: Swap pip
  PIP                                   = 194                           // Momentary: Cycle pip
  PIP_ON                                = 195                           // Discrete:  Set pip on or off
  PIP_FB                                = 195                           // Feedback:  Pip feedback
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback
  DIAL_REDIAL                           = 201                           // Momentary: Redial
  DIAL_OFF_HOOK                         = 202                           // Momentary: Cycle off hook state
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  DIAL_AUTO_ANSWER                      = 204                           // Momentary: Cycle auto answer state
  DIAL_AUDIBLE_RING                     = 205                           // Momentary: Cycle audible ring state
  DIAL_FLASH_HOOK                       = 208                           // Momentary: Flash hook
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay
  DIAL_OFF_HOOK_ON                      = 238                           // Discrete:  Set the dialer off or on hook
  DIAL_OFF_HOOK_FB                      = 238                           // Feedback:  Dialer's hook state feedback
  DIAL_AUTO_ANSWER_ON                   = 239                           // Discrete:  Set auto answer on or off
  DIAL_AUTO_ANSWER_FB                   = 239                           // Feedback:  Auto answer state feedback
  DIAL_AUDIBLE_RING_ON                  = 240                           // Discrete:  Set the audible ring state on or off
  DIAL_AUDIBLE_RING_FB                  = 240                           // Feedback:  Audible ring state feedback

  VOL_LVL                               = 1                             // Level: Volume level (0-255)
  ZOOM_LVL                              = 15                            // Level: Zoom position ( 0=out/Wide, 255= in/Tele)
  FOCUS_LVL                             = 16                            // Level: Focus position (0=near, 255=far)
  IRIS_LVL                              = 17                            // Level: Iris position (0=closed, 255=open)
  ZOOM_SPEED_LVL                        = 18                            // Level: Zoom ramp speed (0-255)
  FOCUS_SPEED_LVL                       = 19                            // Level: Focus ramp speed (0-255)
  IRIS_SPEED_LVL                        = 20                            // Level: Iris ramp speed (0-255)
  PAN_LVL                               = 27                            // Level: Pan position (0=left,255=right)
  TILT_LVL                              = 28                            // Level: Tilt position (0=down, 255=up)
  PAN_SPEED_LVL                         = 29                            // Level: Pan ramp speed (0-255)
  TILT_SPEED_LVL                        = 30                            // Level: Tilt ramp speed (0-255)

  #IF_DEFINED __VIDEOCONFERENCER_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    CAMERACOMPONENTCOUNT_CMD            = 'CAMERACOMPONENTCOUNT'        // CAMERACOMPONENTCOUNT-<count>'
    CAMERACOMPONENTCOUNT_QRY            = '?CAMERACOMPONENTCOUNT'       // getCameraComponentCount, '?CAMERACOMPONENTCOUNT'
    CAMERAPRESETCOUNT_CMD               = 'CAMERAPRESETCOUNT'           // CAMERAPRESETCOUNT-<count>'
    CAMERAPRESETCOUNT_QRY               = '?CAMERAPRESETCOUNT'          // getCameraPresetCount, '?CAMERAPRESETCOUNT'
    CAMERAPRESETPROPERTIES_CMD          = 'CAMERAPRESETPROPERTIES'      // [CAMERAPRESETPROPERTIES-<index>,<displayName>]*'
    CAMERAPRESETPROPERTIES_QRY          = '?CAMERAPRESETPROPERTIES'     // getCameraPresetProperties, '?CAMERAPRESETPROPERTIES'
    CAMERAPRESETPROPERTY_CMD            = 'CAMERAPRESETPROPERTY'        // CAMERAPRESETPROPERTY-<index>,<displayName>'
    CAMERAPRESETPROPERTY_QRY            = '?CAMERAPRESETPROPERTY'       // getCameraPresetProperty, '?CAMERAPRESETPROPERTY-<index>'
    CAMERAPRESET_CMD                    = 'CAMERAPRESET'                // CAMERAPRESET-<index>
    CAMERAPRESET_CMD                    = 'CAMERAPRESET'                // setCameraPreset, 'CAMERAPRESET-<preset>'
    CAMERAPRESET_QRY                    = '?CAMERAPRESET'               // getCameraPreset, '?CAMERAPRESET'
    CLOSESEARCHDB_CMD                   = 'PHONEBOOKCLOSESEARCH'        // closeSearchDB, 'SPEEDDIALCLOSESEARCH-<key>'
    CYCLEINPUTSELECT_CMD                = 'CYCLESELECT'                 // cycleInputSelect(), CYCLESELECT
    DBNEXT_QRY                          = 'PHONEBOOKNEXT'               // queryDBNext, 'SPEEDDIALNEXT-<key>[,<count>]'
    DBPREVIOUS_QRY                      = 'PHONEBOOKPREV'               // queryDBPrevious, 'SPEEDDIALPREV-<key>[,<count>]'
    DELETERECORD_CMD                    = 'PHONEBOOKDELETE'             // deleteRecord, 'SPEEDDIALDELETE-<id>'
    DIAL1_CMD                           = 'DIALID'                      // dial, 'DIALID-<id>'
    DIALDTMF_CMD                        = 'DTMF'                        // dialDTMF, 'DTMF-<digit>'
    DIALERSTATUS_CMD                    = 'DIALERSTATUS'                // Dialer Status, 'DIALERSTATUS'
    DIALERSTATUS_QRY                    = '?DIALERSTATUS'               // getDialerStatus, '?DIALERSTATUS'
    DIALNUMBER_CMD                      = 'DIALNUMBER'                  // dialNumber, 'DIALNUMBER-<number>'
    DIAL_CMD                            = 'DIALINDEX'                   // dial, 'DIALINDEX-<index>'
    INCOMINGCALL_CMD                    = 'INCOMINGCALL'                // incoming call, 'INCOMINGCALL'
    INPUTCOUNT_CMD                      = 'INPUTCOUNT'                  // INPUTCOUNT-<count>
    INPUTCOUNT_QRY                      = '?INPUTCOUNT'                 // getInputCount(), INPUTCOUNT-<count>
    INPUTGROUPSELECT_CMD                = 'INPUTGROUPSELECT'            // setInputGroupSelect(inputGroup), INPUTGROUPSELECT-<inputGroup>
    INPUTGROUPSELECT_QRY                = '?INPUTGROUPSELECT'           // getInputGroupSelect(), ?INPUTGROUPSELECT
    INPUTPROPERTIES_CMD                 = 'INPUTPROPERTIES'             // [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTIES_QRY                 = '?INPUTPROPERTIES'            // getInputProperties(), [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTY_CMD                   = 'INPUTPROPERTY'               // INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTPROPERTY_QRY                   = '?INPUTPROPERTY'              // getInputProperty(index), INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTSELECT_CMD                     = 'INPUTSELECT'                 // setInputSelect(index), INPUTSELECT-<index>
    INPUTSELECT_QRY                     = '?INPUTSELECT'                // getInputSelect(), INPUTSELECT-<index>
    INPUTSOURCE_CMD                     = 'INPUT'                       // setInputSource, INPUT-<sourceSelect>,<inputNumber>
    INPUTSOURCE_QRY                     = '?INPUT'                      // getInputSource, ?INPUT
    PHONEBOOKCAPACITY_CMD               = 'PHONEBOOKCAPACITY'           // PHONEBOOKCAPACITTY-<capacity>
    PHONEBOOKCAPACITY_QRY               = '?PHONEBOOKCAPACITY'          // getPhonebookCapacity, '?SPEEDDIALCAPACITY'
    PHONEBOOKRECORD_CMD                 = 'PHONEBOOKRECORD'             // PHONEBOOKRECORD
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    REFRESH_CMD                         = 'PHONEBOOKREFRESH'            // queryDB, 'PHONEBOOKREFESH-<key>[,<count>]'
    SAVECAMERAPRESET_CMD                = 'CAMERAPRESETSAVE'            // saveCameraPreset, 'CAMERAPRESETSAVE-<preset>'
    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    SEARCHDB_CMD                        = 'PHONEBOOKSEARCH'             // searchDB, 'SPEEDDIALSEARCH-<key>,ID=<id>'
    TRAIN1_ST_CMD                       = 'ON=train()'                  // train, 'ON=train()'
    UPDATERECORD_CMD                    = 'PHONEBOOKUPDATE'             // updateRecord, 'SPEEDDIALUPDATE-<id>,<name>,<number>'
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
  #END_IF // __VIDEOCONFERENCER_COMMANDS__

DEFINE_CONSTANT	//	----- Video Processor : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  VPROC_PRESET                          = 137                           // Momentary: Cycle video processor preset
  ASPECT_RATIO                          = 142                           // Momentary: Cycle aspect ratio
  BRIGHT_UP                             = 148                           // Momentary: Increment brightness
  BRIGHT_DN                             = 149                           // Momentary: Decrement brightness
  COLOR_UP                              = 150                           // Momentary: Increment color
  COLOR_DN                              = 151                           // Momentary: Decrement color
  CONTRAST_UP                           = 152                           // Momentary: Increment contrast
  CONTRAST_DN                           = 153                           // Momentary: Decrement contrast
  SHARP_UP                              = 154                           // Momentary: Increment sharpness
  SHARP_DN                              = 155                           // Momentary: Decrement sharpness
  TINT_UP                               = 156                           // Momentary: Increment tint
  TINT_DN                               = 157                           // Momentary: Decrement tint
  PIP_POS                               = 191                           // Momentary: Cycle pip position
  PIP_SWAP                              = 193                           // Momentary: Swap pip
  PIP                                   = 194                           // Momentary: Cycle pip
  PIP_ON                                = 195                           // Discrete:  Set pip on
  PIP_FB                                = 195                           // Feedback:  Pip feedback
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  PIC_MUTE                              = 210                           // Momentary: Cycle picture/video mute
  PIC_MUTE_ON                           = 211                           // Discrete:  Set picture/video mute on
  PIC_MUTE_FB                           = 211                           // Feedback:  Picture/video mute feedback
  PIC_FREEZE                            = 213                           // Momentary: Cycle freeze
  PIC_FREEZE_ON                         = 214                           // Discrete:  Set freeze on
  PIC_FREEZE_FB                         = 214                           // Feedback:  Freeze feedback
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay

  BRIGHT_LVL                            = 10                            // Level: Brightness level (0-255)
  COLOR_LVL                             = 11                            // Level: Color level (0-255)
  CONTRAST_LVL                          = 12                            // Level: Contrast level (0-255)
  SHARP_LVL                             = 13                            // Level: Sharpness level (0-255)
  TINT_LVL                              = 14                            // Level: Tint level (0-255)

  #IF_DEFINED __VIDEOPROCESSOR_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ACTIVEWINDOW_CMD                    = 'ACTIVEWINDOW'                // setActiveWindow, 'ACTIVEWINDOW-<mss>'
    ACTIVEWINDOW_QRY                    = '?ACTIVEWINDOW'               // getActiveWindow, '?ACTIVEWINDOW'
    ASPECTRATIOCOUNT_CMD                = 'ASPECTRATIOCOUNT'            // ASPECTRATIOCOUNT-<count>'
    ASPECTRATIOCOUNT_QRY                = '?ASPECTRATIOCOUNT'           // getAspectRatioCount(), '?ASPECTRATIOCOUNT'
    ASPECTRATIOPROPERTIES_CMD           = 'ASPECTRATIOPROPERTIES'       // '[ASPECTRATIOPROPERTIES-<index>,<displayName>,<value>]*'
    ASPECTRATIOPROPERTIES_QRY           = '?ASPECTRATIOPROPERTIES'      // getAspectRatioProperties(), '?ASPECTRATIOPROPERTIES'
    ASPECTRATIOPROPERTY_CMD             = 'ASPECTRATIOPROPERTY'         // 'ASPECTRATIOPROPERTY-<index>,<displayName>,<value>'
    ASPECTRATIOPROPERTY_QRY             = '?ASPECTRATIOPROPERTY'        // getAspectRatioProperty(index), '?ASPECTRATIOPROPERTY-<index>'
    ASPECTRATIOSELECT_CMD               = 'ASPECTRATIOSELECT'           // setAspectRatioSelect(index), 'ASPECTRATIOSELECT-<index>'
    ASPECTRATIOSELECT_QRY               = '?ASPECTRATIOSELECT'          // getAspectRatioSelect(), '?ASPECTRATIOSELECT'
    ASPECTRATIO_CMD                     = 'ASPECT'                      // setAspectRatio, 'ASPECT-<aspectRatio>'
    ASPECTRATIO_QRY                     = '?ASPECT'                     // getAspectRatio, '?ASPECT'
    CYCLEINPUTSELECT_CMD                = 'CYCLESELECT'                 // cycleInputSelect(), CYCLESELECT
    INPUTCOUNT_CMD                      = 'INPUTCOUNT'                  // INPUTCOUNT-<count>
    INPUTCOUNT_QRY                      = '?INPUTCOUNT'                 // getInputCount(), INPUTCOUNT-<count>
    INPUTGROUPSELECT_CMD                = 'INPUTGROUPSELECT'            // setInputGroupSelect(inputGroup), INPUTGROUPSELECT-<inputGroup>
    INPUTGROUPSELECT_QRY                = '?INPUTGROUPSELECT'           // getInputGroupSelect(), ?INPUTGROUPSELECT
    INPUTPROPERTIES_CMD                 = 'INPUTPROPERTIES'             // [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTIES_QRY                 = '?INPUTPROPERTIES'            // getInputProperties(), [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTY_CMD                   = 'INPUTPROPERTY'               // INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTPROPERTY_QRY                   = '?INPUTPROPERTY'              // getInputProperty(index), INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTSELECT_CMD                     = 'INPUTSELECT'                 // setInputSelect(index), INPUTSELECT-<index>
    INPUTSELECT_QRY                     = '?INPUTSELECT'                // getInputSelect(), INPUTSELECT-<index>
    INPUTSOURCE_CMD                     = 'INPUT'                       // setInputSource, INPUT-<sourceSelect>,<inputNumber>
    INPUTSOURCE_QRY                     = '?INPUT'                      // getInputSource, ?INPUT
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    SAVEVIDEOPROCESSORPRESET_CMD        = 'VPROCPRESETSAVE'             // saveVideoProcessorPreset, 'VPROCPRESETSAVE-<preset>'
    VIDEOPROCESSORPRESET_CMD            = 'VPROCPRESET'                 // setVideoProcessorPreset, 'VPROCPRESET-<preset>'
    VIDEOPROCESSORPRESET_QRY            = '?VPROCPRESET'                // getVideoProcessorPreset, '?VPROCPRESET'
    VIDEOTYPE_CMD                       = 'VIDEOTYPE'                   // setVideoType, 'VIDEOTYPE-<vt>'
    VIDEOTYPE_QRY                       = '?VIDEOTYPE'                  // getVideoType, '?VIDEOTYPE'
  #END_IF // __VIDEOPROCESSOR_COMMANDS__

DEFINE_CONSTANT	//	----- Video Projector : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle lamp power
  DIGIT_0                               = 10                            // Momentary: Press menu button digit 0
  DIGIT_1                               = 11                            // Momentary: Press menu button digit 1
  DIGIT_2                               = 12                            // Momentary: Press menu button digit 2
  DIGIT_3                               = 13                            // Momentary: Press menu button digit 3
  DIGIT_4                               = 14                            // Momentary: Press menu button digit 4
  DIGIT_5                               = 15                            // Momentary: Press menu button digit 5
  DIGIT_6                               = 16                            // Momentary: Press menu button digit 6
  DIGIT_7                               = 17                            // Momentary: Press menu button digit 7
  DIGIT_8                               = 18                            // Momentary: Press menu button digit 8
  DIGIT_9                               = 19                            // Momentary: Press menu button digit 9
  MENU_PLUS_10                          = 20                            // Momentary: Press menu button plus_10
  MENU_ENTER                            = 21                            // Momentary: Press menu button enter
  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  PWR_ON                                = 27                            // Momentary: Set lamp power on
  PWR_OFF                               = 28                            // Momentary: Set lamp power off
  MENU_CANCEL                           = 43                            // Momentary: Press menu button cancel
  MENU_FUNC                             = 44                            // Momentary: Press menu button menu
  MENU_UP                               = 45                            // Momentary: Press menu up button
  MENU_DN                               = 46                            // Momentary: Press menu down button
  MENU_LT                               = 47                            // Momentary: Press menu left button
  MENU_RT                               = 48                            // Momentary: Press menu right button
  MENU_SELECT                           = 49                            // Momentary: Press menu SELECT button (select current item under cursor)
  MENU_EXIT                             = 50                            // Momentary: Press menu button exit
  MENU_UP_LT                            = 51                            // Momentary: Press menu up left button
  MENU_UP_RT                            = 52                            // Momentary: Press menu up right button
  MENU_DN_LT                            = 53                            // Momentary: Press menu down left button
  MENU_DN_RT                            = 54                            // Momentary: Press menu down right button
  MENU_VIDEO                            = 57                            // Momentary: Press menu button video
  MENU_THUMBS_DN                        = 58                            // Momentary: Press menu button thumbs down
  MENU_THUMBS_UP                        = 59                            // Momentary: Press menu button thumbs up
  MENU_ACCEPT                           = 60                            // Momentary: Press menu button accept
  MENU_REJECT                           = 61                            // Momentary: Press menu button reject
  MENU_LIVE_TV                          = 62                            // Momentary: Press menu button live TV
  MENU_SLEEP                            = 63                            // Momentary: Press menu button sleep
  MENU_PPV                              = 64                            // Momentary: Press menu button PPV
  MENU_FUNCTION                         = 65                            // Momentary: Press menu button function
  MENU_SETUP                            = 66                            // Momentary: Press menu button setup
  MENU_XM                               = 77                            // Momentary: Press menu button xm
  MENU_FM                               = 78                            // Momentary: Press menu button fm
  MENU_AM                               = 79                            // Momentary: Press menu button am
  MENU_CLEAR                            = 80                            // Momentary: Press menu button clear
  MENU_BACK                             = 81                            // Momentary: Press menu button back
  MENU_FORWARD                          = 82                            // Momentary: Press menu button forward
  MENU_ADVANCE                          = 83                            // Momentary: Press menu button advance
  MENU_DIMMER                           = 84                            // Momentary: Press menu button dimmer
  MENU_HOLD                             = 85                            // Momentary: Press menu button hold
  MENU_LIST                             = 86                            // Momentary: Press menu button list
  MENU_LT_PAREN                         = 87                            // Momentary: Press menu button left paren
  MENU_RT_PAREN                         = 88                            // Momentary: Press menu button right paren
  MENU_UNDERSCORE                       = 89                            // Momentary: Press menu button underscore
  MENU_DASH                             = 90                            // Momentary: Press menu button dash
  MENU_ASTERISK                         = 91                            // Momentary: Press menu button asterisk
  MENU_DOT                              = 92                            // Momentary: Press menu button dot
  MENU_POUND                            = 93                            // Momentary: Press menu button pound
  MENU_COMMA                            = 94                            // Momentary: Press menu button comma
  MENU_DIAL                             = 95                            // Momentary: Press menu button dial
  MENU_CONFERENCE                       = 96                            // Momentary: Press menu button conference
  MENU_PLUS_100                         = 97                            // Momentary: Press menu button plus_100
  MENU_PLUS_1000                        = 98                            // Momentary: Press menu button plus_1000
  MENU_DISPLAY                          = 99                            // Momentary: Press menu button display
  MENU_SUBTITLE                         = 100                           // Momentary: Press menu button subtitle
  MENU_INFO                             = 101                           // Momentary: Press menu button info
  MENU_FAVORITES                        = 102                           // Momentary: Press menu button favorites
  MENU_CONTINUE                         = 103                           // Momentary: Press menu button continue
  MENU_RETURN                           = 104                           // Momentary: Press menu button RETURN
  MENU_GUIDE                            = 105                           // Momentary: Press menu button guide
  MENU_PAGE_UP                          = 106                           // Momentary: Press menu button page up
  MENU_PAGE_DN                          = 107                           // Momentary: Press menu button page down
  MENU_DECK_A_B                         = 108                           // Momentary: Press menu button deck AB
  MENU_TV_VCR                           = 109                           // Momentary: Press menu button TV VCR
  MENU_RECORD_SPEED                     = 110                           // Momentary: Press menu button record speed
  MENU_PROGRAM                          = 111                           // Momentary: Press menu button program
  MENU_AB_REPEAT                        = 112                           // Momentary: Press menu button AB repeat
  MENU_HELP                             = 113                           // Momentary: Press menu button help
  MENU_TITLE                            = 114                           // Momentary: Press menu button title
  MENU_TOP_MENU                         = 115                           // Momentary: Press menu button top menu
  MENU_ZOOM                             = 116                           // Momentary: Press menu button zoom
  MENU_ANGLE                            = 117                           // Momentary: Press menu button angle
  MENU_AUDIO                            = 118                           // Momentary: Press menu button audio
  MENU_PREVIEW_INPUT                    = 129                           // Momentary: Press menu button preview input
  MENU_SEND_INPUT                       = 130                           // Momentary: Press menu button send input
  MENU_SEND_GRAPHICS                    = 131                           // Momentary: Press menu button send graphics
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  ASPECT_RATIO                          = 142                           // Momentary: Cycle aspect ratio
  BRIGHT_UP                             = 148                           // Momentary: Increment brightness
  BRIGHT_DN                             = 149                           // Momentary: Decrement brightness
  COLOR_UP                              = 150                           // Momentary: Increment color
  COLOR_DN                              = 151                           // Momentary: Decrement color
  CONTRAST_UP                           = 152                           // Momentary: Increment contrast
  CONTRAST_DN                           = 153                           // Momentary: Decrement contrast
  SHARP_UP                              = 154                           // Momentary: Increment sharpness
  SHARP_DN                              = 155                           // Momentary: Decrement sharpness
  TINT_UP                               = 156                           // Momentary: Increment tint
  TINT_DN                               = 157                           // Momentary: Decrement tint
  PIP_POS                               = 191                           // Momentary: Cycle pip position
  PIP_SWAP                              = 193                           // Momentary: Swap pip
  PIP                                   = 194                           // Momentary: Cycle pip
  PIP_ON                                = 195                           // Discrete:  Set pip on
  PIP_FB                                = 195                           // Feedback:  Pip feedback
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback
  MENU_FLASH                            = 203                           // Momentary: Press menu button flash
  PIC_MUTE                              = 210                           // Momentary: Cycle picture/video mute
  PIC_MUTE_ON                           = 211                           // Discrete:  Set picture/video mute on
  PIC_MUTE_FB                           = 211                           // Feedback:  Picture/video mute feedback
  PIC_FREEZE                            = 213                           // Momentary: Cycle freeze
  PIC_FREEZE_ON                         = 214                           // Discrete:  Set freeze on
  PIC_FREEZE_FB                         = 214                           // Feedback:  Freeze feedback
  MENU_RESET                            = 215                           // Momentary: Press menu button reset
  MENU_INSTANT_REPLAY                   = 218                           // Momentary: Press menu button instant replay

  VOL_LVL                               = 1                             // Level: Volume level (0-255)
  BRIGHT_LVL                            = 10                            // Level: Brightness level (0-255)
  COLOR_LVL                             = 11                            // Level: Color level (0-255)
  CONTRAST_LVL                          = 12                            // Level: Contrast level (0-255)
  SHARP_LVL                             = 13                            // Level: Sharpness level (0-255)
  TINT_LVL                              = 14                            // Level: Tint level (0-255)

  #IF_DEFINED __VIDEOPROJECTOR_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ACTIVEWINDOW_CMD                    = 'ACTIVEWINDOW'                // setActiveWindow, 'ACTIVEWINDOW-<mss>'
    ACTIVEWINDOW_QRY                    = '?ACTIVEWINDOW'               // getActiveWindow, '?ACTIVEWINDOW'
    ASPECTRATIOCOUNT_CMD                = 'ASPECTRATIOCOUNT'            // ASPECTRATIOCOUNT-<count>'
    ASPECTRATIOCOUNT_QRY                = '?ASPECTRATIOCOUNT'           // getAspectRatioCount(), '?ASPECTRATIOCOUNT'
    ASPECTRATIOPROPERTIES_CMD           = 'ASPECTRATIOPROPERTIES'       // '[ASPECTRATIOPROPERTIES-<index>,<displayName>,<value>]*'
    ASPECTRATIOPROPERTIES_QRY           = '?ASPECTRATIOPROPERTIES'      // getAspectRatioProperties(), '?ASPECTRATIOPROPERTIES'
    ASPECTRATIOPROPERTY_CMD             = 'ASPECTRATIOPROPERTY'         // 'ASPECTRATIOPROPERTY-<index>,<displayName>,<value>'
    ASPECTRATIOPROPERTY_QRY             = '?ASPECTRATIOPROPERTY'        // getAspectRatioProperty(index), '?ASPECTRATIOPROPERTY-<index>'
    ASPECTRATIOSELECT_CMD               = 'ASPECTRATIOSELECT'           // setAspectRatioSelect(index), 'ASPECTRATIOSELECT-<index>'
    ASPECTRATIOSELECT_QRY               = '?ASPECTRATIOSELECT'          // getAspectRatioSelect(), '?ASPECTRATIOSELECT'
    ASPECTRATIO_CMD                     = 'ASPECT'                      // setAspectRatio, 'ASPECT-<aspectRatio>'
    ASPECTRATIO_QRY                     = '?ASPECT'                     // getAspectRatio, '?ASPECT'
    COOLDOWNCOUNTER_CMD                 = 'COOLING'                     // 'COOLING-<int>'
    COOLDOWNTIME_CMD                    = 'COOLDOWN'                    // setCoolDownTime, 'COOLDOWN-<time>'
    COOLDOWNTIME_QRY                    = '?COOLDOWN'                   // getCoolDownTime, '?COOLDOWN'
    COUNTERNOTIFICATIONON_CMD           = 'COUNTERNOTIFY'               // setCounterNotificationOn, 'COUNTERNOTIFY-<state>'
    CYCLEINPUTSELECT_CMD                = 'CYCLESELECT'                 // cycleInputSelect(), CYCLESELECT
    INPUTCOUNT_CMD                      = 'INPUTCOUNT'                  // INPUTCOUNT-<count>
    INPUTCOUNT_QRY                      = '?INPUTCOUNT'                 // getInputCount(), INPUTCOUNT-<count>
    INPUTGROUPSELECT_CMD                = 'INPUTGROUPSELECT'            // setInputGroupSelect(inputGroup), INPUTGROUPSELECT-<inputGroup>
    INPUTGROUPSELECT_QRY                = '?INPUTGROUPSELECT'           // getInputGroupSelect(), ?INPUTGROUPSELECT
    INPUTPROPERTIES_CMD                 = 'INPUTPROPERTIES'             // [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTIES_QRY                 = '?INPUTPROPERTIES'            // getInputProperties(), [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
    INPUTPROPERTY_CMD                   = 'INPUTPROPERTY'               // INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTPROPERTY_QRY                   = '?INPUTPROPERTY'              // getInputProperty(index), INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
    INPUTSELECT_CMD                     = 'INPUTSELECT'                 // setInputSelect(index), INPUTSELECT-<index>
    INPUTSELECT_QRY                     = '?INPUTSELECT'                // getInputSelect(), INPUTSELECT-<index>
    INPUTSOURCE_CMD                     = 'INPUT'                       // setInputSource, INPUT-<sourceSelect>,<inputNumber>
    INPUTSOURCE_QRY                     = '?INPUT'                      // getInputSource, ?INPUT
    LAMPTIME_CMD                        = 'LAMPTIME'                    // setLampTime, 'LAMPTIME-<time>'
    LAMPTIME_QRY                        = '?LAMPTIME'                   // getLampTime, '?LAMPTIME'
    PRESSMENUBUTTON10_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-K'
    PRESSMENUBUTTON11_CMD               = 'ALPHA'                       // pressMenuButton, 'ALPHA-L'
    PRESSMENUBUTTON1_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-B'
    PRESSMENUBUTTON2_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-C'
    PRESSMENUBUTTON3_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-D'
    PRESSMENUBUTTON4_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-E'
    PRESSMENUBUTTON5_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-F'
    PRESSMENUBUTTON6_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-G'
    PRESSMENUBUTTON7_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-H'
    PRESSMENUBUTTON8_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-I'
    PRESSMENUBUTTON9_CMD                = 'ALPHA'                       // pressMenuButton, 'ALPHA-J'
    PRESSMENUBUTTON_CMD                 = 'ALPHA'                       // pressMenuButton, 'ALPHA-A'
    PROJECTORPRESET_CMD                 = 'VPROJPRESET'                 // setProjectorPreset, 'VPROJPRESET-<preset>'
    PROJECTORPRESET_QRY                 = '?VPROJPRESET'                // getProjectorPreset, '?VPROJPRESET'
    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    VIDEOTYPE_CMD                       = 'VIDEOTYPE'                   // setVideoType, 'VIDEOTYPE-<vt>'
    VIDEOTYPE_QRY                       = '?VIDEOTYPE'                  // getVideoType, '?VIDEOTYPE'
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
    WARMUPCOUNTER_CMD                   = 'WARMING'                     // 'WARMING-<int>'
    WARMUPTIME_CMD                      = 'WARMUP'                      // setWarmUpTime, 'WARMUP-<time>'
    WARMUPTIME_QRY                      = '?WARMUP'                     // getWarmUpTime, '?WARMUP'
  #END_IF // __VIDEOPROJECTOR_COMMANDS__

DEFINE_CONSTANT	//	----- Video Wall : Channels and Levels

  POWER                                 = 9                             // Momentary: Cycle power
  PWR_ON                                = 27                            // Momentary: Set power on
  PWR_OFF                               = 28                            // Momentary: Set power off
  PAN_UP                                = 132                           // Momentary: Pan up
  PAN_DN                                = 133                           // Momentary: Pan down
  PAN_LT                                = 134                           // Momentary: Pan left
  PAN_RT                                = 135                           // Momentary: Pan right
  BRIGHT_UP                             = 148                           // Momentary: Increment brightness
  BRIGHT_DN                             = 149                           // Momentary: Decrement brightness
  COLOR_UP                              = 150                           // Momentary: Increment color
  COLOR_DN                              = 151                           // Momentary: Decrement color
  CONTRAST_UP                           = 152                           // Momentary: Increment contrast
  CONTRAST_DN                           = 153                           // Momentary: Decrement contrast
  SHARP_UP                              = 154                           // Momentary: Increment sharpness
  SHARP_DN                              = 155                           // Momentary: Decrement sharpness
  TINT_UP                               = 156                           // Momentary: Increment tint
  TINT_DN                               = 157                           // Momentary: Decrement tint
  ZOOM_OUT                              = 158                           // Momentary: Zoom out
  ZOOM_IN                               = 159                           // Momentary: Zoom in
  PIC_MUTE                              = 210                           // Momentary: Cycle picture/video mute
  PIC_MUTE_ON                           = 211                           // Discrete:  Set picture/video mute on
  PIC_MUTE_FB                           = 211                           // Feedback:  Picture/video mute feedback
  PIC_FREEZE                            = 213                           // Momentary: Cycle freeze
  PIC_FREEZE_ON                         = 214                           // Discrete:  Set freeze on
  PIC_FREEZE_FB                         = 214                           // Feedback:  Freeze feedback

  BRIGHT_LVL                            = 10                            // Level: Brightness level (0-255)
  COLOR_LVL                             = 11                            // Level: Color level (0-255)
  CONTRAST_LVL                          = 12                            // Level: Contrast level (0-255)
  SHARP_LVL                             = 13                            // Level: Sharpness level (0-255)
  TINT_LVL                              = 14                            // Level: Tint level (0-255)

  #IF_DEFINED __VIDEOWALL_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    ADDWINDOWCOMPONENT_CMD              = 'WINDOWADD'                   // addWindowComponent, 'WINDOWADD-<index>,<address>'
    IMAGEPOSITION_QRY                   = '?IMAGE'                      // getImagePosition, '?IMAGE'
    IMAGESIZE_QRY                       = '?IMAGE'                      // getImageSize, '?IMAGE'
    IMAGESOURCE_CMD                     = 'IMAGESOURCE'                 // setImageSource, 'IMAGESOURCE-<source>'
    IMAGESOURCE_QRY                     = '?IMAGE'                      // getImageSource, '?IMAGE'
    IMAGE_CMD                           = 'IMAGE'                       // setImage, 'IMAGE-<source>,<x>,<y>,<height>,<width>'
    REMOVEWINDOWCOMPONENT1_CMD          = 'WINDOWREMOVEADDR'            // removeWindowComponent, 'WINDOWREMOVEADDR-<address>'
    REMOVEWINDOWCOMPONENT_CMD           = 'WINDOWREMOVEIDX'             // removeWindowComponent, 'WINDOWREMOVEIDX-<index>'
    SAVEVIDEOWALLCONFIGURATION_CMD      = 'VWALLCONFIGSAVE'             // saveVideoWallConfiguration, 'VWALLCONFIGSAVE-<config>'
    VIDEOWALLCONFIGURATIONLIST_CMD      = 'VWALLCONFIGLIST'             // setVideoWallConfiguration, 'VWALLCONFIGLIST-<config>'
    VIDEOWALLCONFIGURATIONLIST_QRY      = '?VWALLCONFIGLIST'            // queryVideoWallConfiguration, '?VWALLCONFIGLIST'
    VIDEOWALLCONFIGURATION_CMD          = 'VWALLCONFIG'                 // setVideoWallConfiguration, 'VWALLCONFIG-<config>'
    VIDEOWALLCONFIGURATION_QRY          = '?VWALLCONFIG'                // getVideoWallConfiguration, '?VWALLCONFIG'
    WINDOWADDR_CMD                      = 'WINDOWADDR'
    WINDOWCOMPONENTADDRESS_QRY          = '?WINDOWADDR'                 // getWindowComponentAddress, '?WINDOWADDR-<index>'
    WINDOWCOMPONENTINDEX_QRY            = '?WINDOWIDX'                  // getWindowComponentIndex, '?WINDOWIDX-<address>'
    WINDOWPOSITION_QRY                  = '?WINDOW'                     // getWindowPosition, '?WINDOW'
    WINDOWSIZE_QRY                      = '?WINDOW'                     // getWindowSize, '?WINDOW'
    WINDOW_CMD                          = 'WINDOW'                      // setWindow, 'WINDOW-<x>,<y>,<height>,<width>'
    ZORDER1_CMD                         = 'WINDOWZORDER'                // setZOrder, 'WINDOWZORDER-<value>'
    ZORDER_CMD                          = 'WINDOWZORDER'                // setZOrder, 'WINDOWZORDER-<position>'
    ZORDER_QRY                          = '?WINDOWZORDER'               // getZOrder, '?WINDOWZORDER'
  #END_IF // __VIDEOWALL_COMMANDS__

DEFINE_CONSTANT	//	----- Volume Controller : Channels and Levels

  VOL_UP                                = 24                            // Ramping:   Ramp volume up
  VOL_UP_FB                             = 24                            // Feedback:  Volume ramp up feedback
  VOL_DN                                = 25                            // Ramping:   Ramp volume down
  VOL_DN_FB                             = 25                            // Feedback:  Volume ramp down feedback
  VOL_MUTE                              = 26                            // Momentary: Cycle volume mute
  VOL_PRESET                            = 138                           // Momentary: Cycle volume preset
  VOL_MUTE_ON                           = 199                           // Discrete:  Set volume mute
  VOL_MUTE_FB                           = 199                           // Feedback:  Volume mute feedback

  VOL_LVL                               = 1                             // Level: Volume level (0-255)

  #IF_DEFINED __VOLUMECONTROLLER_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    SAVEVOLUMEPRESET_CMD                = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
    VOLUMEPRESET_CMD                    = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
    VOLUMEPRESET_QRY                    = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
  #END_IF // __VOLUMECONTROLLER_COMMANDS__

DEFINE_CONSTANT	//	----- Weather : Channels and Levels

  WEATHER_FORCE_READING                 = 208                           // Momentary: Force new readings
  WEATHER_RAINING                       = 230                           // Feedback:  Weather condition indicates rain
  WEATHER_FREEZING                      = 231                           // Feedback:  Outdoor temperature is at or below freezing
  WEATHER_BAR_RISING                    = 232                           // Feedback:  Barometric pressure rising feedback
  WEATHER_BAR_FALLING                   = 233                           // Feedback:  Barometric pressure falling feedback
  DEVICE_COMMUNICATING                  = 251                           // Feedback:  Device online event
  DATA_INITIALIZED                      = 252                           // Feedback:  Data initialized event

  INDOOR_TEMP_LVL                       = 33                            // Level: Indoor temperature, range is n..m degrees F or C
  OUTDOOR_TEMP_LVL                      = 34                            // Level: Outdoor temperature, range is n..m degrees F or C
  INDOOR_HUMID_LVL                      = 35                            // Level: Indoor humidity range is 0..100 percent
  OUTDOOR_HUMID_LVL                     = 36                            // Level: Outdoor humidity, range is 0..100 percent
  WEATHER_HI_TEMP_LVL                   = 43                            // Level: Today's actual high temperature, range is n..m degrees F or C
  WEATHER_LO_TEMP_LVL                   = 44                            // Level: Today's actual low temperature, range is n..m degrees F or C
  WEATHER_WIND_CHILL_LVL                = 45                            // Level: Wind chill, range is n..m degrees F or C
  WEATHER_HEAT_INDEX_LVL                = 46                            // Level: Heat index, range is n..m degrees F or C
  WEATHER_DEWPOINT_LVL                  = 47                            // Level: Dew point, range is n..m degrees F or C
  WEATHER_BAR_LVL                       = 48                            // Level: Barometric pressure, range is n..m in inches hg or millimeter hg/torr

  #IF_DEFINED __WEATHER_COMMANDS__
  /*|------------------------------------------------------------------------------|
    | Command constants for reference only, they are defined at end of file.       |
    | The #DEFINES is used to exclude from compiler and is never used.             |
    |------------------------------------------------------------------------------|*/

    CHANCEOFPRECIPITATION_QRY           = '?FORECASTCOP'                // getChanceOfPrecipitation, '?FORECASTCOP'
    CURRENTCONDITION_QRY                = '?WEATHERCONDITION'           // getCurrentCondition, '?WEATHERCONDITION'
    FORECASTHIGHTEMPERATURE_QRY         = '?FORECASTHIGH'               // getForecastHighTemperature, '?FORECASTHIGH'
    FORECASTLOWTEMPERATURE_QRY          = '?FORECASTLOW'                // getForecastLowTemperature, '?FORECASTLOW'
    RAINFALL_QRY                        = '?WEATHERRAIN'                // getRainfall, '?WEATHERRAIN-<duration>'
    WEATHERALERT_QRY                    = '?WEATHERALERT'               // getWeatherAlert, '?WEATHERALERT'
    WEATHERCONDITIONS_QRY               = '?FORECASTCONDITION'          // getWeatherConditions, '?FORECASTCONDITION'
    WEATHERSCALE_CMD                    = 'WEATHERSCALE'                // setWeatherScale, 'WEATHERSCALE-<scale>'
    WEATHERSCALE_QRY                    = '?WEATHERSCALE'               // getWeatherScale, '?WEATHERSCALE'
    WINDINFO_QRY                        = '?WEATHERWIND'                // getWindInfo, '?WEATHERWIND'
  #END_IF // __WEATHER_COMMANDS__

DEFINE_CONSTANT	//	----- Command constants

  ACTIVEWINDOW_CMD                      = 'ACTIVEWINDOW'                // setActiveWindow, 'ACTIVEWINDOW-<mss>'
  ACTIVEWINDOW_QRY                      = '?ACTIVEWINDOW'               // getActiveWindow, '?ACTIVEWINDOW'
  ADDAUDIOPROCESSORCOMPONENT_CMD        = 'AUDIOPROCADD'                // addAudioProcessorComponent, 'AUDIOPROCADD-<index>,<address>'
  ADDHVACCOMPONENT_CMD                  = 'HVACADD'                     // addHVACComponent, 'HVACADD-<index>,<hvacAddress>'
  ADDKEYPADCOMPONENT_CMD                = 'KEYPADADD'                   // addKeypadComponent, 'KEYPADADD-<index>,<keypadaddr>'
  ADDKEYPADR_CMD                        = 'KEYPADADDR'                  // setButtonState, 'KEYPADBTN-<index>,<buttonaddr>,<state>'
  ADDKEYPADSYSTEMADDRESS_CMD            = 'KEYPADSYSTEMADDRESS'         // addKeypadSystemAddress, 'KEYPADSYSTEMADDRESS-<index>,<address>'
  ADDKEYPADSYSTEMCOMPONENT_CMD          = 'KEYPADSYSTEMINDEX'           // addKeypadSystemComponent, 'KEYPADSYSTEMADD-<index>,<address>'
  ADDLIGHTCOMPONENT_CMD                 = 'LIGHTADD'                    // addLightComponent, 'LIGHTADD-<index>,<address>'
  ADDLIGHTSYSTEMADDRESS_CMD             = 'LIGHTSYSTEMADDR'             // addLightSystemComponent, 'LIGHTSYSTEMADDR-<index>,<address>'
  ADDOUTPUTSTREAMSINK_CMD               = 'ENCODESINKADD'               // addOutputStreamSink, 'ENCODESINKADD-<url>'
  ADDRFIDSYSTEMADDRESS_CMD              = 'RFIDSYSTEMADDR'              // addRFIDSystemComponent, 'RFIDSYSTEMADDR-<index>,<address>'
  ADDWINDOWCOMPONENT_CMD                = 'WINDOWADD'                   // addWindowComponent, 'WINDOWADD-<index>,<address>'
  ADJUSTEQUALIZEROFFSET_CMD             = 'EQUALIZER_OFFSET'            // setEqualizerOffset overloads, 'EQUALIZER_OFFSET-(band)=(offset),[,EQUALIZER-(band)=(offset)]+'
  ASPECTRATIOCOUNT_CMD                  = 'ASPECTRATIOCOUNT'            // ASPECTRATIOCOUNT-<count>'
  ASPECTRATIOCOUNT_QRY                  = '?ASPECTRATIOCOUNT'           // getAspectRatioCount(), '?ASPECTRATIOCOUNT'
  ASPECTRATIOPROPERTIES_CMD             = 'ASPECTRATIOPROPERTIES'       // '[ASPECTRATIOPROPERTIES-<index>,<displayName>,<value>]*'
  ASPECTRATIOPROPERTIES_QRY             = '?ASPECTRATIOPROPERTIES'      // getAspectRatioProperties(), '?ASPECTRATIOPROPERTIES'
  ASPECTRATIOPROPERTY_CMD               = 'ASPECTRATIOPROPERTY'         // 'ASPECTRATIOPROPERTY-<index>,<displayName>,<value>'
  ASPECTRATIOPROPERTY_QRY               = '?ASPECTRATIOPROPERTY'        // getAspectRatioProperty(index), '?ASPECTRATIOPROPERTY-<index>'
  ASPECTRATIOSELECT_CMD                 = 'ASPECTRATIOSELECT'           // setAspectRatioSelect(index), 'ASPECTRATIOSELECT-<index>'
  ASPECTRATIOSELECT_QRY                 = '?ASPECTRATIOSELECT'          // getAspectRatioSelect(), '?ASPECTRATIOSELECT'
  ASPECTRATIO_CMD                       = 'ASPECT'                      // setAspectRatio, 'ASPECT-<aspectRatio>'
  ASPECTRATIO_QRY                       = '?ASPECT'                     // getAspectRatio, '?ASPECT'
  AUDIOMIXERCROSSPOINTMUTEON_CMD        = 'XPOINTMUTE'                  // setAudioMixerCrosspointMuteOn, 'XPOINTMUTE-<state>,<input>,<output>'
  AUDIOMIXERCROSSPOINTMUTEON_QRY        = '?XPOINTMUTE'                 // isAudioMixerCrosspointMuteOn, '?XPOINTMUTE-<input>,<output>'
  AUDIOMIXERCROSSPOINT_CMD              = 'XPOINT'                      // setAudioMixerCrosspoint, 'XPOINT-<value>,<input>,<output>[,<output>,<output>]'
  AUDIOMIXERCROSSPOINT_QRY              = '?XPOINT'                     // getAudioMixerCrosspoint, '?XPOINT-<input>,<output>'
  AUDIOMIXERPRESET_CMD                  = 'MIXERPRESET'                 // setAudioMixerPreset, 'MIXERPRESET-<preset>'
  AUDIOMIXERPRESET_QRY                  = '?MIXERPRESET'                // getAudioMixerPreset, '?MIXERPRESET'
  AUDIOPROCESSORCOMPONENTADDRESS_QRY    = '?AUDIOPROCADDR'              // getAudioProcessorComponentAddress, '?AUDIOPROCADDR-<index>'
  AUDIOPROCESSORCOMPONENTINDEX_QRY      = '?AUDIOPROCIDX'               // getAudioProcessorComponentIndex, '?AUDIOPROCIDX-<address>'
  AUDIOPROCESSORCROSSPOINTMUTEON_CMD    = 'XPOINTMUTE'                  // setAudioProcessorCrosspointMuteOn, 'XPOINTMUTE-<state>,<input>,<output>'
  AUDIOPROCESSORCROSSPOINTMUTEON_QRY    = '?XPOINTMUTE'                 // isAudioProcessorCrosspointMuteOn, '?XPOINTMUTE-<input>,<output>'
  AUDIOPROCESSORCROSSPOINT_CMD          = 'XPOINT'                      // setAudioProcessorCrosspoint, 'XPOINT-<value>,<input>,<output>[,<output>,<output>]'
  AUDIOPROCESSORCROSSPOINT_QRY          = '?XPOINT'                     // getAudioProcessorCrosspoint, '?XPOINT-<input>,<output>'
  AUDIOPROCESSORINPUT_QRY               = '?INPUT'                      // getAudioProcessorInput, '?INPUT-<output>'
  AUDIOPROCESSOROUTPUT_QRY              = '?OUTPUT'                     // getAudioProcessorOutput, '?OUTPUT-<input>'
  AUDIOPROCESSORPRESET_CMD              = 'AUDIOPROCPRESET'             // setAudioProcessorPreset, 'AUDIOPROCPRESET-<preset>'
  AUDIOPROCESSORPRESET_QRY              = '?AUDIOPROCPRESET'            // getAudioProcessorPreset, '?AUDIOPROCPRESET'
  BACKLIGHTON_CMD                       = 'TEXTBACKLIGHT'               // setBacklightOn, 'TEXTBACKLIGHT-<state>'
  BAND_CMD                              = 'BAND'                        // setBand, 'BAND-<tb>'
  BAND_QRY                              = '?BAND'                       // getBand, '?BAND'
  BUTTONSTATE_CMD                       = 'KEYPADBTN'                   // setButtonState, 'KEYPADBTN-<index>,<buttonaddr>,<state>'
  CAMERACOMPONENTCOUNT_CMD              = 'CAMERACOMPONENTCOUNT'        // CAMERACOMPONENTCOUNT-<count>'
  CAMERACOMPONENTCOUNT_QRY              = '?CAMERACOMPONENTCOUNT'       // getCameraComponentCount, '?CAMERACOMPONENTCOUNT'
  CAMERAPRESETCOUNT_CMD                 = 'CAMERAPRESETCOUNT'           // CAMERAPRESETCOUNT-<count>'
  CAMERAPRESETCOUNT_QRY                 = '?CAMERAPRESETCOUNT'          // getCameraPresetCount, '?CAMERAPRESETCOUNT'
  CAMERAPRESETPROPERTIES_CMD            = 'CAMERAPRESETPROPERTIES'      // [CAMERAPRESETPROPERTIES-<index>,<displayName>]*'
  CAMERAPRESETPROPERTIES_QRY            = '?CAMERAPRESETPROPERTIES'     // getCameraPresetProperties, '?CAMERAPRESETPROPERTIES'
  CAMERAPRESETPROPERTY_CMD              = 'CAMERAPRESETPROPERTY'        // CAMERAPRESETPROPERTY-<index>,<displayName>'
  CAMERAPRESETPROPERTY_QRY              = '?CAMERAPRESETPROPERTY'       // getCameraPresetProperty, '?CAMERAPRESETPROPERTY-<index>'
  CAMERAPRESET_CMD                      = 'CAMERAPRESET'                // setCameraPreset, 'CAMERAPRESET-<preset>'
  CAMERAPRESET_QRY                      = '?CAMERAPRESET'               // getCameraPreset, '?CAMERAPRESET'
  CHANCEOFPRECIPITATION_QRY             = '?FORECASTCOP'                // getChanceOfPrecipitation, '?FORECASTCOP'
  CLEARDISPLAY_CMD                      = 'TEXT'                        // clearDisplay, 'TEXT-'
  CLOSESEARCHDB_CMD                     = 'MEDIADBCLOSESEARCH'          // closeSearchDB, 'MEDIADBCLOSESEARCH-<key>'
  COOLDOWNCOUNTER_CMD                   = 'COOLING'                     // 'COOLING-<int>'
  COOLDOWNTIME_CMD                      = 'COOLDOWN'                    // setCoolDownTime, 'COOLDOWN-<time>'
  COOLDOWNTIME_QRY                      = '?COOLDOWN'                   // getCoolDownTime, '?COOLDOWN'
  COUNTERNOTIFICATIONON_CMD             = 'COUNTERNOTIFY'               // setCounterNotificationOn, 'COUNTERNOTIFY-<state>'
  CURRENTCONDITION_QRY                  = '?WEATHERCONDITION'           // getCurrentCondition, '?WEATHERCONDITION'
  CYCLEINPUTSELECT_CMD                  = 'CYCLESELECT'                 // cycleInputSelect(), CYCLESELECT
  CYCLELIGHTSYSTEMSTATE_CMD             = 'LIGHTSYSTEMSTATE'            // cycleLight, 'LIGHTSYSTEMSTATE-<address>,TOGGLE' toggles state
  CYCLELIGHT_CMD                        = 'LIGHTSTATE'                  // cycleLight, 'LIGHTSTATE-<index>,TOGGLE'
  DBNEXT_QRY                            = 'MEDIADBNEXT'                 // queryDB, 'MEDIADBNEXT-<key>[,<count>]'
  DBPREVIOUS_QRY                        = 'MEDIADBPREV'                 // queryDB, 'MEDIADBPREV-<key>[,<count>]'
  DEBUG_CMD                             = 'DEBUG'                       // setDebugState, 'DEBUG-<state
	DEBUGSTATE_CMD                        = 'DEBUGSTATE'                  // setDebugState, 'DEBUGSTATE-<state
  DEBUGSTATE_QRY                        = '?DEBUG'                      // getDebugState, '?DEBUG'
  DELETERECORD_CMD                      = 'MEDIADBDELETE'               // deleteRecord, 'MEDIADBDELETE-<id>'
  DEVICEDATETIME_CMD                    = 'CLOCK'                       // setDeviceDateTime, 'CLOCK-<mm/dd/yyyy> <hh:mm:ss>'
  DIAL1_CMD                             = 'DIALID'                      // dial, 'DIALID-<id>'
  DIALDTMF_CMD                          = 'DTMF'                        // dialDTMF, 'DTMF-<digit>'
  DIALERSTATUS_CMD                      = 'DIALERSTATUS'                // Dialer Status, 'DIALERSTATUS'
  DIALERSTATUS_QRY                      = '?DIALERSTATUS'               // getDialerStatus, '?DIALERSTATUS'
  DIALNUMBER_CMD                        = 'DIALNUMBER'                  // dialNumber, 'DIALNUMBER-<number>'
  DIAL_CMD                              = 'DIALINDEX'                   // dial, 'DIALINDEX-<index>'
  DISCCAPACITY_QRY                      = '?DISCCAPACITY'               // getDiscCapacity, '?DISCCAPACITY'
  DISCINFO_CMD                          = 'DISCINFO'                    // Disc Info
  DISCINFO_QRY                          = '?DISCINFO'                   // getDiscInfo, '?DISCINFO'
  DISCPROPERTIES_QRY                    = '?DISCPROPS'                  // queryDiscProperties, '?DISCPROPS'
  DISCPROPERTY_QRY                      = '?DISCPROP'                   // queryDiscProperty, '?DISCPROP'
  DISCPROP_CMD                          = 'DISCPROP'                    // Disc Property
  DISC_CMD                              = 'SETDISC'                     // setDisc, 'SETDISC-<disc>'
  EQUALIZERBANDS_CMD                    = 'EQUALIZER_BANDS'             // return 'EQUALIZER_BANDS-(band)[,(band)]+'
  EQUALIZERBANDS_QRY                    = '?EQUALIZER_BANDS'            // getEqualizerBands, '?EQUALIZER_BANDS'
  EQUALIZER_CMD                         = 'EQUALIZER'                   // setEqualizer overloads, 'EQUALIZER-(band)=(gain),[,EQUALIZER-(band)=(gain)]+'
  EQUALIZER_QRY                         = '?EQUALIZER'                  // getEqualizer overloads, '?EQUALIZER', '?EQUALIZER-(band)[,(band)]+'
  FORECASTHIGHTEMPERATURE_QRY           = '?FORECASTHIGH'               // getForecastHighTemperature, '?FORECASTHIGH'
  FORECASTLOWTEMPERATURE_QRY            = '?FORECASTLOW'                // getForecastLowTemperature, '?FORECASTLOW'
  FWVERSION_CMD                         = 'FWVERSION'                   // FWVersion
  FWVERSION_QRY                         = '?FWVERSION'                  // getFWVersion, '?FWVERSION'
  HUMIDIFYSTATE_CMD                     = 'HVACHUMID'                   // setHumidifyState, 'HVACHUMID-<hs>'
  HUMIDIFYSTATE_QRY                     = '?HVACHUMID'                  // getHumidifyState, '?HVACHUMID'
  HUMIDIFYSTATUS_CMD                    = 'HVACHUMIDSTATUS'             // 'HVACHUMIDSTATUS-<state>'
  HUMIDIFYSTATUS_QRY                    = '?HVACHUMIDSTATUS'            // getHumidifyStatus, '?HVACHUMIDSTATUS'
  HVACCOMPONENTADDRESS_CMD              = 'HVACADDR'                    // 'HVACADDR-<index>,<address>'
  HVACCOMPONENTADDRESS_QRY              = '?HVACADDR'                   // getHVACComponentAddress, '?HVACADDR-<index>'
  HVACCOMPONENTINDEX_QRY                = '?HVACIDX'                    // getHVACComponentIndex, '?HVACIDX-<hvacAddress>'
  IMAGEPOSITION_QRY                     = '?IMAGE'                      // getImagePosition, '?IMAGE'
  IMAGESIZE_QRY                         = '?IMAGE'                      // getImageSize, '?IMAGE'
  IMAGESOURCE_CMD                       = 'IMAGESOURCE'                 // setImageSource, 'IMAGESOURCE-<source>'
  IMAGESOURCE_QRY                       = '?IMAGE'                      // getImageSource, '?IMAGE'
  IMAGE_CMD                             = 'IMAGE'                       // setImage, 'IMAGE-<source>,<x>,<y>,<height>,<width>'
  INCOMINGCALL_CMD                      = 'INCOMINGCALL'                // incoming call, 'INCOMINGCALL'
  INPUTCOUNT_CMD                        = 'INPUTCOUNT'                  // INPUTCOUNT-<count>
  INPUTCOUNT_QRY                        = '?INPUTCOUNT'                 // getInputCount(), INPUTCOUNT-<count>
  INPUTGROUPSELECT_CMD                  = 'INPUTGROUPSELECT'            // setInputGroupSelect(inputGroup), INPUTGROUPSELECT-<inputGroup>
  INPUTGROUPSELECT_QRY                  = '?INPUTGROUPSELECT'           // getInputGroupSelect(), ?INPUTGROUPSELECT
  INPUTPROPERTIES_CMD                   = 'INPUTPROPERTIES'             // [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
  INPUTPROPERTIES_QRY                   = '?INPUTPROPERTIES'            // getInputProperties(), [INPUTPROPERTIES-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>]*
  INPUTPROPERTY_CMD                     = 'INPUTPROPERTY'               // INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
  INPUTPROPERTY_QRY                     = '?INPUTPROPERTY'              // getInputProperty(index), INPUTPROPERTY-<index>,<inputGroup>,<signalType>,<deviceLabel>,<displayName>
  INPUTSELECT_CMD                       = 'INPUTSELECT'                 // setInputSelect(index), INPUTSELECT-<index>
  INPUTSELECT_QRY                       = '?INPUTSELECT'                // getInputSelect(), INPUTSELECT-<index>
  INPUTSOURCE_CMD                       = 'INPUT'                       // setInputSource, INPUT-<sourceSelect>,<inputNumber>
  INPUTSOURCE_QRY                       = '?INPUT'                      // getInputSource, ?INPUT
  INPUT_QRY                             = '?INPUT'                      // getInput, '?INPUT-<sl>,<output>'
  IOCHANNELCOUNT_CMD                    = 'IOCHANNELCOUNT'              // setIOChannelCount, 'IOCHANNELCOUNT-<count>'
  IOCHANNELCOUNT_QRY                    = '?IOCHANNELCOUNT'             // getIOChannelCount, '?IOCHANNELCOUNT'
  IOCHANNELDIRECTION_CMD                = 'IOCHANNELDIRECTION'          // setIOChannelDirection, 'IOCHANNELDIRECTION-<channel>,<direction>'
  IOCHANNELDIRECTION_QRY                = '?IOCHANNELDIRECTION'         // getIOChannelDirection, '?IOCHANNELDIRECTION-<channel>'
  IOCHANNELINPUTSENSE_CMD               = 'IOCHANNELINPUTSENSE'         // setIOChannelInputSense, 'IOCHANNELINPUTSENSE-<channel>,<sense>'
  IOCHANNELINPUTSENSE_QRY               = '?IOCHANNELINPUTSENSE'        // getIOChannelState, '?IOCHANNELINPUTSENSE-<channel>'
  IOCHANNELSTATE_CMD                    = 'IOCHANNELSTATE'              // setIOChannelState, 'IOCHANNELSTATE-<channel>,<state>'
  IOCHANNELSTATE_QRY                    = '?IOCHANNELSTATE'             // getIOChannelState, '?IOCHANNELSTATE-<channel>'
  KEYPADCOMPONENTADDRESS_QRY            = '?KEYPADADDR'                 // getKeypadComponentAddress, '?KEYPADADDR-<index>'
  KEYPADCOMPONENTINDEX_QRY              = '?KEYPADIDX'                  // getKeypadComponentIndex, '?KEYPADIDX-<keypadaddr>'
  KEYPADSYSTEMBUTTONSTATE_CMD           = 'KEYPADSYSTEMBUTTONSTATE'     // setKeypadSystemButtonState, 'KEYPADSYSTEMBUTTONSTATE-<address>,<state>'
  KEYPADSYSTEMBUTTONSTATE_QRY           = '?KEYPADSYSTEMBUTTONSTATE'    // getKeypadSystemButtonState, '?KEYPADSYSTEMBUTTONSTATE-<address>'
  KEYPADSYSTEMBUTTONSTATUS_CMD          = 'KEYPADSYSTEMBUTTONSTATUS'    // setKeypadSystemButtonStatus, 'KEYPADSYSTEMBUTTONSTATUS-<address>,<status>'
  KEYPADSYSTEMBUTTONSTATUS_QRY          = '?KEYPADSYSTEMBUTTONSTATUS'   // getKeypadSystemButtonStatus, '?KEYPADSYSTEMBUTTONSTATUS-<address>'
  KEYPADSYSTEMCOMPONENTADDRESS_QRY      = '?KEYPADSYSTEMADDRESS'        // getKeypadSystemComponentAddress, '?KEYPADSYSTEMADDRESS-<index>'
  KEYPADSYSTEMCOMPONENTINDEX_QRY        = '?KEYPADSYSTEMINDEX'          // getKeypadSystemComponentIndex, '?KEYPADSYSTEMINDEX-<address>'
  LAMPTIME_CMD                          = 'LAMPTIME'                    // setLampTime, 'LAMPTIME-<time>'
  LAMPTIME_QRY                          = '?LAMPTIME'                   // getLampTime, '?LAMPTIME'
  LIGHTADDRESS_CMD                      = 'LIGHTADDR'
  LIGHTCOMPONENTADDRESS_QRY             = '?LIGHTADDR'                  // getLightComponentAddress, '?LIGHTADDR-<index>'
  LIGHTCOMPONENTINDEX_QRY               = '?LIGHTIDX'                   // getLightComponentIndex, '?LIGHTIDX-<address>'
  LIGHTLEVEL1_CMD                       = 'LIGHTLEVEL'                  // setLightLevel, 'LIGHTLEVEL-<index>,<level>,<time>'
  LIGHTLEVELRAMP1_CMD                   = 'LIGHTRAMP'                   // setLightLevelRamp, 'LIGHTRAMP-<index>,DOWN'
  LIGHTLEVELRAMP2_CMD                   = 'LIGHTRAMP'                   // setLightLevelRamp, 'LIGHTRAMP-<index>,STOP'
  LIGHTLEVELRAMP_CMD                    = 'LIGHTRAMP'                   // setLightLevelRamp, 'LIGHTRAMP-<index>,UP'
  LIGHTLEVEL_CMD                        = 'LIGHTLEVEL'                  // setLightLevel, 'LIGHTLEVEL-<index>,<level>'
  LIGHTLEVEL_QRY                        = '?LIGHTLEVEL'                 // getLightLevel, '?LIGHTLEVEL-<index>'
  LIGHTON_CMD                           = 'LIGHTSTATE'                  // setLightOn, 'LIGHTSTATE-<index>,<state>'
  LIGHTON_QRY                           = '?LIGHTSTATE'                 // isLightOn, '?LIGHTSTATE-<index>'
  LIGHTSYSTEMADDRESS_CMD                = 'LIGHTSYSTEMADDR'
  LIGHTSYSTEMADDRESS_QRY                = '?LIGHTSYSTEMADDR'            // getLightSystemComponentAddress, '?LIGHTSYSTEMADDR-<index>'
  LIGHTSYSTEMINDEX_CMD                  = 'LIGHTSYSTEMINDEX'
  LIGHTSYSTEMINDEX_QRY                  = '?LIGHTSYSTEMINDEX'           // getLightSystemComponentIndex, '?LIGHTSYSTEMINDEX-<address>'
  LIGHTSYSTEMLEVELRAMP_CMD              = 'LIGHTSYSTEMRAMP'             // setLightLevelRamp, 'LIGHTSYSTEMRAMP-<address>,UP|DOWN|STOP'
  LIGHTSYSTEMLEVEL_CMD                  = 'LIGHTSYSTEMLEVEL'            // setLightLevel, 'LIGHTSYSTEMLEVEL-<address>,<level>[,<interval>]'
  LIGHTSYSTEMLEVEL_QRY                  = '?LIGHTSYSTEMLEVEL'           // getLightLevel, '?LIGHTSYSTEMLEVEL-<address>'
  LIGHTSYSTEMON_QRY                     = '?LIGHTSYSTEMSTATE'           // isLightOn, '?LIGHTSYSTEMSTATE-<address>'
  LIGHTSYSTEMSTATE_CMD                  = 'LIGHTSYSTEMSTATE'            // setLightState, 'LIGHTSYSTEMSTATE-<address>,ON|OFF'
  MEDIADBPROPERTIES_QRY                 = '?MEDIADBPROPS'               // queryMediaDBProperties, '?MEDIADBPROPS-<id>'
  MEDIADBPROPERTY_CMD                   = 'MEDIADBPROP'                 // setMediaDBProperty, 'MEDIADBPROP-<id>,<key>,<value>'
  MEDIADBPROPERTY_QRY                   = '?MEDIADBPROP'                // queryMediaDBProperty, '?MEDIADBPROP-<id>,<key>'
  MEDIADEVICECOUNTERNOTIFICATIONON_CMD  = 'MEDIACOUNTERNOTIFY'          // setMediaDeviceCounterNotificationOn, 'MEDIACOUNTERNOTIFY-<state>'
  MEDIADEVICEPROPERTIES_QRY             = '?DECODEPROPS'                // queryMediaDeviceProperties, '?DECODEPROPS'
  MEDIADEVICEPROPERTY_QRY               = '?DECODEPROP'                 // queryMediaDeviceProperty, '?DECODEPROP-<key>'
  MEDIADEVICESOURCE_CMD                 = 'DECODESOURCE'                // setMediaDeviceSource, 'DECODESOURCE-<url>'
  MEDIADEVICESOURCE_QRY                 = '?DECODESOURCE'               // getMediaDeviceSource, '?DECODESOURCE'
  MOTORPRESET_CMD                       = 'MOTORPRESET'                 // setMotorPreset, 'MOTORPRESET-<preset>'
  MOTORPRESET_QRY                       = '?MOTORPRESET'                // getMotorPreset, '?MOTORPRESET'
  MULTIWINDOWPRESET_CMD                 = 'MULTIWINPRESET'              // setMultiWindowPreset, 'MULTIWINPRESET-<preset>'
  MULTIWINDOWPRESET_QRY                 = '?MULTIWINPRESET'             // getMultiWindowPreset, '?MULTIWINPRESET'
  OKTOARM_QRY                           = '?SECARMABLE'                 // isOKToArm, '?SECARMABLE'
  OUTPUTSTREAMPROPERTIES_QRY            = '?ENCODEPROPS'                // queryOutputStreamProperties, '?ENCODEPROPS'
  OUTPUTSTREAMPROPERTY_CMD              = 'ENCODEPROP'                  // setOutputStreamProperty, 'ENCODEPROP-<key>,<value>'
  OUTPUTSTREAMPROPERTY_QRY              = '?ENCODEPROP'                 // queryOutputStreamProperty, '?ENCODEPROP-<key>'
  OUTPUTSTREAMSINKS_QRY                 = '?ENCODESINK'                 // getOutputStreamSinks, '?ENCODESINK'
  OUTPUTSTREAMSOURCE_CMD                = 'ENCODESOURCE'                // setOutputStreamSource, 'ENCODESOURCE-<url>'
  OUTPUTSTREAMSOURCE_QRY                = '?ENCODESOURCE'               // getOutputStreamSource, '?ENCODESOURCE'
  OUTPUT_QRY                            = '?OUTPUT'                     // getOutput, '?OUTPUT-<sl>,<input>'
  PASSBACKON_CMD                        = 'PASSBACK'                    // setPassbackOn, 'PASSBACK-<state>'
  PASSTHRU_CMD                          = 'PASSTHRU'                    // passThru, 'PASSTHRU-<buffer>'
  PHONEBOOKCAPACITY_CMD                 = 'PHONEBOOKCAPACITY'           // PHONEBOOKCAPACITTY-<capacity>
  PHONEBOOKCAPACITY_QRY                 = '?PHONEBOOKCAPACITY'          // getPhonebookCapacity, '?SPEEDDIALCAPACITY'
  PHONEBOOKRECORD_CMD                   = 'PHONEBOOKRECORD'             // PHONEBOOKRECORD
  PLAYPOSITION1_CMD                     = 'PLAYPOSITION'                // setPlayPosition, 'PLAYPOSITION-<title>,<track>,<counter>'
  PLAYPOSITION2_CMD                     = 'PLAYPOSITION'                // setPlayPosition, 'PLAYPOSITION-<track>'
  PLAYPOSITION3_CMD                     = 'PLAYPOSITION'                // setPlayPosition, 'PLAYPOSITION-<track>,<counter>'
  PLAYPOSITION_CMD                      = 'PLAYPOSITION'                // setPlayPosition, 'PLAYPOSITION-<title>,<track>'
  POINTSTATE_CMD                        = 'SECPOINTSTATE'               // setPointState, 'SECPOINTSTATE-<point>,<state>,<password>'
  POINTSTATE_QRY                        = '?SECPOINTSTATE'              // getPointState, '?SECPOINTSTATE-<point>'
  POINTSTATUS_QRY                       = '?SECPOINTSTATUS'             // getPointStatus, '?SECPOINTSTATUS-<point>'
  POOLSPAAUXON_CMD                      = 'POOLAUX'                     // setPoolSpaAuxOn, 'POOLAUX-<aux name>,<state>'
  POOLSPAAUXON_QRY                      = '?POOLAUX'                    // isPoolSpaAuxOn, '?POOLAUX-<aux name>'
  POOLSPATEMPERATURESCALE_CMD           = 'POOLSCALE'                   // setPoolSpaTemperatureScale, 'POOLSCALE-<ts>'
  POOLSPATEMPERATURESCALE_QRY           = '?POOLSCALE'                  // getPoolSpaTemperatureScale, '?POOLSCALE'
  MENU_ALPHA_A_CMD                      = 'ALPHA-A'                     // pressMenuButton, 'ALPHA-A'
  MENU_ALPHA_B_CMD                      = 'ALPHA-B'                     // pressMenuButton, 'ALPHA-B'
  MENU_ALPHA_C_CMD                      = 'ALPHA-C'                     // pressMenuButton, 'ALPHA-C'
  MENU_ALPHA_D_CMD                      = 'ALPHA-D'                     // pressMenuButton, 'ALPHA-D'
  MENU_ALPHA_E_CMD                      = 'ALPHA-E'                     // pressMenuButton, 'ALPHA-E'
  MENU_ALPHA_F_CMD                      = 'ALPHA-F'                     // pressMenuButton, 'ALPHA-F'
  MENU_ALPHA_G_CMD                      = 'ALPHA-G'                     // pressMenuButton, 'ALPHA-G'
  MENU_ALPHA_H_CMD                      = 'ALPHA-H'                     // pressMenuButton, 'ALPHA-H'
  MENU_ALPHA_I_CMD                      = 'ALPHA-I'                     // pressMenuButton, 'ALPHA-I'
  MENU_ALPHA_J_CMD                      = 'ALPHA-J'                     // pressMenuButton, 'ALPHA-J'
  MENU_ALPHA_K_CMD                      = 'ALPHA-K'                     // pressMenuButton, 'ALPHA-K'
  MENU_ALPHA_L_CMD                      = 'ALPHA-L'                     // pressMenuButton, 'ALPHA-L'
  MENU_ALPHA_M_CMD                      = 'ALPHA-M'                     // pressMenuButton, 'ALPHA-M'
  MENU_ALPHA_N_CMD                      = 'ALPHA-N'                     // pressMenuButton, 'ALPHA-N'
  MENU_ALPHA_O_CMD                      = 'ALPHA-O'                     // pressMenuButton, 'ALPHA-O'
  MENU_ALPHA_P_CMD                      = 'ALPHA-P'                     // pressMenuButton, 'ALPHA-P'
  MENU_ALPHA_Q_CMD                      = 'ALPHA-Q'                     // pressMenuButton, 'ALPHA-Q'
  MENU_ALPHA_R_CMD                      = 'ALPHA-R'                     // pressMenuButton, 'ALPHA-R'
  MENU_ALPHA_S_CMD                      = 'ALPHA-S'                     // pressMenuButton, 'ALPHA-S'
  MENU_ALPHA_T_CMD                      = 'ALPHA-T'                     // pressMenuButton, 'ALPHA-T'
  MENU_ALPHA_U_CMD                      = 'ALPHA-U'                     // pressMenuButton, 'ALPHA-U'
  MENU_ALPHA_V_CMD                      = 'ALPHA-V'                     // pressMenuButton, 'ALPHA-V'
  MENU_ALPHA_W_CMD                      = 'ALPHA-W'                     // pressMenuButton, 'ALPHA-W'
  MENU_ALPHA_X_CMD                      = 'ALPHA-X'                     // pressMenuButton, 'ALPHA-X'
  MENU_ALPHA_Y_CMD                      = 'ALPHA-Y'                     // pressMenuButton, 'ALPHA-Y'
  MENU_ALPHA_Z_CMD                      = 'ALPHA-Z'                     // pressMenuButton, 'ALPHA-Z'
  PROCESSREADERSBYTAG_CMD               = 'READERSBYTAG'                // processReadersByTagEvent, 'READERSBYTAG- READERSBYTAG-<tagId>,<tagName>,<tagInfo>,<tagPercentPower>,<readerCount>,<readerIndex> [,<readerAddress>,<tagSignalStrength>]*'
  PROCESSREADERSTATUS_CMD               = 'READERSTATUS'                // processReaderStatusEvent, 'READERSTATUS-<readerAddress>,<readerStatus>,<lastError>'
  PROCESSREADERS_CMD                    = 'READERS'                     // processReadersEvent, 'READERS-<readerAddress1>[,readerAddressN>]*'
  PROCESSTAGACQUIRED_CMD                = 'TAGACQUIRED'                 // processTagAcquiredEvent, 'TAGACQUIRED-<readerAddress>,<tagId>,<tagName>,<tagInfo>,<timestamp>, <tagSignalStrength>,<tagPercentPower>'
  PROCESSTAGBUTTON_CMD                  = 'TAGBUTTON'                   // processTagLostEvent, 'TAGBUTTON-<tagId>,<tagName>,<tagInfo>,<state>'
  PROCESSTAGINFO_CMD                    = 'TAGINFO'                     // processTagInfoEvent, 'TAGINFO-<tagId>,<tagName>,<tagInfo>'
  PROCESSTAGLOST_CMD                    = 'TAGLOST'                     // processTagLostEvent, 'TAGLOST-<readerAddress>,<tagId>,<tagName>,<tagInfo>,<lastTimestamp>'
  PROCESSTAGSBYREADER_CMD               = 'TAGSBYREADER'                // processTagsByReaderEvent, 'TAGSBYREADER-<readerAddress>,<tagCount>,<tagIndex> [,<tagId>,<tagName>,<tagInfo>,<tagSignalStrength>,<tagPercentPower>]*'
  PROCESSTAGSIGNALSTRENGTH_CMD          = 'TAGSIGNALSTRENGTH'           // processTagAcquiredEvent, 'TAGSIGNALSTRENGTH-<readerAddress>,<tagId>,<tagName>,<tagInfo>,<timestamp>, <tagSignalStrength>,<tagPercentPower>'
  PROJECTORPRESET_CMD                   = 'VPROJPRESET'                 // setProjectorPreset, 'VPROJPRESET-<preset>'
  PROJECTORPRESET_QRY                   = '?VPROJPRESET'                // getProjectorPreset, '?VPROJPRESET'
  PROPERTY_CMD                          = 'PROPERTY'                    // setProperty, 'PROPERTY-<key>,<value>'
  PROPERTY_QRY                          = '?PROPERTY'                   // getProperty, '?PROPERTY-<key>'
  RAINFALL_QRY                          = '?WEATHERRAIN'                // getRainfall, '?WEATHERRAIN-<duration>'
  READERAUTOPOLL_CMD                    = 'READERAUTOPOLL'              // SetReaderAutoPoll, 'READERAUTOPOLL-<readerAddress>,<autoPoll>[,<interval>]'
  READERSBYTAG_QRY                      = '?READERSBYTAG'               // getReadersByTag, '?READERSBYTAG-<tagId>'
  READERSTATUS_QRY                      = '?READERSTATUS'               // getReaderStatus, '?READERSTATUS-<readerAddress>'
  READERS_QRY                           = '?READERS'                    // getReaders, '?READERS'
  REFRESH_CMD                           = 'MEDIADBREFRESH'              // queryDB, 'MEDIADBREFESH-<key>[,<count>]'
  REINITIALIZE_CMD                      = 'REINIT'                      // reinitialize, 'REINIT'
  RELAYCHANNELCOUNT_CMD                 = 'RELAYCHANNELCOUNT'           // setRELAYChannelCount, 'RELAYCHANNELCOUNT-<count>'
  RELAYCHANNELCOUNT_QRY                 = '?RELAYCHANNELCOUNT'          // getRELAYChannelCount, '?RELAYCHANNELCOUNT'
  RELAYCHANNELSTATE_CMD                 = 'RELAYCHANNELSTATE'           // setRELAYChannelState, 'RELAYCHANNELSTATE-<channel>,<state>'
  RELAYCHANNELSTATE_QRY                 = '?RELAYCHANNELSTATE'          // getRELAYChannelState, '?RELAYCHANNELSTATE-<channel>'
  REMOVEAUDIOPROCESSORCOMPONENT1_CMD    = 'AUDIOPROCREMOVEIDX'          // removeAudioProcessorComponent, 'AUDIOPROCREMOVEIDX-<index>'
  REMOVEAUDIOPROCESSORCOMPONENT_CMD     = 'AUDIOPROCREMOVEADDR'         // removeAudioProcessorComponent, 'AUDIOPROCREMOVEADDR-<address>'
  REMOVEHVACCOMPONENT1_CMD              = 'HVACREMOVEADDR'              // removeHVACComponent, 'HVACREMOVEADDR-<hvacAddress>'
  REMOVEHVACCOMPONENT_CMD               = 'HVACREMOVEIDX'               // removeHVACComponent, 'HVACREMOVEIDX-<index>'
  REMOVEKEYPADCOMPONENT1_CMD            = 'KEYPADREMOVEADDR'            // removeKeypadComponent, 'KEYPADREMOVEADDR-<keypadaddr>'
  REMOVEKEYPADCOMPONENT_CMD             = 'KEYPADREMOVEIDX'             // removeKeypadComponent, 'KEYPADREMOVEIDX-<index>'
  REMOVEKEYPADSYSTEMCOMPONENT1_CMD      = 'KEYPADSYSTEMREMOVEADDRESS'   // removeKeypadSystemComponent, 'KEYPADSYSTEMREMOVEADDRESS-<address>'
  REMOVEKEYPADSYSTEMCOMPONENT_CMD       = 'KEYPADSYSTEMREMOVEINDEX'     // removeKeypadSystemComponent, 'KEYPADSYSTEMREMOVEINDEX-<index>'
  REMOVELIGHTCOMPONENT1_CMD             = 'LIGHTREMOVEADDR'             // removeLightComponent, 'LIGHTREMOVEADDR-<address>'
  REMOVELIGHTCOMPONENT_CMD              = 'LIGHTREMOVEIDX'              // removeLightComponent, 'LIGHTREMOVEIDX-<index>'
  REMOVELIGHTSYSTEMADDRESS_CMD          = 'LIGHTSYSTEMREMOVEADDRESS'    // removeLightSystemComponent, 'LIGHTSYSTEMREMOVEADDRESS-<address>'
  REMOVELIGHTSYSTEMINDEX_CMD            = 'LIGHTSYSTEMREMOVEINDEX'      // removeLightSystemComponent, 'LIGHTSYSTEMREMOVEINDEX-<index>'
  REMOVEOUTPUTSTREAMSINK_CMD            = 'ENCODESINKREMOVE'            // removeOutputStreamSink, 'ENCODESINKREMOVE-<url>'
  REMOVERFIDSYSTEMADDRESS_CMD           = 'RFIDSYSTEMREMOVEADDRESS'     // removeRFIDSystemComponent, 'RFIDSYSTEMREMOVEADDRESS-<address>'
  REMOVERFIDSYSTEMINDEX_CMD             = 'RFIDSYSTEMREMOVEINDEX'       // removeRFIDSystemComponent, 'RFIDSYSTEMREMOVEINDEX-<index>'
  REMOVEWINDOWCOMPONENT1_CMD            = 'WINDOWREMOVEADDR'            // removeWindowComponent, 'WINDOWREMOVEADDR-<address>'
  REMOVEWINDOWCOMPONENT_CMD             = 'WINDOWREMOVEIDX'             // removeWindowComponent, 'WINDOWREMOVEIDX-<index>'
  RFIDSYSTEMADDRESS_CMD                 = 'RFIDSYSTEMADDR'
  RFIDSYSTEMADDRESS_QRY                 = '?RFIDSYSTEMADDR'             // getRFIDSystemComponentAddress, '?RFIDSYSTEMADDR-<index>'
  RFIDSYSTEMINDEX_CMD                   = 'RFIDSYSTEMINDEX'
  RFIDSYSTEMINDEX_QRY                   = '?RFIDSYSTEMINDEX'            // getRFIDSystemComponentIndex, '?RFIDSYSTEMINDEX-<address>'
  SAVEAUDIOMIXERPRESET_CMD              = 'MIXERPRESETSAVE'             // saveAudioMixerPreset, 'MIXERPRESETSAVE-<preset>'
  SAVEAUDIOPROCESSORPRESET_CMD          = 'AUDIOPROCPRESETSAVE'         // saveAudioProcessorPreset, 'AUDIOPROCPRESETSAVE-<preset>'
  SAVECAMERAPRESET_CMD                  = 'CAMERAPRESETSAVE'            // saveCameraPreset, 'CAMERAPRESETSAVE-<preset>'
  SAVEMULTIWINDOWPRESET_CMD             = 'MULTIWINPRESETSAVE'          // saveMultiWindowPreset, 'MULTIWINPRESETSAVE-<preset>'
  SAVESWITCHERPRESET_CMD                = 'SWITCHPRESETSAVE'            // saveSwitcherPreset, 'SWICTHPRESETSAVE-<preset>'
  SAVEVIDEOPROCESSORPRESET_CMD          = 'VPROCPRESETSAVE'             // saveVideoProcessorPreset, 'VPROCPRESETSAVE-<preset>'
  SAVEVIDEOWALLCONFIGURATION_CMD        = 'VWALLCONFIGSAVE'             // saveVideoWallConfiguration, 'VWALLCONFIGSAVE-<config>'
  SAVEVOLUMEPRESET_CMD                  = 'VOLPRESETSAVE'               // saveVolumePreset, 'VOLPRESETSAVE-<preset>'
  SEARCHDB_CMD                          = 'MEDIADBSEARCH'               // searchDB, 'MEDIADBSEARCH-<key>,<search type>=<search string>,RETURN=<return type>'
  SECURITYSTATE_CMD                     = 'SECSTATE'                    // setSecurityState, 'SECSTATE-<state>,<password>'
  SECURITYSTATE_QRY                     = '?SECSTATE'                   // getSecurityState, '?SECSTATE'
  SECURITYSTATUS_QRY                    = '?SECSTATUS'                  // getSecurityStatus, '?SECSTATUS'
  SLIDE_CMD                             = 'SLIDE'                       // setSlide, 'SLIDE-<slide>'
  SLIDE_QRY                             = '?SLIDE'                      // getSlide, '?SLIDE'
  STATIONPRESETCOUNT_CMD                = 'STATIONPRESETCOUNT'          // 'STATIONPRESETCOUNT-<count>'
  STATIONPRESETCOUNT_QRY                = '?STATIONPRESETCOUNT'         // getStationPresetCount, '?STATIONPRESETCOUNT'
  STATIONPRESETPROPERTIES_CMD           = 'STATIONPRESETPROPERTIES'     // '[STATIONPRESETPROPERTIES-<index>,<displayName>,<station>]*'
  STATIONPRESETPROPERTIES_QRY           = '?STATIONPRESETPROPERTIES'    // getStationPresetProperties, '?STATIONPRESETPROPERTIES'
  STATIONPRESETPROPERTY_CMD             = 'STATIONPRESETPROPERTY'       // 'STATIONPRESETPROPERTY-<index>,<displayName>,<station>'
  STATIONPRESETPROPERTY_QRY             = '?STATIONPRESETPROPERTY'      // getStationPresetProperty(index), '?STATIONPRESETPROPERTY-<index>'
  STATIONPRESETSELECT_CMD               = 'STATIONPRESETSELECT'         // 'STATIONPRESETSELECT-<index>'
  STATIONPRESETSELECT_QRY               = '?STATIONPRESETSELECT'        // getStationPresetSelect, '?STATIONPRESETSELECT'
  STATIONPRESET_CMD                     = 'TUNERPRESET'                 // setStationPreset, 'TUNERPRESET-<preset>'
  STATIONPRESET_QRY                     = '?TUNERPRESET'                // getStationPreset, '?TUNERPRESET'
  STATION_CMD                           = 'XCH'                         // setStation, 'XCH-<station>'
  STATION_QRY                           = '?XCH'                        // getStation, '?XCH'
  SURROUNDMODE_CMD                      = 'SURROUND'                    // setSurroundMode, 'SURROUND-<mode>'
  SURROUNDMODE_QRY                      = '?SURROUND'                   // getSurroundMode, '?SURROUND'
	SURROUNDMODES_CMD                     = 'SURROUNDMODES'               // SurroundModes, 'SURROUNDMODES-<mode>,<mode>...'
	SURROUNDMODES_QRY                     = '?SURROUNDMODES'              // getSurroundMode, '?SURROUND'
  SWITCHAUDIOPROCESSORINPUTTOOUTPUT_CMD = 'AI<input>O<output,output,>'
  SWITCHERPRESET_CMD                    = 'SWITCHPRESET'                // setSwitcherPreset, 'SWITCHPRESET-<preset>'
  SWITCHERPRESET_QRY                    = '?SWITCHPRESET'               // getSwitcherPreset, '?SWITCHPRESET'
  SWITCHINPUTTOOUTPUT1_CMD              = 'AI<input>O<output,output,>'
  SWITCHINPUTTOOUTPUT2_CMD              = 'CL<sl>I<input>O<output,output>'
  SWITCHINPUTTOOUTPUT3_CMD              = 'VI<input>O<output,output,>'
  SWITCHINPUTTOOUTPUT_CMD               = 'CI<input>O<output,output>'
  SWITCH_LAUDIOI_CMD                    = 'SWITCH-LAUDIOI'
  TAGINFO_QRY                           = '?TAGINFO'                    // getTagInfo, '?TAGINFO-<tagId>'
  TAGSBYREADER_QRY                      = '?TAGSBYREADER'               // getTagsByReader, '?TAGSBYREADER-<readerAddress>'
  TAPECOUNTERNOTIFICATION_CMD           = 'TAPECOUNTERNOTIFY'           // setTapeCounterNotification, 'TAPECOUNTERNOTIFY-<state>'
  TEMPERATURESCALE_CMD                  = 'HVACSCALE'                   // setTemperatureScale, 'HVACSCALE-<ts>'
  TEMPERATURESCALE_QRY                  = '?HVACSCALE'                  // getTemperatureScale, '?HVACSCALE'
  TEXTDISPLAY_CMD                       = 'TEXT'                        // setTextDisplay, 'TEXT-<text>'
  TEXTLINE_CMD                          = 'TEXT'                        // setTextLine, 'TEXT-<line>,<text>'
  TEXT_CMD                              = 'TEXT'                        // setText, 'TEXT-<line>,<column>,<text>'
  TITLECOUNTERNOTIFICATIONON_CMD        = 'TITLECOUNTERNOTIFY'          // setTitleCounterNotificationOn, 'TITLECOUNTERNOTIFY-<state>'
  TITLECOUNTER_CMD                      = 'TITLECOUNTER'                // Title Counter
  TITLEINFO_CMD                         = 'TITLEINFO'                   // Title Info
  TITLEINFO_QRY                         = '?TITLEINFO'                  // getTitleInfo, '?TITLEINFO'
  TITLEPROPERTIES_QRY                   = '?TITLEPROPS'                 // queryTitleProperties, '?TITLEPROPS'
  TITLEPROPERTY_QRY                     = '?TITLEPROP'                  // queryTitleProperty, '?TITLEPROP'
  TITLEPROP_CMD                         = 'TITLEPROP'                   // Disc Property
  TRACKCOUNTERNOTIFICATIONON_CMD        = 'TRACKCOUNTERNOTIFY'          // setTrackCounterNotificationOn, 'TRACKCOUNTERNOTIFY-<state>'
  TRACKINFO_QRY                         = '?TRACKINFO'                  // getTrackInfo, '?TRACKINFO'
  TRACKPROPERTIES_QRY                   = '?TRACKPROPS'                 // queryTrackProperties, '?TRACKPROPS'
  TRACKPROPERTY_QRY                     = '?TRACKPROP'                  // queryTrackProperty, '?TRACKPROP'
  TRACKPROP_CMD                         = 'TRACKPROP'                   // Track Property
  TRAIN1_ST_CMD                         = 'ON=train()'                  // train, 'ON=train()'
  TUNERBANDPROPERTIES_CMD               = 'TUNERBANDPROPERTIES'         // '[TUNERBANDPROPERTIES-<index>,<displayName>,<tunerBand>]*'
  TUNERBANDPROPERTIES_QRY               = '?TUNERBANDPROPERTIES'        // getTunerBandProperties, '?TUNERBANDPROPERTIES'
  TUNERBANDPROPERTYCOUNT_CMD            = 'TUNERBANDPROPERTYCOUNT'      // 'TUNERBANDPROPERTYCOUNT-<count>'
  TUNERBANDPROPERTYCOUNT_QRY            = '?TUNERBANDPROPERTYCOUNT'     // getTunerBandPropertyCount, '?TUNERBANDPROPERTYCOUNT'
  TUNERBANDPROPERTY_CMD                 = 'TUNERBANDPROPERTY'           // 'TUNERBANDPROPERTY-<index>,<displayName>,<tunerBand>'
  TUNERBANDPROPERTY_QRY                 = '?TUNERBANDPROPERTY'          // getTunerBandProperty(index), '?TUNERBANDPROPERTY-<index>'
  TUNERBANDSELECT_CMD                   = 'TUNERBANDSELECT'             // 'TUNERBANDSELECT-<index>'
  TUNERBANDSELECT_QRY                   = '?TUNERBANDSELECT'            // getTunerBandSelect, '?TUNERBANDSELECT'
  TUNERCOMPONENTPROPERTIES_CMD          = 'TUNERCOMPONENTPROPERTIES'    // '[TUNERCOMPONENTPROPERTIES-<index>,<displayName>,<value>]*'
  TUNERCOMPONENTPROPERTIES_QRY          = '?TUNERCOMPONENTPROPERTIES'   // getTunerComponentProperties, '?TUNERCOMPONENTPROPERTIES'
  TUNERCOMPONENTPROPERTYCOUNT_CMD       = 'TUNERCOMPONENTPROPERTYCOUNT'   // 'TUNERCOMPONENTPROPERTYCOUNT-<count>'
  TUNERCOMPONENTPROPERTYCOUNT_QRY       = '?TUNERCOMPONENTPROPERTYCOUNT'   // getTunerComponentPropertyCount, '?TUNERCOMPONENTPROPERTYCOUNT'
  TUNERCOMPONENTPROPERTY_CMD            = 'TUNERCOMPONENTPROPERTY'      // 'TUNERCOMPONENTPROPERTY-<index>,<displayName>,<value>'
  TUNERCOMPONENTPROPERTY_QRY            = '?TUNERCOMPONENTPROPERTY'     // getTunerComponentProperty(index), '?TUNERCOMPONENTPROPERTY-<index>'
  TUNERCOMPONENTSELECT_CMD              = 'TUNERCOMPONENTSELECT'        // 'TUNERCOMPONENTSELECT-<index>'
  TUNERCOMPONENTSELECT_QRY              = '?TUNERCOMPONENTSELECT'       // getTunerComponentSelect, '?TUNERCOMPONENTSELECT'
  UPDATERECORD_CMD                      = 'MEDIADBUPDATE'               // updateRecord, 'MEDIADBUPDATE-<id>,<name>,<record type>[,<url>]'
  UPSALARMS_CMD                         = 'UPSALARMS'                   // UPSALARMS-<alarms>
  UPSALARMS_QRY                         = '?UPSALARMS'                  // getAlarms, '?UPSALARMS'
  UPSBACKUPPROPERTY_CMD                 = 'UPSBACKUPPROPERTY'           // UPSBACKUPPROPERTY-<bu-prop>,<bu-value>
  UPSBACKUPPROPERTY_QRY                 = '?UPSBACKUPPROPERTY'          // getBackupProperty, '?UPSBACKUPPROPERTY-<bu-prop>'
  UPSBACKUPSTATUS_CMD                   = 'UPSBACKUPSTATUS'             // UPSBACKUPSTATUS-<backupstatus>
  UPSBACKUPSTATUS_QRY                   = '?UPSBACKUPSTATUS'            // getBackupStatus, '?UPSBACKUPSTATUS'
  UPSBACKUPSUPPORTED_CMD                = 'UPSBACKUPSUPPORTED'          // UPSBACKUPSUPPORTED-<TRUE|FALSE>
  UPSBACKUPSUPPORTED_QRY                = '?UPSBACKUPSUPPORTED'         // isBackupSupported, '?UPSBACKUPSUPPORTED'
  UPSBACKUPTIMER_CMD                    = 'UPSBACKUPTIMER'              // UPSBACKUPTIMER-<timer>,<value>
  UPSBACKUPTIMER_QRY                    = '?UPSBACKUPTIMER'             // getBackupTimer, '?UPSBACKUPTIMER-<timer>'
  UPSINPUTPROPERTY_CMD                  = 'UPSINPUTPROPERTY'            // UPSINPUTPROPERTY-<input-prop>,<value>
  UPSINPUTPROPERTY_QRY                  = '?UPSINPUTPROPERTY'           // getInputProperty, '?UPSINPUTPROPERTY-<input-prop>'
  UPSOUTLETCOUNT_CMD                    = 'UPSOUTLETCOUNT'              // UPSOUTLETCOUNT-<count>
  UPSOUTLETCOUNT_QRY                    = '?UPSOUTLETCOUNT'             // getOutletCount, '?UPSOUTLETCOUNT'
  UPSOUTLETPROPERTIES_CMD               = 'UPSOUTLETPROPERTIES'         // UPSOUTLETPROPERTIES-[<outlet_number>,<outlet-prop>,<value>]
  UPSOUTLETPROPERTIES_QRY               = '?UPSOUTLETPROPERTIES'        // getOutletProperties, '?UPSOUTLETPROPERTIES-<outlet_number>'
  UPSOUTLETPROPERTY_CMD                 = 'UPSOUTLETPROPERTY'           // UPSOUTLETPROPERTY-<outlet_number>,<outlet-prop>,<value>    1+
  UPSOUTLETPROPERTY_QRY                 = '?UPSOUTLETPROPERTY'          // getOutletProperty, '?UPSOUTLETPROPERTY-<outlet_number>,<outlet-prop>'
  UPSOUTLETSTATE_CMD                    = 'UPSOUTLETSTATE'              // setOutletState, 'UPSOUTLETSTATE-<outlet_number>,<outlet-state>'
  UPSOUTLETSTATE_QRY                    = '?UPSOUTLETSTATE'             // getOutletState, '?UPSOUTLETSTATE-<outlet_number>'
  UPSSTATE_CMD                          = 'UPSSTATE'                    // setState, 'UPSSTATE-<ups-state>[,<delay>]'
  UPSSTATE_QRY                          = '?UPSSTATE'                   // getState, '?UPSSTATE-<ups_state>'
  UPSSTATUS_CMD                         = 'UPSSTATUS'                   // UPSSTATUS-<state>
  UPSSTATUS_QRY                         = '?UPSSTATUS'                  // getStatus, '?UPSSTATUS'
  UPSTEMPERATURESCALE_CMD               = 'UPSTEMPERATURESCALE'         // setTemperatureScale, 'UPSTEMPERATURESCALE-<tempscale>'
  UPSTEMPERATURESCALE_QRY               = '?UPSTEMPERATURESCALE'        // getTemperatureScale, '?UPSTEMPERATURESCALE'
  UPSTESTRESULT_CMD                     = 'UPSTESTRESULT'               // UPSTESTRESULT-<test>,<testresult>
  UPSTESTRESULT_QRY                     = '?UPSTESTRESULT'              // getTestResult, '?UPSTESTRESULT'
  UPSTESTSSUPPORTED_CMD                 = 'UPSTESTSSUPPORTED'           // UPSTESTSSUPPORTED-<test>[,<test>]*
  UPSTESTSSUPPORTED_QRY                 = '?UPSTESTSSUPPORTED'          // getTestsSupported, '?UPSTESTSSUPPORTED'
  UPSTEST_CMD                           = 'UPSTEST'                     // setTest, 'UPSTEST-<test>,<action>'
  VERSION_CMD                           = 'VERSION'                     // Version
  VERSION_QRY                           = '?VERSION'                    // getVersion, '?VERSION'
  VIDEOPROCESSORPRESET_CMD              = 'VPROCPRESET'                 // setVideoProcessorPreset, 'VPROCPRESET-<preset>'
  VIDEOPROCESSORPRESET_QRY              = '?VPROCPRESET'                // getVideoProcessorPreset, '?VPROCPRESET'
  VIDEOTYPE_CMD                         = 'VIDEOTYPE'                   // setVideoType, 'VIDEOTYPE-<vt>'
  VIDEOTYPE_QRY                         = '?VIDEOTYPE'                  // getVideoType, '?VIDEOTYPE'
  VIDEOWALLCONFIGURATIONLIST_CMD        = 'VWALLCONFIGLIST'             // setVideoWallConfiguration, 'VWALLCONFIGLIST-<config>'
  VIDEOWALLCONFIGURATIONLIST_QRY        = '?VWALLCONFIGLIST'            // queryVideoWallConfiguration, '?VWALLCONFIGLIST'
  VIDEOWALLCONFIGURATION_CMD            = 'VWALLCONFIG'                 // setVideoWallConfiguration, 'VWALLCONFIG-<config>'
  VIDEOWALLCONFIGURATION_QRY            = '?VWALLCONFIG'                // getVideoWallConfiguration, '?VWALLCONFIG'
  VOLUMEPRESET_CMD                      = 'VOLPRESET'                   // setVolumePreset, 'VOLPRESET-<preset>'
  VOLUMEPRESET_QRY                      = '?VOLPRESET'                  // getVolumePreset, '?VOLPRESET'
  WARMUPCOUNTER_CMD                     = 'WARMING'                     // 'WARMING-<int>'
  WARMUPTIME_CMD                        = 'WARMUP'                      // setWarmUpTime, 'WARMUP-<time>'
  WARMUPTIME_QRY                        = '?WARMUP'                     // getWarmUpTime, '?WARMUP'
  WEATHERALERT_QRY                      = '?WEATHERALERT'               // getWeatherAlert, '?WEATHERALERT'
  WEATHERCONDITIONS_QRY                 = '?FORECASTCONDITION'          // getWeatherConditions, '?FORECASTCONDITION'
  WEATHERSCALE_CMD                      = 'WEATHERSCALE'                // setWeatherScale, 'WEATHERSCALE-<scale>'
  WEATHERSCALE_QRY                      = '?WEATHERSCALE'               // getWeatherScale, '?WEATHERSCALE'
  WINDINFO_QRY                          = '?WEATHERWIND'                // getWindInfo, '?WEATHERWIND'
  WINDOWADDR_CMD                        = 'WINDOWADDR'
  WINDOWCOMPONENTADDRESS_QRY            = '?WINDOWADDR'                 // getWindowComponentAddress, '?WINDOWADDR-<index>'
  WINDOWCOMPONENTINDEX_QRY              = '?WINDOWIDX'                  // getWindowComponentIndex, '?WINDOWIDX-<address>'
  WINDOWPOSITION_QRY                    = '?WINDOW'                     // getWindowPosition, '?WINDOW'
  WINDOWSIZE_QRY                        = '?WINDOW'                     // getWindowSize, '?WINDOW'
  WINDOW_CMD                            = 'WINDOW'                      // setWindow, 'WINDOW-<x>,<y>,<height>,<width>'
  ZORDER1_CMD                           = 'WINDOWZORDER'                // setZOrder, 'WINDOWZORDER-<value>'
  ZORDER_CMD                            = 'WINDOWZORDER'                // setZOrder, 'WINDOWZORDER-<position>'
  ZORDER_QRY                            = '?WINDOWZORDER'               // getZOrder, '?WINDOWZORDER'

DEFINE_FUNCTION CHAR[DUET_MAX_HDR_LEN] DuetPackCmdHeader(CHAR cHdr[])
{
/*|------------------------------------------------------------------------------|
  | DuetPackCmdHeader: Package header for module SEND_COMMAND or SEND_STRING     |
  |                                                                              |
  | Note: Adds the command header to the string and adds the command if missing. |
  |       This function assumes the standard Duet command separator '-'          |
  |                                                                              |
    |    Parameters:                                                               |
  |       cHdr Header of command                                                 |
  |    Returns:                                                                  |
  |       Packed header with command separator added if missing                  |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR cSep[1]

  cSep                                  =	'-'

  IF(RIGHT_STRING(cHdr,LENGTH_ARRAY(cSep)) != cSep)
  {
    RETURN "cHdr,cSep";
  }

  RETURN cHdr;
}

DEFINE_FUNCTION CHAR[DUET_MAX_CMD_LEN] DuetPackCmdParam(CHAR cCmd[], CHAR cParam[])
{
/*|------------------------------------------------------------------------------|
  | DuetPackCmdParam: Package parameter for module SEND_COMMAND or SEND_STRING   |
  |                                                                              |
  | Note: Wraps the parameter in double-quotes if it contains the separator.     |
  |       This function assumes the standard Duet parameter separator ','        |
  |                                                                              |
  |    Parameters:                                                               |
  |       cCmd   Header of command                                               |
  |       cParam Parameter of command                                            |
  |    Returns:                                                                  |
  |       Packed parameter wrapped in double-quotes if needed, added to the      |
  |       command                                                                |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR cTemp[DUET_MAX_CMD_LEN]
  STACK_VAR CHAR cTempParam[DUET_MAX_CMD_LEN]
  STACK_VAR CHAR cCmdSep[1]
  STACK_VAR CHAR cParamSep[1]
  STACK_VAR INTEGER nLoop

  cCmdSep                               =	'-'
  cParamSep                             =	','

  // Not the first param?  Add the param separator
  cTemp                                 =	cCmd

  IF(FIND_STRING(cCmd,cCmdSep,1) != (LENGTH_ARRAY(cCmd)-LENGTH_ARRAY(cCmdSep)+1))
  {
    cTemp                               =	"cTemp,cParamSep"
  }
  // Escape any quotes
  FOR(nLoop   = 1; nLoop <= LENGTH_ARRAY(cParam); nLoop++)
  {
    IF(cParam[nLoop]         == '"')
    {
      cTempParam                        =	"cTempParam,'"'"
    }

    cTempParam                          =	"cTempParam,cParam[nLoop]"
  }

  // Add the param, wrapped in double-quotes if needed

  IF(FIND_STRING(cTempParam,cParamSep,1) > 0)
  {
    cTemp                               =	"cTemp,'"',cTempParam,'"'"
  }
  ELSE
  {
    cTemp                               =	"cTemp,cTempParam"
  }

  RETURN cTemp;
}

DEFINE_FUNCTION CHAR[DUET_MAX_CMD_LEN] DuetPackCmdParamArray(CHAR cCmd[], CHAR cParams[][])
{
/*|------------------------------------------------------------------------------|
  | DuetPackCmdParamArray: To package parameters for module SEND_COMMAND or      |
  |       SEND_STRING                                                            |
  |                                                                              |
  |    Notes : Wraps the parameter in double-quotes if it contains the           |
  |       separator and separates them using the separator sequence This         |
  |       function assumes the standard Duet parameter separator ','             |
  |                                                                              |
  |    Parameters:                                                               |
  |       cCmd    sndcmd/str to which parameter will be added                    |
  |       cParams sndcmd/str parameter array                                     |
  |    Returns:                                                                  |
  |       Packed parameters wrapped in double-quotes if needed                   |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR    cTemp[DUET_MAX_CMD_LEN]
  STACK_VAR INTEGER nLoop
  STACK_VAR INTEGER nMax
  STACK_VAR CHAR cCmdSep[1]
  STACK_VAR CHAR cParamSep[1]

  cCmdSep                               =	'-'
  cParamSep                             =	','

  nMax                                  =	LENGTH_ARRAY(cParams)

  IF(nMax                    == 0)
  {
    nMax                                =	MAX_LENGTH_ARRAY(cParams)
  }

  cTemp                                 =	cCmd

  FOR(nLoop   = 1; nLoop <= nMax; nLoop++)
  {
    cTemp                               =	DuetPackCmdParam(cTemp,cParams[nLoop])
  }

  RETURN cTemp;
}

DEFINE_FUNCTION CHAR[DUET_MAX_CMD_LEN] DuetPackCmdSimple(CHAR cHdr[], CHAR cParam[])
{
/*|------------------------------------------------------------------------------|
  | DuetPackCmdSimple: To package header and 1 parameter for module SEND_COMMAND |
  |       or SEND_STRING                                                         |
  |                                                                              |
  |    Notes : Adds the command header to the string and adds the command if     |
  |       missing. This function assumes the standard Duet command separator     |
  |       '-'. This function also adds a parameter to the command                |
  |                                                                              |
  |    Parameters:                                                               |
  |       cHdr   sndcmd/str header                                               |
  |       cParam sndcmd/str parameter                                            |
  |    Returns:                                                                  |
  |       Packed header with command separator added if missing and parameter    |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR cCmd[DUET_MAX_CMD_LEN]

  cCmd                                  =	DuetPackCmdHeader(cHdr)
  cCmd                                  =	DuetPackCmdParam(cCmd,cParam)

  RETURN cCmd;
}

DEFINE_FUNCTION CHAR[DUET_MAX_HDR_LEN] DuetParseCmdHeader(CHAR cCmd[])
{
/*|------------------------------------------------------------------------------|
  | DuetParseCmdHeader: Parse out parameters from module SEND_COMMAND or         |
  |       SEND_STRING                                                            |
  |                                                                              |
  |    Notes : Parses the strings sent to or from modules extracting the command |
  |       header. Command separating CHARacter assumed to be '-', Duet standard  |
  |                                                                              |
  |    Parameters:                                                               |
  |       cCmd sndcmd/str data                                                   |
  |    Returns:                                                                  |
  |       Parsed property/method name, still includes the leading '?' if present |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR cTemp[DUET_MAX_HDR_LEN]
  STACK_VAR CHAR cSep[1]

  cSep                                  =	'-'

  // Assume the argument to be the command
  cTemp                                 =	cCmd

  // If we find the seperator, remove it from the command

  IF(FIND_STRING(cCmd,cSep,1) > 0)
  {
    cTemp                               =	REMOVE_STRING(cCmd,cSep,1)

    IF(LENGTH_ARRAY(cTemp))
    {
      cTemp                             =	LEFT_STRING(cTemp,LENGTH_ARRAY(cTemp)-LENGTH_ARRAY(cSep))
    }
  }

  // Did not find seperator, argument is the command (like ?SOMETHING)
  ELSE
  {
    cCmd                                =	""
  }

  RETURN cTemp;
}

DEFINE_FUNCTION CHAR[DUET_MAX_PARAM_LEN] DuetParseCmdParam(CHAR cCmd[])
{
/*|------------------------------------------------------------------------------|
  | DuetParseCmdParam: Parse out parameters from module SEND_COMMAND or          |
  |       SEND_STRING                                                            |
  |                                                                              |
  |    Notes : Parses the strings sent to or from modules extracting the         |
  |       parameters. A single param is picked of the cmd string and removed,    |
  |       through the separator. The separator is not returned from the          |
  |       function. If the first character of the param is a double quote, the   |
  |       function will remove up to (and including) the next double-quote and   |
  |       the separator without spaces. The double quotes will then be stripped  |
  |       from the parameter before it is returned. If the                       |
  |       double-quote/separator sequence is not found, the function will remove |
  |       up to (and including) the separator CHARacter and the leading double   |
  |       quote will not be removed. If the separator is not found, the entire   |
  |       remained of the command is removed. Command separating character       |
  |       assumed to be ',', Duet standard                                       |
  |                                                                              |
  |    Parameters:                                                               |
  |       cCmd sndcmd/str data                                                   |
  |    Returns:                                                                  |
  |       Parsed parameter from the front of the string not including the        |
  |           separator                                                          |
  |------------------------------------------------------------------------------|*/

  STACK_VAR CHAR cTemp[DUET_MAX_PARAM_LEN]
  STACK_VAR CHAR cSep[1]
  STACK_VAR CHAR chC
  STACK_VAR INTEGER nLoop
  STACK_VAR INTEGER nState
  STACK_VAR CHAR bInquotes
  STACK_VAR CHAR bDone

  cSep                                  =	','

  // Reset state
  nState                                =	1; //ST_START
  bInquotes                             =	FALSE;
  bDone                                 =	FALSE;

  // Loop the command and escape it
  FOR(nLoop   = 1; nLoop <= LENGTH_ARRAY(cCmd); nLoop++)
  {
    // Grab CHARacters and process it based on state machine
    chC                                 =	cCmd[nLoop];

    SWITCH(nState)
    {
      // Start or STRING: end of string bails us out

      CASE 1: //ST_START
      {
        // Starts with a quote?
        // If so, skip it, set flag and move to collect.

        IF(chC               == '"')
        {
          nState                        =	2; //ST_COLLECT
          bInquotes                     =	TRUE;
        }

        // Starts with a comma?  Empty param

        ELSE IF(chC          == ',')
        {
          // I am done
          bDone                         =	TRUE;
        }

        // Not a quote or a comma?  Add it to the string and move to collection
        ELSE
        {
          cTemp                         =	"cTemp, chC"
          nState                        =	2; //ST_COLLECT
        }

        BREAK;
      }

      // Collect string.

      CASE 2: //ST_COLLECT
      {
        // If in quotes, just grab the CHARacters

        IF(bInquotes)
        {
          // Ah...found a quote, jump to end quote state

          IF(chC             == '"' )
          {
            nState                      =	3; //ST_END_QUOTE

            BREAK;
          }
        }

        // Not in quotes, look for commas

        ELSE IF(chC          == ',')
        {
          // I am done
          bDone                         =	TRUE;

          BREAK;
        }

        // Not in quotes, look for quotes (this would be wrong)
        // But instead of barfing, I will just add the quote (below)

        ELSE IF(chC          == '"' )
        {
          // I will check to see if it should be escaped

          IF(nLoop < LENGTH_ARRAY(cCmd))
          {
            // If this is 2 uqotes back to back, just INCLUDE the one

            IF(cCmd[nLoop+1]            =	'"')
            {
              nLoop++;
            }
          }
        }

        // Add CHARacter to collection
        cTemp                           =	"cTemp,chC"

        BREAK;
      }

      // End Quote

      CASE 3: //ST_END_QUOTE
      {
        // Hit a comma

        IF(chC               == ',')
        {
          // I am done
          bDone                         =	TRUE;
        }

        // OK, found a quote right after another quote.  So this is escaped.

        ELSE IF(chC          == '"')
        {
          cTemp                         =	"cTemp,chC"
          nState                        =	2; //ST_COLLECT
        }

        BREAK;
      }
    }

    // OK, if end of string or done, process and exit

    IF(bDone                 == TRUE || nLoop >= LENGTH_ARRAY(cCmd))
    {
      // remove cTemp from cCmd
      cCmd                              =	MID_STRING(cCmd, nLoop + 1, LENGTH_ARRAY(cCmd) - nLoop)

      // cTemp is done

      RETURN cTemp;
    }
  }

  // Well...we should never hit this

  RETURN "";
}

#END_IF //  __SNAPI_CONST__

/*|------------------------------------------------------------------------------|
  |                                                                              |
  |      !!! END OF PROGRAM --- DO NOT PUT ANY CODE BELOW THIS COMMENT  !!!      |
  |                                                                              |
  |------------------------------------------------------------------------------|*/